/*  
 *  NMF2map.c
 * 
 *  $Author: gael $, $Date: 2008-09-15 12:33:00 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_list.h>
#include <gdl/gdl_machine.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_runtime.h>
#include <gdl/gdl_snp_genome.h>
#include <gdl/gdl_snp_chromosome.h>
#include <gdl/gdl_snp.h>

#include "nucleoMiner.h"

#include "const.h"
#include "header.c"
#include "util.c"

#include "tiling_hmm.h"

static gdl_string * PROGRAM = "NMF2map";

static int help_flag    = 0;
static int verbose_flag = 0;

static gdl_string * CONFIG = NULL;
static gdl_string * OUTPUT  = NULL;
static size_t ITER_MAX      = 25;
static double EPSILON       = 1.e-8;
static char NA_CHAR       = '?';

static struct option long_options[] =
{
   /* These options set a flag. */
   {"help", no_argument,       &help_flag, 1},
   {"verbose", no_argument,       &verbose_flag, 1},
   {"brief",   no_argument,       &verbose_flag, 0},
   /* These options don't set a flag.
      We distinguish them by their indices. */
   {"config",   required_argument, 0, 'c'},        // the configuration file
   {"output",   required_argument, 0, 'o'},       // the output stem
   {"iter-max",   required_argument, 0, 'x'},     // the maximum number if iterations for the EM in each window
   {"epsilon",   required_argument, 0, 'e'},      // the convergence threshold for the EM in each window
   {"na-string",   required_argument, 0, 'n'},      // the character used to indicate missing genotype
   {0, 0, 0, 0}
};
 
static int
parse_argument (int argc, char *argv[])
{
	int c;
     
   while (1)
     {
      
       /* getopt_long stores the option index here. */
       int option_index = 0;
 
       c = getopt_long (argc, argv, "c:x:a:o:e:n:",
                        long_options, &option_index);
 
       /* Detect the end of the options. */
       if (c == -1)
         break;
         
       switch (c)
         {
         case 0:
           /* If this option set a flag, do nothing else now. */
           if (long_options[option_index].flag != 0)
             break;
           printf ("option %s", long_options[option_index].name);
           if (optarg)
             printf (" with arg %s", optarg);
           printf ("\n");
           break;
 
         case 'c':
           CONFIG = gdl_string_clone (optarg);
           break;
         case 'o':
           OUTPUT = gdl_string_clone (optarg);
           break;
         case 'x':
           ITER_MAX = atoi (optarg);
           break; 
         case 'e':
           EPSILON = (double)atof (optarg);
           break;
         case 'n':
           NA_CHAR = optarg[0];
           break;
         case '?':
           GDL_ERROR_VAL ("Unknown arguments", GDL_EINVAL, -1);
         default:
           GDL_ERROR_VAL ("Bad arguments", GDL_EINVAL, -1);
         }
     }
   /* Print any remaining command line arguments (not options). 
   if (optind < argc)
     {
       printf ("non-option ARGV-elements: ");
       while (optind < argc)
         printf ("%s ", argv[optind++]);
       putchar ('\n');
     }*/
}

static int
check_argument (void)
{
	if (CONFIG == 0)
	{
		GDL_ERROR_VAL ("No configuration file defined",
		               GDL_FAILURE,
		               1);
	}
	if (OUTPUT == 0)
	{
		GDL_ERROR_VAL ("No output stem defined",
		               GDL_FAILURE,
		               1);	
	}
	
	return GDL_SUCCESS;
}

static int
help (void)
{
   print_header (PROGRAM);
   printf ("\n");
   printf ("*These options set a flag:\n");
   printf ("\n");
   printf ("--help             Display a brief help on program usage\n");
   printf ("--verbose          Output message on standard output to see what the program is doing\n");
   printf ("\n");
   printf ("*These options require arguments:\n");
   printf ("\n");
   printf ("Long        Short   Description\n");
   printf ("--config     -c    The configuration file\n");
   printf ("--output     -o    The name/path to the output directory\n");
   printf ("--iter-max   -x    The maximum number of iterations for the EM\n");
   printf ("--epsilon    -e    The convergence threshold for the EM\n");
   printf ("--na-string  -n    The character used to indicate missing genotype (default is ?)\n");
}

static double
HMMbackward (const gdl_snp_data * data,
             const size_t i,
             double * beta[],
             const double scale[],
             const double R[],
             const size_t keep[],
             const size_t KL)
{
	int l = KL - 1;
	
	if (gdl_snp_data_is_missing (data, i, keep[l]))
	{
		beta[l][0]=beta[l][1]=1;
	}
	else
	{
		switch(gdl_snp_data_get (data, i, keep[l]))
		{
			case 0:
				beta[l][0] = 1;
				beta[l][1] = 0;
				break;
			case 1:
				beta[l][0] = 0;
				beta[l][1] = 1;
				break;
		}
	}
	
	for(l = KL - 2; l >= 0; l--)
	{
		if (gdl_snp_data_is_missing (data, i, keep[l]))
		{
			beta[l][0] = 0.5 * (1.0 - R[l]) * beta[l+1][0] + 0.5 * R[l] * beta[l+1][1];
			beta[l][1] = 0.5 * R[l] * beta[l+1][0] + 0.5 * (1.0 - R[l]) * beta[l+1][1]; 
		}
		else
		{
			switch(gdl_snp_data_get (data, i, keep[l]))
			{
				case 0:
					beta[l][0] = (1.0 - R[l]) * beta[l+1][0] + R[l-1] * beta[l+1][1];
					beta[l][1] = 0;
					break;
				case 1:
					beta[l][0] = 0;
					beta[l][1] = R[l] * beta[l+1][0] + (1.0 - R[l]) * beta[l+1][1];
					break;
			}
		}
		beta[l][0] /= scale[l+1];
		beta[l][1] /= scale[l+1];
	}
}

static double
HMMforward (const gdl_snp_data * data,
            const size_t i,
            double * alpha[],
            double scale[],
            const double R[],
            const size_t keep[],
            const size_t KL)
{
	size_t l = 0;
	double log_i = 0;
	
	if (gdl_snp_data_is_missing (data, i, keep[l]))
	{
		alpha[0][0]=alpha[0][1]=0.5;
	}
	else
	{
		switch(gdl_snp_data_get (data, i, keep[l]))
		{
			case 0:
				alpha[l][0] = 1;
				alpha[l][1] = 0;
				break;
			case 1:
				alpha[l][0] = 0;
				alpha[l][1] = 1;
				break;
		}
	}
	log_i += log(alpha[l][0] + alpha[l][1]);
	for(l = 1; l < KL; l++)
	{
		if (gdl_snp_data_is_missing (data, i, keep[l]))
		{
			alpha[l][0] = 0.5 * (1.0 - R[l-1]) * alpha[l-1][0] + 0.5 * R[l-1] * alpha[l-1][1];
			alpha[l][1] = 0.5 * R[l-1] * alpha[l-1][0] + 0.5 * (1.0 - R[l-1]) * alpha[l-1][1]; 
		}
		else
		{
			switch(gdl_snp_data_get (data, i, keep[l]))
			{
				case 0:
					alpha[l][0] = (1.0 - R[l-1]) * alpha[l-1][0] + R[l-1] * alpha[l-1][1];
					alpha[l][1] = 0;
					break;
				case 1:
					alpha[l][0] = 0;
					alpha[l][1] = R[l-1] * alpha[l-1][0] + (1.0 - R[l-1]) * alpha[l-1][1];
					break;
			}
		}
		scale[l] = alpha[l][0] + alpha[l][1];
		log_i += log(scale[l]);
		alpha[l][0] /= scale[l];
		alpha[l][1] /= scale[l];
	}
	
	return log_i;
}   

static void
ComputeR (double R[], double * z[], const size_t N, const size_t L)
{
	size_t i,l;
	for(l = 0; l < L-1; l++)
	{
		R[l] = 0;
		for(i = 0; i < N; i++)
		{
			R[l] += z[i][l]*(1-z[i][l+1]) + (1-z[i][l])*z[i][l+1];
		}
		R[l] /= N;
		//fprintf (stderr, "R[%d] = %g\n", l, R[l]);
	}
}
/**
 * Perform EM to estimate recombination rate
 * between pair of markers (of no missing geno,
 * then estimate recombination rate only)
 * 
 * Assume a single population
 */ 
static int
EMrecMap(const gdl_snp_chromosome * chrom, 
         const size_t max_iter,     // EM max number of iterations
         const double epsilon)      // EM covergence precision      
{
	const gdl_snp_data * data = chrom->pops[0]; 
	const size_t N = data->N;
	const size_t L = data->L;
	size_t i, j, k, l, * keep, KL;
	size_t ** midx;	
	double tot, ** alpha, ** beta, ** Z, * Rold, * Rnew, * scale, eps; 
	
	/* Alloc workspace */
	keep = GDL_CALLOC (size_t, L);
	Rold = GDL_CALLOC (double, L);
	Rnew = GDL_CALLOC (double, L);
	midx = GDL_CALLOC (size_t *, N);
	Z    = GDL_MATRIX_ALLOC (double, N, L);
	alpha = GDL_MATRIX_ALLOC (double, L, 2);
	beta  = GDL_MATRIX_ALLOC (double, L, 2); 
	scale = GDL_CALLOC (double, L);
	
	// First, keep only markers with a reasonable amount of datapoints
	KL = 0;
	for(l = 0; l < L; l++)
	{
		for(k = i = 0; i < N; i++)
		{
			if (!gdl_snp_data_is_missing (data, i, l))
			{
				k++;
			}
		}
		if (k > N/2)
		{
			keep[KL] = l;
			KL++;
		}
	}
	// for reach individual store the missing data
	// sites if any
	// first look if exists missing data
	for(i = 0; i < N; i++)
	{
		size_t nm = 0;
		for(l = 0; l < KL; l++)
		{
			if (gdl_snp_data_is_missing (data, i, keep[l]))
			{
				nm++;
				Z[i][l] = 0.5;
			}
			else
			{
				Z[i][l] = gdl_snp_data_get (data, i, keep[l]);
			}
		}
		if (nm)
		{
			midx[i] = GDL_CALLOC (size_t, nm+1);
			midx[i][0] = nm;
			nm = 1;
			for(l = 0; l < KL; l++)
			{
				if (gdl_snp_data_is_missing (data, i, keep[l]))
				{
					midx[i][nm++] = l;
				}
			}
		}
	}
	// get a first estimate of R ignoring missing genotypes
	ComputeR(Rnew, Z, N, KL);
	memcpy (Rold, Rnew, sizeof(double)*KL);
	// allocate memory
	for(j = 0; j < max_iter; j++)
	{
		for(k = 0; k < N; k++)
		{
			if (midx[k])
			{
				// E-step
				tot = HMMforward (data, k, alpha, scale, Rnew, keep, KL);
				HMMbackward (data, k, beta, scale, Rnew, keep, KL);
				//fprintf (stderr, "NMISS = %d\n", midx[k][0]);
				for(l = 1; l <= midx[k][0]; l++)
				{
					// Problem with some Z values ??? check forward backward....
					tot = alpha[midx[k][l]][0]*beta[midx[k][l]][0]+alpha[midx[k][l]][1]*beta[midx[k][l]][1];
					if (tot > 1.e-20)
					{
						Z[k][midx[k][l]] = alpha[midx[k][l]][1]*beta[midx[k][l]][1]/tot;
					}
					else
					{
						//fprintf (stderr, "PREV = %g NEXT = %g\n", R[midx[k][l]-1],R[midx[k][l]]);
						Z[k][midx[k][l]] = 0.5;
					}
					//fprintf (stderr, "UPDATE Z[%d][%d] = %g (a0 = %g, b0 = %g, a1 = %g, b1 = %g, tot =  %g)\n", k, midx[k][l], Z[k][midx[k][l]], alpha[midx[k][l]][0], beta[midx[k][l]][0], alpha[midx[k][l]][1], beta[midx[k][l]][1], tot);
				} 
			}	
		}
		// loop on missing genotypes and input them
		// M-step
		ComputeR(Rnew, Z, N, KL);
		eps = 0.;
		for(l = 0; l < KL-1; l++)
		{
			eps += fabs(Rnew[l]-Rold[l]);
		}
		fprintf (stderr, " %d epsilon = %e\n", j, eps);
		if (eps < epsilon)
		{
			break;	
		}
		else
		{
			memcpy(Rold, Rnew, sizeof(double)*KL);
		}
	}
	/* 
	 * Now, impute genotypes for the other markers
	 */
	// First, extend the estimates of the recombination
	// rate to intervals with stretches of missing loci
	// 
	// Assume rec rate to be constant over the interval and
	// use physical distance to scale it properly
	// 
	// Assume also that the first loci was kept
	//
	long pos_prev, pos_next;
	for(l = 0; l < KL - 1; l++)
	{
		pos_prev = chrom->chrom->snps[keep[l]]->position;
		pos_next = chrom->chrom->snps[keep[l+1]]->position;
		Rold[l] = Rnew[l] / (pos_next - pos_prev + 1);
	}
	k = 0;
	for(l = 0; l < L - 1; l++)
	{
		if (k < KL -1 && keep[k+1] == l)
		{
			k++;
		}
		pos_prev = chrom->chrom->snps[l]->position;
		pos_next = chrom->chrom->snps[l+1]->position;
		Rnew[l] = Rold[k] * (pos_next - pos_prev + 1);
	}
	/**
	 * Output the genetic marker map
	 */
	FILE * stream;
	gdl_string * filename = gdl_string_sprintf ("%s.gmap", chrom->name);
	stream = gdl_fileopen (filename, "w");
	k = 0;
	gdl_string * last_field = NULL;
	for(l = 0; l < L; l++)
	{
		if (k < KL -1 && keep[k+1] == l)
		{
			last_field = "*\n";
			k++;
		}
		else if (last_field)
		{
			last_field  = "\n";	
		}
		else
		{
			last_field = "*\n";
		}
		fprintf (stream, "%s\t%s\t%d\t%e\t%e\t%s", chrom->name, chrom->chrom->snps[l]->rs, chrom->chrom->snps[l]->position, Rnew[l], Rold[k], last_field);
		fflush (stream);
	}
	gdl_fileclose (filename, stream);
	gdl_string_free (filename);
	
	// Now, init the Z matrix for the entire dataset
	for(i = 0; i < N; i++)
	{
		for(l = 0; l < L; l++)
		{
			if (gdl_snp_data_is_missing (data, i, l))
			{
				Z[i][l] = 0.5;
			}
			else
			{
				Z[i][l] = gdl_snp_data_get (data, i, l);
			}
		}
	}
	// Run the HMM
	for(l = 0; l < L; l++) keep[l] = l;
	// output the imputed genotypes Z
	size_t * nna = GDL_CALLOC (size_t, L);
	for(i = 0; i < N; i++)
	{
		tot = HMMforward (data, i, alpha, scale, Rnew, keep, L);
		HMMbackward (data, i, beta, scale, Rnew, keep, L);
		for(l = 0; l < L; l++)
		{
			if (gdl_snp_data_is_missing (data, i, l))
			{
				tot = alpha[l][0]*beta[l][0]+alpha[l][1]*beta[l][1];
				if (tot > 1.e-20)
				{
					Z[i][l] = alpha[l][1]*beta[l][1]/tot;
				}
				else
				{
					Z[i][l] = 0.5;
				}
				nna[l]++;
				//fprintf (stdout, " %1.2f", Z[i][l]);
			}
//			else
//			{
//				//fprintf (stdout, " %d", (int)Z[i][l]);
//			}
		}
		//fprintf (stdout, "\n");
		//fflush (stdout);
	}
	/**
	 * Output the genotypes
	 */
	filename = gdl_string_sprintf ("%s.geno", chrom->name);
	stream   = gdl_fileopen (filename, "w");
	for(l = 0; l < L; l++)
	{
		// 1 = status (imputed or not)
		fprintf (stream, "%d", (nna[l]) ? 1 : 0);
		// 2 = marker name
		fprintf (stream, " %s", chrom->chrom->snps[l]->rs);
		// 3 = genotypes 
		if (nna[l])
		{
			for(i = 0; i < N; i++)
			{
				fprintf (stream, " %1.2f", Z[i][l]);
			}
		}
		else
		{
			for(i = 0; i < N; i++)
			{
				fprintf (stream, " %d", (int)Z[i][l]);
			}
		}
		fprintf (stream, "\n");
		fflush (stream);
	}
	gdl_fileclose (filename, stream);
	gdl_string_free (filename);
	/* Clean workspace */
	GDL_FREE (keep);
	GDL_FREE (Rold);
	GDL_FREE (Rnew);
	GDL_MATRIX_FREE (alpha, L);
	GDL_MATRIX_FREE (beta, L); 
	GDL_FREE (scale);
	GDL_MATRIX_FREE (midx, N);
	GDL_MATRIX_FREE (Z, N);
}

int
main (int argc, char *argv[])
{
  int status;
  
  parse_argument (argc, argv);
  
  if (help_flag)
  {
  	  exit (help());
  }
  
  status = check_argument ();
  
  if (status == GDL_SUCCESS)
  {
  		size_t c;
  		FILE * stream;
  		DIR * dir;
  		gdl_string * file;
  		gdl_runtime * runtime;
  		gdl_snp_data_format * F;
  		gdl_snp_genome * G;
  		
  		runtime = runtime_setup();
		
		F = gdl_snp_data_format_alloc (1);
			
		F->na_char = NA_CHAR;
		
		if ((dir=opendir(OUTPUT))==0)
		{
			mkdir (OUTPUT, S_IRWXU);
		}
		
		fprintf (stderr,"--\nRead genotypes");
		fflush (stderr);	
		G = gdl_snp_genome_alloc (CONFIG, F, OUTPUT);
		fprintf (stderr," [ DONE ]\n--\n");
		fflush (stderr);
		
		const size_t NC = gdl_snp_genome_size (G);
		for(c = 0; c < NC; c++)
		{
			gdl_snp_chromosome * chrom = gdl_snp_genome_get (G, c);
			fprintf (stderr,">chromosome %s\n", G->chroms[c]);
			fflush (stderr);
			chrom->name = gdl_string_clone (G->chroms[c]);
			EMrecMap(chrom, ITER_MAX, EPSILON);
			gdl_snp_chromosome_free (chrom);
		}
		fprintf (stderr,"--\n[ END ]\n--\n");
		fflush (stderr);
	 	runtime_setdown (runtime);
  }
  
  exit(0);
}
