/*  
 *  NMalign.c
 * 
 *  $Author: gael $, $Date: 2008-10-17 16:07:07 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_list.h>
#include <gdl/gdl_machine.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_sort_long.h>
#include <gdl/gdl_runtime.h>
#include <gdl/gdl_vector.h>
#include <gdl/gdl_matrix.h>
#include <gdl/gdl_multireg.h>
#include <gdl/gdl_genome_coord.h>

#include "nucleoMiner.h"

#include "const.h"
#include "header.c"
#include "util.c"

#include "tiling_array.h"
#include "tiling_hmm.h"
#include "tiling_nucleo.h"
#include "align_nucleo.h"

static gdl_string * PROGRAM = "NMalign";

static int help_flag       = 0;
static int verbose_flag    = 0;
static int print_flag      = 0;

static gdl_string * INPUT   = NULL;
static gdl_string * CONFIG  = NULL;
static gdl_string * OUTPUT  = NULL;
static gdl_string * CHROM   = NULL;
static size_t NCLASS        = 3;
static long TILING          = 4;
static long NUCLEO_MIN      = 31;
static double PRB_MAPPED_MIN   = 1.0;
static double PRB_OVERLAP_MIN  = 0.85;

static struct option long_options[] =
{
   /* These options set a flag. */
   {"help", no_argument,       &help_flag, 1},
   {"verbose", no_argument,       &verbose_flag, 1},
   {"brief",   no_argument,       &verbose_flag, 0},
   {"print",   no_argument,       &print_flag, 1},
   /* These options don't set a flag.
    * We distinguish them by their indices. */
   {"input",   required_argument, 0, 'i'},   // the config file
   {"config",   required_argument, 0, 'c'},   // the config file
   {"output",  required_argument, 0, 'o'},   // the output stem
   {"chrom",   required_argument, 0, 's'},   // the name of the chromosome to focus on
   {"nclass",  required_argument, 0, 'n'},    
   {"tiling",  required_argument, 0, 't'},
   {"nucmin",  required_argument, 0, 'm'}, 
   {"prb-mapped",  required_argument, 0, 'k'},
   {"prb-overlap",  required_argument, 0, 'l'},  
   {0, 0, 0, 0}
};
 
static int
parse_argument (int argc, char *argv[])
{
	int c;
     
   while (1)
     {
      
       /* getopt_long stores the option index here. */
       int option_index = 0;
 
       c = getopt_long (argc, argv, "i:o:c:s:n:t:m:k:l:",
                        long_options, &option_index);
 
       /* Detect the end of the options. */
       if (c == -1)
         break;
         
       switch (c)
         {
         case 0:
           /* If this option set a flag, do nothing else now. */
           if (long_options[option_index].flag != 0)
             break;
           printf ("option %s", long_options[option_index].name);
           if (optarg)
             printf (" with arg %s", optarg);
           printf ("\n");
           break;
 
         case 'i':
           INPUT = gdl_string_clone (optarg);
           break;
         case 'c':
           CONFIG = gdl_string_clone (optarg);
           break;
         case 'o':
           OUTPUT = gdl_string_clone (optarg);
           break;
         case 's':
           CHROM = gdl_string_clone (optarg);
           break;
         case 'n':
           NCLASS = (size_t) atoi (optarg);
           break;
         case 't':
           TILING = atol (optarg);
           break;
         case 'm':
           NUCLEO_MIN = atol (optarg);
           break;
         case 'k':
           PRB_MAPPED_MIN = (double)atof (optarg);
           break;
         case 'l':
           PRB_OVERLAP_MIN = (double)atof (optarg);
           break; 
         case '?':
           GDL_ERROR_VAL ("Unknown arguments", GDL_EINVAL, -1);
         default:
           GDL_ERROR_VAL ("Bad arguments", GDL_EINVAL, -1);
         }
     }
   /* Print any remaining command line arguments (not options). 
   if (optind < argc)
     {
       printf ("non-option ARGV-elements: ");
       while (optind < argc)
         printf ("%s ", argv[optind++]);
       putchar ('\n');
     }*/
}

static int
check_argument (void)
{
	if (CONFIG == 0)
	{
		GDL_ERROR_VAL ("No input file defined",
		               GDL_FAILURE,
		               1);
	}
	if (OUTPUT == 0)
	{
		GDL_ERROR_VAL ("No output stem defined",
		               GDL_FAILURE,
		               1);	
	}
	if (print_flag && INPUT == 0)
	{
		GDL_ERROR_VAL ("No input directory defined",
		               GDL_FAILURE,
		               1);	
	}
	
	return GDL_SUCCESS;
}

static int
help (void)
{
   print_header (PROGRAM);
   printf ("\n");
   printf ("%s implements an algorithm to align two nucleosome occupancy profiles\n", PROGRAM);
   printf ("\n");
   printf ("*These options set a flag:\n");
   printf ("\n");
   printf ("--help\tDisplay a brief help on program usage\n");
   printf ("--verbose\tOutput message on standard output to see what the program is doing\n");
   printf ("\n");
   printf ("--print\tOutput a summary of the alignments in plain text file\n");
   printf ("\n");
   printf ("*These options require arguments:\n");
   printf ("\n");
   printf ("--input or -i\tThe input directory (use with --print)\n");
   printf ("--config or -c\tThe input configuration file (see manual)\n");
   printf ("--output or -o\tThe name/path to the output directory\n");
   printf ("--chrom or -s\tThe name of the chromosome to look at\n");
   printf ("--nclass or -n\tThe number of nucleosome classes (default is 3)\n");
   printf ("--tiling or -t\tThe size of the tiling step (e.g 4bp)\n");
   printf ("--nucmin or -m\tThe minimal number of probes covering a nucleosome\n");
   printf ("--prb-mapped or -k\tThe minimal fraction of nucleosomal probes that can be mapped on both the ref and the qry (default is 1.0). Otherwise nucleosome is discarded\n");
   printf ("--prb-overlap or -l\tThe minimal fraction of overlap between ref and qry nucleosomes to be considered as aligned (default is 0.85)\n");
}

int
main (int argc, char *argv[])
{
  int status;
  
  parse_argument (argc, argv);
  
  if (help_flag)
  {
  	  exit (help());
  }
  
  status = check_argument ();
  
  if (status == GDL_SUCCESS)
  {
  		size_t c,w;
  		int status; 
  		gdl_string * file;
  		FILE * stream;
  		DIR * dir;
  		gdl_runtime * runtime;
  		gdl_tiling_nucleo_hmm * hmm;
  		align_nucleo_config * conf;
  		
  		runtime = runtime_setup();
  		
		conf = align_nucleo_config_alloc (CONFIG);
		conf->hmm_nclass = NCLASS;
		
		if ((dir=opendir(OUTPUT))==0)
		{
			mkdir (OUTPUT, S_IRWXU);
		}
		
		if (print_flag)
		{
			// Loop on chromosomes
			for(c = 0; c < conf->ref_data->nchrom; c++)
			{
				if (CHROM && strcmp(CHROM, conf->ref_data->chroms[c]->name))
	            {
	               continue;
	            }
				file = gdl_string_sprintf ("%s/%s.nal", INPUT, conf->ref_data->chroms[c]->name);
		 		if (gdl_isfile (file))
		 		{ 
			  		align_nucleo_alignment * nucleo_align;
			  		fprintf (stdout, "Process chromosome %s: wait...", conf->ref_data->chroms[c]->name);
		 	   		fflush(stdout);
			   		stream = gdl_fileopen (file, "r");
			  		nucleo_align = align_nucleo_alignment_fread (stream, conf);
			   		gdl_fileclose (file, stream);
		 	   		gdl_string_free (file);
		 	 		if (!nucleo_align)
			 		{
			 			GDL_ERROR_VAL ("Unable to read nucleosome alignment", GDL_FAILURE, GDL_FAILURE); 
			 		}
					file   = gdl_string_sprintf ("%s/%s.nuc", OUTPUT, conf->ref_data->chroms[c]->name);
			 		stream = gdl_fileopen (file, "w");
			 		align_nucleo_alignment_fprintf (stream, nucleo_align);
			 		gdl_fileclose (file, stream);
			 		gdl_string_free (file);
			 		fprintf (stdout, "\b\b\b\b\b\b\b[ DONE ]\n");
			 		fflush(stdout);
			 		align_nucleo_alignment_free (nucleo_align);
		 		}
			}
		}
		else
		{
			// Loop on chromosomes
			for(c = 0; c < conf->ref_data->nchrom; c++)
			{
	            if (CHROM && strcmp(CHROM, conf->ref_data->chroms[c]->name))
	            {
	               continue;
	            }
				align_nucleo_chromosome ref_chrom;
				// load the data 
				ref_chrom.data = conf->ref_data->chroms[c];
			 	fprintf (stdout, "Process chromosome %s: wait...", ref_chrom.data->name);
			 	fflush(stdout);
			 	// load the hmm results
			 	ref_chrom.hmm  = load_chrom_hmm (conf->ref_hmm_dir, ref_chrom.data->name, (NCLASS==3) ? 1 : 2, NCLASS);
			 	// anchor the hmm results on the data
			 	align_nucleo_chromosome_anchor_probe (&ref_chrom);
			 	// anchor the qry data and hmm results on the ref
			 	align_nucleo_chromosome_anchor_qry (conf, &ref_chrom, NCLASS, TILING);
			 	align_nucleo_chromosome_anchor_nucleo (&ref_chrom, TILING, NUCLEO_MIN, PRB_MAPPED_MIN, NCLASS);
			 	align_nucleo_alignment * nucleo_align = align_nucleo_chromosome_align_nucleo (&ref_chrom, PRB_OVERLAP_MIN);
			 	file   = gdl_string_sprintf ("%s/%s.nal", OUTPUT, ref_chrom.data->name);
			 	stream = gdl_fileopen (file, "w");
			 	align_nucleo_alignment_fwrite (stream, conf, nucleo_align);
			 	gdl_fileclose (file, stream);
			 	fprintf (stdout, "\b\b\b\b\b\b\b[ DONE ]\n", ref_chrom.data->name);
			 	fflush(stdout);
			 	align_nucleo_alignment_free (nucleo_align);
			}
		}
		
		runtime_setdown (runtime);
  }
  exit(0);
}

