/*
 *  NMalign2.c
 *
 *  $Author: gael $, $Date: 2008-10-17 16:07:07 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 *
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 */
#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_list.h>
#include <gdl/gdl_machine.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_sort_long.h>
#include <gdl/gdl_runtime.h>
#include <gdl/gdl_vector.h>
#include <gdl/gdl_matrix.h>
#include <gdl/gdl_multireg.h>
#include <gdl/gdl_genome_coord.h>

#include "nucleoMiner.h"

#include "const.h"
#include "header.c"
#include "util.c"

#include "tiling_array.h"
#include "tiling_hmm.h"
#include "tiling_nucleo.h"
#include "align_nucleo2.h"

static gdl_string * PROGRAM = "NMalign2";

static int help_flag       = 0;
static int verbose_flag    = 0;
static int print_flag      = 0;

static gdl_string * INPUT   = NULL;
static gdl_string * CONFIG  = NULL;
static gdl_string * OUTPUT  = NULL;
static gdl_string * CHROM   = NULL;
static size_t TILING        = 4;
static size_t NUCLEO_MIN    = 31;
static double COVERAGE      = 0.5;
static size_t OVERLAP_MIN   = 50;


static struct option long_options[] =
{
   /* These options set a flag. */
   {"help", no_argument,       &help_flag, 1},
   {"verbose", no_argument,       &verbose_flag, 1},
   {"brief",   no_argument,       &verbose_flag, 0},
   {"print",   no_argument,       &print_flag, 1},
   /* These options don't set a flag.
    * We distinguish them by their indices. */
   {"input",   required_argument, 0, 'i'},   // the config file
   {"config",   required_argument, 0, 'c'},   // the config file
   {"output",  required_argument, 0, 'o'},   // the output stem
   {"chrom",   required_argument, 0, 's'},   // the name of the chromosome to focus on
   {"tiling",  required_argument, 0, 't'},
   {"ovrmin",  required_argument, 0, 'v'},
   {"coverage",  required_argument, 0, 'g'},
   {"nucmin",  required_argument, 0, 'n'},
   {0, 0, 0, 0}
};

static int
parse_argument (int argc, char *argv[])
{
	int c;

   while (1)
     {

       /* getopt_long stores the option index here. */
       int option_index = 0;

       c = getopt_long (argc, argv, "i:o:c:s:t:v:g:n:",
                        long_options, &option_index);

       /* Detect the end of the options. */
       if (c == -1)
         break;

       switch (c)
         {
         case 0:
           /* If this option set a flag, do nothing else now. */
           if (long_options[option_index].flag != 0)
             break;
           printf ("option %s", long_options[option_index].name);
           if (optarg)
             printf (" with arg %s", optarg);
           printf ("\n");
           break;

         case 'i':
           INPUT = gdl_string_clone (optarg);
           break;
         case 'c':
           CONFIG = gdl_string_clone (optarg);
           break;
         case 'o':
           OUTPUT = gdl_string_clone (optarg);
           break;
         case 's':
           CHROM = gdl_string_clone (optarg);
           break;
         case 'v':
           OVERLAP_MIN = (size_t) atoi (optarg);
           break;
         case 'g':
           COVERAGE = (double) atof (optarg);
           break;
         case 't':
           TILING = (size_t)atoi (optarg);
           break;
         case 'n':
           NUCLEO_MIN = (size_t)atoi (optarg);
           break;
         case '?':
           GDL_ERROR_VAL ("Unknown arguments", GDL_EINVAL, -1);
         default:
           GDL_ERROR_VAL ("Bad arguments", GDL_EINVAL, -1);
         }
     }
   /* Print any remaining command line arguments (not options).
   if (optind < argc)
     {
       printf ("non-option ARGV-elements: ");
       while (optind < argc)
         printf ("%s ", argv[optind++]);
       putchar ('\n');
     }*/
}

static int
check_argument (void)
{
	if (CONFIG == 0)
	{
		GDL_ERROR_VAL ("No input file defined",
		               GDL_FAILURE,
		               1);
	}
	if (OUTPUT == 0)
	{
		GDL_ERROR_VAL ("No output stem defined",
		               GDL_FAILURE,
		               1);
	}
	if (print_flag && INPUT == 0)
	{
		GDL_ERROR_VAL ("No input directory defined",
		               GDL_FAILURE,
		               1);
	}

	return GDL_SUCCESS;
}

static int
help (void)
{
   print_header (PROGRAM);
   printf ("\n");
   printf ("%s implements an algorithm to align two nucleosome occupancy profiles\n", PROGRAM);
   printf ("\n");
   printf ("*These options set a flag:\n");
   printf ("\n");
   printf ("--help\tDisplay a brief help on program usage\n");
   printf ("--verbose\tOutput message on standard output to see what the program is doing\n");
   printf ("\n");
   printf ("--print\tOutput a summary of the alignments in plain text file\n");
   printf ("\n");
   printf ("*These options require arguments:\n");
   printf ("\n");
   printf ("--input or -i\tThe input directory (use with --print)\n");
   printf ("--config or -c\tThe input configuration file (see manual)\n");
   printf ("--output or -o\tThe name/path to the output directory\n");
   printf ("--chrom or -s\tThe name of the chromosome to look at\n");
   printf ("--ovrmin or -v\tThe minimal number of bp in common between two nucleosomes\n");
   printf ("--coverage or -g\tThe coverage fraction w.r.t to the average size of the nucleosome\n");
   printf ("\n");
   printf ("*These following options must fit the HMM parameters\n");
   printf ("\n");
   printf ("--tiling or -t\tThe tiling step size (in bp, default is 4)\n");
   printf ("--nucmin or -n\tThe min number of probes within a nucleosome (default is 31)\n");
}

int
main (int argc, char *argv[])
{
  int status;

  parse_argument (argc, argv);

  if (help_flag)
  {
  	  exit (help());
  }

  status = check_argument ();

  if (status == GDL_SUCCESS)
  {
  		size_t c,w;
  		int status;
  		gdl_string * file;
  		FILE * stream, * stream2;
  		DIR * dir;
  		gdl_runtime * runtime;
  		gdl_tiling_nucleo_hmm * hmm;
  		nm_align_nucleo_config * conf;

  		runtime = runtime_setup();

		conf = nm_align_nucleo_config_alloc (CONFIG);

		if ((dir=opendir(OUTPUT))==0)
		{
			mkdir (OUTPUT, S_IRWXU);
		}

		// Step 1 - create a blank alignment object

		// Step 2 - loop on the ref chromosomes, create it and load
		//          the corresponding qry chromosomes
		//          All of this is stored in the alignment object

		if (verbose_flag)
		{
			fprintf (stdout, "--\nStep 1: Load Reference HMM files\n--\n");
		 	fflush(stdout);
		}

		nm_align_nucleo_workspace * work = nm_align_nucleo_workspace_alloc (conf);

		if (verbose_flag)
		{
			fprintf (stdout, "--\nStep 2: Load corresponding Query HMM files\n--\n");
		 	fflush(stdout);
		}

		for(c = 0; c < work->nref_chrom; c++)
		{
            if (CHROM && strcmp(CHROM, work->ref_chroms[c]->name))
            {
               continue;
            }
			if (verbose_flag)
			{
				fprintf (stdout, "Process chromosome %s: wait...", work->ref_chroms[c]->name);
			 	fflush(stdout);
			}
			nm_align_nucleo_chromosome * ref_chrom = work->ref_chroms[c];
			/**
		 	 * Here use the alignment to store the query chromosomes. If the query chrom
		 	 * has already been seen, just re-upload it
		 	 */
		 	nm_align_nucleo_workspace_anchor_qry (work, ref_chrom, 3, TILING, 0);

		 	if (verbose_flag)
			{
			 	fprintf (stdout, "\b\b\b\b\b\b\b[ DONE ]\n", work->ref_chroms[c]->name);
			 	fflush(stdout);
			}
		}

		nm_align_nucleo_workspace_init_qry (work);

		/**
		 * Do the calling of the nucleosomes
		 */
		if (verbose_flag)
		{
			fprintf (stdout, "--\nStep 3: Call nucleosomes on the Reference\n--\n");
		 	fflush(stdout);
		}
		for(c = 0; c < work->nref_chrom; c++)
		{
			nm_align_nucleo_chromosome_anchor_nucleo (work->ref_chroms[c], TILING, NUCLEO_MIN, 3, 0);
		}
		if (verbose_flag)
		{
			fprintf (stdout, "--\nStep 4: Call nucleosomes on the Query\n--\n");
		 	fflush(stdout);
		}
		for(c = 0; c < work->nqry_chrom; c++)
		{
			nm_align_nucleo_chromosome_anchor_nucleo (work->qry_chroms[c], TILING, NUCLEO_MIN, 3, 1);
		}
		if (verbose_flag)
		{
			fprintf (stdout, "--\nStep 4: Map Query nucleosomes on the Reference\n--\n");
		 	fflush(stdout);
		}

		nm_align_nucleo_chromosome_map_qry (work, stderr);

		if (verbose_flag)
		{
			fprintf (stdout, "--\nStep 5: Align nucleosomes\n--\n");
		 	fflush(stdout);
		}
		for(c = 0; c < work->nref_chrom; c++)
		{
            if (verbose_flag)
			{
				fprintf (stdout, "Process chromosome %s: wait...", work->ref_chroms[c]->name);
			 	fflush(stdout);
			}

			nm_align_nucleo_chromosome * ref_chrom = work->ref_chroms[c];

			if (gdl_list_size (ref_chrom->_mapped_nucleosomes))
			{
				size_t i;
				gdl_list_itr * itr = gdl_list_iterator_front (ref_chrom->_mapped_nucleosomes);

				do
				{
					nm_align_nucleo_nucleosome * qry_nuc = (nm_align_nucleo_nucleosome *) gdl_list_iterator_value (itr);
					long qry_start = qry_nuc->ref_start_pos;
					long qry_end   = qry_nuc->ref_end_pos;
					long max = 0, overlap = 0;
					int i_max = -1;

					for(i = 0; i < ref_chrom->n_nucleosome; i++)
					{
						if (ref_chrom->nucleosomes[i]->align)
						{
							continue;
						}
						if (ref_chrom->nucleosomes[i]->ref_start_pos > qry_end)
						{
							break;
						}
						long ref_start = ref_chrom->nucleosomes[i]->ref_start_pos;
						long ref_end   = ref_chrom->nucleosomes[i]->ref_end_pos;
						if (ref_start >= qry_start && ref_end <= qry_end)
						{
							// qqqqqqqqqqqqqqqq
							//    rrrrrrrrrrr
							overlap = (ref_end-ref_start);
						}
						else if ((ref_start <= qry_start && ref_end >= qry_start) && ref_end <= qry_end)
						{
							//       qqqqqqqqqq
							// rrrrrrrrrrr
							overlap = (ref_end-qry_start);
						}
						else if ((ref_start >= qry_start && ref_start <= qry_end) && ref_end >= qry_end)
						{
							// qqqqqqqqqq
							//      rrrrrrrrrrrrrr
							overlap = (qry_end-ref_start);
						}
						else if (qry_start >= ref_start && qry_end <= ref_end)
						{
							//    qqqqqqqqqq
							//  rrrrrrrrrrrrrrr
							overlap = (qry_end-qry_start);
						}
						else
						{
							continue;
						}
						//if (overlap > OVERLAP_MIN && ((double)overlap) >= 0.5*COVERAGE*((qry_end-qry_start)+(ref_end-ref_start)))
						if (overlap > OVERLAP_MIN)
						{
							if (overlap > max)
							{
								max   = overlap;
								i_max = i;
							}
						}
					}
					if (i_max != -1)
					{
						ref_chrom->nucleosomes[i_max]->align = qry_nuc;
						qry_nuc->align     = ref_chrom->nucleosomes[i_max];
						qry_nuc->align_idx = i_max;
					}
					else
					{
						qry_nuc->align_idx = -1;
					}
				}
				while(gdl_list_iterator_next (itr));

				gdl_list_iterator_free (itr);

				// save it in text format
				file   = gdl_string_sprintf ("%s/%s_ref.txt", OUTPUT, conf->chroms[c]);
				stream = gdl_fileopen (file, "w");
				for(i = 0; i < ref_chrom->n_nucleosome; i++)
				{
					if (ref_chrom->nucleosomes[i]->align)
					{
						fprintf (stream, "m %s %d %d %d %d %d %d\n", ref_chrom->name, ref_chrom->nucleosomes[i]->chunk_idx, ref_chrom->nucleosomes[i]->ref_start_pos, ref_chrom->nucleosomes[i]->ref_end_pos, ref_chrom->nucleosomes[i]->align->chunk_idx, ref_chrom->nucleosomes[i]->align->ref_start_pos, ref_chrom->nucleosomes[i]->align->ref_end_pos);
					}
					else
					{
						fprintf (stream, "u %s %d %d %d NA NA NA\n", ref_chrom->name, ref_chrom->nucleosomes[i]->chunk_idx, ref_chrom->nucleosomes[i]->ref_start_pos, ref_chrom->nucleosomes[i]->ref_end_pos);
					}
				}
				gdl_fileclose (file, stream);
				// save it in binary format
				file   = gdl_string_sprintf ("%s/%s_ref.db", OUTPUT, conf->chroms[c]);
				stream = gdl_fileopen (file, "w");
				nm_align_nucleo_chromosome_fwrite (stream, ref_chrom);
				gdl_fileclose (file, stream);

				if (verbose_flag)
				{
				 	fprintf (stdout, "\b\b\b\b\b\b\b[ DONE ]\n");
				 	fflush(stdout);
				}
			}
			else
			{
				if (verbose_flag)
				{
				 	fprintf (stdout, "\b\b\b\b\b\b\b[ DONE ]\n");
				 	fprintf (stdout, " !! NO NUCLEOSOME MAPPED ON THAT CHROMOSOME !!\n");
				 	fflush(stdout);
				}
			}
		}

		runtime_setdown (runtime);
  }
  exit(0);
}

