/*  
 *  NManofdr.c
 * 
 *  $Author: veyrieras $, $Date: 2008-11-14 15:05:01 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */

#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_list.h>
#include <gdl/gdl_machine.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_sort_long.h>
#include <gdl/gdl_runtime.h>
#include <gdl/gdl_vector.h>
#include <gdl/gdl_matrix.h>
#include <gdl/gdl_multireg.h>
#include <gdl/gdl_genome_coord.h>
#include <gdl/gdl_gfeatures.h>
#include <gdl/gdl_gfeatures_gff3.h>

#include "nucleoMiner.h"

#include "const.h"
#include "header.c"
#include "util.c"

#include "tiling_array.h"
#include "tiling_nucleo.h"
#include "tiling_feature.h"
#include "align_nucleo.h"

static gdl_string * PROGRAM = "NManofdr";

static int help_flag        = 0;
static int verbose_flag     = 0;

static gdl_string * INPUT   = NULL;
static gdl_string * CONFIG  = NULL;
static gdl_string * OUTPUT  = NULL;

static struct option long_options[] =
{
   /* These options set a flag. */
   {"help", no_argument,           &help_flag, 1},
   {"verbose", no_argument,        &verbose_flag, 1},
   {"brief",   no_argument,        &verbose_flag, 0},
   /* These options don't set a flag.
      We distinguish them by their indices. */
   {"input",   required_argument, 0, 'i'},    // the config file
   {"config",   required_argument, 0, 'c'},    // the config file
   {"output",  required_argument, 0, 'o'},    // the output stem
   {0, 0, 0, 0}
};
 
 
static int
parse_argument (int argc, char *argv[])
{
	int c;
     
   while (1)
     {
      
       /* getopt_long stores the option index here. */
       int option_index = 0;
 
       c = getopt_long (argc, argv, "i:c:o:",
                        long_options, &option_index);
 
       /* Detect the end of the options. */
       if (c == -1)
         break;
         
       switch (c)
         {
         case 0:
           /* If this option set a flag, do nothing else now. */
           if (long_options[option_index].flag != 0)
             break;
           printf ("option %s", long_options[option_index].name);
           if (optarg)
             printf (" with arg %s", optarg);
           printf ("\n");
           break;
 
         case 'i':
           INPUT = gdl_string_clone (optarg);
           break;
         case 'o':
           OUTPUT = gdl_string_clone (optarg);
           break;
         case 'c':
           CONFIG = gdl_string_clone (optarg);
           break;
         case '?':
           GDL_ERROR_VAL ("Unknown arguments", GDL_EINVAL, -1);
         default:
           GDL_ERROR_VAL ("Bad arguments", GDL_EINVAL, -1);
         }
     }
   /* Print any remaining command line arguments (not options). 
   if (optind < argc)
     {
       printf ("non-option ARGV-elements: ");
       while (optind < argc)
         printf ("%s ", argv[optind++]);
       putchar ('\n');
     }*/
}

static int
check_argument (void)
{
	if (INPUT == 0)
	{
		GDL_ERROR_VAL ("No input directory defined",
		               GDL_FAILURE,
		               1);
	}
	if (OUTPUT == 0)
	{
		GDL_ERROR_VAL ("No output stem defined",
		               GDL_FAILURE,
		               1);	
	}
	
	return GDL_SUCCESS;
}

static int
help (void)
{
   print_header (PROGRAM);
   printf ("\n");
   printf ("%s summarizes results obtained with NManova\n", PROGRAM);
   printf ("   by computing overlap score with given features\n"); 
   printf ("\n");
   printf ("*These options set a flag:\n");
   printf ("\n");
   printf ("--help\tDisplay a brief help on program usage\n");
   printf ("\n");
   printf ("*These options require arguments:\n");
   printf ("\n");
   printf ("--input or -i\tThe input directory (use with --print)\n");
   printf ("--config or -c\tThe input configuration file (see manual)\n");
   printf ("--output or -o\tThe name/path to the output directory\n");
//   printf ("--nclass or -n\tThe number of nucleosome classes (default is 3)\n");
}

int
main (int argc, char *argv[])
{
  int status;
  
  parse_argument (argc, argv);
  
  if (help_flag)
  {
  	  exit (help());
  }
  
  status = check_argument ();
  
  if (status == GDL_SUCCESS)
  {
  		size_t c;
  		int i, j, status; 
  		gdl_string * file;
  		FILE * stream;
  		gdl_runtime * runtime;
  		align_nucleo_config * conf;
  		align_nucleo_nucleosome * ref_nuc, * qry_nuc;
		align_nucleo_alignment * align;
		double * p_val = 0;
		size_t np_val  = 0;
	  
  		runtime = runtime_setup();
  		
  		// read the configuration file
  		conf = align_nucleo_config_alloc (CONFIG);
		conf->hmm_nclass = 3;
		// Loop on chromosomes
		for(c = 0; c < conf->ref_data->nchrom; c++)
		{
			file = gdl_string_sprintf ("%s/%s.nal", INPUT, conf->ref_data->chroms[c]->name);
		 	if (gdl_isfile (file))
		 	{ 
			    fprintf (stderr, "Load p-values from chromosome %s: wait...", conf->ref_data->chroms[c]->name);
		 		fflush(stderr);
			 	
			 	stream = gdl_fileopen (file, "r");
			 	align = align_nucleo_alignment_fread (stream, conf);
			 	gdl_fileclose (file, stream);
		 	 	gdl_string_free (file);
		 	 	
		 	 	if (!align)
			 	{
			 		GDL_ERROR_VAL ("Unable to read nucleosome alignment", GDL_FAILURE, GDL_FAILURE); 
			 	}
			 	
			 	size_t nok = 0;
			 	for(i = 0; i < align->size; i++)
				{
					ref_nuc = align->ref[i];
					if (!ref_nuc || !ref_nuc->align || ref_nuc->ignore == 'y' || ref_nuc->pval == -1)
							continue;
					nok++;
				}
				if (p_val)
				{
					double * tmp = GDL_CALLOC (double, np_val + nok);
					memcpy(tmp, p_val, sizeof(double)*(np_val));
					GDL_FREE (p_val);
					p_val = tmp;
				}
				else
				{
					p_val = GDL_CALLOC (double, nok);	
				}
				nok = 0;
				for(i = 0; i < align->size; i++)
				{
					ref_nuc = align->ref[i];
					if (!ref_nuc || !ref_nuc->align || ref_nuc->ignore == 'y' || ref_nuc->pval == -1)
							continue;
					p_val[np_val + nok++] = ref_nuc->pval;
				}
				np_val += nok; 
				
				align_nucleo_alignment_free (align);
			 	fprintf (stderr, "\b\b\b\b\b\b\b[ OK ]\n");
		 		fflush(stderr);
		 	}	 	
		}
		fprintf (stderr, "--\n");
		fprintf (stderr, "#p-values = %d\n", np_val);
		fprintf (stderr, "--\n");
		fprintf (stderr, "Sort p-values: wait...");
		fflush(stderr);
		gdl_sort (p_val, 1, np_val);
		fprintf (stderr, "\b\b\b\b\b\b\b[ OK ]\n");
		fflush(stderr);
		fprintf (stderr, "Compute FDR: wait...");
		double fdr[6]   = {1.e-4,1.e-3,1.e-2,0.05,0.1,0.2};
		size_t fdrpv[6] = {0,0,0,0,0,0};
		size_t fdrok[6] = {0,0,0,0,0,0};
		for(i = np_val - 1; i >= 0; i--)
		{
			for(j = 0; j < 6; j++)
			{
				if (!fdrok[j] && p_val[i] < (fdr[j]*(double)(i+1))/(double)np_val)
				{
					fdrok[j] = 1;
					fdrpv[j] = i;
				}				
			}
			if (fdrok[0])
				break;
		}
		fprintf (stderr, "\b\b\b\b\b\b\b[ OK ]\n");
		fflush(stderr);
		
		stream = gdl_fileopen (OUTPUT, "w");
		fprintf (stream, "###\n");
		fprintf (stream, "# Output of NManofdr\n");
		fprintf (stream, "###\n");
		fprintf (stream, "FDR\tPval\t#SNEP\t#NUC\t%%SNEP\n");
		for(j = 0; j < 6; j++)
		{
			fprintf (stream, "%e\t%e\t%d\t%d\t%1.5f\n", fdr[j], p_val[fdrpv[j]], fdrpv[j]+1, np_val, (double)(fdrpv[j]+1)/np_val); 
		}
		gdl_fileclose (OUTPUT, stream);
		
		runtime_setdown (runtime);
  }
  
  exit(0);
}
