/*  
 *  NManofoc.c
 * 
 *  $Author: veyrieras $, $Date: 2008-11-14 15:05:01 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */

#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_list.h>
#include <gdl/gdl_machine.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_sort_long.h>
#include <gdl/gdl_runtime.h>
#include <gdl/gdl_vector.h>
#include <gdl/gdl_matrix.h>
#include <gdl/gdl_multireg.h>
#include <gdl/gdl_genome_coord.h>
#include <gdl/gdl_gfeatures.h>
#include <gdl/gdl_gfeatures_gff3.h>

#include "nucleoMiner.h"

#include "const.h"
#include "header.c"
#include "util.c"

#include "tiling_array.h"
#include "tiling_nucleo.h"
#include "tiling_feature.h"
#include "align_nucleo.h"

static gdl_string * PROGRAM = "NManofoc";

static int help_flag        = 0;
static int verbose_flag     = 0;

static gdl_string * INPUT   = NULL;
static gdl_string * CONFIG  = NULL;
static gdl_string * OUTPUT  = NULL;
static gdl_string * FEATURE = NULL;
static gdl_string * FEATURE_ID = NULL;
static gdl_string * SEQ_ID  = NULL;

static struct option long_options[] =
{
   /* These options set a flag. */
   {"help", no_argument,           &help_flag, 1},
   {"verbose", no_argument,        &verbose_flag, 1},
   {"brief",   no_argument,        &verbose_flag, 0},
   /* These options don't set a flag.
      We distinguish them by their indices. */
   {"input",   required_argument, 0, 'i'},    // the config file
   {"config",   required_argument, 0, 'c'},    // the config file
   {"output",  required_argument, 0, 'o'},    // the output stem
   {"feature",  required_argument, 0, 'f'},   // the output stem
   {"feature-id",  required_argument, 0, 'l'},   // the output stem
   {"seq-id",  required_argument, 0, 's'},   // the output stem
   {0, 0, 0, 0}
};
 
 
static int
parse_argument (int argc, char *argv[])
{
	int c;
     
   while (1)
     {
      
       /* getopt_long stores the option index here. */
       int option_index = 0;
 
       c = getopt_long (argc, argv, "i:c:o:f:l:s:",
                        long_options, &option_index);
 
       /* Detect the end of the options. */
       if (c == -1)
         break;
         
       switch (c)
         {
         case 0:
           /* If this option set a flag, do nothing else now. */
           if (long_options[option_index].flag != 0)
             break;
           printf ("option %s", long_options[option_index].name);
           if (optarg)
             printf (" with arg %s", optarg);
           printf ("\n");
           break;
 
         case 'i':
           INPUT = gdl_string_clone (optarg);
           break;
         case 'o':
           OUTPUT = gdl_string_clone (optarg);
           break;
         case 'c':
           CONFIG = gdl_string_clone (optarg);
           break;
         case 'f':
           FEATURE = gdl_string_clone (optarg);
           break;
         case 'l':
           FEATURE_ID = gdl_string_clone (optarg);
           break;
         case 's':
           SEQ_ID = gdl_string_clone (optarg);
           break;
         case '?':
           GDL_ERROR_VAL ("Unknown arguments", GDL_EINVAL, -1);
         default:
           GDL_ERROR_VAL ("Bad arguments", GDL_EINVAL, -1);
         }
     }
   /* Print any remaining command line arguments (not options). 
   if (optind < argc)
     {
       printf ("non-option ARGV-elements: ");
       while (optind < argc)
         printf ("%s ", argv[optind++]);
       putchar ('\n');
     }*/
}

static int
check_argument (void)
{
	if (INPUT == 0)
	{
		GDL_ERROR_VAL ("No input directory defined",
		               GDL_FAILURE,
		               1);
	}
//	if (OUTPUT == 0)
//	{
//		GDL_ERROR_VAL ("No output stem defined",
//		               GDL_FAILURE,
//		               1);	
//	}
	
	return GDL_SUCCESS;
}


static const gdl_string *
rec_id (const gdl_gfeatures_gff3_record * rec)
{
	if (gdl_hashtable_lookup (rec->attributes, "ID"))
		return gdl_hashtable_lookup (rec->attributes, "ID");
	else if (rec->nparent)
		return gdl_hashtable_lookup (rec->parents[0]->attributes, "ID");
}

static double
nuc_weight (const align_nucleo_nucleosome * ref_nuc,
            const long region_start,
            const long region_end)
{
	if (ref_nuc->ref_start_pos > region_end || ref_nuc->ref_end_pos < region_start)
	{
		return 0.0;	
	}
	if (ref_nuc->ref_start_pos >= region_start && ref_nuc->ref_end_pos <= region_end)
	{
		return 1.0;
	}
	else if (ref_nuc->ref_start_pos < region_start)
	{
		return ((double)(ref_nuc->ref_end_pos-region_start))/(ref_nuc->ref_end_pos-ref_nuc->ref_start_pos);	
	}
	else
	{
		return ((double)(region_end-ref_nuc->ref_start_pos))/(ref_nuc->ref_end_pos-ref_nuc->ref_start_pos);	
	}
}

static int
help (void)
{
   print_header (PROGRAM);
   printf ("\n");
   printf ("%s summarizes results obtained with NManova\n", PROGRAM);
   printf ("   by computing overlap score with given features\n"); 
   printf ("\n");
   printf ("*These options set a flag:\n");
   printf ("\n");
   printf ("--help\tDisplay a brief help on program usage\n");
   printf ("\n");
   printf ("*These options require arguments:\n");
   printf ("\n");
   printf ("--input or -i\tThe input directory (use with --print)\n");
   printf ("--config or -c\tThe input configuration file (see manual)\n");
   printf ("--output or -o\tThe name/path to the output directory\n");
   printf ("--nclass or -n\tThe number of nucleosome classes (default is 3)\n");
   printf ("--seqid or -s\tThe name of the chromosome to look at\n");
   printf ("--feature or -f\tA GFF (version 3) annotation file\n");
   printf ("--feature-id or -l\tA file listing the features to consider\n");
}

int
main (int argc, char *argv[])
{
  int status;
  
  parse_argument (argc, argv);
  
  if (help_flag)
  {
  	  exit (help());
  }
  
  status = check_argument ();
  
  if (status == GDL_SUCCESS)
  {
  		size_t c,w,snep_id=0;
  		int status; 
  		gdl_string * file;
  		FILE * stream;
  		gdl_runtime * runtime;
  		align_nucleo_config * conf;
  		gdl_gfeatures * feature;
  		gdl_tiling_data * rdata = 0, * qdata = 0;
  		gdl_hashtable * selected_features = 0;
  		int i, j, start_idx, end_idx;
		long start, end;
		align_nucleo_nucleosome * ref_nuc, * qry_nuc;
		align_nucleo_alignment * align;
	   gdl_gfeatures_landmark * chrom_feature;
	   gdl_tiling_gene_expcmp * expcmp = 0;
  		gdl_tiling_gene_expcmp_entry * exp_entry;
  	
  		runtime = runtime_setup();
  		
  		// Read the feature
		feature = gdl_gfeatures_alloc (gdl_gfeatures_gff3, FEATURE);
		 
		conf = align_nucleo_config_alloc (CONFIG);
		conf->hmm_nclass = 3;
		
		if (FEATURE_ID)
		{
			selected_features = read_selected_features (FEATURE_ID);
		}
		
		// Loop on chromosomes
		for(c = 0; c < conf->ref_data->nchrom; c++)
		{
			if (SEQ_ID && strcmp(conf->ref_data->chroms[c]->name, SEQ_ID))
			{
				continue;
			}
			file = gdl_string_sprintf ("%s/%s.nal", INPUT, conf->ref_data->chroms[c]->name);
		 	if (gdl_isfile (file))
		 	{ 
			   fprintf (stderr, "Process chromosome %s: wait...", conf->ref_data->chroms[c]->name);
		 		fflush(stderr);
			 	
			 	stream = gdl_fileopen (file, "r");
			 	align = align_nucleo_alignment_fread (stream, conf);
			 	gdl_fileclose (file, stream);
		 	 	gdl_string_free (file);
		 	 	
		 	 	if (!align)
			 	{
			 		GDL_ERROR_VAL ("Unable to read nucleosome alignment", GDL_FAILURE, GDL_FAILURE); 
			 	}
			 	
			 	chrom_feature = gdl_gfeatures_lookup (feature, conf->ref_data->chroms[c]->name);
			 	
			 	if (!chrom_feature)
			 	{
			 		GDL_ERROR_VAL (gdl_string_sprintf ("Unable to get features for chromosome %s", conf->ref_data->chroms[c]->name), GDL_FAILURE, GDL_FAILURE); 
			 	}
				for(i = 0; i < align->size; i++)
				{
					ref_nuc = align->ref[i];
					if (!ref_nuc || !ref_nuc->align || ref_nuc->ignore == 'y')
							continue;
					long bin_start = ref_nuc->ref_start_pos;
					long bin_end   = ref_nuc->ref_end_pos;
					double score = 0;
					size_t f;
				 	for(f = 0; f < chrom_feature->size; f++)
					{
						gdl_gfeatures_gff3_record * rec = (gdl_gfeatures_gff3_record *) chrom_feature->records[f];
						if (rec->end < bin_start)
							continue;
						if (rec->start > bin_end)
							break;
						if (is_a_selected_feature (feature, rec, selected_features))
						{
							if (rec->start >= bin_start && rec->start <= bin_end)
							{
								if (rec->end <= bin_end)
									score += 1;
								else
									score += ((double)(bin_end-rec->start))/(rec->end-rec->start);
							}
							else if (rec->end >= bin_start && rec->end <= bin_end)
							{
								score += ((double)(rec->end-bin_start))/(rec->end-rec->start);
							}
						}
					}
					printf ("%s %d %d %g %g %g\n", conf->ref_data->chroms[c]->name, ref_nuc->ref_start_pos, ref_nuc->ref_end_pos, score, ref_nuc->pval, ref_nuc->gamma);
				}
				align_nucleo_alignment_free (align);
			 	fprintf (stderr, "\b\b\b\b\b\b\b[ OK ]\n");
		 		fflush(stderr);
		 	}	 	
		}
		gdl_gfeatures_free (feature);
  	
		runtime_setdown (runtime);
  }
  
  exit(0);
}
