/*  
 *  NManout.c
 * 
 *  $Author: veyrieras $, $Date: 2008-11-14 10:45:04 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_list.h>
#include <gdl/gdl_machine.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_sort_long.h>
#include <gdl/gdl_runtime.h>
#include <gdl/gdl_vector.h>
#include <gdl/gdl_matrix.h>
#include <gdl/gdl_multireg.h>
#include <gdl/gdl_genome_coord.h>
#include <gdl/gdl_gfeatures.h>
#include <gdl/gdl_gfeatures_gff3.h>

#include "nucleoMiner.h"

#include "const.h"
#include "header.c"
#include "util.c"

#include "tiling_array.h"
#include "tiling_hmm.h"
#include "tiling_nucleo.h"
#include "tiling_feature.h"
#include "align_nucleo.h"

static gdl_string * PROGRAM = "NManout";

static int help_flag       = 0;
static int verbose_flag    = 0;
static int filter_flag     = 0;
static int update_flag     = 0;
static int only_main_feature_flag = 0;
static int snep_tiling_flag = 0;
static int align_file_flag = 0;

static gdl_string * INPUT   = NULL;
static gdl_string * OUTPUT  = NULL;
static gdl_string * CONFIG  = NULL;
static gdl_string * EXPCMP  = NULL;
static size_t      NCLASS      = 3;
static gdl_string * SEQ_ID     = NULL;
static gdl_string * FEATURE    = NULL;
static long WINDOW = 10000;
static double PVAL_MAX = 0.01;
static double PROB_MIN = 0.25;
static size_t NEIGHBOUR = 2;

static struct option long_options[] =
{
   /* These options set a flag. */
   {"help", no_argument,           &help_flag, 1},
   {"verbose", no_argument,        &verbose_flag, 1},
   {"brief",   no_argument,        &verbose_flag, 0},
   {"main-feature-only",   no_argument,  &only_main_feature_flag, 1},
//   {"snep-tiling",   no_argument,        &snep_tiling_flag, 1},
   {"summary",   no_argument,        &align_file_flag, 1},
   /* These options don't set a flag.
      We distinguish them by their indices. */
   {"input",   required_argument, 0, 'i'},    // the config file
   {"output",  required_argument, 0, 'o'},    // the output stem
   {"config",   required_argument, 0, 'c'},   // the name of the chromosome to focus on
   {"nclass",  required_argument, 0, 'n'},    
   {"seqid", required_argument, 0, 's'},
   {"feature", required_argument, 0, 'f'},
   {"window", required_argument, 0, 'w'},
   {"pv-cutoff", required_argument, 0, 'p'},
   {"prob-min", required_argument, 0, 'm'},
   {"expcmp", required_argument, 0, 'e'},
   {"neighbour", required_argument, 0, 'b'},
   {0, 0, 0, 0}
};
 
 
static int
parse_argument (int argc, char *argv[])
{
	int c;
     
   while (1)
     {
      
       /* getopt_long stores the option index here. */
       int option_index = 0;
 
       c = getopt_long (argc, argv, "i:o:c:n:s:f:w:p:m:e:b:",
                        long_options, &option_index);
 
       /* Detect the end of the options. */
       if (c == -1)
         break;
         
       switch (c)
         {
         case 0:
           /* If this option set a flag, do nothing else now. */
           if (long_options[option_index].flag != 0)
             break;
           printf ("option %s", long_options[option_index].name);
           if (optarg)
             printf (" with arg %s", optarg);
           printf ("\n");
           break;
 
         case 'i':
           INPUT = gdl_string_clone (optarg);
           break;
         case 'o':
           OUTPUT = gdl_string_clone (optarg);
           break;
         case 'c':
           CONFIG = gdl_string_clone (optarg);
           break;
         case 'f':
           FEATURE = gdl_string_clone (optarg);
           break;
         case 'e':
           EXPCMP = gdl_string_clone (optarg);
           break;
         case 'n':
           NCLASS = (size_t) atoi (optarg);
           break;
         case 's':
           SEQ_ID = gdl_string_clone (optarg);
           break;
         case 'w':
           WINDOW = atol (optarg);
           break;
         case 'p':
           PVAL_MAX = (double) atof (optarg);
           break;
         case 'm':
           PROB_MIN = (double) atof (optarg);
           break;
         case 'b':
         	NEIGHBOUR = (size_t) atoi(optarg);
         	break;
         case '?':
           GDL_ERROR_VAL ("Unknown arguments", GDL_EINVAL, -1);
         default:
           GDL_ERROR_VAL ("Bad arguments", GDL_EINVAL, -1);
         }
     }
   /* Print any remaining command line arguments (not options). 
   if (optind < argc)
     {
       printf ("non-option ARGV-elements: ");
       while (optind < argc)
         printf ("%s ", argv[optind++]);
       putchar ('\n');
     }*/
}

static int
check_argument (void)
{
	if (INPUT == 0)
	{
		GDL_ERROR_VAL ("No input directory defined",
		               GDL_FAILURE,
		               1);
	}
	if (CONFIG == 0)
	{
		GDL_ERROR_VAL ("No configuration file defined",
		               GDL_FAILURE,
		               1);
	}
	if (OUTPUT == 0)
	{
		GDL_ERROR_VAL ("No output stem defined",
		               GDL_FAILURE,
		               1);	
	}
	
	return GDL_SUCCESS;
}

static int
help (void)
{
   print_header (PROGRAM);
   printf ("\n");
   printf ("%s summarizes results obtained with NManova\n", PROGRAM);
   printf ("\n");
   printf ("*These options set a flag:\n");
   printf ("\n");
   printf ("--help\tDisplay a brief help on program usage\n");
   printf ("--main-feature-only\tWhen --feature is used, output only parent features\n");
//   printf ("--snep-tiling\tOutput a file reporting SNEP occupancy as a the input tiling data set\n");
   printf ("--summary\tOutput a single file with only the aligned nucleosomes and their statistics\n");
   printf ("\n");
   printf ("*These options require arguments:\n");
   printf ("\n");
   printf ("--input or -i\tThe input directory (use with --print)\n");
   printf ("--config or -c\tThe input configuration file (see manual)\n");
   printf ("--output or -o\tThe name/path to the output directory\n");
   printf ("--nclass or -n\tThe number of nucleosome classes (default is 3)\n");
   printf ("--seqid or -s\tThe name of the chromosome to look at\n");
   printf ("--feature or -f\tA GFF (version 3) annotation file\n");
   printf ("--window or -w\tThe window (bp) around SNEP to look at\n");
   printf ("--pv-cutoff or -p\tThe p-value cut-off for calling SNEP\n");
   printf ("--prob-min or -m\tThe threshold [0,1] to consider aligned nuleosomes as k\n");
   printf ("--expcmp or -e\tA file giving a table of differentially expressed genes and their features\n");
   printf ("--neighbour or -b\tThe number of nucleosomes to look at around each SNEP\n");
}

static int
print_features (FILE * stream,
                gdl_gfeatures * feature,
                gdl_gfeatures_landmark * chrom_feature,
                const long nuc_start,
                const long nuc_end,
                const long window,
                gdl_tiling_gene_expcmp * expcmp)
{
	const long start = nuc_start - window;
	const long end   = nuc_start + window;
	int start_idx, end_idx, j; 
	
	gdl_gfeatures_gff3_within (chrom_feature, start, end, &start_idx, &end_idx);
	
	if (start_idx == -1) return 0;
	for(j = start_idx; j <= end_idx; j++)
	{
		gdl_gfeatures_gff3_record * rec = (gdl_gfeatures_gff3_record *) chrom_feature->records[j];
		if (gdl_hashtable_lookup (rec->attributes, "ID"))
		{
			fprintf (stream, "   %s %s %s", gdl_dictionary_get (feature->dico, "type", rec->type_d), gdl_hashtable_lookup (rec->attributes, "ID"), gdl_hashtable_lookup (rec->attributes, "Alias"));
		}
		else
		{
			if (only_main_feature_flag)
			{
				continue;
			}
			fprintf (stream, "   %s %s %s", gdl_dictionary_get (feature->dico, "type", rec->type_d), gdl_hashtable_lookup (rec->parents[0]->attributes, "ID"), gdl_hashtable_lookup (rec->attributes, "Alias"));
		}
		fprintf (stream, " %d %d %c", rec->start, rec->end, rec->strand);
		long pos2nuc = (rec->start+rec->end)/2-(nuc_start+nuc_end)/2;
		fprintf (stream, " %d", pos2nuc);
		if (rec->strand == '+')
		{
			pos2nuc = rec->start-(nuc_start+nuc_end)/2;
			fprintf (stream, " %d", pos2nuc);
			pos2nuc = rec->end-(nuc_start+nuc_end)/2;
			fprintf (stream, " %d", pos2nuc);
		}
		else
		{
			pos2nuc = rec->end-(nuc_start+nuc_end)/2;
			fprintf (stream, " %d", pos2nuc);
			pos2nuc = rec->start-(nuc_start+nuc_end)/2;
			fprintf (stream, " %d", pos2nuc);
		}
		if ( (rec->start >= nuc_start && rec->end <= nuc_end)
		   ||(nuc_start >= rec->start && nuc_end <= rec->end))
		{
			fprintf (stream, " 1");	
		}
		else
		{
			fprintf (stream, " 0");
		}
		if (expcmp && !strcmp(gdl_dictionary_get (feature->dico, "type", rec->type_d), "gene")) 
		{
			 const gdl_string * gene_id        = gdl_hashtable_lookup (rec->attributes, "ID");
			 gdl_tiling_gene_expcmp_entry * ge = gdl_tiling_gene_expcmp_lookup (expcmp, gene_id);
			 if (ge)
			 {
			 	fprintf (stream, " %g %g %g %d %s", ge->fold, ge->statistic, ge->pvalue, ge->is_significant, ge->comment);
			 }
			 else
			 {
			 	fprintf (stream, " - - - - -");
			 }
		}
		else
		{
			fprintf (stream, " - - - - -");
		}
		fprintf (stream, "\n");
		fflush(stream);
	}
	
	return end_idx - start_idx + 1; 
}    
       
print_context (FILE * stream,
               align_nucleo_alignment * align,
               align_nucleo_nucleosome * nuc,
               const size_t i,
               const unsigned char which,
               const size_t n)
{
	int j, nj;
	double * pr = GDL_MALLOC (double, 4);
	
	nuc->_seen='y';
	if (nuc->align)
	{
		nuc->align->_seen='y';		
	}
	for(nj = 0, j = i-1;j >= 0 && nj < n;j--)
	{
		if (align->ref[j] && align->ref[j]->_seen!='y')
		{
			align_nucleo_nucleosome_state_prob (align->ref[j], align->ref[j]->align, pr);
			if (align->ref[j]->align)
			{
				align->ref[j]->align->_seen='y';
				fprintf (stream, "   %s aligned %d %d %g %g %g\n",(j < i) ? "upstream" : "downstream", align->ref[j]->ref_start_pos, align->ref[j]->ref_end_pos, align->ref[j]->gamma, align->ref[j]->pval, pr[3]);
			}
			else
			{
				fprintf (stream, "   %s insertion %d %d - - %g\n",(j < i) ? "upstream" : "downstream", align->ref[j]->ref_start_pos, align->ref[j]->ref_end_pos, pr[1]);	
			}
			align->ref[j]->_seen='y';
			nj++;
		}
		else if (align->qry[j] && align->qry[j]->_seen!='y')
		{
			align_nucleo_nucleosome_state_prob (align->qry[j], align->qry[j]->align, pr);
			if (align->qry[j]->align)
			{
				align->qry[j]->align->_seen='y';
				fprintf (stream, "   %s aligned %d %d %g %g %g\n",(j < i) ? "upstream" : "downstream", align->qry[j]->ref_start_pos, align->qry[j]->ref_end_pos, align->qry[j]->gamma, align->qry[j]->pval, pr[3]);
			}
			else
			{
				fprintf (stream, "   %s deletion %d %d - - %g\n",(j < i) ? "upstream" : "downstream", align->qry[j]->ref_start_pos, align->qry[j]->ref_end_pos, pr[1]);	
			}
			align->qry[j]->_seen='y';
			nj++;		
		}
	}
	for(nj = 0, j++;nj < n;j++)
	{
		if (align->ref[j])
		{
			align->ref[j]->_seen='n';
			if (align->ref[j]->align) align->ref[j]->align->_seen='n';
		}
		else
		{
			align->qry[j]->_seen='n';
			if (align->qry[j]->align) align->qry[j]->align->_seen='n';
		}
		nj++;
	}
	for(nj = 0, j = i+1;j < align->size && nj < n;j++)
	{
		if (align->ref[j] && align->ref[j]->_seen!='y')
		{
			align_nucleo_nucleosome_state_prob (align->ref[j], align->ref[j]->align, pr);
			if (align->ref[j]->align)
			{
				align->ref[j]->align->_seen='y';
				fprintf (stream, "   %s aligned %d %d %g %g %g\n",(j < i) ? "upstream" : "downstream", align->ref[j]->ref_start_pos, align->ref[j]->ref_end_pos, align->ref[j]->gamma, align->ref[j]->pval, pr[3]);
			}
			else
			{
				fprintf (stream, "   %s insertion %d %d - - %g\n",(j < i) ? "upstream" : "downstream", align->ref[j]->ref_start_pos, align->ref[j]->ref_end_pos, pr[1]);	
			}
			align->ref[j]->_seen='y';
			nj++;
		}
		else if (align->qry[j] && align->qry[j]->_seen!='y')
		{
			align_nucleo_nucleosome_state_prob (align->qry[j], align->qry[j]->align, pr);
			if (align->qry[j]->align)
			{
				align->qry[j]->align->_seen='y';
				fprintf (stream, "   %s aligned %d %d %g %g %g\n",(j < i) ? "upstream" : "downstream", align->qry[j]->ref_start_pos, align->qry[j]->ref_end_pos, align->qry[j]->gamma, align->qry[j]->pval, pr[3]);
			}
			else
			{
				fprintf (stream, "   %s deletion %d %d - - %g\n",(j < i) ? "upstream" : "downstream", align->qry[j]->ref_start_pos, align->qry[j]->ref_end_pos, pr[1]);	
			}
			align->qry[j]->_seen='y';
			nj++;						
		}
	}
	for(nj = 0, j--;nj < n;j--)
	{
		if (align->ref[j])
		{
			align->ref[j]->_seen='n';
			if (align->ref[j]->align) align->ref[j]->align->_seen='n';
		}
		else
		{
			align->qry[j]->_seen='n';
			if (align->qry[j]->align) align->qry[j]->align->_seen='n';
		}
		nj++;
	}
	nuc->_seen='n';
	if (nuc->align)
	{
		nuc->align->_seen='n';		
	}
	
	GDL_FREE (pr);
}

int 
print_seq (FILE * seq_stream,
           gdl_seqdb * seqdb,
           const gdl_string * seq_id,
           const long start,
           const long end,
           const gdl_string * header)
{           
	gdl_seq_io * seq_io = gdl_seq_io_alloc (gdl_seq_io_type_get ("fasta"));
	gdl_seq * seq = gdl_seqdb_id2seq_lookup (seqdb, seq_id);
	if (!seq)
	{
		fprintf(stderr, "[ WARNING ] Sequence ID %s not found in sequence database\n", seq_id);
		return GDL_EINVAL;
	}
	gdl_string * sub_seq;
	if (start < end)
	{
		sub_seq = gdl_seq_extract_string (seq, start-1, end-1);
	}
	else
	{
		sub_seq = gdl_seq_extract_string (seq, end-1, start-1);
		sub_seq = gdl_seq_reverse_str (sub_seq);	
	}
	fprintf(seq_stream, ">%s\n", header);
	gdl_seq_io_string_dump (seq_stream, seq_io, sub_seq);
	fflush(seq_stream);
	gdl_string_free (sub_seq);
	gdl_seq_io_free (seq_io);				
}

int
main (int argc, char *argv[])
{
  int status;
  
  parse_argument (argc, argv);
  
  if (help_flag)
  {
  	  exit (help());
  }
  
  status = check_argument ();
  
  if (status == GDL_SUCCESS)
  {
  		size_t c,w,snep_id=0;
  		int status; 
  		gdl_string * file;
  		FILE * stream, * ins_stream, * del_stream, * nuc_stream, * seq_stream, * tl_stream, * gff_stream;
  		DIR * dir;
  		gdl_runtime * runtime;
  		gdl_tiling_nucleo_hmm * hmm;
  		align_nucleo_config * conf;
  		gdl_gfeatures * feature;
  		gdl_tiling_gene_expcmp * expcmp = 0;
  		
  		runtime = runtime_setup();
  	
  	   conf = align_nucleo_config_alloc (CONFIG);
		conf->hmm_nclass = NCLASS;
		
		if (align_file_flag)
		{	
			//Loop on chromosomes
			for(c = 0; c < conf->ref_data->nchrom; c++)
			{
				if (SEQ_ID && strcmp(conf->ref_data->chroms[c]->name, SEQ_ID))
				{
					continue;
				}
				file = gdl_string_sprintf ("%s/%s.nal", INPUT, conf->ref_data->chroms[c]->name);
			 	if (gdl_isfile (file))
			 	{ 
				   align_nucleo_alignment * align;
				   gdl_gfeatures_landmark * chrom_feature;
				   stream = gdl_fileopen (file, "r");
				 	align = align_nucleo_alignment_fread (stream, conf);
				 	gdl_fileclose (file, stream);
			 	 	gdl_string_free (file);
			 	 	if (!align)
				 	{
				 		GDL_ERROR_VAL ("Unable to read nucleosome alignment", GDL_FAILURE, GDL_FAILURE); 
				 	}
				 	size_t i,ref_chunk,qry_chunk,out_chunk=0,old_chunk=0;
				 	align_nucleo_nucleosome * nucleo;
				 	for(i = 0; i < align->size; i++)
					{
						nucleo = align->ref[i];
						if (!nucleo || !nucleo->align)
							continue;
						if (i==0)
						{
							ref_chunk=nucleo->chunk_idx;
							qry_chunk=nucleo->align->chunk_idx;
						}
						else if (ref_chunk!=nucleo->chunk_idx)
						{
							ref_chunk=nucleo->chunk_idx;
							qry_chunk=nucleo->align->chunk_idx;
							out_chunk++;
						}
						else if (qry_chunk!=nucleo->align->chunk_idx)
						{
							ref_chunk=nucleo->chunk_idx;
							qry_chunk=nucleo->align->chunk_idx;
							out_chunk++;
						}
						old_chunk = out_chunk;
						printf("%s %d %d %d %d %d",conf->ref_data->chroms[c]->name,out_chunk,nucleo->ref_start_pos,nucleo->ref_end_pos,nucleo->align->ref_start_pos,nucleo->align->ref_end_pos);
						if (nucleo->ignore=='n' || nucleo->pval!=-1)
							printf(" %g", nucleo->pval);
						else
							printf(" -1");
						if (nucleo->gamma > 0)
							printf(" 0");
						else
							printf(" 1");
						printf(" %g", nucleo->gamma);
						printf(" %d %d\n", nucleo->status, nucleo->align->status);
					}
					align_nucleo_alignment_free (align);
			 	}
			}
			exit(0);
		}			
		
		// Read the feature
		feature = gdl_gfeatures_alloc (gdl_gfeatures_gff3, FEATURE);
		
		if (EXPCMP)
		{
			expcmp = gdl_tiling_gene_expcmp_alloc (EXPCMP);
		}
		
		file = gdl_string_sprintf ("%s_snep_table.txt", OUTPUT);
		nuc_stream = gdl_fileopen (file, "w");
		gdl_string_free (file);
		file = gdl_string_sprintf ("%s_snep.gff", OUTPUT);
		gff_stream = gdl_fileopen (file, "w");
		fprintf (gff_stream, "##gff-version 3\n");
		fprintf (gff_stream, "#\n");
		fprintf (gff_stream, "# This file has been generated by NManout\n");
		fprintf (gff_stream, "# p-value cut-off = %e\n", PVAL_MAX);
		fprintf (gff_stream, "#\n");
		gdl_string_free (file);
//		file = gdl_string_sprintf ("%s_ins.txt", OUTPUT);
//		ins_stream = gdl_fileopen (file, "w");
//		gdl_string_free (file);
//		file = gdl_string_sprintf ("%s_del.txt", OUTPUT);
//		del_stream = gdl_fileopen (file, "w");
//		gdl_string_free (file);
		if (conf->ref_seq || conf->qry_seq)
		{
			file = gdl_string_sprintf ("%s_snep_seq.txt", OUTPUT);
			seq_stream = gdl_fileopen (file, "w");
			gdl_string_free (file);
		}
		//if (snep_tiling_flag)
		//{
			file = gdl_string_sprintf ("%s_snep_tiling.txt", OUTPUT);
			tl_stream = gdl_fileopen (file, "w");
			gdl_string_free (file);
		//}
		// Loop on chromosomes
		for(c = 0; c < conf->ref_data->nchrom; c++)
		{
			if (SEQ_ID && strcmp(conf->ref_data->chroms[c]->name, SEQ_ID))
			{
				continue;
			}
			file = gdl_string_sprintf ("%s/%s.nal", INPUT, conf->ref_data->chroms[c]->name);
		 	if (gdl_isfile (file))
		 	{ 
			   align_nucleo_alignment * align;
			   gdl_gfeatures_landmark * chrom_feature;
			   
			   fprintf (stdout, "Process chromosome %s: wait...", conf->ref_data->chroms[c]->name);
		 		fflush(stdout);
			 	
			 	stream = gdl_fileopen (file, "r");
			 	align = align_nucleo_alignment_fread (stream, conf);
			 	gdl_fileclose (file, stream);
		 	 	gdl_string_free (file);
		 	 	
		 	 	if (!align)
			 	{
			 		GDL_ERROR_VAL ("Unable to read nucleosome alignment", GDL_FAILURE, GDL_FAILURE); 
			 	}
			 	
			 	chrom_feature = gdl_gfeatures_lookup (feature, conf->ref_data->chroms[c]->name);
			 	
			 	if (!chrom_feature)
			 	{
			 		GDL_ERROR_VAL (gdl_string_sprintf ("Unable to get features for chromosome %s", conf->ref_data->chroms[c]->name), GDL_FAILURE, GDL_FAILURE); 
			 	}
			 	
			 	size_t i, * idx = GDL_MALLOC (size_t, 4);
				double sd, * pr = GDL_MALLOC (double, 4);
				int j, start_idx, end_idx;
				long start, end;
				align_nucleo_nucleosome * ref_nucleo, * qry_nucleo;
				align_nucleo_nucleosome * nuc_left, * nuc_right;
				
				for(i = 0; i < align->size; i++)
				{
					if (align->ref[i]) align->ref[i]->_seen='n';
					if (align->qry[i]) align->qry[i]->_seen='n';
				}
				
				for(i = 0; i < align->size; i++)
				{
					if (align->ref[i] && align->ref[i]->seen != 'y' && align->ref[i]->ignore != 'y')
					{
						ref_nucleo = align->ref[i];
						ref_nucleo->seen = 'y';
						align_nucleo_nucleosome_state_prob (ref_nucleo, ref_nucleo->align, pr);
						gdl_sort_index (idx, pr, 1, 4);
						//fprintf (stream, "%s %d %c %d %d %d %d %g %g %g", align->ref_chrom->data->name, ref_nucleo->chunk_idx, ref_nucleo->ignore, ref_nucleo->outlier, ref_nucleo->ref_start_pos, ref_nucleo->ref_end_pos, ref_nucleo->status, ref_nucleo->delta, ref_nucleo->indel, 1.0-ref_nucleo->prob[0]);
						if (ref_nucleo->align)
						{
							//if (idx[3] != 3)
							//	continue;
							qry_nucleo = ref_nucleo->align;
							qry_nucleo->seen = 'y';
							if (ref_nucleo->pval != -1.0 && ref_nucleo->pval <= PVAL_MAX)
							{
								start = GDL_MIN (ref_nucleo->ref_start_pos, qry_nucleo->ref_start_pos);
								end   = GDL_MAX (ref_nucleo->ref_end_pos, qry_nucleo->ref_end_pos);
								fprintf (nuc_stream, ">%s SNEP_%d %d %d %g %g %g %g\n", conf->ref_data->chroms[c]->name, snep_id, start, end, ref_nucleo->gamma, ref_nucleo->pval, pr[3], ref_nucleo->delta);
								fprintf (gff_stream, "%s\tNucleoMiner\tSNEP\t%d\t%d\t%g\t+\t.\tID=SNEP_%d;p_value=%e;align_prob=%e;align_delta=%e\n", conf->ref_data->chroms[c]->name, start, end, ref_nucleo->gamma, snep_id, ref_nucleo->pval, pr[3], ref_nucleo->delta);
								fflush (gff_stream);
								fprintf (nuc_stream, "-- Features\n");
								print_features (nuc_stream, feature, chrom_feature, start, end, WINDOW, expcmp);
								if (NEIGHBOUR)
								{
									fprintf (nuc_stream, "-- Context\n");
									print_context (nuc_stream,  align, ref_nucleo, i, 'r', NEIGHBOUR);
								}
								if (conf->ref_seq)
								{
									gdl_string * header = gdl_string_sprintf("SNEP_%d |REF|ID=%s|start=%d|end=%d|",snep_id,conf->ref_data->chroms[c]->name, ref_nucleo->ref_start_pos, ref_nucleo->ref_end_pos);
									print_seq (seq_stream, conf->ref_seq, conf->ref_data->chroms[c]->name, ref_nucleo->ref_start_pos, ref_nucleo->ref_end_pos, header);
									gdl_string_free (header);
								}
								if (conf->qry_seq)
								{
									const long qry_start = qry_nucleo->chrom->hmm->chunks[qry_nucleo->chunk_idx]->probes[qry_nucleo->start_idx]->position;
									const long qry_end   = qry_nucleo->chrom->hmm->chunks[qry_nucleo->chunk_idx]->probes[qry_nucleo->end_idx]->position;
									const gdl_string * qry_seqid = qry_nucleo->chrom->data->name;
									gdl_string * header = gdl_string_sprintf("SNEP_%d |QRY|ID=%s|start=%d|end=%d|",snep_id, qry_seqid, qry_start, qry_end);
									print_seq (seq_stream, conf->qry_seq, qry_seqid, qry_start, qry_end, header);
									gdl_string_free (header);
								}
								snep_id++;
							}
							//if (snep_tiling_flag)
							//{
								const gdl_tiling_nucleo_chunk * ref_chunk = ref_nucleo->chrom->hmm->chunks[ref_nucleo->chunk_idx];
								for(j = ref_nucleo->start_idx; j <= ref_nucleo->end_idx; j++)
								{
									double tl_pr = 1.0/(ref_nucleo->end_idx-ref_nucleo->start_idx+1);
									fprintf (tl_stream, "%s %d %g", conf->ref_data->chroms[c]->name, ref_chunk->probes[j]->position, tl_pr);
									if (ref_nucleo->pval != -1.0 && ref_nucleo->pval <= PVAL_MAX)
									{
										if (ref_nucleo->gamma > 0)
										{
											fprintf (tl_stream, " 0 %g\n", tl_pr);
										}
										else
										{
											fprintf (tl_stream, " %g 0\n", tl_pr);
										}
									}
									else
									{
										fprintf (tl_stream, " 0 0\n");
									}
								}
								fflush(tl_stream);
							//}
						}
						else if (pr[1] >= PROB_MIN && idx[3]==1)
						{
							// compute insertion % size
							double score = 0;
							align_nucleo_alignment_ref_flanking_qry (align, i, &nuc_left, &nuc_right);
							if (nuc_left->ref_end_pos > ref_nucleo->ref_start_pos)
							{
								if (nuc_left->ref_start_pos >= ref_nucleo->ref_start_pos)
								{
									score += ((double)(GDL_MIN(nuc_left->ref_end_pos, ref_nucleo->ref_end_pos)-nuc_left->ref_start_pos))/(ref_nucleo->ref_end_pos-ref_nucleo->ref_start_pos);
								}
								else
								{
									score += ((double)(GDL_MIN(nuc_left->ref_end_pos, ref_nucleo->ref_end_pos)-ref_nucleo->ref_start_pos))/(ref_nucleo->ref_end_pos-ref_nucleo->ref_start_pos);
								}
							}
							if (nuc_right->ref_start_pos < ref_nucleo->ref_end_pos)
							{
								if (nuc_right->ref_end_pos <= ref_nucleo->ref_end_pos)
								{
									score += ((double)(nuc_right->ref_end_pos-GDL_MAX(nuc_right->ref_start_pos, ref_nucleo->ref_start_pos)))/(ref_nucleo->ref_end_pos-ref_nucleo->ref_start_pos);
								}
								else
								{
									score += ((double)(ref_nucleo->ref_end_pos-GDL_MAX(nuc_right->ref_start_pos, ref_nucleo->ref_start_pos)))/(ref_nucleo->ref_end_pos-ref_nucleo->ref_start_pos);
								}	
							}
							//fprintf (ins_stream, ">%s INS %d %d %g %g %g %g\n", conf->ref_data->chroms[c]->name, ref_nucleo->ref_start_pos, ref_nucleo->ref_end_pos, ref_nucleo->delta, ref_nucleo->indel, score, pr[1]);
							//print_features (ins_stream, feature, chrom_feature, ref_nucleo->ref_start_pos, ref_nucleo->ref_end_pos, WINDOW, expcmp);
						}						
					}
					else if (align->qry[i] && align->qry[i]->seen != 'y' && align->qry[i]->ignore != 'y')
					{
						qry_nucleo = align->qry[i];
						qry_nucleo->seen = 'y';
						align_nucleo_nucleosome_state_prob (qry_nucleo->align, qry_nucleo, pr);
						gdl_sort_index (idx, pr, 1, 4);
						if (qry_nucleo->align)
						{
							//if (idx[3] != 3)
							//	continue;
							ref_nucleo = qry_nucleo->align;
							ref_nucleo->seen = 'y';
							if (ref_nucleo->pval != -1.0 && ref_nucleo->pval <= PVAL_MAX)
							{
								start = GDL_MIN (ref_nucleo->ref_start_pos, qry_nucleo->ref_start_pos);
								end   = GDL_MAX (ref_nucleo->ref_end_pos, qry_nucleo->ref_end_pos);
								fprintf (nuc_stream, ">%s SNEP_%d %d %d %g %g %g %g\n", conf->ref_data->chroms[c]->name, snep_id, start, end, qry_nucleo->gamma, qry_nucleo->pval, pr[3], ref_nucleo->delta);
								fprintf (gff_stream, "%s\tNucleoMiner\tSNEP\t%d\t%d\t%g\t+\t.\tID=SNEP_%d;p_value=%e;align_prob=%e;align_delta=%e\n", conf->ref_data->chroms[c]->name, start, end, ref_nucleo->gamma, snep_id, ref_nucleo->pval, pr[3], ref_nucleo->delta);
								fflush (gff_stream);
								fprintf (nuc_stream, "-- Features\n");
								print_features (nuc_stream, feature, chrom_feature, start, end, WINDOW, expcmp);
								if (NEIGHBOUR)
								{
									fprintf (nuc_stream, "-- Context\n");
									print_context (nuc_stream, align, qry_nucleo, i, 'q', NEIGHBOUR);
								}
								if (conf->ref_seq)
								{
									gdl_string * header = gdl_string_sprintf("SNEP_%d |REF|ID=%s|start=%d|end=%d|",snep_id,conf->ref_data->chroms[c]->name, ref_nucleo->ref_start_pos, ref_nucleo->ref_end_pos);
									print_seq (seq_stream, conf->ref_seq, conf->ref_data->chroms[c]->name, ref_nucleo->ref_start_pos, ref_nucleo->ref_end_pos, header);
									gdl_string_free (header);
								}
								if (conf->qry_seq)
								{
									const long qry_start = qry_nucleo->chrom->hmm->chunks[qry_nucleo->chunk_idx]->probes[qry_nucleo->start_idx]->position;
									const long qry_end   = qry_nucleo->chrom->hmm->chunks[qry_nucleo->chunk_idx]->probes[qry_nucleo->end_idx]->position;
									const gdl_string * qry_seqid = qry_nucleo->chrom->data->name;
									gdl_string * header = gdl_string_sprintf("SNEP_%d |QRY|ID=%s|start=%d|end=%d|",snep_id, qry_seqid, qry_start, qry_end);
									print_seq (seq_stream, conf->qry_seq, qry_seqid, qry_start, qry_end, header);
									gdl_string_free (header);
								}
								snep_id++;
							}
							//if (snep_tiling_flag)
							//{
								const gdl_tiling_nucleo_chunk * ref_chunk = ref_nucleo->chrom->hmm->chunks[ref_nucleo->chunk_idx];
								for(j = ref_nucleo->start_idx; j <= ref_nucleo->end_idx; j++)
								{
									double tl_pr = 1.0/(ref_nucleo->end_idx-ref_nucleo->start_idx+1);
									fprintf (tl_stream, "%s %d %g", conf->ref_data->chroms[c]->name, ref_chunk->probes[j]->position, tl_pr);
									if (ref_nucleo->pval != -1.0 && ref_nucleo->pval <= PVAL_MAX)
									{
										if (ref_nucleo->gamma > 0)
										{
											fprintf (tl_stream, " 0 %g\n", tl_pr);
										}
										else
										{
											fprintf (tl_stream, " %g 0\n", tl_pr);
										}
									}
									else
									{
										fprintf (tl_stream, " 0 0\n");
									}
								}
								fflush(tl_stream);
							//}
						}
						else if (pr[2] >= PROB_MIN && idx[3]==2)
						{
							// compute insertion % size
							double score = 0;
							align_nucleo_alignment_qry_flanking_ref (align, i, &nuc_left, &nuc_right);
							if (nuc_left->ref_end_pos > qry_nucleo->ref_start_pos)
							{
								if (nuc_left->ref_start_pos >= qry_nucleo->ref_start_pos)
								{
									score += ((double)(GDL_MIN(nuc_left->ref_end_pos, qry_nucleo->ref_end_pos)-nuc_left->ref_start_pos))/(qry_nucleo->ref_end_pos-qry_nucleo->ref_start_pos);
								}
								else
								{
									score += ((double)(GDL_MIN(nuc_left->ref_end_pos, qry_nucleo->ref_end_pos)-qry_nucleo->ref_start_pos))/(qry_nucleo->ref_end_pos-qry_nucleo->ref_start_pos);
								}
							}
							if (nuc_right->ref_start_pos < qry_nucleo->ref_end_pos)
							{
								if (nuc_right->ref_end_pos <= qry_nucleo->ref_end_pos)
								{
									score += ((double)(nuc_right->ref_end_pos-GDL_MAX(nuc_right->ref_start_pos, qry_nucleo->ref_start_pos)))/(qry_nucleo->ref_end_pos-qry_nucleo->ref_start_pos);
								}
								else
								{
									score += ((double)(qry_nucleo->ref_end_pos-GDL_MAX(nuc_right->ref_start_pos, qry_nucleo->ref_start_pos)))/(qry_nucleo->ref_end_pos-qry_nucleo->ref_start_pos);
								}	
							}
							//fprintf (del_stream, ">%s DEL %d %d %g %g %g %g\n", conf->ref_data->chroms[c]->name, qry_nucleo->ref_start_pos, qry_nucleo->ref_end_pos, qry_nucleo->delta, qry_nucleo->indel, score, pr[2]);
							//print_features (del_stream, feature, chrom_feature, qry_nucleo->ref_start_pos, qry_nucleo->ref_end_pos, WINDOW, expcmp);
						}
					}
				}
				GDL_FREE (pr);
			 	
			 	align_nucleo_alignment_free (align);
		 		fprintf (stdout, "\b\b\b\b\b\b\b[ OK ]\n");
		 		fflush(stdout);	 	
		 	}	 	
		 	
		}
		
		file = gdl_string_sprintf ("%s_snep_table.txt", OUTPUT);
		gdl_fileclose (file, nuc_stream);
		gdl_string_free (file);
		file = gdl_string_sprintf ("%s_snep.gff", OUTPUT);
		gdl_fileclose (file, gff_stream);
		gdl_string_free (file);
//		file = gdl_string_sprintf ("%s_ins.txt", OUTPUT);
//		gdl_fileclose (file, ins_stream);
//		gdl_string_free (file);
//		file = gdl_string_sprintf ("%s_del.txt", OUTPUT);
//		gdl_fileclose (file, del_stream);
//		gdl_string_free (file);
		if (conf->ref_seq || conf->qry_seq)
		{
			file = gdl_string_sprintf ("%s_snep_seq.txt", OUTPUT);
			gdl_fileclose (file, seq_stream);
			gdl_string_free (file);
		}
//		if (snep_tiling_flag)
//		{
		file = gdl_string_sprintf ("%s_snep_tiling.txt", OUTPUT);
		gdl_fileclose (file, tl_stream);
		gdl_string_free (file);
//		}
		
		gdl_gfeatures_free (feature);
		runtime_setdown (runtime);
  }
  
  exit(0);
}

