/*  
 *  NManout2.c
 * 
 *  $Author: veyrieras $, $Date: 2008-11-14 10:45:04 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_list.h>
#include <gdl/gdl_machine.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_sort_long.h>
#include <gdl/gdl_runtime.h>
#include <gdl/gdl_vector.h>
#include <gdl/gdl_matrix.h>
#include <gdl/gdl_multireg.h>
#include <gdl/gdl_genome_coord.h>
#include <gdl/gdl_gfeatures.h>
#include <gdl/gdl_gfeatures_gff3.h>

#include "nucleoMiner.h"

#include "const.h"
#include "header.c"
#include "util.c"

#include "tiling_array.h"
#include "tiling_hmm.h"
#include "tiling_nucleo.h"
#include "tiling_feature.h"
#include "align_nucleo2.h"

static gdl_string * PROGRAM = "NManout2";

static int help_flag       = 0;
static int verbose_flag    = 0;
static int filter_flag     = 0;
static int update_flag     = 0;
static int only_main_feature_flag = 0;
static int snep_tiling_flag = 0;
static int align_file_flag = 0;

static gdl_string * INPUT   = NULL;
static gdl_string * OUTPUT  = NULL;
static gdl_string * DATA    = NULL;
static gdl_string * EXPCMP  = NULL;
static gdl_string * REF_SEQ = NULL;
static gdl_string * QRY_SEQ = NULL;

static gdl_string * SEQ_ID     = NULL;
static gdl_string * FEATURE    = NULL;
static long WINDOW             = 2500;
static double PVAL_MAX         = 0.01;
static size_t NEIGHBOUR        = 2;
static size_t TILING           = 4;
static size_t PROBE_MIN        = 10;

static struct option long_options[] =
{
   /* These options set a flag. */
   {"help", no_argument,           &help_flag, 1},
   {"verbose", no_argument,        &verbose_flag, 1},
   {"brief",   no_argument,        &verbose_flag, 0},
   {"main-feature-only",   no_argument,  &only_main_feature_flag, 1},
//   {"snep-tiling",   no_argument,        &snep_tiling_flag, 1},
   {"summary",   no_argument,        &align_file_flag, 1},
   /* These options don't set a flag.
      We distinguish them by their indices. */
   {"input",   required_argument, 0, 'i'},    
   {"output",  required_argument, 0, 'o'},   
   {"data",   required_argument, 0, 'd'},   
   {"seqid", required_argument, 0, 's'},
   {"feature", required_argument, 0, 'f'},
   {"window", required_argument, 0, 'w'},
   {"pv-cutoff", required_argument, 0, 'p'},
   {"expcmp", required_argument, 0, 'e'},
   {"neighbour", required_argument, 0, 'b'},
   {"ref-seq", required_argument, 0, 'r'},
   {"qry-seq", required_argument, 0, 'q'},
   {"tiling", required_argument, 0, 't'},
   {"probe-min", required_argument, 0, 'm'},
   {0, 0, 0, 0}
};
 
 
static int
parse_argument (int argc, char *argv[])
{
	int c;
     
   while (1)
     {
      
       /* getopt_long stores the option index here. */
       int option_index = 0;
 
       c = getopt_long (argc, argv, "i:o:d:s:f:w:p:e:b:r:q:t:m:",
                        long_options, &option_index);
 
       /* Detect the end of the options. */
       if (c == -1)
         break;
         
       switch (c)
         {
         case 0:
           /* If this option set a flag, do nothing else now. */
           if (long_options[option_index].flag != 0)
             break;
           printf ("option %s", long_options[option_index].name);
           if (optarg)
             printf (" with arg %s", optarg);
           printf ("\n");
           break;
 
         case 'i':
           INPUT = gdl_string_clone (optarg);
           break;
         case 'o':
           OUTPUT = gdl_string_clone (optarg);
           break;
         case 'd':
           DATA = gdl_string_clone (optarg);
           break;
         case 'f':
           FEATURE = gdl_string_clone (optarg);
           break;
         case 'e':
           EXPCMP = gdl_string_clone (optarg);
           break;
         case 's':
           SEQ_ID = gdl_string_clone (optarg);
           break;
         case 'w':
           WINDOW = atol (optarg);
           break;
         case 'p':
           PVAL_MAX = (double) atof (optarg);
           break;
         case 'b':
         	NEIGHBOUR = (size_t) atoi(optarg);
         	break;
         case 'r':
           REF_SEQ = gdl_string_clone (optarg);
           break;
         case 'q':
           QRY_SEQ = gdl_string_clone (optarg);
           break;
         case 't':
           TILING = (size_t) atoi (optarg);
           break;
         case 'm':
           PROBE_MIN = (size_t) atoi (optarg);
           break;  
         case '?':
           GDL_ERROR_VAL ("Unknown arguments", GDL_EINVAL, -1);
         default:
           GDL_ERROR_VAL ("Bad arguments", GDL_EINVAL, -1);
         }
     }
   /* Print any remaining command line arguments (not options). 
   if (optind < argc)
     {
       printf ("non-option ARGV-elements: ");
       while (optind < argc)
         printf ("%s ", argv[optind++]);
       putchar ('\n');
     }*/
}

static int
check_argument (void)
{
	if (INPUT == 0)
	{
		GDL_ERROR_VAL ("No input directory defined",
		               GDL_FAILURE,
		               1);
	}
	if (DATA == 0)
	{
		GDL_ERROR_VAL ("No data file defined",
		               GDL_FAILURE,
		               1);
	}
	if (OUTPUT == 0)
	{
		GDL_ERROR_VAL ("No output stem defined",
		               GDL_FAILURE,
		               1);	
	}
	if ((REF_SEQ && !QRY_SEQ) || (!REF_SEQ && QRY_SEQ))
	{
		GDL_ERROR_VAL ("You must provide query and reference sequence file together",
		               GDL_FAILURE,
		               1);
	}
	
	return GDL_SUCCESS;
}

static int
help (void)
{
   print_header (PROGRAM);
   printf ("\n");
   printf ("%s summarizes results obtained with NManova\n", PROGRAM);
   printf ("\n");
   printf ("*These options set a flag:\n");
   printf ("\n");
   printf ("--help\tDisplay a brief help on program usage\n");
   printf ("--main-feature-only\tWhen --feature is used, output only parent features\n");
//   printf ("--snep-tiling\tOutput a file reporting SNEP occupancy as a the input tiling data set\n");
   printf ("--summary\tOutput a single file with only the aligned nucleosomes and their statistics\n");
   printf ("\n");
   printf ("*These options require arguments:\n");
   printf ("\n");
   printf ("--input or -i\tThe input directory (use with --print)\n");
   printf ("--data or -d\tThe data file\n");
   printf ("--output or -o\tThe name/path to the output directory\n");
   printf ("--nclass or -n\tThe number of nucleosome classes (default is 3)\n");
   printf ("--seqid or -s\tThe name of the chromosome to look at\n");
   printf ("--feature or -f\tA GFF (version 3) annotation file\n");
   printf ("--window or -w\tThe window (bp) around SNEP to look at\n");
   printf ("--pv-cutoff or -p\tThe p-value cut-off for calling SNEP\n");
   printf ("--prob-min or -m\tThe threshold [0,1] to consider aligned nuleosomes as k\n");
   printf ("--expcmp or -e\tA file giving a table of differentially expressed genes and their features\n");
   printf ("--neighbour or -b\tThe number of nucleosomes to look at around each SNEP\n");
   printf ("--ref-seq or -r\tThe reference sequence file\n");
   printf ("--qry-seq or -r\tThe query sequence file\n");
   printf ("--tiling or -t\tThe tiling step (default is 4bp)\n");
   printf ("--probe-min or -m\tMinimum number of probes within the overlap (default is 10)\n");
}

static int
print_features (FILE * stream,
                gdl_gfeatures * feature,
                gdl_gfeatures_landmark * chrom_feature,
                const long nuc_start,
                const long nuc_end,
                const long window,
                gdl_tiling_gene_expcmp * expcmp)
{
	const long start = nuc_start - window;
	const long end   = nuc_start + window;
	int start_idx, end_idx, j; 
	
	gdl_gfeatures_gff3_within (chrom_feature, start, end, &start_idx, &end_idx);
	
	if (start_idx == -1) return 0;
	for(j = start_idx; j <= end_idx; j++)
	{
		gdl_gfeatures_gff3_record * rec = (gdl_gfeatures_gff3_record *) chrom_feature->records[j];
		if (gdl_hashtable_lookup (rec->attributes, "ID"))
		{
			fprintf (stream, "   %s %s %s", gdl_dictionary_get (feature->dico, "type", rec->type_d), gdl_hashtable_lookup (rec->attributes, "ID"), gdl_hashtable_lookup (rec->attributes, "Alias"));
		}
		else
		{
			if (only_main_feature_flag)
			{
				continue;
			}
			fprintf (stream, "   %s %s %s", gdl_dictionary_get (feature->dico, "type", rec->type_d), gdl_hashtable_lookup (rec->parents[0]->attributes, "ID"), gdl_hashtable_lookup (rec->attributes, "Alias"));
		}
		fprintf (stream, " %d %d %c", rec->start, rec->end, rec->strand);
		long pos2nuc = (rec->start+rec->end)/2-(nuc_start+nuc_end)/2;
		fprintf (stream, " %d", pos2nuc);
		if (rec->strand == '+')
		{
			pos2nuc = rec->start-(nuc_start+nuc_end)/2;
			fprintf (stream, " %d", pos2nuc);
			pos2nuc = rec->end-(nuc_start+nuc_end)/2;
			fprintf (stream, " %d", pos2nuc);
		}
		else
		{
			pos2nuc = rec->end-(nuc_start+nuc_end)/2;
			fprintf (stream, " %d", pos2nuc);
			pos2nuc = rec->start-(nuc_start+nuc_end)/2;
			fprintf (stream, " %d", pos2nuc);
		}
		if ( (rec->start >= nuc_start && rec->end <= nuc_end)
		   ||(nuc_start >= rec->start && nuc_end <= rec->end))
		{
			fprintf (stream, " 1");	
		}
		else
		{
			fprintf (stream, " 0");
		}
		if (expcmp && !strcmp(gdl_dictionary_get (feature->dico, "type", rec->type_d), "gene")) 
		{
			 const gdl_string * gene_id        = gdl_hashtable_lookup (rec->attributes, "ID");
			 gdl_tiling_gene_expcmp_entry * ge = gdl_tiling_gene_expcmp_lookup (expcmp, gene_id);
			 if (ge)
			 {
			 	fprintf (stream, " %g %g %g %d %s", ge->fold, ge->statistic, ge->pvalue, ge->is_significant, ge->comment);
			 }
			 else
			 {
			 	fprintf (stream, " - - - - -");
			 }
		}
		else
		{
			fprintf (stream, " - - - - -");
		}
		fprintf (stream, "\n");
		fflush(stream);
	}
	
	return end_idx - start_idx + 1; 
}    
       
print_context (FILE * stream,
               nm_align_nucleo_chromosome * align,
               nm_align_nucleo_nucleosome * nuc,
               const size_t i,
               const size_t n)
{
	int j, nj;
	
	for(nj = 0, j = i-1;j >= 0 && nj < n;j--,nj++)
	{
		fprintf (stream, "   downstream %d %d %d %g %g\n", (align->nucleosomes[j]->align!=0), align->nucleosomes[j]->ref_start_pos, align->nucleosomes[j]->ref_end_pos, align->nucleosomes[j]->gamma, align->nucleosomes[j]->pval);
	}
	for(nj = 0, j = i+1;j < align->n_nucleosome && nj < n;j++,nj++)
	{
		fprintf (stream, "   upstream %d %d %d %g %g\n", (align->nucleosomes[j]->align!=0), align->nucleosomes[j]->ref_start_pos, align->nucleosomes[j]->ref_end_pos, align->nucleosomes[j]->gamma, align->nucleosomes[j]->pval);
	}
}

int 
print_seq (FILE * seq_stream,
           gdl_seqdb * seqdb,
           const gdl_string * seq_id,
           const long start,
           const long end,
           const gdl_string * header)
{           
	gdl_seq_io * seq_io = gdl_seq_io_alloc (gdl_seq_io_type_get ("fasta"));
	gdl_seq * seq = gdl_seqdb_id2seq_lookup (seqdb, seq_id);
	if (!seq)
	{
		fprintf(stderr, "[ WARNING ] Sequence ID %s not found in sequence database\n", seq_id);
		return GDL_EINVAL;
	}
	gdl_string * sub_seq;
	if (start < end)
	{
		sub_seq = gdl_seq_extract_string (seq, start-1, end-1);
	}
	else
	{
		sub_seq = gdl_seq_extract_string (seq, end-1, start-1);
		sub_seq = gdl_seq_reverse_str (sub_seq);	
	}
	fprintf(seq_stream, ">%s\n", header);
	gdl_seq_io_string_dump (seq_stream, seq_io, sub_seq);
	fflush(seq_stream);
	gdl_string_free (sub_seq);
	gdl_seq_io_free (seq_io);				
}

static int
txt_output (FILE * stream, const gdl_tiling_data * data, const size_t chrom_idx, const nm_align_nucleo_chromosome * align, const size_t i)
{
	nm_align_nucleo_nucleosome * ref_nuc = align->nucleosomes[i];
	if (ref_nuc->align != 0)
	{
		size_t j;
		nm_align_nucleo_nucleosome * qry_nuc = align->nucleosomes[i]->align;
		// Get the boundaries of the aligned region
		long nuc_start = GDL_MAX (ref_nuc->ref_start_pos, qry_nuc->ref_start_pos);
		long nuc_end   = GDL_MIN (ref_nuc->ref_end_pos, qry_nuc->ref_end_pos);
		int nprobe = 0, probe_start_idx = -1, probe_end_idx = -1;
		for(j = 0; j < data->chroms[chrom_idx]->nprobe; j++)
		{
			if (data->chroms[chrom_idx]->positions[j] >= nuc_start && data->chroms[chrom_idx]->positions[j] <= nuc_end)
			{
				if (!nprobe) 
				{
					probe_start_idx = j;
				}
				nprobe++;
			}
			else if (nprobe && j > 0)
			{
				probe_end_idx = j - 1;
				break;
			}
		}
		if ((probe_start_idx != -1 && probe_end_idx != -1) && nprobe > PROBE_MIN)
		{
			size_t a, g, p, na;
			double * x  = GDL_CALLOC (double, 4);
			size_t * nx = GDL_CALLOC (size_t, 4); 
			
			fprintf (stream, "%s %d %d", data->chroms[chrom_idx]->name, nuc_start, nuc_end);
			
			for(p = probe_start_idx; p <= probe_end_idx; p++)
			{
				for(g = 0; g < 2; g++) // loop on manip for strain A
				{
					na = data->narrayByGroup[g]; 
					for(a = 0; a < na; a++) // loop on replicates
					{
						x[g] += gdl_tiling_data_get_val (data, chrom_idx, p, g, a);
						nx[g]++;
					}
				}
				for(g = 2; g < 4; g++) // loop on manip for strain B
				{
					na = data->narrayByGroup[g];
					for(a = 0; a < na; a++) // loop on replicates
					{
						x[g] += gdl_tiling_data_get_val (data, chrom_idx, p, g, a);
						nx[g]++;
					}
				}
			}
			
			fprintf (stream, " %d %d %d %g %g %d %d %d %g %g", ref_nuc->chunk_idx, ref_nuc->ref_start_pos, ref_nuc->ref_end_pos, x[0]/nx[0], x[1]/nx[1], qry_nuc->chunk_idx, qry_nuc->ref_start_pos, qry_nuc->ref_end_pos, x[2]/nx[2], x[3]/nx[3]);
			fprintf (stream, " %g %e\n", ref_nuc->gamma, ref_nuc->pval);
			fflush (stream); 
			
			GDL_FREE (x);
			GDL_FREE (nx);
		}
	}
}

int
main (int argc, char *argv[])
{
  int status;
  
  parse_argument (argc, argv);
  
  if (help_flag)
  {
  	  exit (help());
  }
  
  status = check_argument ();
  
  if (status == GDL_SUCCESS)
  {
  		size_t c,w,snep_id=0;
  		int status; 
  		gdl_string * file;
  		FILE * stream, * ins_stream, * del_stream, * nuc_stream, * seq_stream, * tl_stream, * gff_stream;
  		DIR * dir;
  		gdl_runtime * runtime;
  		gdl_gfeatures * feature;
  		gdl_tiling_gene_expcmp * expcmp = 0;
  		gdl_seqdb * ref_seq = 0, * qry_seq = 0;
  		
  		runtime = runtime_setup();
  		
  		// read the data
  		stream = gdl_fileopen (DATA, "r");
		gdl_tiling_data * data = gdl_tiling_data_fread (stream);
		gdl_fileclose (DATA, stream);
  		
		// Read the feature
		feature = gdl_gfeatures_alloc (gdl_gfeatures_gff3, FEATURE);
		
		if (EXPCMP)
		{
			expcmp = gdl_tiling_gene_expcmp_alloc (EXPCMP);
		}
		
		if (REF_SEQ)
		{
			gdl_seq_io * seq_io = gdl_seq_io_alloc (gdl_seq_io_type_get ("fasta"));
			
			gdl_seq_io_open (seq_io, REF_SEQ);
			gdl_seq_io_parse (seq_io);
			ref_seq = gdl_seq_io_close (seq_io);
			gdl_seqdb_id2seq_create (ref_seq);
			
			gdl_seq_io_open (seq_io, QRY_SEQ);
			gdl_seq_io_parse (seq_io);
			qry_seq = gdl_seq_io_close (seq_io);
			gdl_seqdb_id2seq_create (qry_seq);
			
			gdl_seq_io_free (seq_io);
		}
		
		// Open output streams
		file = gdl_string_sprintf ("%s_snep_table.txt", OUTPUT);
		nuc_stream = gdl_fileopen (file, "w");
		gdl_string_free (file);
		file = gdl_string_sprintf ("%s_snep.gff", OUTPUT);
		gff_stream = gdl_fileopen (file, "w");
		fprintf (gff_stream, "##gff-version 3\n");
		fprintf (gff_stream, "#\n");
		fprintf (gff_stream, "# This file has been generated by NManout\n");
		fprintf (gff_stream, "# p-value cut-off = %e\n", PVAL_MAX);
		fprintf (gff_stream, "#\n");
		gdl_string_free (file);

		if (REF_SEQ)
		{
			file = gdl_string_sprintf ("%s_snep_seq.txt", OUTPUT);
			seq_stream = gdl_fileopen (file, "w");
			gdl_string_free (file);
			
			
		}
		
		file = gdl_string_sprintf ("%s_snep_tiling.txt", OUTPUT);
		tl_stream = gdl_fileopen (file, "w");
		gdl_string_free (file);
		
		// Loop on chromosomes
		for(c = 0; c < data->nchrom; c++)
		{
			if (SEQ_ID && strcmp(data->chroms[c]->name, SEQ_ID))
			{
				continue;
			}
			
			file = gdl_string_sprintf ("%s/%s.db", INPUT, data->chroms[c]->name);
		 	
		 	if (gdl_isfile (file))
		 	{ 
			   nm_align_nucleo_chromosome * align;
			   gdl_gfeatures_landmark * chrom_feature;
			   
			    fprintf (stderr, "Process chromosome %s: wait...", data->chroms[c]->name);
		 		fflush(stderr);
			 	
			 	stream = gdl_fileopen (file, "r");
			 	align = nm_align_nucleo_chromosome_fread (stream);
			 	gdl_fileclose (file, stream);
		 	 	gdl_string_free (file);
		 	 	
		 	 	if (!align)
			 	{
			 		GDL_ERROR_VAL ("Unable to read nucleosome alignment", GDL_FAILURE, GDL_FAILURE); 
			 	}
			 	
			 	chrom_feature = gdl_gfeatures_lookup (feature, data->chroms[c]->name);
			 	
			 	if (!chrom_feature)
			 	{
			 		GDL_ERROR_VAL (gdl_string_sprintf ("Unable to get features for chromosome %s", data->chroms[c]->name), GDL_FAILURE, GDL_FAILURE); 
			 	}
			 	
			 	size_t i;
				double sd;
				int j, start_idx, end_idx;
				long start, end;
				nm_align_nucleo_nucleosome * ref_nucleo, * qry_nucleo;
				
				for(i = 0; i < align->n_nucleosome; i++)
				{
					if (!align->nucleosomes[i]->align)
					{
						continue;
					}
					
					ref_nucleo = align->nucleosomes[i];
					
					qry_nucleo = ref_nucleo->align;
					
					txt_output (stdout, data, c, align, i);
					
					if (ref_nucleo->pval != -1.0 && ref_nucleo->pval <= PVAL_MAX)
					{
						start = GDL_MIN (ref_nucleo->ref_start_pos, qry_nucleo->ref_start_pos);
						end   = GDL_MAX (ref_nucleo->ref_end_pos, qry_nucleo->ref_end_pos);
						fprintf (nuc_stream, ">%s SNEP_%d %d %d %g %g\n", data->chroms[c]->name, snep_id, start, end, ref_nucleo->gamma, ref_nucleo->pval);
						fprintf (gff_stream, "%s\tNucleoMiner\tSNEP\t%d\t%d\t%g\t+\t.\tID=SNEP_%d;p_value=%e\n", data->chroms[c]->name, start, end, ref_nucleo->gamma, snep_id, ref_nucleo->pval);
						fflush (gff_stream);
						fprintf (nuc_stream, "-- Features\n");
						print_features (nuc_stream, feature, chrom_feature, start, end, WINDOW, expcmp);
						if (NEIGHBOUR)
						{
							fprintf (nuc_stream, "-- Context\n");
							print_context (nuc_stream,  align, ref_nucleo, i, NEIGHBOUR);
						}
						if (ref_seq)
						{
							gdl_string * header = gdl_string_sprintf("SNEP_%d |REF|ID=%s|start=%d|end=%d|",snep_id, data->chroms[c]->name, ref_nucleo->ref_start_pos, ref_nucleo->ref_end_pos);
							print_seq (seq_stream, ref_seq, data->chroms[c]->name, ref_nucleo->ref_start_pos, ref_nucleo->ref_end_pos, header);
							gdl_string_free (header);
						}
						if (qry_seq)
						{
							const long qry_start = qry_nucleo->qry_start_pos;
							const long qry_end   = qry_nucleo->qry_end_pos;
							const gdl_string * qry_seqid = qry_nucleo->qry_seq_name;
							gdl_string * header = gdl_string_sprintf("SNEP_%d |QRY|ID=%s|start=%d|end=%d|",snep_id, qry_seqid, qry_start, qry_end);
							print_seq (seq_stream, qry_seq, qry_seqid, qry_start, qry_end, header);
							gdl_string_free (header);
						}
						snep_id++;
					}
					double tl_pr = ((double)TILING)/(ref_nucleo->ref_end_pos-ref_nucleo->ref_start_pos+1);
					for(j = ref_nucleo->ref_start_pos; j <= ref_nucleo->ref_end_pos; j += TILING)
					{
						fprintf (tl_stream, "%s %d %g", data->chroms[c]->name, j, tl_pr);
						if (ref_nucleo->pval != -1.0 && ref_nucleo->pval <= PVAL_MAX)
						{
							if (ref_nucleo->gamma > 0)
							{
								fprintf (tl_stream, " 0 %g\n", tl_pr);
							}
							else
							{
								fprintf (tl_stream, " %g 0\n", tl_pr);
							}
						}
						else
						{
							fprintf (tl_stream, " 0 0\n");
						}
					}
					fflush(tl_stream);
				}
				nm_align_nucleo_chromosome_free (align);
		 		fprintf (stderr, "\b\b\b\b\b\b\b[ OK ]\n");
		 		fflush(stderr);
		 	}
		}
		
		file = gdl_string_sprintf ("%s_snep_table.txt", OUTPUT);
		gdl_fileclose (file, nuc_stream);
		gdl_string_free (file);
		file = gdl_string_sprintf ("%s_snep.gff", OUTPUT);
		gdl_fileclose (file, gff_stream);
		gdl_string_free (file);
//		file = gdl_string_sprintf ("%s_ins.txt", OUTPUT);
//		gdl_fileclose (file, ins_stream);
//		gdl_string_free (file);
//		file = gdl_string_sprintf ("%s_del.txt", OUTPUT);
//		gdl_fileclose (file, del_stream);
//		gdl_string_free (file);
		if (REF_SEQ && QRY_SEQ)
		{
			file = gdl_string_sprintf ("%s_snep_seq.txt", OUTPUT);
			gdl_fileclose (file, seq_stream);
			gdl_string_free (file);
		}
//		if (snep_tiling_flag)
//		{
		file = gdl_string_sprintf ("%s_snep_tiling.txt", OUTPUT);
		gdl_fileclose (file, tl_stream);
		gdl_string_free (file);
//		}
		
		gdl_gfeatures_free (feature);
		runtime_setdown (runtime);
  }
  
  exit(0);
}

