/*  
 *  NManova.c
 * 
 *  $Author: gael $, $Date: 2008-10-17 16:07:07 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */

#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_list.h>
#include <gdl/gdl_machine.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_sort_long.h>
#include <gdl/gdl_runtime.h>
#include <gdl/gdl_vector.h>
#include <gdl/gdl_matrix.h>
#include <gdl/gdl_multireg.h>
#include <gdl/gdl_genome_coord.h>

#include "nucleoMiner.h"

#include "const.h"
#include "header.c"
#include "util.c"

#include "tiling_array.h"
#include "tiling_hmm.h"
#include "tiling_nucleo.h"
#include "align_nucleo.h"

static gdl_string * PROGRAM = "NManova";

static int help_flag       = 0;
static int verbose_flag    = 0;
static int filter_flag     = 0;
static int update_flag     = 0;
static int extract_flag    = 0;
static int no_probe_effect_flag = 0;

static gdl_string * INPUT   = NULL;
static gdl_string * OUTPUT  = NULL;
static gdl_string * CONFIG  = NULL;
static size_t NCLASS        = 3;
static double PRB_COVERAGE_MIN = 0.5;
static double PRB_ANOVA_MIN    = 10;
static long LINKER_SIZE_MIN    = 50;
static size_t PERMUT           = 0;
static gdl_string * RM_REGION  = NULL;
static gdl_string * SEQ_ID     = NULL;
static long START = -1;
static long END   = -1;

static struct option long_options[] =
{
   /* These options set a flag. */
   {"help", no_argument,           &help_flag, 1},
   {"verbose", no_argument,        &verbose_flag, 1},
   {"brief",   no_argument,        &verbose_flag, 0},
   {"filter",   no_argument,       &filter_flag, 1},
   {"update",   no_argument,       &update_flag, 1},
   {"extract",   no_argument,      &extract_flag, 1},
   {"no-probe-effect",   no_argument,      &no_probe_effect_flag, 1},
   /* These options don't set a flag.
      We distinguish them by their indices. */
   {"input",   required_argument, 0, 'i'},    // the config file
   {"output",  required_argument, 0, 'o'},    // the output stem
   {"config",   required_argument, 0, 'c'},   // the name of the chromosome to focus on
   {"nclass",  required_argument, 0, 'n'},    
   {"prb-coverage",  required_argument, 0, 'j'},
   {"prb-anova",  required_argument, 0, 'k'},
   {"linker-min", required_argument, 0, 'r'},
   {"permut", required_argument, 0, 'p'},
   {"rm-region", required_argument, 0, 'g'},
   {"seqid", required_argument, 0, 'd'},
   {"start", required_argument, 0, 's'},
   {"end", required_argument, 0, 'e'},
   {0, 0, 0, 0}
};
 
 
static int
parse_argument (int argc, char *argv[])
{
	int c;
     
   while (1)
     {
      
       /* getopt_long stores the option index here. */
       int option_index = 0;
 
       c = getopt_long (argc, argv, "i:o:c:n:t:m:j:k:r:p:g:s:e:d:",
                        long_options, &option_index);
 
       /* Detect the end of the options. */
       if (c == -1)
         break;
         
       switch (c)
         {
         case 0:
           /* If this option set a flag, do nothing else now. */
           if (long_options[option_index].flag != 0)
             break;
           printf ("option %s", long_options[option_index].name);
           if (optarg)
             printf (" with arg %s", optarg);
           printf ("\n");
           break;
 
         case 'i':
           INPUT = gdl_string_clone (optarg);
           break;
         case 'o':
           OUTPUT = gdl_string_clone (optarg);
           break;
         case 'c':
           CONFIG = gdl_string_clone (optarg);
           break;
         case 'n':
           NCLASS = (size_t) atoi (optarg);
           break;
         case 'j':
           PRB_COVERAGE_MIN = (double)atof (optarg);
           break; 
         case 'k':
           PRB_ANOVA_MIN = (double)atof (optarg);
           break; 
         case 'r':
           LINKER_SIZE_MIN = atol (optarg);
           break;
         case 'p':
           PERMUT = (size_t)atoi (optarg);
           break;
         case 'g':
           RM_REGION = gdl_string_clone (optarg);
           break;
         case 'd':
           SEQ_ID = gdl_string_clone (optarg);
           break;
         case 's':
           START = atol (optarg);
           break;
         case 'e':
           END = atol (optarg);
           break;   
         case '?':
           GDL_ERROR_VAL ("Unknown arguments", GDL_EINVAL, -1);
         default:
           GDL_ERROR_VAL ("Bad arguments", GDL_EINVAL, -1);
         }
     }
   /* Print any remaining command line arguments (not options). 
   if (optind < argc)
     {
       printf ("non-option ARGV-elements: ");
       while (optind < argc)
         printf ("%s ", argv[optind++]);
       putchar ('\n');
     }*/
}

static int
check_argument (void)
{
	if (INPUT == 0)
	{
		GDL_ERROR_VAL ("No input directory defined",
		               GDL_FAILURE,
		               1);
	}
	if (CONFIG == 0)
	{
		GDL_ERROR_VAL ("No configuration file defined",
		               GDL_FAILURE,
		               1);
	}
	if (!extract_flag && OUTPUT == 0)
	{
		GDL_ERROR_VAL ("No output stem defined",
		               GDL_FAILURE,
		               1);	
	}
	if (extract_flag && (START<0 || END<01))
	{
		GDL_ERROR_VAL ("Invalid sequence start/end defined",
		               GDL_FAILURE,
		               1);		
	}
	if (extract_flag && SEQ_ID==0)
	{
		GDL_ERROR_VAL ("No sequence id defined",
		               GDL_FAILURE,
		               1);		
	}
	
	return GDL_SUCCESS;
}

static int
help (void)
{
   print_header (PROGRAM);
   printf ("\n");
   printf ("%s carries out ANOVA for each aligned nucleosome\n", PROGRAM);
   printf ("\n");
   printf ("*These options set a flag:\n");
   printf ("\n");
   printf ("--help\tDisplay a brief help on program usage\n");
   printf ("--verbose\tOutput message on standard output to see what the program is doing\n");
   printf ("\n");
   printf ("--filter\tFilter the nucleosome alignment by removing regions with unexpected alignments\n");
   printf ("--update\tUpdate the binary files generated by NMalign\n");
   printf ("--extract\tJust extract the data used to do the ANOVA\n");
   printf ("--no-probe-effect\tDo not include a probe effect in the model\n");
   printf ("\n");
   printf ("*These options require arguments:\n");
   printf ("\n");
   printf ("--config or -c\tThe input configuration file (see manual)\n");
   printf ("--input or -i\tThe name/path to the directory containing files (.nal) generated by NMalign\n");
   printf ("--output or -o\tThe name/path to the output directory\n");
   printf ("--nclass or -n\tThe number of nucleosome classes (default is 3)\n");
   printf ("--prb-coverage or -j\tThe minimal fraction of observed probes spanning the nucleosome (default is 0.5)\n");
   printf ("--prb-anova or -k\tThe minimal number of shared probes for the aligned nucleosome (default is 10)\n");
   printf ("--permut or -p\tThe number of permutations to do to compute the empirical p-value by permutations (default is 0)\n");
   printf ("--seqid or -d\tThe name of the chromosome/sequence to focus on\n");
   printf ("--start or -s\tThe starting position on the chromosome (use with --extract)\n");
   printf ("--end or -e\tThe ending position on the chromosome (use with --extract)\n");
   return 0;
}

int
main (int argc, char *argv[])
{
  int status;
  
  parse_argument (argc, argv);
  
  if (help_flag)
  {
  	  exit (help());
  }
  
  status = check_argument ();
  
  if (status == GDL_SUCCESS)
  {
  		size_t c,w;
  		int status; 
  		gdl_string * file;
  		FILE * stream;
  		DIR * dir;
  		gdl_runtime * runtime;
  		gdl_tiling_nucleo_hmm * hmm;
  		align_nucleo_config * conf;
  		_rm_region * rm_region = 0;
  		
  		runtime = runtime_setup();

  	
  	    conf = align_nucleo_config_alloc (CONFIG);
		conf->hmm_nclass = NCLASS;
		
		if (OUTPUT && (dir=opendir(OUTPUT))==0)
		{
			mkdir (OUTPUT, S_IRWXU);
		}
		
		if (!extract_flag && RM_REGION)
		{
			rm_region = read_rm_region (RM_REGION);
		}
		
		// Loop on chromosomes
		for(c = 0; c < conf->ref_data->nchrom; c++)
		{
			if (SEQ_ID && strcmp(conf->ref_data->chroms[c]->name, SEQ_ID))
			{
				continue;
			}
			file = gdl_string_sprintf ("%s/%s.nal", INPUT, conf->ref_data->chroms[c]->name);
		 	if (gdl_isfile (file))
		 	{ 
			   align_nucleo_alignment * nucleo_align;
			   fprintf (stdout, "Process chromosome %s: wait...", conf->ref_data->chroms[c]->name);
		 	   fflush(stdout);
			   stream = gdl_fileopen (file, "r");
			   nucleo_align = align_nucleo_alignment_fread (stream, conf);
			   gdl_fileclose (file, stream);
		 	   gdl_string_free (file);
		 	 	
		 	   if (!nucleo_align)
			 	{
			 		GDL_ERROR_VAL ("Unable to read nucleosome alignment", GDL_FAILURE, GDL_FAILURE); 
			 	}
			 	if (filter_flag)
			 	{
			 		align_nucleo_linker_alignment_filter (nucleo_align);
			 	}
			 	if (extract_flag)
			 	{
			 		align_nucleo_alignment_anova_extract (conf, nucleo_align->ref_chrom, nucleo_align, PRB_COVERAGE_MIN, PRB_ANOVA_MIN, no_probe_effect_flag, runtime->rng, START, END, stdout);
			 	}
			 	else
			 	{
			 	    align_nucleo_linker_alignment_statistics (conf, nucleo_align);
				    align_nucleo_alignment_pval (conf, nucleo_align->ref_chrom, nucleo_align, PRB_COVERAGE_MIN, PRB_ANOVA_MIN, PERMUT, no_probe_effect_flag, runtime->rng);
				 	if (rm_region)
				 	{
				 		_rm_region_perform (nucleo_align, rm_region);
				 	}
				 	file   = gdl_string_sprintf ("%s/%s.nuc", OUTPUT, nucleo_align->ref_chrom->data->name);
				 	stream = gdl_fileopen (file, "w");
				 	align_nucleo_alignment_fprintf (stream, nucleo_align);
				 	gdl_fileclose (file, stream);
				 	gdl_string_free (file);
	//			 	file   = gdl_string_sprintf ("%s/%s.lkr", OUTPUT, nucleo_align->ref_chrom->data->name);
	//			 	stream = gdl_fileopen (file, "w");
	//			 	if (!no_anova_flag)
	//			 	{
	//		 			align_nucleo_linker_alignment_pval (conf, nucleo_align->ref_chrom, nucleo_align, LINKER_SIZE_MIN, PRB_COVERAGE_MIN);
	//			 	}
	//		 		align_nucleo_linker_alignment_fprintf (stream, nucleo_align);
	//		 		gdl_fileclose (file, stream);
	//			 	gdl_string_free (file);
				 	if (update_flag)
				 	{
				 		file = gdl_string_sprintf ("%s/%s.nal", OUTPUT, conf->ref_data->chroms[c]->name);
				 		stream = gdl_fileopen (file, "w");
				 		align_nucleo_alignment_fwrite (stream, conf, nucleo_align);
				 		gdl_fileclose (file, stream);
			 	 		gdl_string_free (file);
				 	}	
			 		align_nucleo_alignment_free (nucleo_align);
			 		fprintf (stdout, "\b\b\b\b\b\b\b[ OK ]\n");
			 		fflush(stdout);	 
			 	}	
		 	}
		}
		runtime_setdown (runtime);
  }
  exit(0);
}

