/*  
 *  NManova2.c
 * 
 *  $Author: gael $, $Date: 2008-10-17 16:07:07 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_list.h>
#include <gdl/gdl_machine.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_sort_long.h>
#include <gdl/gdl_runtime.h>
#include <gdl/gdl_vector.h>
#include <gdl/gdl_matrix.h>
#include <gdl/gdl_multireg.h>

#include <gdl/gdl_genome_coord.h>

#include "nucleoMiner.h"

#include "const.h"
#include "header.c"
#include "util.c"

#include "tiling_array.h"
#include "tiling_hmm.h"
#include "tiling_nucleo.h"
#include "align_nucleo2.h"

static gdl_string * PROGRAM = "NManova2";

static int help_flag       = 0;
static int verbose_flag    = 0;
static int print_flag      = 0;

static gdl_string * INPUT   = NULL;
static gdl_string * ALIGN   = NULL;
static gdl_string * OUTPUT  = NULL;
static size_t PROBE_MIN     = 10;

static struct option long_options[] =
{
   /* These options set a flag. */
   {"help", no_argument,       &help_flag, 1},
   {"verbose", no_argument,       &verbose_flag, 1},
   {"brief",   no_argument,       &verbose_flag, 0},
   /* These options don't set a flag.
    * We distinguish them by their indices. */
   {"input",   required_argument, 0, 'i'},   // the config file
   {"align",   required_argument, 0, 'a'},   // the alignment directory
   {"output",  required_argument, 0, 'o'},   // the output stem
   {"prb-min",  required_argument, 0, 'p'},   
   {0, 0, 0, 0}
};
 
static int
parse_argument (int argc, char *argv[])
{
	int c;
     
   while (1)
     {
      
       /* getopt_long stores the option index here. */
       int option_index = 0;
 
       c = getopt_long (argc, argv, "i:o:a:p:",
                        long_options, &option_index);
 
       /* Detect the end of the options. */
       if (c == -1)
         break;
         
       switch (c)
         {
         case 0:
           /* If this option set a flag, do nothing else now. */
           if (long_options[option_index].flag != 0)
             break;
           printf ("option %s", long_options[option_index].name);
           if (optarg)
             printf (" with arg %s", optarg);
           printf ("\n");
           break;
 
         case 'i':
           INPUT = gdl_string_clone (optarg);
           break;
         case 'a':
           ALIGN = gdl_string_clone (optarg);
           break;
         case 'o':
           OUTPUT = gdl_string_clone (optarg);
           break;
         case 'p':
           PROBE_MIN = (size_t) atoi (optarg);
           break;
         case '?':
           GDL_ERROR_VAL ("Unknown arguments", GDL_EINVAL, -1);
         default:
           GDL_ERROR_VAL ("Bad arguments", GDL_EINVAL, -1);
         }
     }
   /* Print any remaining command line arguments (not options). 
   if (optind < argc)
     {
       printf ("non-option ARGV-elements: ");
       while (optind < argc)
         printf ("%s ", argv[optind++]);
       putchar ('\n');
     }*/
}

static int
check_argument (void)
{
	if (INPUT == 0)
	{
		GDL_ERROR_VAL ("No input data file defined",
		               GDL_FAILURE,
		               1);
	}
	if (ALIGN == 0)
	{
		GDL_ERROR_VAL ("No alignment directory defined",
		               GDL_FAILURE,
		               1);
	}
	if (OUTPUT == 0)
	{
		GDL_ERROR_VAL ("No output stem defined",
		               GDL_FAILURE,
		               1);	
	}
	
	return GDL_SUCCESS;
}

static int
help (void)
{
   print_header (PROGRAM);
   printf ("\n");
   printf ("%s implements an algorithm to align two nucleosome occupancy profiles\n", PROGRAM);
   printf ("\n");
   printf ("*These options set a flag:\n");
   printf ("\n");
   printf ("--help\tDisplay a brief help on program usage\n");
   printf ("--verbose\tOutput message on standard output to see what the program is doing\n");
   printf ("\n");
   printf ("*These options require arguments:\n");
   printf ("\n");
   printf ("--input or -i\tThe input directory\n");
   printf ("--align or -a\tThe alignment directory (as generated by NMalign2)\n");
   printf ("--output or -o\tThe name/path to the output directory\n");
   printf ("--prb-min or -p\tThe minimum number of probes within a nucleosome\n");
}

static size_t 
ANOVA_SETUP (const gdl_tiling_data * data,
			 const size_t chrom_idx,
			 const size_t probe_start_idx,
             const size_t probe_end_idx,
             gdl_matrix * X0,
             gdl_matrix * X1,
             gdl_vector * y,
             const size_t P,
             const gdl_boolean no_probe_effect_flag,
             FILE * stream)
{
	size_t g,na,a,i,j,np;
	int pidx,p;
	double yi;
	const gdl_tiling_chromosome * chrom = data->chroms[chrom_idx];
	
	for(np = i = 0, p = probe_start_idx; p <= probe_end_idx; p++, np++)
	{
		for(g = 0; g < 2; g++) // loop on manip for strain A
		{
			na = data->narrayByGroup[g]; 
			for(a = 0; a < na; a++, i++) // loop on replicates
			{
				yi = gdl_tiling_data_get_val (data, chrom_idx, p, g, a);
				gdl_matrix_set (X0, i, 0, 1.);
				gdl_matrix_set (X0, i, 1, 1.);
				if (!g) 
				{
					gdl_matrix_set (X0, i, 2, 1.);
					gdl_matrix_set (X1, i, 0, 1.);
				}
				if (!no_probe_effect_flag && np < P - 1) 
				{
					gdl_matrix_set (X0, i, 3 + np, 1.);
				}
				gdl_vector_set (y, i, yi);
				if (stream) 
				{
					fprintf (stream, "A %s_%d %s %d %g\n", chrom->name, chrom->positions[p], (g==0) ? "A" : "B", np, yi);
//					fprintf (stream, "%g", yi);
//					for(j = 0; j < X0->size2; j++)
//					{
//						fprintf (stream, " %1.0f", gdl_matrix_get (X0, i, j));
//					}
//					fprintf (stream, " %1.0f\n", gdl_matrix_get (X1, i, 0));
				}
			}
		}
		for(g = 2; g < 4; g++) // loop on manip for strain B
		{
			na = data->narrayByGroup[g];
			for(a = 0; a < na; a++, i++) // loop on replicates
			{
				yi = gdl_tiling_data_get_val (data, chrom_idx, p, g, a);
				gdl_matrix_set (X0, i, 0, 1.);
				if (g==2)
				{
					gdl_matrix_set (X0, i, 2, 1.);
				}
				if (!no_probe_effect_flag && np < P - 1) 
				{
					gdl_matrix_set (X0, i, 3 + np, 1.);
				}
				gdl_vector_set (y, i, yi);
				if (stream) 
				{
					fprintf (stream, "B %s_%d %s %d %g\n", chrom->name, chrom->positions[p], (g==2) ? "A" : "B", np, yi);
//					fprintf (stream, "%g", yi);
//					for(j = 0; j < X0->size2; j++)
//					{
//						fprintf (stream, " %1.0f", gdl_matrix_get (X0, i, j));
//					}
//					fprintf (stream, " %1.0f\n", gdl_matrix_get (X1, i, 0));
				}
			}
		}
	}
}

static void
ANOVA (const gdl_tiling_data * data,
	   const size_t chrom_idx,
       nm_align_nucleo_nucleosome * ref_nucleo,
       nm_align_nucleo_nucleosome * qry_nucleo,
       const size_t probe_start_idx,
       const size_t probe_end_idx,
       const gdl_boolean no_probe_effect,
       const size_t permut,
       const gdl_rng * rng,
       FILE * stream)
{
	size_t i, j, N, M, P;
	gdl_multireg_linear_workspace * W;
	gdl_matrix * X0, * X1;
	gdl_vector * y;
	const gdl_tiling_chromosome * chrom = data->chroms[chrom_idx];
	
	//printf ("START = %d END = %d\n", probe_end_idx, probe_start_idx);
	
	P = probe_end_idx - probe_start_idx + 1;
	N = P * chrom->narray;
	M = 3;
	
	if (no_probe_effect)
	{
		X0 = gdl_matrix_calloc (N, M);
	}
	else
	{
		X0 = gdl_matrix_calloc (N, M + P - 1);
	}
	X1 = gdl_matrix_calloc (N, 1);
	y  = gdl_vector_alloc (N);
	
	if (no_probe_effect)
	{
		W  = gdl_multireg_linear_workspace_alloc (N, M + 1);
	}
	else
	{
		W  = gdl_multireg_linear_workspace_alloc (N, M + P);
	}
	
	if (stream) 
	{
		fprintf(stream, "<nucleosome_alignment chrom=\"%s\" ref_start=\"%d\" ref_end=\"%d\" qry_start=\"%d\" qry_end=\"%d\" shared-probes=\"%d\">\n", chrom->name,ref_nucleo->ref_start_pos,ref_nucleo->ref_end_pos,qry_nucleo->ref_start_pos,qry_nucleo->ref_end_pos, P);
		fprintf(stream, "<data>\n");
	}
	
	ANOVA_SETUP (data, chrom_idx, probe_start_idx, probe_end_idx, X0, X1, y, P, no_probe_effect, stream);
	
	if (permut)
	{
		gdl_multireg_linear_eval_permut (W, X0, X1, y, permut, rng);
	}
	else
	{
		//printf ("EVAL X0 [%d, %d] ", X0->size1, X0->size2);
		gdl_multireg_linear_eval (W, X0, X1, y);
		//printf (" : OK\n");
		//fflush(stdout);
	}
	
	qry_nucleo->pval  = gdl_multireg_linear_eval_pvalue(W);
	ref_nucleo->pval  = qry_nucleo->pval;
	qry_nucleo->gamma = gdl_vector_get (gdl_multireg_linear_coefficients(W), X0->size2);
	ref_nucleo->gamma = qry_nucleo->gamma;
	
	if (stream) 
	{
		fprintf(stream, "</data>\n");
		fprintf(stream, "<effect name=\"gamma\" value=\"%e\" pvalue=\"%e\"/></p-value>\n", qry_nucleo->gamma, qry_nucleo->pval);
		fprintf(stream, "</nucleosome_alignment>\n");
		fflush(stream);
	}
	
	// clean
	gdl_vector_free (y);
	gdl_matrix_free (X1);
	gdl_matrix_free (X0);
	gdl_multireg_linear_workspace_free (W);
}

int
main (int argc, char *argv[])
{
  int status;
  
  parse_argument (argc, argv);
  
  if (help_flag)
  {
  	  exit (help());
  }
  
  status = check_argument ();
  
  if (status == GDL_SUCCESS)
  {
  		size_t c,i,j,k;
  		int status; 
  		gdl_string * file;
  		FILE * stream;
  		DIR * dir;
  		gdl_runtime * runtime;
  		
  		runtime = runtime_setup();
  		
		if ((dir=opendir(OUTPUT))==0)
		{
			mkdir (OUTPUT, S_IRWXU);
		}
		
		if (verbose_flag)
		{
			fprintf (stdout, "--\nStep 1: Load the data file\n--\n");
		 	fflush(stdout);
		}
		
		stream = gdl_fileopen (INPUT, "r");
		gdl_tiling_data * data = gdl_tiling_data_fread (stream);
		gdl_fileclose (INPUT, stream);
		
		// check if the data format is ok
		if (data->ngroup != 4)
		{
			GDL_ERROR_VAL ("Data must contain 4 groups (2 x 2)", GDL_EINVAL, GDL_EINVAL);	
		} 
		
		if (verbose_flag)
		{
			fprintf (stdout, "--\nStep 2: Perform the ANOVA\n--\n");
		 	fflush(stdout);
		}
		
		for(c = 0; c < data->nchrom; c++)
		{
			nm_align_nucleo_chromosome * align;
			gdl_tiling_chromosome      * chrom = data->chroms[c];
			
			fprintf (stdout, "Process chromosomes %s: wait...", chrom->name);
			fflush(stdout);
			
			// Load the corresponding alignment 
			file   = gdl_string_sprintf ("%s/%s_ref.db", ALIGN, chrom->name);
			stream = gdl_fileopen (file, "r");
			align = nm_align_nucleo_chromosome_fread (stream);
			gdl_fileclose (file, stream);
			gdl_string_free (file);
			// Loop on aligned chromosomes and do the ANOVA
			for(i = 0; i < align->n_nucleosome; i++)
			{
				nm_align_nucleo_nucleosome * ref_nuc = align->nucleosomes[i];
				if (ref_nuc->align != 0)
				{
					nm_align_nucleo_nucleosome * qry_nuc = align->nucleosomes[i]->align;
					// Get the boundaries of the aligned region
					long nuc_start = GDL_MAX (ref_nuc->ref_start_pos, qry_nuc->ref_start_pos);
					long nuc_end   = GDL_MIN (ref_nuc->ref_end_pos, qry_nuc->ref_end_pos);
					int nprobe = 0, probe_start_idx = -1, probe_end_idx = -1;
					for(j = 0; j < chrom->nprobe; j++)
					{
						if (chrom->positions[j] >= nuc_start && chrom->positions[j] <= nuc_end)
						{
							if (!nprobe) 
							{
								probe_start_idx = j;
							}
							nprobe++;
						}
						else if (nprobe && j > 0)
						{
							probe_end_idx = j - 1;
							break;
						}
					}
					if ((probe_start_idx != -1 && probe_end_idx != -1) && nprobe > PROBE_MIN)
					{
						//printf ("NUCLEOSOME %d %d ==> %d, %d (%d)\n", nuc_start, nuc_end, probe_start_idx, probe_end_idx, nprobe);
						ANOVA (data, c, ref_nuc, qry_nuc, probe_start_idx, probe_end_idx, 0, 0, runtime->rng, (verbose_flag) ? stdout : 0);
					}
					else
					{
						ref_nuc->pval  = qry_nuc->pval  = -1;
						ref_nuc->gamma = qry_nuc->gamma = GDL_NAN;
					}
				}
			}
			
			// save the corresponding alignment 
			file   = gdl_string_sprintf ("%s/%s.db", OUTPUT, chrom->name);
			stream = gdl_fileopen (file, "w");
			nm_align_nucleo_chromosome_fwrite (stream, align);
			gdl_fileclose (file, stream);
			gdl_string_free (file);
			
			nm_align_nucleo_chromosome_free (align);
			
			fprintf (stdout, "\b\b\b\b\b\b\b[ DONE ]\n");
			fflush(stdout);
		}
		
		runtime_setdown (runtime);
  }
  exit(0);
}

