/*  
 *  NManobin2.c
 * 
 *  $Author: veyrieras $, $Date: 2008-11-14 15:04:39 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_list.h>
#include <gdl/gdl_machine.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_sort_long.h>
#include <gdl/gdl_runtime.h>
#include <gdl/gdl_vector.h>
#include <gdl/gdl_matrix.h>
#include <gdl/gdl_multireg.h>
#include <gdl/gdl_genome_coord.h>
#include <gdl/gdl_gfeatures.h>
#include <gdl/gdl_gfeatures_gff3.h>
#include <gdl/gdl_gfeatures_pos.h>

#include "nucleoMiner.h"

#include "const.h"
#include "header.c"
#include "util.c"

#include "tiling_array.h"
#include "tiling_hmm.h"
#include "tiling_nucleo.h"
#include "tiling_feature.h"
#include "align_nucleo2.h"

static gdl_string * PROGRAM = "NManobin2";

static int help_flag       = 0;
static int verbose_flag    = 0;
static int sliding_flag    = 0;

static gdl_string * INPUT   = NULL;
static gdl_string * CONFIG  = NULL;
static size_t       NCLASS  = 3;
static gdl_string * SEQ_ID  = NULL;
static gdl_string * FEATURE = NULL;
static gdl_string * FEATURE_ID = NULL;
static gdl_string * DATA    = NULL;
static gdl_string * POS_F    = NULL;
static long WINDOW          = 1000;
static double PVAL_SNEP     = 0.01;

static struct option long_options[] =
{
   /* These options set a flag. */
   {"help", no_argument,           &help_flag, 1},
   {"verbose", no_argument,        &verbose_flag, 1},
   {"brief",   no_argument,        &verbose_flag, 0},
   {"sliding",   no_argument,      &sliding_flag, 1},
   /* These options don't set a flag.
      We distinguish them by their indices. */
   {"input",   required_argument, 0, 'i'},   
   {"output",  required_argument, 0, 'o'},   
   {"seqid", required_argument, 0, 's'},
   {"feature", required_argument, 0, 'f'},
   {"feature-id", required_argument, 0, 'l'},
   {"data", required_argument, 0, 'd'},
   {"window", required_argument, 0, 'w'},
   {"pv-cutoff", required_argument, 0, 'p'},
   {"position", required_argument, 0, 'x'},
   {0, 0, 0, 0}
};
 
 
static int
parse_argument (int argc, char *argv[])
{
	int c;
     
   while (1)
     {
      
       /* getopt_long stores the option index here. */
       int option_index = 0;
 
       c = getopt_long (argc, argv, "i:o:s:f:d:w:p:d:l:x:",
                        long_options, &option_index);
 
       /* Detect the end of the options. */
       if (c == -1)
         break;
         
       switch (c)
         {
         case 0:
           /* If this option set a flag, do nothing else now. */
           if (long_options[option_index].flag != 0)
             break;
           printf ("option %s", long_options[option_index].name);
           if (optarg)
             printf (" with arg %s", optarg);
           printf ("\n");
           break;
 
         case 'i':
           INPUT = gdl_string_clone (optarg);
           break;
         case 'f':
           FEATURE = gdl_string_clone (optarg);
           break;
         case 'l':
           FEATURE_ID = gdl_string_clone (optarg);
           break;
         case 'd':
           DATA = gdl_string_clone (optarg);
           break;
         case 's':
           SEQ_ID = gdl_string_clone (optarg);
           break;
         case 'w':
           WINDOW = atol (optarg);
           break;
         case 'p':
           PVAL_SNEP = (double) atof (optarg);
           break;
         case 'x':
           POS_F = gdl_string_clone (optarg);
           break;
         case '?':
           GDL_ERROR_VAL ("Unknown arguments", GDL_EINVAL, -1);
         default:
           GDL_ERROR_VAL ("Bad arguments", GDL_EINVAL, -1);
         }
     }
   /* Print any remaining command line arguments (not options). 
   if (optind < argc)
     {
       printf ("non-option ARGV-elements: ");
       while (optind < argc)
         printf ("%s ", argv[optind++]);
       putchar ('\n');
     }*/
}

static int
check_argument (void)
{
	if (INPUT == 0)
	{
		GDL_ERROR_VAL ("No input directory defined",
		               GDL_FAILURE,
		               1);
	}
	if (DATA == 0)
	{
		GDL_ERROR_VAL ("No data file defined",
		               GDL_FAILURE,
		               1);
	}
//	if (OUTPUT == 0)
//	{
//		GDL_ERROR_VAL ("No output stem defined",
//		               GDL_FAILURE,
//		               1);	
//	}
	
	return GDL_SUCCESS;
}

static int
help (void)
{
   print_header (PROGRAM);
   printf ("\n");
   printf ("%s summarizes results obtained with NManova\n", PROGRAM);
   printf ("   using bin or sliding windows along the genome\n"); 
   printf ("\n");
   printf ("*These options set a flag:\n");
   printf ("\n");
   printf ("--help\tDisplay a brief help on program usage\n");
   printf ("--sliding\tUse a sliding window instead of bin\n");
   printf ("\n");
   printf ("*These options require arguments:\n");
   printf ("\n");
   printf ("--input or -i\tThe input directory (use with --print)\n");
   printf ("--data or -d\tThe data file\n");
   printf ("--output or -o\tThe name/path to the output directory\n");
   printf ("--seqid or -s\tThe name of the chromosome to look at\n");
   printf ("--feature or -f\tA GFF (version 3) annotation file\n");
   printf ("--feature-id or -l\tA file listing the features to consider\n");
   printf ("--pv-cutoff or -p\tThe p-value cut-off for calling SNEP\n");
   printf ("--window or -w\tThe bin/window size\n");
   printf ("--position or -x\tA file providing only positions per landmark\n");
}

int
main (int argc, char *argv[])
{
  int status;
  
  parse_argument (argc, argv);
  
  if (help_flag)
  {
  	  exit (help());
  }
  
  status = check_argument ();
  
  if (status == GDL_SUCCESS)
  {
  		size_t c,w,snep_id=0;
  		int status; 
  		gdl_string * file;
  		FILE * stream;
  		DIR * dir;
  		gdl_runtime * runtime;
  		gdl_gfeatures * feature;
  		gdl_gfeatures_landmark * chrom_feature;
  		gdl_gfeatures * pos_feature = 0;
  		gdl_gfeatures_landmark * pos_chrom_feature = 0;
  		gdl_hashtable * selected_features = 0;
  		
  		runtime = runtime_setup();
  		
  		stream = gdl_fileopen (DATA, "r");
		gdl_tiling_data * data = gdl_tiling_data_fread (stream);
		gdl_fileclose (DATA, stream);
  		
  		// Read the feature
		feature = gdl_gfeatures_alloc (gdl_gfeatures_gff3, FEATURE);
		if (FEATURE_ID)
		{
			selected_features = read_selected_features (FEATURE_ID);
		}
		if (POS_F)
		{
			pos_feature = gdl_gfeatures_alloc (gdl_gfeatures_pos, POS_F);
		}
		
		for(c = 0; c < data->nchrom; c++)
		{
			nm_align_nucleo_chromosome * align;
			gdl_tiling_chromosome      * chrom = data->chroms[c];
			int i, j, start_idx, end_idx;
			long start, end;
			nm_align_nucleo_nucleosome * nucleo;
			long win;
			double nuc_score, snep_score, score, data_score;
			
			if (SEQ_ID && strcmp(chrom->name, SEQ_ID))
			{
				continue;
			}
			
			fprintf (stderr, "Process chromosomes %s: wait...", chrom->name);
			fflush(stderr);
			
			// Load the corresponding alignment 
			file   = gdl_string_sprintf ("%s/%s.db", INPUT, chrom->name);
			stream = gdl_fileopen (file, "r");
			align = nm_align_nucleo_chromosome_fread (stream);
			gdl_fileclose (file, stream);
			gdl_string_free (file);
  			 	
			chrom_feature = gdl_gfeatures_lookup (feature, chrom->name);
		 	
		 	if (!chrom_feature)
		 	{
		 		GDL_ERROR_VAL (gdl_string_sprintf ("Unable to get features for chromosome %s", chrom->name), GDL_FAILURE, GDL_FAILURE);
		 	}
		 	
		 	gdl_gfeatures_gff3_record * chrom_rec = 0;
			gdl_gfeatures_pos_record  * pos_rec   = 0;
			i = 0;
			while(chrom_rec==0 && i < chrom_feature->size)
			{
				gdl_gfeatures_gff3_record * rec = (gdl_gfeatures_gff3_record *)chrom_feature->records[i];
				if (gdl_gfeatures_gff3_record_is_of_type (feature, rec, "chromosome"))
				{
					chrom_rec = rec;
					break;
				}
				i++;
			}
			if (chrom_rec==0)
			{
				GDL_ERROR_VAL (gdl_string_sprintf ("No feature chromosome for landmark %s", chrom->name), GDL_FAILURE, GDL_FAILURE);
			}
			
			if (pos_feature)
			{
				pos_chrom_feature = gdl_gfeatures_lookup (pos_feature, chrom->name);
			}
			
			long chrom_start = chrom_rec->start;
			long chrom_end   = chrom_rec->end;
			for(win = chrom_start; win <= chrom_end-WINDOW; win += WINDOW)
			{
				long bin_start = win;
				long bin_end   = win+WINDOW;
				nuc_score  = 0;
				snep_score = 0;
				for(i = 0; i < align->n_nucleosome; i++)
				{
					score  = 0;
					nucleo = align->nucleosomes[i];
					if (!nucleo->align || nucleo->ignore == 'y' || nucleo->pval == -1.)
					{
						continue;
					}
					if (nucleo->ref_start_pos >= bin_start && nucleo->ref_start_pos <= bin_end)
					{
						if (nucleo->ref_end_pos <= bin_end)
							score = 1;
						else
							score = ((double)(bin_end-nucleo->ref_start_pos))/(nucleo->ref_end_pos-nucleo->ref_start_pos);
					}
					else if (nucleo->ref_end_pos >= bin_start && nucleo->ref_end_pos <= bin_end)
					{
						score = ((double)(nucleo->ref_end_pos-bin_start))/(nucleo->ref_end_pos-nucleo->ref_start_pos);
					}
					else if (nucleo->ref_start_pos > bin_end)
					{
						break;
					}
					//printf ("%d %d %d %d %g\n", bin_start, bin_end, nucleo->ref_start_pos, nucleo->ref_end_pos, score);
					nuc_score += score;
					if (nucleo->pval <= PVAL_SNEP)
						snep_score += score;
				}
				fprintf (stdout, "%s %d %d %g %g", chrom_feature->seqid, bin_start, bin_end, nuc_score, snep_score);
				if (selected_features)
				{
					int start_idx, end_idx;
					gdl_gfeatures_gff3_within (chrom_feature, bin_start, bin_end, &start_idx, &end_idx);
					score = 0;
					for(i = start_idx; i != -1 && i <= end_idx; i++)
					{		
						gdl_gfeatures_gff3_record * rec = (gdl_gfeatures_gff3_record *) chrom_feature->records[i];
						if (is_a_selected_feature (feature, rec, selected_features))
						{
							if (rec->start >= bin_start && rec->start <= bin_end)
							{
								if (rec->end <= bin_end)
									score += 1;
								else
									score += ((double)(bin_end-rec->start))/(rec->end-rec->start);
							}
							else if (rec->end >= bin_start && rec->end <= bin_end)
							{
								score += ((double)(rec->end-bin_start))/(rec->end-rec->start);
							}
						}
					}
					fprintf (stdout, " %g", score);
				}
				if (pos_chrom_feature)
				{
					score = 0;
					for(i = 0; i < pos_chrom_feature->size; i++)
					{
						gdl_gfeatures_pos_record * rec = pos_chrom_feature->records[i];
						if (rec->position >= bin_start && rec->position <= bin_end)
						{
							score++;
						}
					}
					fprintf (stdout, " %g", score);
				}
				fprintf (stdout, "\n");
				fflush(stdout);
				win += (sliding_flag) ? 1 : WINDOW; 
			}
		 	nm_align_nucleo_chromosome_free (align);
		 	fprintf (stderr, "\b\b\b\b\b\b\b[ OK ]\n");
	 		fflush(stderr);
		}
		gdl_gfeatures_free (feature);
		runtime_setdown (runtime);
  }
  
  exit(0);
}

