/*  
 *  NMcl2tab.c
 * 
 *  $Author: gael $, $Date: 2008-10-17 16:07:07 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_list.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_eqtl_genome.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_statistics_double.h>

#include "nucleoMiner.h"

#include "header.c"

static gdl_string * PROGRAM = "NMcl2tab";

static int help_flag    = 0;
static int verbose_flag = 0;
static int rotate_flag  = 1;
static int CbyC_flag    = 0;
static int qqnorm_flag  = 0;

static gdl_string * CELFILENAMES = NULL;
static gdl_string * CELFILEDIR   = NULL;
static gdl_string * PROBEFILE    = NULL;
static gdl_string * OUTPUT		 = NULL;
static gdl_string * QQNORM		 = NULL;
static int UNIQNESS = -1;

static struct option long_options[] =
{
   /* These options set a flag. */
   {"help", no_argument,       &help_flag, 1},
   {"verbose", no_argument,       &verbose_flag, 1},
   {"brief",   no_argument,       &verbose_flag, 0},
   {"no-rotate",   no_argument,       &rotate_flag, 0},
   {"qqnorm",   required_argument, 0, 'n'},      // if and how you want to quantile-quantile normalize before returning data
   {"CbyC", no_argument,       &CbyC_flag, 1},
   /* These options don't set a flag.
      We distinguish them by their indices. */
   {"cel",   required_argument, 0, 'c'},      // a file listing the CEL files
   {"cel-dir",   required_argument, 0, 'd'},  // the path to the directory containing the CEL files
   {"prb",   required_argument, 0, 'p'},      // the genomic location of the probes to keep
   {"output",   required_argument, 0, 'o'},      // the genomic location of the probes to keep
   {0, 0, 0, 0}
};
 
static int
parse_argument (int argc, char *argv[])
{
	int c;
     
   while (1)
     {
      
       /* getopt_long stores the option index here. */
       int option_index = 0;
 
       c = getopt_long (argc, argv, "c:d:o:n:p:u:",
                        long_options, &option_index);
 
       /* Detect the end of the options. */
       if (c == -1)
         break;
         
       switch (c)
         {
         case 0:
           /* If this option set a flag, do nothing else now. */
           if (long_options[option_index].flag != 0)
             break;
           printf ("option %s", long_options[option_index].name);
           if (optarg)
             printf (" with arg %s", optarg);
           printf ("\n");
           break;
 
         case 'c':
           CELFILENAMES = gdl_string_clone (optarg);
           break;
         case 'd':
           CELFILEDIR = gdl_string_clone (optarg);
           break;
         case 'p': 
      	   PROBEFILE = gdl_string_clone (optarg);
      	   break;
      	 case 'n': 
      	   QQNORM = gdl_string_clone (optarg);
      	   qqnorm_flag = 1;
      	   break;
      	 case 'o':
           OUTPUT = gdl_string_clone (optarg);
           break;
         case 'u':
           UNIQNESS =  atoi (optarg);
           break;
         case '?':
           GDL_ERROR_VAL ("Unknown arguments", GDL_EINVAL, -1);
         default:
           GDL_ERROR_VAL ("Bad arguments", GDL_EINVAL, -1);
         }
     }
   /* Print any remaining command line arguments (not options). 
   if (optind < argc)
     {
       printf ("non-option ARGV-elements: ");
       while (optind < argc)
         printf ("%s ", argv[optind++]);
       putchar ('\n');
     }*/
}

static int
check_argument (void)
{
	if (CELFILENAMES == 0)
	{
		GDL_ERROR_VAL ("No CEL location file defined",
		               GDL_FAILURE,
		               1);
	}
	if (PROBEFILE == 0)
	{
		GDL_ERROR_VAL ("No probe file defined",
		               GDL_FAILURE,
		               1);	
	}
	if (CELFILEDIR == 0)
	{
		fprintf (stderr, "--\nWARNING : no CEL dir defined: set to current directory\n--\n");
		CELFILEDIR = ".";
	}
	if (qqnorm_flag  && \
		( 	(strcmp(QQNORM, "qq") != 0) && \
			(strcmp(QQNORM, "lqq") != 0) && \
			(strcmp(QQNORM, "qql") !=0 ) )\
		)
	{
		GDL_ERROR_VAL ("No or wrong normalization type defined for -n option",
		               GDL_FAILURE,
		               1);
	}
	return GDL_SUCCESS;
}

static int
help (void)
{
   print_header (PROGRAM);
   printf ("\n");
   printf ("*These options set a flag:\n");
   printf ("\n");
   printf ("--help\tDisplay a brief help on program usage\n");
   printf ("--verbose\tOutput message on standard output to see what the program is doing\n");
   printf ("\n");
   printf ("--no-rotate\tNo rotation of the x,y coordinates of the array\n");
   printf ("--CbyC\tExport data chromosome by chromosome in the -o directory\n");
   printf ("\n");
   printf ("*These options require arguments:\n");
   printf ("\n");
   printf ("--cel or -c\tA file listing the name of the CEL files to consider\n");
   printf ("--cel-dir or -d\tThe path to the directory containing the CEL files\n");
   printf ("--prb or -p\tThe file which gives the genomic and x,y location of the probes\n");
   printf ("--output or -o\tThe output file or directory\n");
   printf ("--qqnorm or -n\tQuantile-Quantile normalization. 'qq' = qqnorm, 'lqq' = log2 transformation followed by qqnorm, 'qql' = qqnorm followed by log2 transformation\n");
   printf ("\n");
}

typedef struct
{
	size_t x;
	size_t y;
	long   pos;	
} probe;

probe *
probe_sscanf (const gdl_string * line)
{
	size_t i,j,n=strlen(line);
	gdl_string * tok;
	probe * p;
	
	p = GDL_MALLOC (probe, 1);
	
	i=j=0;
	// chromosome name
	tok = gdl_string_next_token (line, n, &i, &j);
	gdl_string_free (tok);
	// position
	tok = gdl_string_next_token (line, n, &i, &j);
	p->pos = atol (tok);
	gdl_string_free (tok);
	// x
	tok = gdl_string_next_token (line, n, &i, &j);
	p->x = atoi (tok);
	gdl_string_free (tok);
	// y
	tok = gdl_string_next_token (line, n, &i, &j);
	p->y = atoi (tok);
	gdl_string_free (tok);
	
	return p;
}

typedef struct
{
	size_t size;
	gdl_string * name;
	probe      ** probes;
} probe_set_chrom;

probe_set_chrom *
probe_set_chrom_alloc (const gdl_string * name)
{
	probe_set_chrom * c;
	
	c = GDL_CALLOC (probe_set_chrom, 1);
	
	c->name = gdl_string_clone (name);
	
	return c;
}

void
probe_set_chrom_init (probe_set_chrom * c, gdl_list * buf)
{
	size_t i,size;
	gdl_list_itr * itr;
	
	size=gdl_list_size(buf);
	
	c->size   = size;
	c->probes = GDL_MALLOC (probe *, size);
	
	itr = gdl_list_iterator_front (buf);
	i=0;
	do
	{
		c->probes[i] = (probe *)gdl_list_iterator_value(itr);
		i++;
	}
	while(gdl_list_iterator_next(itr));
	gdl_list_iterator_free (itr);
}

typedef struct
{
	size_t size;
	probe_set_chrom ** chroms;
} probe_set_genome;

probe_set_genome *
probe_set_genome_alloc (gdl_list * buf)
{
	size_t i,size;
	gdl_list_itr * itr;
	probe_set_genome * g;	
	
	g = GDL_MALLOC (probe_set_genome, 1);
	
	size=gdl_list_size(buf);
	
	g->size   = size;
	g->chroms = GDL_MALLOC (probe_set_chrom *, size);
	
	itr = gdl_list_iterator_front (buf);
	i=0;
	do
	{
		g->chroms[i] = (probe_set_chrom *)gdl_list_iterator_value(itr);
		i++;
	}
	while(gdl_list_iterator_next(itr));
	gdl_list_iterator_free (itr);
	
	return g;
}

probe_set_genome *
probe_set_genome_read (const gdl_string * filename)
{
	size_t i,j,k,l,n;
	double m;
	gdl_string * line, * tok;
	FILE * stream;
	probe_set_chrom * chrom=0;
	gdl_list * genome_buf;
	gdl_list * chrom_buf;
	probe * prb;
		
	
	stream     = gdl_fileopen (filename, "r");
	genome_buf = gdl_list_alloc (gdl_list_default);
	line=0;
	gdl_getline (&line, &n, stream);//skip header
	gdl_string_free (line);line=0;
   while(gdl_getline (&line, &n, stream)!=-1)
   {
   	i=j=0;
   	tok=gdl_string_next_token (line, n, &i, &j);
   	if (!chrom || strcmp (tok, chrom->name))
   	{
   		if (chrom)
   		{
   			probe_set_chrom_init (chrom, chrom_buf);
   			fprintf(stderr, "Chromosome %s : %d probes [ OK ]\n", chrom->name, chrom->size);
   			fflush(stderr);
   			gdl_list_push_back (genome_buf, chrom, 0);
   			gdl_list_free(chrom_buf);
   		}
   		chrom = probe_set_chrom_alloc(tok);
   		chrom_buf = gdl_list_alloc (gdl_list_default);
   	}
   	gdl_string_free (tok);
   	prb = probe_sscanf(line);
   	gdl_list_push_back (chrom_buf, prb, 0);
   	gdl_string_free (line);line=0;
   }
   gdl_fileclose (filename, stream);
   // last chrom
   probe_set_chrom_init (chrom, chrom_buf);
	fprintf(stderr, "Chromosome %s : %d probes [ OK ]\n", chrom->name, chrom->size);
	fflush(stderr);
	gdl_list_push_back (genome_buf, chrom, 0);
	gdl_list_free(chrom_buf);

	probe_set_genome * genome;
	
	genome = probe_set_genome_alloc (genome_buf);

	gdl_list_free (genome_buf);
	
	return genome;	
}

typedef struct
{
	gdl_string * filename;
	gdl_string * expname;
} array_description;

array_description *
array_description_alloc (gdl_string * filename, gdl_string * expname)
{
	array_description * a;
	
	a = GDL_MALLOC (array_description, 1);
	
	a->expname  = expname;
	a->filename = filename;
	
	return a;	
} 

typedef struct
{
	size_t nX;
	size_t nY;
	double * data;
} array_data;

array_data *
array_data_alloc (const size_t nX, const size_t nY)
{
	array_data * d;
	
	d = GDL_MALLOC (array_data, 1);
	
	d->nX = nX;
	d->nY = nY;
	d->data = GDL_MALLOC (double, nX*nY);
	
	return d;
}

void
array_data_set (array_data * d, const size_t x, const size_t y, const double v)
{
	d->data[y*d->nY+x]=v;
}

double
array_data_get (const array_data * d, const size_t x, const size_t y)
{
	return d->data[y*d->nY+x];
}

array_data *
array_data_read (const gdl_string * filename)
{
	size_t i,j,k,l,n,nX=0,nY=0,x,y;
	double m;
	gdl_string * line, * tok;
	FILE * stream;
	array_data * array=0;
	
	stream = gdl_fileopen (filename, "r");
	// first, get the number of arrays
	line=0;
   while(gdl_getline (&line, &n, stream)!=-1)
   {
   	if (n >= 8 && !strncmp(line, "[HEADER]", 8))
   	{
   		for(i=0;i<3;i++)
   		{
	   		gdl_string_free (line);line=0;
	   		gdl_getline (&line, &n, stream);
   		}
   		for(i=0;i<n && line[i]!='=';i++);
   		nX = atoi(&line[i+1]);
   		gdl_string_free (line);line=0;
	   	gdl_getline (&line, &n, stream);
	   	for(i=0;i<n && line[i]!='=';i++);
   		nY = atoi(&line[i+1]);
   		// printf ("nX %d nY %d\n", nX, nY);
   	}
   	else if ((nX && nY) && (n >= 11 && !strncmp(line, "[INTENSITY]", 11)))
   	{
   		// skip the two first lines
   		for(i=0;i<3;i++)
   		{
	   		gdl_string_free (line);line=0;
	   		gdl_getline (&line, &n, stream);
   		}
   		array = array_data_alloc (nX, nY);
   		for(i=0;i<nX;i++)
   		{
		   	for(j=0;j<nY;j++)
		   	{
			   	k=l=0;
			   	tok = gdl_string_next_token (line, n, &k, &l);
			   	x   = (size_t) atoi(tok); gdl_string_free (tok);
			   	tok = gdl_string_next_token (line, n, &k, &l);
			   	y   = (size_t) atoi(tok); gdl_string_free (tok);
			   	tok = gdl_string_next_token (line, n, &k, &l);
			   	m   = (double) atof(tok); gdl_string_free (tok);
			   	//printf ("%d %d %g\n", x, y, m);
			   	array_data_set (array, x, y, m);
			   	gdl_string_free (line);line=0;
	   			gdl_getline (&line, &n, stream);
		   	} 
   		}
   	}
   	gdl_string_free (line);line=0;
   	if (array) break;
   }
	
	gdl_fileclose (filename, stream);
	
	return array;	
	
}

typedef struct
{
	size_t size;
	array_description ** descs;
	array_data ** arrays;
} array_design;

array_design *
array_design_alloc (const size_t size)
{
	array_design * a;
	
	a = GDL_MALLOC (array_design, 1);
	
	a->size   = size;
	a->descs  = GDL_MALLOC (array_description *, size);
	a->arrays = GDL_MALLOC (array_data *, size);
	
	return a;
}

array_design *
array_design_read (const gdl_string * filename)
{
	size_t i,j,k,n,size;
	gdl_string * line, * tok;
	long offset;
	FILE * stream;
	array_design * design;
	
	stream = gdl_fileopen (filename, "r");
	offset = ftell (stream); // set offset at beginning of stream
	// first, get the number of arrays
	line=0;size=0;
   while(gdl_getline (&line, &n, stream)!=-1)
   {
   	size++;
   	gdl_string_free (line);line=0;
   }
	fseek (stream, offset, SEEK_SET); // back to offset position
	// alloc the design
	design = array_design_alloc (size);
	// read the design
	line=0;k=0;
   while(gdl_getline (&line, &n, stream)!=-1)
   {
   	i=j=0;
   	design->descs[k++] = array_description_alloc (gdl_string_next_token (line, n, &i, &j), gdl_string_next_token (line, n, &i, &j));
   	gdl_string_free (line);line=0;
   }
	gdl_fileclose (filename, stream);
	
	return design;
}

int
array_design_import (array_design * design, const gdl_string * cel_dir)
{
	size_t i;
	gdl_string * url;
	
	for(i=0;i<design->size;i++)
	{
		url = gdl_string_sprintf ("%s/%s", cel_dir, design->descs[i]->filename);
		fprintf (stderr, "For experiment %s:\n  Read data from CEL file %s:wait...", design->descs[i]->expname, design->descs[i]->filename);
		design->arrays[i] = array_data_read (url);
		fprintf (stderr, "\b\b\b\b\b\b\b [ OK ]\n");
		gdl_string_free (url);
	}
	
	return GDL_SUCCESS;
} 

int
export_data (const gdl_string *output, const array_design * design, const probe_set_genome * probe_set, const gdl_boolean CbyCmode)
{
	size_t i,j,k;
	FILE * stream;
	
	if (!CbyCmode)
	{
		stream = gdl_fileopen(output, "w");
		// prine the header
		fprintf (stream, "chromosome\tposition");
		for(i=0;i<design->size;i++) fprintf (stream, "\t%s", design->descs[i]->expname);
		fprintf (stream, "\n");
		fflush(stream);
		for(i=0;i<probe_set->size;i++) // for each chromosome
		{
			const probe_set_chrom * chrom = probe_set->chroms[i]; // shortcut pointer
			for(j=0;j<chrom->size;j++) // for each probe
			{
				const probe * prb = chrom->probes[j]; // shortcut pointer
				fprintf (stream, "%s\t%ld", chrom->name, prb->pos);
				for(k=0;k<design->size;k++) // for each experiment
				{
					if (rotate_flag)
						fprintf (stream, "\t%e", array_data_get(design->arrays[k], prb->y, design->arrays[k]->nX-1-prb->x));
					else
						fprintf (stream, "\t%e", array_data_get(design->arrays[k], prb->x, prb->y));
				}	
				fprintf (stream, "\n");
				fflush(stream);
			}
		}
		gdl_fileclose(output, stream);
		return GDL_SUCCESS;
	}
	else
	{
		DIR *dir;
		gdl_string * outpfile;
		
		if (output && (dir=opendir(output))==0) /* if cannot open then create */
		{
			mkdir (output, S_IRWXU);
		}
		
		for(i=0;i<probe_set->size;i++) // for each chromosome
		{
			const probe_set_chrom * chrom = probe_set->chroms[i]; // shortcut pointer
			outpfile = gdl_string_sprintf("%s/%s.txt", output, chrom->name);
			stream = gdl_fileopen(outpfile, "w");
			// prine the header
			fprintf (stream, "chromosome\tposition");
			for(j=0;j<design->size;j++) fprintf (stream, "\t%s", design->descs[j]->expname);
			fprintf (stream, "\n");
			fflush(stream);
			for(j=0;j<chrom->size;j++) // for each probe
			{
				const probe * prb = chrom->probes[j]; // shortcut pointer
				fprintf (stream, "%s\t%ld", chrom->name, prb->pos);
				for(k=0;k<design->size;k++) // for each experiment
				{
					if (rotate_flag)
						fprintf (stream, "\t%e", array_data_get(design->arrays[k], prb->y, design->arrays[k]->nX-1-prb->x));
					else
						fprintf (stream, "\t%e", array_data_get(design->arrays[k], prb->x, prb->y));
				}	
				fprintf (stream, "\n");
				fflush(stream);
			}
			gdl_fileclose(outpfile, stream);
		}
		return GDL_SUCCESS;	
	}
}

int
qqnorm (array_design * design)
{
  size_t i,j,r;
  const size_t nX = design->arrays[0]->nX;
  const size_t nY = design->arrays[0]->nY;
  const size_t n = nX*nY;
  double thisrank;
  double mean;
  
  size_t ** sorted_index = GDL_MATRIX_ALLOC(size_t , design->size, n); // sorted_index[0] = index of smallest data value

  // get sorted indexes
  for(i=0;i<design->size;i++) // for each experiment
  {
  	gdl_sort_index(sorted_index[i], design->arrays[i]->data, 1, n);
  }
  
  for(r=0;r<n;r++) // for each rank
  {
  	// get values of that rank from each experiment, and compute their mean
  	
  	for(i=0, thisrank=0;i<design->size;i++) 
  		thisrank += design->arrays[i]->data[ sorted_index[i][r] ];
  	thisrank /= design->size;
  	// replace
  	for(i=0;i<design->size;i++) design->arrays[i]->data[ sorted_index[i][r] ] = thisrank;
  }
  // print
//  for(i=0;i<design->size;i++)
//  {
//     for(j=0;j<n;j++)
//       printf("%.1f\t", design->arrays[i]->data[j]);
//   printf("\n");
//  }
  
  GDL_MATRIX_FREE(sorted_index, design->size);
  return GDL_SUCCESS;
}

int
apply_log2 (array_design * design)
{
	size_t i,j;
	const size_t nX = design->arrays[0]->nX;
    const size_t nY = design->arrays[0]->nY;
    const size_t n = nX*nY;
  
	for(i=0;i<design->size;i++) 
	for(j=0;j<n;j++)
		design->arrays[i]->data[j] = log(design->arrays[i]->data[j])/M_LN2;
	
	return(GDL_SUCCESS);
}
	
int
main (int argc, char *argv[])
{
  int status;
  
  parse_argument (argc, argv);
  
  if (help_flag)
  {
  	  exit (help());
  }
  
  status = check_argument ();
  
  if (status == GDL_SUCCESS)
  {
  	     fprintf(stdout, "--\n");
		 fprintf(stdout, " extractAffy\n");
		 fprintf(stdout, "--\n");
		 fprintf(stdout, " CEL description from file [ %s ]\n", CELFILENAMES);
		 fprintf(stdout, " CEL directory             [ %s ]\n", CELFILEDIR);
		 fprintf(stdout, " Probe file                [ %s ]\n", PROBEFILE);
		 fprintf(stdout, "--\n");
		 
		 probe_set_genome * probe_set;
		 
		 fprintf(stdout, " Read probe set\n");
		 fprintf(stdout, "--\n");
		 probe_set = probe_set_genome_read (PROBEFILE);
		 if (!probe_set)
		 {
		 	  fprintf (stderr, "Unable to read probe set from [ %s ]\n", PROBEFILE);
		 	  GDL_ERROR_VAL ("IO file error", GDL_FAILURE, 1);
		 }
		 
		 array_design * design;
		 
		 fprintf(stdout, "--\n");
		 fprintf(stdout, " Import data from .CEL files\n");
		 fprintf(stdout, "--\n");
		 design = array_design_read (CELFILENAMES);
		 if (!design)
		 {
		 	  fprintf (stderr, "Unable to read the array desing from [ %s ]\n", CELFILENAMES);
		 	  GDL_ERROR_VAL ("IO file error", GDL_FAILURE, 1);
		 }
		 status = array_design_import (design, CELFILEDIR);
		 if (status != GDL_SUCCESS)
		 {
		 	  fprintf (stderr, "Unable to import data from CEL files in directory  [ %s ]\n", CELFILEDIR);
		 	  GDL_ERROR_VAL ("IO file error", GDL_FAILURE, 1);
		 }
		 if (qqnorm_flag)
		 {
		 	fprintf(stdout, "--\n");
		    fprintf(stdout, " Normalizing data\n");
		 	fprintf(stdout, "--\n");
		 	if ( strcmp(QQNORM, "lqq") == 0) apply_log2(design);
		 	qqnorm(design);
		 	if ( strcmp(QQNORM, "qql") == 0) apply_log2(design);
		 }	 
		 fprintf(stdout, "--\n");
		 fprintf(stdout, " Export data\n");
		 fprintf(stdout, "--\n");
		 FILE * stream;
//		 stream = gdl_fileopen (OUTPUT, "w");
		 export_data (OUTPUT, design, probe_set, CbyC_flag);
//		 gdl_fileclose (OUTPUT, stream);
		 
  }
  
  exit(0);
}
