/*  
 *  NMhmmvit.c
 * 
 *  $Author: gael $, $Date: 2008-09-15 12:32:58 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_list.h>
#include <gdl/gdl_machine.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_runtime.h>

#include "nucleoMiner.h"

#include "const.h"
#include "header.c"
#include "util.c"

#include "tiling_array.h"
#include "tiling_hmm.h"
#include "tiling_nucleo.h"

static gdl_string * PROGRAM = "NMhmmvit";

static int help_flag       = 0;
static int verbose_flag    = 0;
static int rm_outlier_flag = 0;
static int print_flag = 0;
static int heterogeneous_flag = 0;

static gdl_string * INPUT   = NULL;
static gdl_string * RESULT  = NULL;
static gdl_string * OUTPUT  = NULL;
static gdl_string * ATLAS   = NULL;
static size_t NPROBE_MAX    = 38;
static size_t NPROBE_MIN    = 31;
static size_t TILING_SIZE   = 4;
static size_t CHUNK_MIN     = 100;
static size_t REMOVE_MIN    = 50;
static gdl_string * CHROM   = NULL;

static struct option long_options[] =
{
   /* These options set a flag. */
   {"help", no_argument,       &help_flag, 1},
   {"verbose", no_argument,       &verbose_flag, 1},
   {"brief",   no_argument,       &verbose_flag, 0},
   {"rm-outlier",   no_argument,       &rm_outlier_flag, 1},
   {"print",   no_argument,       &print_flag, 1},
   {"heterogeneous",   no_argument,       &heterogeneous_flag, 1},
   /* These options don't set a flag.
      We distinguish them by their indices. */
   {"input",   required_argument, 0, 'i'},        // the data file
   {"result",   required_argument, 0, 'r'},        // the result directory
   {"output",   required_argument, 0, 'o'},       // the output stem
   {"nmax",   required_argument, 0, 'x'},         // maximum number of probes for a nucleosome
   {"nmin",   required_argument, 0, 'n'},         // minimum number of probes for a nucleosome
   {"tiling",   required_argument, 0, 't'},       // tiling physical size
   {"chrom",   required_argument, 0, 'c'},        // the name of the chromosome to focus on
   {"chunk-min",   required_argument, 0, 'k'},      // the convergence threshold for the EM in each window
   {"remove-min",   required_argument, 0, 'm'},      // the convergence threshold for the EM in each window
   {"atlas",   required_argument, 0, 'a'},      // the convergence threshold for the EM in each window
   {0, 0, 0, 0}
};
 
 
static int
parse_argument (int argc, char *argv[])
{
	int c;
     
   while (1)
     {
      
       /* getopt_long stores the option index here. */
       int option_index = 0;
 
       c = getopt_long (argc, argv, "i:o:r:c:x:n:t:m:k:a:",
                        long_options, &option_index);
 
       /* Detect the end of the options. */
       if (c == -1)
         break;
         
       switch (c)
         {
         case 0:
           /* If this option set a flag, do nothing else now. */
           if (long_options[option_index].flag != 0)
             break;
           printf ("option %s", long_options[option_index].name);
           if (optarg)
             printf (" with arg %s", optarg);
           printf ("\n");
           break;
 
         case 'i':
           INPUT = gdl_string_clone (optarg);
           break;
         case 'r':
           RESULT = gdl_string_clone (optarg);
           break;
         case 'o':
           OUTPUT = gdl_string_clone (optarg);
           break;
         case 'a':
           ATLAS = gdl_string_clone (optarg);
           break;
         case 'x':
           NPROBE_MAX = (size_t)atoi(optarg);
           break; 
         case 'n':
           NPROBE_MIN = (size_t)atoi(optarg);
           break;
         case 't':
           TILING_SIZE = atoi (optarg);
           break;
         case 'c':
           CHROM = gdl_string_clone (optarg);
           break; 
         case 'm':
           REMOVE_MIN = atoi (optarg);
           break; 
         case 'k':
           CHUNK_MIN = (size_t)atoi (optarg);
           break;
         case '?':
           GDL_ERROR_VAL ("Unknown arguments", GDL_EINVAL, -1);
         default:
           GDL_ERROR_VAL ("Bad arguments", GDL_EINVAL, -1);
         }
     }
   /* Print any remaining command line arguments (not options). 
   if (optind < argc)
     {
       printf ("non-option ARGV-elements: ");
       while (optind < argc)
         printf ("%s ", argv[optind++]);
       putchar ('\n');
     }*/
}

static int
check_argument (void)
{
	if (INPUT == 0)
	{
		GDL_ERROR_VAL ("No input file defined",
		               GDL_FAILURE,
		               1);
	}
	if (!print_flag && RESULT == 0)
	{
		GDL_ERROR_VAL ("No result directory defined",
		               GDL_FAILURE,
		               1);
	}
	if (!print_flag && OUTPUT == 0)
	{
		GDL_ERROR_VAL ("No output stem defined",
		               GDL_FAILURE,
		               1);	
	}
	
	return GDL_SUCCESS;
}

static int
help (void)
{
   print_header (PROGRAM);
   printf ("\n");
   printf ("*These options set a flag:\n");
   printf ("\n");
   printf ("--help\tDisplay a brief help on program usage\n");
   printf ("--verbose\tOutput message on standard output to see what the program is doing\n");
   printf ("\n");
   printf ("--rm-outlier\tRemove probes with obvious bad estimates\n");
   printf ("--print\tPrint in human-readable format the output of NMhmmfit\n");
   printf ("\n");
   printf ("*These options require arguments:\n");
   printf ("\n");
   printf ("--input or -i\tThe input db file representing the tiling array dataset(s) (see NMtdb)\n");
   printf ("--result or -r\tThe directory containing the files generated by NMhmmfit\n");
   printf ("--output or -o\tThe name/path to the output directory\n");
   printf ("--nmax or -x\tThe maximum number of probes for a well-localized nucleosome\n");
   printf ("--nmin or -n\tThe minimum number of probes for a well-localized nucleosome\n");
   printf ("--tiling or -t\tThe size of the tiling step (e.g 4bp)\n");
   printf ("--chunk-min or -l\tThe minimal size of probe chunks to be included\n");
   printf ("--remove-min or -m\tThe minimal size of...\n");
   printf ("--chrom or -c\tThe name of the chromosome to focus on (optional)\n");
}

static gdl_tiling_chromosome *
gdl_tiling_hmm_nucleo_atlas_chrom (const gdl_tiling_data * atlas_calls, const gdl_string * name)
{
	size_t c;
	for(c = 0; c < atlas_calls->nchrom; c++)
	{
		if (!strcmp(atlas_calls->chroms[c]->name, name))
			return atlas_calls->chroms[c];
	}
	return 0;
} 

int
main (int argc, char *argv[])
{
  int status;
  
  parse_argument (argc, argv);
  
  if (help_flag)
  {
  	  exit (help());
  }
  
  status = check_argument ();
  
  if (status == GDL_SUCCESS)
  {
  		size_t c,w;
  		int status; 
  		gdl_string * file;
  		FILE * stream;
  		DIR * dir;
  		gdl_runtime * runtime;
  		
  		runtime = runtime_setup();
  		
  		 gdl_tiling_data * data, * atlas_genome = 0;;
		 
		 stream = gdl_fileopen (INPUT, "r");
		 data = gdl_tiling_data_fread (stream);
		 gdl_fileclose (INPUT, stream);
		 
		 if (print_flag)
		 {
		 	 gdl_tiling_nucleo_hmm * hmm;
			 hmm = gdl_tiling_nucleo_hmm_alloc (gdl_tiling_data_get_dim (data), NPROBE_MIN, NPROBE_MAX);
			 if (heterogeneous_flag)
		    {
		       hmm->heterogeneous = gdl_true;
		    }
			 // Loop on chromosomes
			 for(c = 0; c < data->nchrom; c++)
			 {
			 	 const gdl_tiling_chromosome * chrom = data->chroms[c];
			 	 if (CHROM && strcmp(CHROM, chrom->name))
			 	 	continue;
			 	 fprintf (stderr, ">%s|nprobe=%d|\n", chrom->name, chrom->nprobe);
			 	 gdl_tiling_nucleo_hmm_chrom_viterbi_start (hmm, chrom, TILING_SIZE, CHUNK_MIN, REMOVE_MIN, rm_outlier_flag, RESULT);
			 	 gdl_tiling_nucleo_hmm_chrom_viterbi_fprintf (stdout, hmm);
			 	 gdl_tiling_nucleo_hmm_chrom_viterbi_end (hmm, chrom, RESULT);
			 }
			 gdl_tiling_nucleo_hmm_free (hmm);
		 }
		 else
		 {
			 if ((dir=opendir(OUTPUT))==0)
				{
					mkdir (OUTPUT, S_IRWXU);
				}
			 
			 if (ATLAS)
			 {
			 	stream       = gdl_fileopen (ATLAS, "r");
			 	atlas_genome = gdl_tiling_data_fread (stream);
			   gdl_fileclose (ATLAS, stream);
			 }
			 
			 gdl_tiling_nucleo_hmm * hmm;
			 
			 hmm = gdl_tiling_nucleo_hmm_alloc (gdl_tiling_data_get_dim (data), NPROBE_MIN, NPROBE_MAX);
			 gdl_tiling_nucleo_hmm_set_logger (hmm, stderr);
			 if (heterogeneous_flag)
		    {
		       hmm->heterogeneous = gdl_true;
		    }
			 // Loop on chromosomes
			 for(c = 0; c < data->nchrom; c++)
			 {
			 	 const gdl_tiling_chromosome * chrom = data->chroms[c];
			 	  if (CHROM && strcmp(CHROM, chrom->name))
			 	 	continue;
			 	 const gdl_tiling_chromosome * atlas = (atlas_genome) ? gdl_tiling_hmm_nucleo_atlas_chrom (atlas_genome, chrom->name) : 0;
			 	 fprintf (stderr, ">%s|nprobe=%d|\n", chrom->name, chrom->nprobe);
			 	 gdl_tiling_nucleo_hmm_chrom_viterbi_start (hmm, chrom, TILING_SIZE, CHUNK_MIN, REMOVE_MIN, rm_outlier_flag, RESULT);
			 	 gdl_tiling_nucleo_hmm_chrom_viterbi_perform (hmm);
			 	 file   = gdl_string_sprintf("%s/%s.nuc",OUTPUT,chrom->name);
			 	 stream = gdl_fileopen (file, "w");
			 	 gdl_tiling_nucleo_hmm_chrom_viterbi_summarize (stream, hmm, atlas);
			 	 gdl_fileclose (file, stream);
			 	 gdl_string_free (file);
			 	 gdl_tiling_nucleo_hmm_chrom_viterbi_end (hmm, chrom, RESULT);
			 }
			 
			 gdl_tiling_nucleo_hmm_free (hmm);
		 }
		 
		 runtime_setdown (runtime);
  }
  
  exit(0);
}

