/*  
 *  NMmpa.c
 * 
 *  $Author: gael $, $Date: 2008-09-15 12:33:00 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_list.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_string.h>
#include <gdl/gdl_hash.h>

#include <nucleoMiner.h>

#include "header.c"

static gdl_string * PROGRAM = "NMmpa";

static int help_flag    = 0;
static int verbose_flag = 0;
static int rotate_flag = 0;

static gdl_string * INPUT  = NULL;
static gdl_string * OUTPUT = NULL;
static gdl_string * CHROM_FILE = NULL;
static size_t ARRAY_SIZE   = 0;

static struct option long_options[] =
 {
   /* These options set a flag. */
   {"help", no_argument,       &help_flag, 1},
   {"verbose", no_argument,       &verbose_flag, 1},
   {"brief",   no_argument,       &verbose_flag, 0},
   {"rotate",   no_argument,       &rotate_flag, 1},
   /* These options don't set a flag.
      We distinguish them by their indices. */
   {"input",   required_argument, 0, 'i'},
   {"output",   required_argument, 0, 'o'},
   {"array-size",   required_argument, 0, 'a'},
   {"chrom-file",   required_argument, 0, 'c'},
   {0, 0, 0, 0}
 };
 
static int
parse_argument (int argc, char *argv[])
{
	int c;
     
   while (1)
     {
      
       /* getopt_long stores the option index here. */
       int option_index = 0;
 
       c = getopt_long (argc, argv, "i:o:a:c:",
                        long_options, &option_index);
 
       /* Detect the end of the options. */
       if (c == -1)
         break;
         
       switch (c)
         {
         case 0:
           /* If this option set a flag, do nothing else now. */
           if (long_options[option_index].flag != 0)
             break;
           printf ("option %s", long_options[option_index].name);
           if (optarg)
             printf (" with arg %s", optarg);
           printf ("\n");
           break;
 
         case 'i':
           INPUT = gdl_string_clone (optarg);
           break;
         case 'o':
           OUTPUT = gdl_string_clone (optarg);
           break;
         case 'a': 
      	   ARRAY_SIZE = atoi(optarg);
      	   break;
         case 'c': 
      	   CHROM_FILE = gdl_string_clone (optarg);
      	   break;
         case '?':
           GDL_ERROR_VAL ("Unknown arguments", GDL_EINVAL, -1);
         default:
           GDL_ERROR_VAL ("Bad arguments", GDL_EINVAL, -1);
         }
     }
   /* Print any remaining command line arguments (not options). 
   if (optind < argc)
     {
       printf ("non-option ARGV-elements: ");
       while (optind < argc)
         printf ("%s ", argv[optind++]);
       putchar ('\n');
     }*/
}

static int
check_argument (void)
{
	if (INPUT == 0)
	{
		GDL_ERROR_VAL ("No input file defined",
		               GDL_FAILURE,
		               1);
	}
	
	return GDL_SUCCESS;
}

static int
help (void)
{
   print_header (PROGRAM);
   printf ("\n");
   printf ("--help\tDisplay a brief help on program usage\n");
   printf ("--verbose\tOutput message on standard output to see what the program is doing\n");
   printf ("\n");
   printf ("--rotate\tRotate the x,y coordinates of the array\n");
   printf ("\n");
   printf ("--input or -i\tThe input file\n");
   printf ("--output or -o\tThe output directory\n");
   printf ("--array-size or -a\tThe number of rows/columns on the array\n");
   printf ("--chrom-file or -c\tA file which lists the sequence ids to work on\n");
}

typedef struct
{
	gdl_string * refseq;   // the refseq id
	long          start;   // the starting position of the hit on the chromosome
	size_t        length;  // the length of the hit
	size_t strand;
} tiling_probe_hit;

typedef struct
{
  gdl_string  * id;
  size_t        length;
  tiling_probe_hit * hits;
} tiling_probe;

tiling_probe *
tiling_probe_alloc (gdl_string  * probe_id, const size_t probe_length)
{
	tiling_probe * p;
	
	p = GDL_MALLOC (tiling_probe, 1);
	
	p->id = probe_id;
	p->length = probe_length;
	p->hits   = GDL_CALLOC (tiling_probe_hit, 2); // forward + reverse
	
	return p;
}

static gdl_hashtable * 
refseq2chrom (const gdl_string * filename)
{
	size_t i,j,n;
	gdl_string * line, * refseq;
	FILE * stream;
	gdl_hashtable * r2c;
	
	r2c = gdl_hashtable_alloc (gdl_string_interface, 0);
	
	stream = gdl_fileopen(filename, "r");
   line   = 0;
   while(gdl_getline (&line, &n, stream)!=-1)
   {
   	i=j=0;
   	refseq=gdl_string_next_token (line, n, &i, &j);
   	gdl_hashtable_add (r2c, refseq, gdl_string_next_token (line, n, &i, &j), 1);
   	gdl_string_free (refseq);
   	gdl_string_free (line);line=0;
   }
   gdl_fileclose(filename, stream);
  
	return r2c;
}
// Assume that the str id of the probes is "X_Y", where
// X,Y are the coordinates of the probe on the chip
// then return ARRAY_SIZE*X+Y+1
size_t
get_probe_idx (const gdl_string * xy_name, size_t * array_size, const gdl_boolean is_rotated)
{
	size_t i;
	size_t x,y,t;
	gdl_string * tmp;
	
	for(i=0;xy_name[i]!='_';i++);
	tmp=gdl_string_alloc(i);
   strncpy(tmp, xy_name, i);
   x=atoi(tmp);
	gdl_string_free (tmp);
	y=atoi(&xy_name[i+1]);
	
	if (is_rotated)
	{
		t = y;
		y = x;
		x = (*array_size)-1-t;
		return y*(*array_size)+x+1;
	}
	else
		return y*(*array_size)+x+1;
}

static gdl_string *
get_outfilename (const gdl_string * odir, const gdl_string * refseq, gdl_hashtable * r2c, const size_t strand)
{
	return gdl_string_sprintf ("%s/%s%c.prb", odir, (r2c) ? gdl_hashtable_lookup (r2c, refseq) : refseq, (strand==1) ? '+' : '-');
}

int
main (int argc, char *argv[])
{
  int status;
  
  parse_argument (argc, argv);
  
  if (help_flag)
  {
  	  exit (help());
  }
  
  status = check_argument ();
  
  if (status == GDL_SUCCESS)
  {
  	  int status; 
  	  long hit_start;
  	  size_t i,j,k,n,probe_length,hit_strand,hit_size,pm_size,pos_qur,unique,probe_idx=0,match;
  	  gdl_string * ofile, * line, * tok, * probe_id, * refseq;
  	  tiling_probe_hit * hit;
  	  gdl_list * hit_buffer;
  	  gdl_list_itr * hit_itr;
  	  gdl_hashtable * r2c=0;
  	  FILE * stream, * ostream;
  	  
  	  DIR * dir;
		
	  if ((dir=opendir(OUTPUT))==0)
	  {
	     mkdir (OUTPUT, S_IRWXU);
	  }
  	  
  	  if (CHROM_FILE)
  	  {
  	  		r2c = refseq2chrom (CHROM_FILE);
  	  }
  	
  	  stream = gdl_fileopen(INPUT, "r");
  	  line       = 0;
  	  hit_buffer = 0;
  	  probe_id   = 0;
  	  pm_size    = 0;
  	  match      = 0;
  	  status = gdl_getline (&line, &n, stream);
  	  do
	  {
	  	   if (line[0]=='>')
	  	   {
	  	   	i=j=0;
	  	   	probe_id=gdl_string_next_token (&line[2], n-2, &i, &j);
	  	   	tok     =gdl_string_next_token (&line[2], n-2, &i, &j);
	  	   	if (!strcmp(tok, "Reverse"))
	  	   		hit_strand=-1;	
	  	   	else
	  	   		hit_strand=1;	  	   		
	  	   	gdl_string_free(tok);
	  	   	for(i=j;i<n-2 && line[i]!='=';i++);
	  	   	i++;
	  	   	probe_length=atoi(&line[i]);
	  	   	// loop on matches...
	  	   	gdl_string_free (line);line=0;
	  	   	if (!hit_buffer)
	  	   		hit_buffer = gdl_list_alloc (gdl_list_default);
	  	   	while((status=gdl_getline (&line, &n, stream)) != -1 && line[0]!='>')
	  	   	{
	  	   		i=j=0;
	  	   		hit = GDL_MALLOC (tiling_probe_hit, 1);
	  	   		hit->strand = hit_strand;
	  	   		hit->refseq = gdl_string_next_token (line, n, &i, &j);
	  	   		tok    = gdl_string_next_token (line, n, &i, &j);
	  	   		hit->start = atol(tok);
	  	   		gdl_string_free (tok);
	  	   		tok = gdl_string_next_token (line, n, &i, &j);
	  	   		pos_qur   = atoi(tok);
	  	   		pos_qur   = (hit_strand==-1) ? probe_length-pos_qur+1 : pos_qur;
	  	   		gdl_string_free (tok);
	  	   		tok = gdl_string_next_token (line, n, &i, &j);
	  	   		hit->length = atoi(tok);
	  	   		gdl_string_free (tok);
	  	   		pm_size += (pos_qur==1 && hit->length==probe_length);
	  	   		gdl_list_push_back (hit_buffer, hit, 0);
	  	   		gdl_string_free (line);line=0;
	  	   	}
	  	   	match++;
	  	   }
	  	   else
	  	   {
	  	   	gdl_string_free (line);line=0;
	  	   	status=gdl_getline (&line, &n, stream);
	  	   }
	  	   if (match==2)
	  	   {
	  	   	hit_size=gdl_list_size (hit_buffer);
	  	   	probe_idx = (ARRAY_SIZE) ? get_probe_idx (probe_id, &ARRAY_SIZE, rotate_flag) : probe_idx+1;
	  	   	if (hit_size)
	  	   	{
	  	   		unique = (pm_size==1) ? ((hit_size==1) ? 0 : 1) : (2 + (pm_size>1));
             	hit_itr = gdl_list_iterator_front (hit_buffer);
	  	   		do
	  	   		{
	  	   			hit     = gdl_list_iterator_value (hit_itr);
	  	   			ofile   = get_outfilename (OUTPUT, hit->refseq, r2c, hit->strand);
	  	   			ostream = gdl_fileopen (ofile, "a");
	  	   			fprintf (ostream, "%s %d %d %d %d\n", probe_id, probe_idx, hit->start, hit->length, unique);
	  	   			fflush(ostream);
	  	   			gdl_fileclose (ofile, ostream);
	  	   			gdl_string_free (ofile);
	  	   			gdl_string_free (hit->refseq);
	  	   			GDL_FREE (hit);
	  	   		}
	  	   		while (gdl_list_iterator_next(hit_itr));
	  	   		gdl_list_iterator_free (hit_itr);
	  	   	}
	  	   	gdl_list_free (hit_buffer);
	  	   	hit_buffer=0;
	  	   	pm_size=0;
	  	   	match=0;
	  	   }
	  	   gdl_string_free (probe_id);
	  	   probe_id=0;	
	  }
	  while(status!=-1);
  }	 
}
