/*  
 *  NMmprb.c
 * 
 *  $Author: gael $, $Date: 2008-10-17 16:07:07 $, $Version$
 *
 *  Libgdl : a C library for statistical genetics
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */

#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_list.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_runtime.h>
#include <gdl/gdl_gfeatures.h>
#include <gdl/gdl_gfeatures_gff3.h>
#include <gdl/gdl_genome_coord.h>

#include "nucleoMiner.h"

#include "const.h"
#include "header.c"
#include "util.c"

#include "tiling_array.h"

static gdl_string * PROGRAM = "NMmprb";

static int help_flag      = 0;
static int verbose_flag   = 0;
static int reference_flag = 0;

static gdl_string * REF_PRB     = NULL;
static gdl_string * QRY_PRB     = NULL;
static gdl_string * OUTPUT       = NULL;
static gdl_string * REF_QRY      = NULL;
static size_t WINDOW = 4;

static struct option long_options[] =
{
   /* These options set a flag. */
   {"help", no_argument,       &help_flag, 1},
   {"verbose", no_argument,       &verbose_flag, 1},
   {"brief",   no_argument,       &verbose_flag, 0},
   /* These options don't set a flag.
      We distinguish them by their indices. */
   {"ref",   required_argument, 0, 'r'},        
   {"qry",   required_argument, 0, 'q'},       
   {"output",   required_argument, 0, 'o'},        
   {"ref2qry",   required_argument, 0, 'c'},
   {"window",   required_argument, 0, 'w'},    
   {0, 0, 0, 0}
};
 
static int
parse_argument (int argc, char *argv[])
{
	int c;
     
   while (1)
     {
      
       /* getopt_long stores the option index here. */
       int option_index = 0;
 
       c = getopt_long (argc, argv, "r:q:c:o:w:",
                        long_options, &option_index);
 
       /* Detect the end of the options. */
       if (c == -1)
         break;
         
       switch (c)
         {
         case 0:
           /* If this option set a flag, do nothing else now. */
           if (long_options[option_index].flag != 0)
             break;
           printf ("option %s", long_options[option_index].name);
           if (optarg)
             printf (" with arg %s", optarg);
           printf ("\n");
           break;
 		
 		 case 'o':
           OUTPUT = gdl_string_clone (optarg);
           break;
         case 'r':
           REF_PRB = gdl_string_clone (optarg);
           break;
         case 'q':
           QRY_PRB = gdl_string_clone (optarg);
           break;
         case 'c':
           REF_QRY  = gdl_string_clone (optarg);
           break;
         case 'w':
           WINDOW  = (size_t) atoi (optarg);
           break;
         case '?':
           GDL_ERROR_VAL ("Unknown arguments", GDL_EINVAL, -1);
         default:
           GDL_ERROR_VAL ("Bad arguments", GDL_EINVAL, -1);
         }
     }
   /* Print any remaining command line arguments (not options). 
   if (optind < argc)
     {
       printf ("non-option ARGV-elements: ");
       while (optind < argc)
         printf ("%s ", argv[optind++]);
       putchar ('\n');
     }*/
}

static int
check_argument (void)
{
	if (REF_PRB == 0 && QRY_PRB == 0)
	{
		GDL_ERROR_VAL ("No reference/query probe file defined",
		               GDL_FAILURE,
		               1);
	}
	if (OUTPUT == 0)
	{
		GDL_ERROR_VAL ("No output file defined",
		               GDL_FAILURE,
		               1);
	}
	return GDL_SUCCESS;
}

static int
help (void)
{
   print_header (PROGRAM);
   printf ("\n");
   printf ("%s keep only the probes that share the same genomic and array coordinates\n", PROGRAM);
   printf ("   between the reference and the query\n");
   printf ("\n");
   printf ("*These options set a flag:\n");
   printf ("\n");
   printf ("--help\tDisplay a brief help on program usage\n");
   printf ("--verbose\tOutput message on standard output to see what the program is doing\n");
   printf ("\n");
   printf ("*These options require arguments:\n");
   printf ("\n");
   printf ("--ref or -r\tThe reference probes\n");
   printf ("--qry or -q\tThe query probes\n");
   printf ("--output or -o\tThe output file\n");
   printf ("--ref2qry or -c\tThe reference-query genomic alignment\n");
   printf ("--window or -w\tThe window around the reference probe position\n");
}

typedef struct
{
	size_t x;
	size_t y;
	long   pos;	
} probe;

probe *
probe_sscanf (const gdl_string * line)
{
	size_t i,j,n=strlen(line);
	gdl_string * tok;
	probe * p;
	
	p = GDL_MALLOC (probe, 1);
	
	i=j=0;
	// chromosome name
	tok = gdl_string_next_token (line, n, &i, &j);
	gdl_string_free (tok);
	// position
	tok = gdl_string_next_token (line, n, &i, &j);
	p->pos = atol (tok);
	gdl_string_free (tok);
	// x
	tok = gdl_string_next_token (line, n, &i, &j);
	p->x = atoi (tok);
	gdl_string_free (tok);
	// y
	tok = gdl_string_next_token (line, n, &i, &j);
	p->y = atoi (tok);
	gdl_string_free (tok);
	
	return p;
}

typedef struct
{
	size_t size;
	gdl_string * name;
	probe      ** probes;
} probe_set_chrom;

probe_set_chrom *
probe_set_chrom_alloc (const gdl_string * name)
{
	probe_set_chrom * c;
	
	c = GDL_CALLOC (probe_set_chrom, 1);
	
	c->name = gdl_string_clone (name);
	
	return c;
}

void
probe_set_chrom_init (probe_set_chrom * c, gdl_list * buf)
{
	size_t i,size;
	gdl_list_itr * itr;
	
	size=gdl_list_size(buf);
	
	c->size   = size;
	c->probes = GDL_MALLOC (probe *, size);
	
	itr = gdl_list_iterator_front (buf);
	i=0;
	do
	{
		c->probes[i] = (probe *)gdl_list_iterator_value(itr);
		i++;
	}
	while(gdl_list_iterator_next(itr));
	gdl_list_iterator_free (itr);
}

typedef struct
{
	size_t size;
	probe_set_chrom ** chroms;
} probe_set_genome;

probe_set_genome *
probe_set_genome_alloc (gdl_list * buf)
{
	size_t i,size;
	gdl_list_itr * itr;
	probe_set_genome * g;	
	
	g = GDL_MALLOC (probe_set_genome, 1);
	
	size=gdl_list_size(buf);
	
	g->size   = size;
	g->chroms = GDL_MALLOC (probe_set_chrom *, size);
	
	itr = gdl_list_iterator_front (buf);
	i=0;
	do
	{
		g->chroms[i] = (probe_set_chrom *)gdl_list_iterator_value(itr);
		i++;
	}
	while(gdl_list_iterator_next(itr));
	gdl_list_iterator_free (itr);
	
	return g;
}

probe_set_genome *
probe_set_genome_read (const gdl_string * filename)
{
	size_t i,j,k,l,n;
	double m;
	gdl_string * line, * tok;
	FILE * stream;
	probe_set_chrom * chrom=0;
	gdl_list * genome_buf;
	gdl_list * chrom_buf;
	probe * prb;
		
	
	stream     = gdl_fileopen (filename, "r");
	genome_buf = gdl_list_alloc (gdl_list_default);
	line=0;
	gdl_getline (&line, &n, stream);//skip header
	gdl_string_free (line);line=0;
   while(gdl_getline (&line, &n, stream)!=-1)
   {
   	i=j=0;
   	tok=gdl_string_next_token (line, n, &i, &j);
   	if (!chrom || strcmp (tok, chrom->name))
   	{
   		if (chrom)
   		{
   			probe_set_chrom_init (chrom, chrom_buf);
   			fprintf(stderr, "Chromosome %s : %d probes [ OK ]\n", chrom->name, chrom->size);
   			fflush(stderr);
   			gdl_list_push_back (genome_buf, chrom, 0);
   			gdl_list_free(chrom_buf);
   		}
   		chrom = probe_set_chrom_alloc(tok);
   		chrom_buf = gdl_list_alloc (gdl_list_default);
   	}
   	gdl_string_free (tok);
   	prb = probe_sscanf(line);
   	gdl_list_push_back (chrom_buf, prb, 0);
   	gdl_string_free (line);line=0;
   }
   gdl_fileclose (filename, stream);
   // last chrom
   probe_set_chrom_init (chrom, chrom_buf);
	fprintf(stderr, "Chromosome %s : %d probes [ OK ]\n", chrom->name, chrom->size);
	fflush(stderr);
	gdl_list_push_back (genome_buf, chrom, 0);
	gdl_list_free(chrom_buf);

	probe_set_genome * genome;
	
	genome = probe_set_genome_alloc (genome_buf);

	gdl_list_free (genome_buf);
	
	return genome;	
}

static void
merge_ref_qry (const probe_set_genome * ref_prb,
               const probe_set_genome * qry_prb,
               const gdl_genome_coord * ref_qry_coord,
               const size_t window,
               FILE * stream)
{
	int qry_idx;
	size_t a,c,p,qp,i;
	long ref_chunk;
	
	for(c=0;c<ref_prb->size;c++)
	{
		const probe_set_chrom * ref_chrom = ref_prb->chroms[c];
		const gdl_chunk_coord_cluster * chunk_cluster = gdl_genome_coord_lookup (ref_qry_coord, "ref", ref_chrom->name);
		if (!chunk_cluster)
		{
			fprintf (stderr, "[ WARNING] NMmprb: no landmark %s in ref to qry coordinates\n", ref_chrom->name);
			fflush (stderr);
			continue;
		}
		ref_chunk=0;
		for(p=0;p<ref_chrom->size;p++)
		{
			long ref_pos = ref_chrom->probes[p]->pos;
			ref_chunk    = gdl_chunk_coord_cluster_goto (chunk_cluster, ref_pos, ref_chunk);
			if (ref_chunk == -1) 
			{
				fprintf (stderr, "[ WARNING] NMmprb: unable to map position %d on landmark %s (unmapped region)\n", ref_pos, ref_chrom->name);
				fflush (stderr);
				ref_chunk=0;
				continue;
			}
			const gdl_string * qry_name = gdl_dictionary_get (chunk_cluster->seqid_dico, "qry", chunk_cluster->chunks[ref_chunk]->qry_id);
			qry_idx = -1;
			for(i = 0; i < qry_prb->size; i++)
			{
				if (!strcmp(qry_prb->chroms[i]->name, qry_name))
				{
					qry_idx = i;
					break;
				}
			}
			if (qry_idx == -1)
			{
				fprintf (stderr, "[ WARNING] NMmprb: no landmark %s in qry data\n", qry_name);
				fflush (stderr);
				continue;
			}
			long qry_pos = gdl_chunk_coord_cluster_map (chunk_cluster, ref_pos, ref_chunk);
			const probe_set_chrom * qry_chrom = qry_prb->chroms[qry_idx];
			for(qp = 0; qp < qry_chrom->size; qp++)
			{
				if (qry_chrom->probes[qp]->pos==-1)
					continue;
				if (fabs((double)(qry_chrom->probes[qp]->pos-qry_pos)) <= window)
				{
					if ((qry_chrom->probes[qp]->x == ref_chrom->probes[p]->x) && (qry_chrom->probes[qp]->y == ref_chrom->probes[p]->y)) 
					{
					   fprintf (stream, "%s %d %d %d\n", ref_chrom->name, ref_chrom->probes[p]->pos, ref_chrom->probes[p]->x, ref_chrom->probes[p]->y);
					   fflush(stream);
					   qry_chrom->probes[qp]->pos=-1;
					   break;
					}
				}
			}
			if (qp == qry_chrom->size)
			{
				fprintf (stderr, "[ WARNING] NMmprb: unable to map position %d on landmark %s (no probe at position %d on qry %s)\n", ref_pos, ref_chrom->name, qry_pos, qry_chrom->name);
				fflush (stderr);
				continue;
			}
		}	
	}
}

static void
merge_test (const gdl_genome_coord * ref_qry_coord)
{
	int qry_idx;
	size_t a,c,p,qp,i;
	long ref_chunk;
	
	const gdl_chunk_coord_cluster * chunk_cluster = gdl_genome_coord_lookup (ref_qry_coord, "ref", "chrIV");
	ref_chunk=0;
	for(p = 892545;p <= 893000; p += 4)
	{
		ref_chunk = gdl_chunk_coord_cluster_goto (chunk_cluster, p, ref_chunk);
		if (ref_chunk == -1) 
		{
			fprintf (stderr, "[ WARNING] NMmprb: unable to map position %d on landmark chrIV (unmapped region)\n", p);
			fflush (stderr);
			ref_chunk=0;
			continue;
		}
		const gdl_string * qry_name = gdl_dictionary_get (chunk_cluster->seqid_dico, "qry", chunk_cluster->chunks[ref_chunk]->qry_id);
		printf ("chrIV %d %d\n", p, gdl_chunk_coord_cluster_map (chunk_cluster, p, ref_chunk));
	}
}

int
main (int argc, char *argv[])
{
  int status;
  
  parse_argument (argc, argv);
  
  if (help_flag)
  {
  	  exit (help());
  }
  
  status = check_argument ();
  
  if (status == GDL_SUCCESS)
  {
  		int status; 
  		gdl_string * file;
  		FILE * stream;
  		gdl_runtime * runtime;
  		probe_set_genome * ref_prb, * qry_prb;
  		gdl_genome_coord * ref_qry_coord = 0;
  		
  		runtime = runtime_setup();
  	
  		ref_qry_coord = gdl_genome_coord_alloc (REF_QRY);
  		
//  		merge_test (ref_qry_coord);
//  		
//  		exit(0);
  		
	    ref_prb = probe_set_genome_read (REF_PRB);
  	
  		qry_prb = probe_set_genome_read (QRY_PRB);
  	
	    stream = gdl_fileopen (OUTPUT, "w");
	 	
	    merge_ref_qry (ref_prb, qry_prb, ref_qry_coord, WINDOW, stream);
	    
	    gdl_fileclose (OUTPUT, stream);
		
	 	runtime_setdown (runtime);
	
	}
   exit(0);
}
