/*
 *  NMnucoc.c
 *
 *  $Author: gael $, $Date: 2008-10-17 16:07:07 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 *
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 */
#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_list.h>
#include <gdl/gdl_machine.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_sort_long.h>
#include <gdl/gdl_runtime.h>
#include <gdl/gdl_vector.h>
#include <gdl/gdl_matrix.h>
#include <gdl/gdl_multireg.h>
#include <gdl/gdl_genome_coord.h>

#include "nucleoMiner.h"

#include "const.h"
#include "header.c"
#include "util.c"

#include "tiling_array.h"
#include "tiling_hmm.h"
#include "tiling_nucleo.h"
#include "align_nucleo2.h"

static gdl_string * PROGRAM = "NMnucoc";

static int help_flag       = 0;
static int verbose_flag    = 0;

static gdl_string * INPUT   = NULL;
static gdl_string * CONFIG  = NULL;
static gdl_string * OUTPUT  = NULL;
static gdl_string * CHROM   = NULL;
static gdl_string * REGION  = NULL;
static size_t TILING        = 4;
static size_t NUCLEO_MIN    = 31;
static size_t WINDOW        = 1000;

static struct option long_options[] =
{
   /* These options set a flag. */
   {"help", no_argument,       &help_flag, 1},
   {"verbose", no_argument,       &verbose_flag, 1},
   {"brief",   no_argument,       &verbose_flag, 0},
   /* These options don't set a flag.
    * We distinguish them by their indices. */
   {"input",   required_argument, 0, 'i'},   // the config file
   {"output",  required_argument, 0, 'o'},
   {"config",  required_argument, 0, 'c'},   // the config file
   {"chrom",   required_argument, 0, 's'},   // the name of the chromosome to focus on
   {"tiling",  required_argument, 0, 't'},
   {"nucmin",  required_argument, 0, 'n'},
   {"window",  required_argument, 0, 'w'},
   {"region",  required_argument, 0, 'r'},
   {0, 0, 0, 0}
};

static int
parse_argument (int argc, char *argv[])
{
	int c;

   while (1)
     {

       /* getopt_long stores the option index here. */
       int option_index = 0;

       c = getopt_long (argc, argv, "i:c:s:t:n:w:o:r:",
                        long_options, &option_index);

       /* Detect the end of the options. */
       if (c == -1)
         break;

       switch (c)
         {
         case 0:
           /* If this option set a flag, do nothing else now. */
           if (long_options[option_index].flag != 0)
             break;
           printf ("option %s", long_options[option_index].name);
           if (optarg)
             printf (" with arg %s", optarg);
           printf ("\n");
           break;

         case 'i':
           INPUT = gdl_string_clone (optarg);
           break;
         case 'c':
           CONFIG = gdl_string_clone (optarg);
           break;
         case 's':
           CHROM = gdl_string_clone (optarg);
           break;
         case 'o':
			OUTPUT = gdl_string_clone (optarg);
			break;
         case 't':
           TILING = (size_t)atoi (optarg);
           break;
         case 'n':
           NUCLEO_MIN = (size_t)atoi (optarg);
           break;
         case 'w':
			WINDOW = (size_t)atoi (optarg);
			break;
         case 'r':
			REGION = gdl_string_clone (optarg);
			break;
         case '?':
           GDL_ERROR_VAL ("Unknown arguments", GDL_EINVAL, -1);
         default:
           GDL_ERROR_VAL ("Bad arguments", GDL_EINVAL, -1);
         }
     }
   /* Print any remaining command line arguments (not options).
   if (optind < argc)
     {
       printf ("non-option ARGV-elements: ");
       while (optind < argc)
         printf ("%s ", argv[optind++]);
       putchar ('\n');
     }*/
}

static int
check_argument (void)
{
	if (CONFIG == 0)
	{
		GDL_ERROR_VAL ("No input file defined",
		               GDL_FAILURE,
		               1);
	}
	if (OUTPUT == 0)
	{
		GDL_ERROR_VAL ("No output file defined",
					   GDL_FAILURE,
					   1);
	}

	return GDL_SUCCESS;
}

static int
help (void)
{
   print_header (PROGRAM);
   printf ("\n");
   printf ("%s implements computes nucleosome density along the aligned genomes\n", PROGRAM);
   printf ("\n");
   printf ("*These options set a flag:\n");
   printf ("\n");
   printf ("--help\tDisplay a brief help on program usage\n");
   printf ("--verbose\tOutput message on standard output to see what the program is doing\n");
   printf ("\n");
   printf ("--print\tOutput a summary of the alignments in plain text file\n");
   printf ("\n");
   printf ("*These options require arguments:\n");
   printf ("\n");
   printf ("--input or -i\tThe input directory (use with --print)\n");
   printf ("--config or -c\tThe input configuration file (see manual)\n");
   printf ("--output or -o\tThe output file\n");
   printf ("--chrom or -s\tThe name of the chromosome to look at\n");
   printf ("--window or -w\tThe sliding window size\n");
   printf ("\n");
   printf ("*These following options must fit the HMM parameters\n");
   printf ("\n");
   printf ("--tiling or -t\tThe tiling step size (in bp, default is 4)\n");
   printf ("--nucmin or -n\tThe min number of probes within a nucleosome (default is 31)\n");
}

typedef struct
{
	gdl_string * name;
	size_t size;
	long * starts;
	long * ends;
} chrom_regions;

static gdl_hashtable *
read_chrom_regions (const gdl_string * filename)
{
	gdl_hashtable * regions;
	size_t i,j,n,b;
	gdl_string * line=0, *tok;
	chrom_regions * chrom_r=0;
	FILE * stream;

	regions = gdl_hashtable_alloc (gdl_hash_default, 0);

    stream = gdl_fileopen (filename, "r");

	line=0;
    while (gdl_getline (&line, &n, stream) != -1)
	{
		i = j = 0;
		gdl_string * chrom_name = gdl_string_next_token (line, n, &i, &j);
		if ((chrom_r=gdl_hashtable_lookup(regions, chrom_name))==0)
		{
			chrom_r = GDL_CALLOC (chrom_regions, 1);
			chrom_r->name   = chrom_name;
			chrom_r->starts = GDL_MALLOC (long, 1);
			chrom_r->ends   = GDL_MALLOC (long, 1);
			gdl_hashtable_add (regions, chrom_name, chrom_r, 0);
		}
		else
		{
			gdl_string_free (chrom_name);
		}
		if (chrom_r->size)
		{
			long * ltmp = GDL_MALLOC (long, chrom_r->size+1);
			memcpy(ltmp, chrom_r->starts, sizeof(long)*chrom_r->size);
			GDL_FREE (chrom_r->starts);
			chrom_r->starts=ltmp;
			long * jtmp = GDL_MALLOC (long, chrom_r->size+1);
			memcpy(jtmp, chrom_r->ends, sizeof(long)*chrom_r->size);
			GDL_FREE (chrom_r->ends);
			chrom_r->ends=jtmp;
		}
		tok = gdl_string_next_token (line, n, &i, &j);
		chrom_r->starts[chrom_r->size] = atol (tok);
		gdl_string_free (tok);
		tok = gdl_string_next_token (line, n, &i, &j);
		chrom_r->ends[chrom_r->size] = atol (tok);
		(chrom_r->size)++;
		gdl_string_free (tok);
		gdl_string_free (line);
		line=0;
	}
	gdl_fileclose (filename, stream);

	return regions;
}

int
main (int argc, char *argv[])
{
  int status;

  parse_argument (argc, argv);

  if (help_flag)
  {
  	  exit (help());
  }

  status = check_argument ();

  if (status == GDL_SUCCESS)
  {
  		size_t i,j_r,j_q,j_c,jj,c,w;
  		int status;
  		gdl_string * file;
  		FILE * stream, * stream2;
  		DIR * dir;
  		gdl_runtime * runtime;
  		gdl_tiling_nucleo_hmm * hmm;
  		nm_align_nucleo_config * conf;
  		gdl_hashtable * regions = 0;

  		runtime = runtime_setup();

		conf = nm_align_nucleo_config_alloc (CONFIG);

		if ((dir=opendir(OUTPUT))==0)
		{
			mkdir (OUTPUT, S_IRWXU);
		}

		if (REGION)
		{
			regions = read_chrom_regions (REGION);
		}

		if (verbose_flag)
		{
			fprintf (stdout, "--\nStep 1: Load Reference HMM files\n--\n");
		 	fflush(stdout);
		}

		nm_align_nucleo_workspace * work = nm_align_nucleo_workspace_alloc (conf);

		if (verbose_flag)
		{
			fprintf (stdout, "--\nStep 2: Load corresponding Query HMM files\n--\n");
		 	fflush(stdout);
		}

		for(c = 0; c < work->nref_chrom; c++)
		{
            if (CHROM && strcmp(CHROM, work->ref_chroms[c]->name))
            {
               continue;
            }
			if (verbose_flag)
			{
				fprintf (stdout, "Process chromosome %s: wait...", work->ref_chroms[c]->name);
			 	fflush(stdout);
			}
			nm_align_nucleo_chromosome * ref_chrom = work->ref_chroms[c];
			/**
		 	 * Here use the alignment to store the query chromosomes. If the query chrom
		 	 * has already been seen, just re-upload it
		 	 */
		 	nm_align_nucleo_workspace_anchor_qry (work, ref_chrom, 3, TILING, 0);

		 	if (verbose_flag)
			{
			 	fprintf (stdout, "\b\b\b\b\b\b\b[ DONE ]\n", work->ref_chroms[c]->name);
			 	fflush(stdout);
			}
		}

		nm_align_nucleo_workspace_init_qry (work);

		/**
		 * Do the calling of the nucleosomes
		 */
		if (verbose_flag)
		{
			fprintf (stdout, "--\nStep 3: Call nucleosomes on the Reference\n--\n");
		 	fflush(stdout);
		}
		for(c = 0; c < work->nref_chrom; c++)
		{
			nm_align_nucleo_chromosome_anchor_nucleo (work->ref_chroms[c], TILING, NUCLEO_MIN, 3, 0);
		}
		if (verbose_flag)
		{
			fprintf (stdout, "--\nStep 4: Call nucleosomes on the Query\n--\n");
		 	fflush(stdout);
		}
		for(c = 0; c < work->nqry_chrom; c++)
		{
			nm_align_nucleo_chromosome_anchor_nucleo (work->qry_chroms[c], TILING, NUCLEO_MIN, 3, 1);
		}
		if (verbose_flag)
		{
			fprintf (stdout, "--\nStep 4: Map Query nucleosomes on the Reference\n--\n");
		 	fflush(stdout);
		}

		nm_align_nucleo_chromosome_map_qry (work, stderr);

		if (verbose_flag)
		{
			fprintf (stdout, "--\nStep 5: Compute occupancy\n--\n");
		 	fflush(stdout);
		}

		stream = gdl_fileopen(OUTPUT, "w");
		fprintf(stream, "chrom window_start window_end window_mid window_weight ref_occupancy qry_occupancy\n");
		fflush(stream);

		for(c = 0; c < work->nref_chrom; c++)
		{
            if (verbose_flag)
			{
				fprintf (stdout, "Process chromosome %s: wait...", work->ref_chroms[c]->name);
			 	fflush(stdout);
			}

			nm_align_nucleo_chromosome * ref_chrom = work->ref_chroms[c];
			chrom_regions * chrom_region = 0;

			if (regions)
			{
				chrom_region = gdl_hashtable_lookup(regions, ref_chrom->name);
			}

			// First create a ordered vector of query nucleosomes
			size_t NQRY_NUC = 0;
			nm_align_nucleo_nucleosome ** qry_nucleosomes = 0;

			if (gdl_list_size (ref_chrom->_mapped_nucleosomes))
			{
				double * qry_pos;
				size_t * qry_idx;
				gdl_list_itr * itr = gdl_list_iterator_front (ref_chrom->_mapped_nucleosomes);

				NQRY_NUC        = gdl_list_size (ref_chrom->_mapped_nucleosomes);
				qry_nucleosomes = GDL_MALLOC (nm_align_nucleo_nucleosome *, NQRY_NUC);

				qry_pos = GDL_MALLOC (double, NQRY_NUC);
				qry_idx = GDL_MALLOC (size_t, NQRY_NUC);

				i=0;
				do
				{
					nm_align_nucleo_nucleosome * qry_nuc = (nm_align_nucleo_nucleosome *) gdl_list_iterator_value (itr);
					qry_nucleosomes[i] = qry_nuc;
					qry_pos[i]         = qry_nuc->ref_start_pos;
					i++;
				}
				while(gdl_list_iterator_next (itr));
				gdl_list_iterator_free (itr);

				gdl_sort_index (qry_idx, qry_pos, 1, NQRY_NUC);

				nm_align_nucleo_nucleosome ** qntmp = GDL_MALLOC (nm_align_nucleo_nucleosome *, NQRY_NUC);
				for(i = 0; i < NQRY_NUC; i++)
				{
					qntmp[i] = qry_nucleosomes[qry_idx[i]];
				}
				GDL_FREE (qry_nucleosomes);

				qry_nucleosomes = qntmp;

				GDL_FREE (qry_pos);
				GDL_FREE (qry_idx);
			}

			// Now loop on ref chrom chunks and compute the occupancy

			for (i = 0; i < ref_chrom->hmm->size; i++)
		    {
				const gdl_tiling_nucleo_chunk * chunk = ref_chrom->hmm->chunks[i];
				long chunk_start  = chunk->probes[0]->position;
				long chunk_end    = chunk->probes[chunk->size-1]->position;
				long chunk_length = (chunk_end-chunk_start+1);
				size_t nwindow;
				if (chunk_length < WINDOW)
				{
					nwindow=1;
				}
				else
				{
					nwindow = chunk_length-WINDOW+1;
				}

				for(w = j_r = j_q = j_c = 0; w < nwindow; w++)
				{
					long window_start = chunk_start  + w;
					long window_end   = GDL_MIN (window_start + WINDOW - 1, chunk_end);
					if (chrom_region)
					{
						size_t ok = 0;
						for(jj = j_c; jj < chrom_region->size; jj++)
						{
							if (window_start >= chrom_region->starts[jj] && window_end <= chrom_region->ends[jj])
							{
								ok=1;
								break;
							}
							else if (window_end <= chrom_region->starts[jj])
							{
								break;
							}
							if (window_start > chrom_region->starts[jj])
							{
								(j_c)++;
							}
						}
						if (!ok)
						{
							continue;
						}
					}
					double win_weight = ((double)(window_end-window_start+1))/WINDOW;
					double ref_occ    = 0;
					double qry_occ    = 0;
					// Ref Occupancy
					for(jj = j_r;jj < ref_chrom->n_nucleosome; jj++)
					{
						long ref_start  = ref_chrom->nucleosomes[jj]->ref_start_pos;
						long ref_end    = ref_chrom->nucleosomes[jj]->ref_end_pos;
						long ref_length = (ref_end-ref_start+1);
						if (ref_start >= window_start && ref_start < window_end)
						{
							if (ref_end <= window_end)
							{
								ref_occ += ref_length;
							}
							else
							{
								ref_occ += window_end-ref_start;
							}
						}
						else if (ref_end < window_end && ref_end >= window_start)
						{
							if (ref_start >= window_start)
							{
								ref_occ += ref_length;
							}
							else
							{
								ref_occ += ref_end-window_start;
							}
						}
						if ((ref_end < window_start+1) && (j_r < ref_chrom->n_nucleosome-1))
						{
							j_r++;
						}
						if (ref_start > window_end)
						{
							break;
						}
					}
					// Qry Occupancy
					for(jj = j_q;jj < NQRY_NUC; jj++)
					{
						long qry_start  = qry_nucleosomes[jj]->ref_start_pos;
						long qry_end    = qry_nucleosomes[jj]->ref_end_pos;
						long qry_length = (qry_end-qry_start+1);
						if (qry_start >= window_start && qry_start < window_end)
						{
							if (qry_end <= window_end)
							{
								qry_occ += qry_length;
							}
							else
							{
								qry_occ += window_end-qry_start;
							}
						}
						else if (qry_end < window_end && qry_end >= window_start)
						{
							if (qry_start >= window_start)
							{
								qry_occ += qry_length;
							}
							else
							{
								qry_occ += qry_end-window_start;
							}
						}
						if ((qry_end < window_start+1) && (j_q < NQRY_NUC-1))
						{
							j_q++;
						}
						if (qry_start > window_end)
						{
							break;
						}
					}
					// output
					fprintf(stream, "%s %d %d %d %g %g %g\n", ref_chrom->name, window_start, window_end, (window_start+window_end)/2, win_weight, ref_occ, qry_occ);
					fflush(stream);
				}
		    }
			GDL_FREE (qry_nucleosomes);
		}

		gdl_fileclose(OUTPUT, stream);

		runtime_setdown (runtime);
  }
  exit(0);
}

