/*
 *  NMt2feat2.c
 *
 *  $Author: gael $, $Date: 2008-09-15 12:32:59 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 *
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 */

#include <math.h>
#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_list.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_runtime.h>
#include <gdl/gdl_gfeatures.h>
#include <gdl/gdl_gfeatures_gff3.h>

#include "nucleoMiner.h"

#include "const.h"
#include "header.c"
#include "util.c"

#include "tiling_array.h"

static gdl_string * PROGRAM = "NMt2feat2";

static int help_flag      = 0;
static int verbose_flag   = 0;
static int print_flag     = 0;
static int start_flag     = 0;
static int end_flag       = 0;
static int upstream_flag  = 0;

static gdl_string * INPUT       = NULL;
static gdl_string * OUTPUT      = NULL;
static gdl_string * FEATURE     = NULL;
static gdl_string * FEATURE_ID  = NULL;
static long WINDOW              = 1000;
static int NON_ZERO             = -1;
static double ZERO_VALUE        = 0;

static struct option long_options[] =
{
   /* These options set a flag. */
   {"help", no_argument,       &help_flag, 1},
   {"verbose", no_argument,       &verbose_flag, 1},
   {"brief",   no_argument,       &verbose_flag, 0},
   {"start",   no_argument,       &start_flag, 1},
   {"end",     no_argument,       &end_flag, 1},
   {"upstream",     no_argument,    &upstream_flag, 1},
   /* These options don't set a flag.
      We distinguish them by their indices. */
   {"input",   required_argument, 0, 'i'},         // the data file
   {"output",   required_argument, 0, 'o'},        // the output file
   {"feature",   required_argument, 0, 'f'},      // genomic features
   {"feature-id",   required_argument, 0, 'l'},      // genomic features
   {"window",   required_argument, 0, 'w'},      // genomic features
   {"non-zero-col",   required_argument, 0, 'c'},      // push only signal if the given column of the array has signal
   {"zero-value",   required_argument, 0, 'z'},      // push only signal if the given column of the array has signal
   {0, 0, 0, 0}
};

static int
parse_argument (int argc, char *argv[])
{
	int c;

   while (1)
     {

       /* getopt_long stores the option index here. */
       int option_index = 0;

       c = getopt_long (argc, argv, "i:o:f:l:w:z:c:",
                        long_options, &option_index);

       /* Detect the end of the options. */
       if (c == -1)
         break;

       switch (c)
         {
         case 0:
           /* If this option set a flag, do nothing else now. */
           if (long_options[option_index].flag != 0)
             break;
           printf ("option %s", long_options[option_index].name);
           if (optarg)
             printf (" with arg %s", optarg);
           printf ("\n");
           break;

         case 'i':
           INPUT = gdl_string_clone (optarg);
           break;
         case 'o':
           OUTPUT = gdl_string_clone (optarg);
           break;
         case 'f':
           FEATURE = gdl_string_clone (optarg);
           break;
         case 'l':
           FEATURE_ID = gdl_string_clone (optarg);
           break;
         case 'w':
           WINDOW = atoi (optarg);
           break;
         case 'c':
           NON_ZERO = atoi (optarg);
           break;
         case 'z':
           ZERO_VALUE = (double)atof (optarg);
           break;
         case '?':
           GDL_ERROR_VAL ("Unknown arguments", GDL_EINVAL, -1);
         default:
           GDL_ERROR_VAL ("Bad arguments", GDL_EINVAL, -1);
         }
     }
   /* Print any remaining command line arguments (not options).
   if (optind < argc)
     {
       printf ("non-option ARGV-elements: ");
       while (optind < argc)
         printf ("%s ", argv[optind++]);
       putchar ('\n');
     }*/
}

static int
check_argument (void)
{
	if (INPUT == 0)
	{
		GDL_ERROR_VAL ("No input file defined",
		               GDL_FAILURE,
		               1);
	}
	if (OUTPUT == 0)
	{
		GDL_ERROR_VAL ("No output stem defined",
		               GDL_FAILURE,
		               1);
	}
	if (FEATURE == 0)
	{
		GDL_ERROR_VAL ("No feature file defined",
		               GDL_FAILURE,
		               1);
	}

	return GDL_SUCCESS;
}

static int
help (void)
{
	print_header (PROGRAM);
	printf ("\n");
    printf ("%s computes tiling signal profiles around genomic feature boundaries\n", PROGRAM);
    printf ("\n");
    printf ("*These options set a flag:\n");
    printf ("\n");
    printf ("--help\tDisplay a brief help on program usage\n");
    printf ("--verbose\tOutput message on standard output to see what the program is doing\n");
    printf ("\n");
    printf ("--start\tLook at the 5'end of the feature\n");
    printf ("--end\tLook at the 3'end of the feature\n");
    printf ("--upstream\tlook only at the upstream region of the feature boundary\n");
    printf ("\n");
    printf ("*These options require arguments:\n");
    printf ("\n");
    printf ("--input or -i\tThe input tiling dataset file (.db, as converted by NMtdb)\n");
    printf ("--output or -o\tThe stem name of the output file(s)\n");
    printf ("--feature or -f\tThe path to the feature file in gff format (version 3)\n");
    printf ("--feature-id or -l\tThe path to the feature file type and/or ids (see manual)\n");
	printf ("--window or -w\tThe window size (in bp) around the each feature boundary\n");
    printf ("--non-zero-col or -c\tRemove features for which the corresponding column in the dataset indicates no signal (see --zero-value)\n");
    printf ("--zero-value or -z\tThe minimal value to claim for no signal\n");
}

void
push_signal (double * signal, const gdl_tiling_chromosome * chrom_nucleo, const size_t p)
{
	size_t i;
	for(i = 0; i < chrom_nucleo->narray; i++)
	{
		signal[i] += GDL_TILING_CHROM_GET (chrom_nucleo, p, i);
	}
	signal[i] += 1;
}

int
print_avg_signal (FILE * stream, double * signal, const gdl_string * which, const size_t NA)
{
	size_t j;

	fprintf (stream, "%s", which);
	for(j = 0; j < NA; j++)
	{
		fprintf (stream, " %g", signal[j]);
	}
	fprintf (stream, " %d", (int)floor(signal[j]));
	fprintf (stream, "\n");
}

int
main (int argc, char *argv[])
{
  int status;

  parse_argument (argc, argv);

  if (help_flag)
  {
  	  exit (help());
  }

  status = check_argument ();

  if (status == GDL_SUCCESS)
  {
  		int status;
  		gdl_string * file;
  		FILE * stream, * matstream;
  		gdl_runtime * runtime;
  		gdl_hashtable * selected_features = 0;

  		runtime = runtime_setup();

	 fprintf(stderr, "--\n");
	 fprintf(stderr, " NMt2feat2\n");
	 fprintf(stderr, "--\n");

	 gdl_tiling_data * nucleo;

	 stream  = gdl_fileopen (INPUT, "r");
	 nucleo  = gdl_tiling_data_fread (stream);
	 gdl_fileclose (INPUT, stream);

	 if (!nucleo)
	 	 GDL_ERROR_VAL ("Unable to read data file", GDL_FAILURE, GDL_FAILURE);

	// Read the feature
	gdl_gfeatures * feature = gdl_gfeatures_alloc (gdl_gfeatures_gff3, FEATURE);

	if (FEATURE_ID)
		selected_features = read_selected_features (FEATURE_ID);

	const size_t NCHROM = gdl_gfeatures_size (feature);
	size_t i,c,f,p,p0,push;
	double * avg_signal;

	const size_t NA = gdl_tiling_data_get_dim (nucleo);

	stream = gdl_fileopen (OUTPUT, "w");
	for (c = 0; c < NCHROM; c++)
	{
		gdl_gfeatures_landmark * chrom_feature = gdl_gfeatures_get (feature, c);
		const int chrom_nucidx = gdl_tiling_data_chrom_lookup (nucleo, chrom_feature->seqid);
		if (chrom_nucidx == -1)
		{
			fprintf (stderr, "Chromosome (landmark) %s is not present in the tiling data set\n", chrom_feature->seqid);
			continue;
		}
		const gdl_tiling_chromosome * chrom_nucleo = nucleo->chroms[chrom_nucidx];
		const size_t NPROBE = chrom_nucleo->nprobe;
		for(p = f = 0; f < chrom_feature->size; f++)
		{
			gdl_gfeatures_gff3_record * rec = (gdl_gfeatures_gff3_record *) chrom_feature->records[f];
			if (is_a_selected_feature (feature, rec, selected_features))
			{
				avg_signal = GDL_CALLOC (double, NA+1);
				if ((start_flag && (rec->strand == '+' || rec->strand == '?')) || (end_flag && rec->strand == '-'))
				{
					for(;p < NPROBE && chrom_nucleo->positions[p] < rec->start - WINDOW;p++);
					p0=p;
					push=1;
					// do we have signal form the selected column
					if (NON_ZERO>=0)
					{
						double z=0;
						for(;p < NPROBE && chrom_nucleo->positions[p] < rec->start + WINDOW;p++)
						{
							z+=GDL_TILING_CHROM_GET (chrom_nucleo, p, NON_ZERO);
						}
						p=p0;
						if (z <= ZERO_VALUE)
							push=0;
					}
					if (push)
					{
						for(;p < NPROBE && chrom_nucleo->positions[p] < rec->start + WINDOW;p++)
						{
							long relative_pos = (rec->strand == '+' || rec->strand == '?') ? chrom_nucleo->positions[p] - rec->end :  rec->end - chrom_nucleo->positions[p];
							if (upstream_flag && relative_pos > 0)
								continue;
							push_signal(avg_signal, chrom_nucleo, p);
						}
					}
					p=p0;
				}
				else if ((end_flag && (rec->strand == '+' || rec->strand == '?')) || (start_flag && rec->strand == '-'))
				{
					for(;p < NPROBE && chrom_nucleo->positions[p] < rec->end - WINDOW;p++);
					p0=p;
					push=1;
					// do we have signal form the selected column
					if (NON_ZERO>=0)
					{
						double z=0;
						for(;p < NPROBE && chrom_nucleo->positions[p] < rec->start + WINDOW;p++)
						{
							z+=GDL_TILING_CHROM_GET (chrom_nucleo, p, NON_ZERO);
						}
						p=p0;
						if (z <= ZERO_VALUE)
							push=0;
					}
					if (push)
					{
						for(;p < NPROBE && chrom_nucleo->positions[p] < rec->end + WINDOW;p++)
						{
							long relative_pos = (rec->strand == '+' || rec->strand == '?') ? chrom_nucleo->positions[p] - rec->end :  rec->end - chrom_nucleo->positions[p];
							if (upstream_flag && relative_pos > 0)
								continue;
							push_signal(avg_signal, chrom_nucleo, p);
						}
					}
					p=p0;
				}
				else if (!end_flag && !start_flag)
				{
					for(;p < NPROBE && chrom_nucleo->positions[p] < rec->start - WINDOW;p++);
					p0=p;
					push=1;
					// do we have signal form the selected column
					if (NON_ZERO>=0)
					{
						double z=0;
						for(;p < NPROBE && chrom_nucleo->positions[p] < rec->end + WINDOW;p++)
						{
							z+=GDL_TILING_CHROM_GET (chrom_nucleo, p, NON_ZERO);
						}
						p=p0;
						if (z <= ZERO_VALUE)
							push=0;
					}
					if (push)
					{
						for(;p < NPROBE && chrom_nucleo->positions[p] < rec->end + WINDOW;p++)
						{
							push_signal(avg_signal, chrom_nucleo, p);
						}
					}
					p=p0;
				}
				const gdl_string * name = 0;
				if (gdl_hashtable_lookup (rec->attributes, "ID")==0)
				{
					name = gdl_hashtable_lookup (rec->parents[0]->attributes, "ID");
				}
				else
				{
					name = gdl_hashtable_lookup (rec->attributes, "ID");
				}
				print_avg_signal (stream, avg_signal, name, NA);
				GDL_FREE (avg_signal);
			}
		}
	}

	gdl_fileclose (file, stream);

	gdl_string_free (file);

	gdl_tiling_data_free (nucleo);

	runtime_setdown (runtime);

  }
  exit(0);
}
