/*  
 *  NMt2featcis.c
 * 
 *  $Author: gael $, $Date: 2008-09-15 12:32:59 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_list.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_runtime.h>
#include <gdl/gdl_gfeatures.h>
#include <gdl/gdl_gfeatures_gff3.h>

#include "nucleoMiner.h"

#include "const.h"
#include "header.c"
#include "util.c"

#include "tiling_array.h"

static gdl_string * PROGRAM = "NMt2featcis";

static int help_flag      = 0;
static int verbose_flag   = 0;
static int print_flag     = 0;
static int square_flag    = 0;
//static int no_within_flag = 0;
static int no_overlap_flag = 0;

static gdl_string * INPUT       = NULL;
static gdl_string * OUTPUT      = NULL;
static gdl_string * FEATURE     = NULL;
static gdl_string * FEATURE_ID  = NULL;
static gdl_string * EXTRACT_C   = NULL;
static long UPSTREAM_W          = 10000;
static long DOWNSTREAM_W        = 10000;
static long DOWNSTREAM_N        = 100;
static long UPSTREAM_N          = 100;
static size_t INSIDE            = 10;
static int COL_FROM             = -1;
static int COL_TO               = -1;
static double COL_THRESH        = 0;

static struct option long_options[] =
{
   /* These options set a flag. */
   {"help", no_argument,            &help_flag, 1},
   {"verbose", no_argument,         &verbose_flag, 1},
   {"brief",   no_argument,         &verbose_flag, 0},
   {"square",  no_argument,      &square_flag, 1},
   //{"within",  no_argument,   &no_within_flag, 1},
   {"no-overlap", no_argument,  &no_overlap_flag, 1},
   /* These options don't set a flag.
      We distinguish them by their indices. */
   {"input",   required_argument, 0, 'i'},          // the data file
   {"output",   required_argument, 0, 'o'},         // the output file
   {"feature",   required_argument, 0, 'f'},        // genomic features
   {"feature-id",   required_argument, 0, 'l'},     // genomic features
   {"up-window",   required_argument, 0, 'u'},      // genomic features
   {"down-window",   required_argument, 0, 'd'},    // genomic features
   {"up-nclass",   required_argument, 0, 'p'},      // genomic features
   {"down-nclass",   required_argument, 0, 'w'},    // genomic features
   {"inside",   required_argument, 0, 's'},         // push only signal if the given column of the array has signal
   {"col-from",   required_argument, 0, 'j'},       // push only signal if the given column of the array has signal
   {"col-to",   required_argument, 0, 'k'},         // push only signal if the given column of the array has signal
   {"col-thresh",   required_argument, 0, 't'},         // push only signal if the given column of the array has signal
   {"extract",   required_argument, 0, 'e'},         // push only signal if the given column of the array has signal
   {0, 0, 0, 0}
};
 
static int
parse_argument (int argc, char *argv[])
{
	int c;
     
   while (1)
     {
      
       /* getopt_long stores the option index here. */
       int option_index = 0;
 
       c = getopt_long (argc, argv, "i:o:f:l:u:d:p:w:s:j:k:t:e:",
                        long_options, &option_index);
 
       /* Detect the end of the options. */
       if (c == -1)
         break;
         
       switch (c)
         {
         case 0:
           /* If this option set a flag, do nothing else now. */
           if (long_options[option_index].flag != 0)
             break;
           printf ("option %s", long_options[option_index].name);
           if (optarg)
             printf (" with arg %s", optarg);
           printf ("\n");
           break;
 
         case 'i':
           INPUT = gdl_string_clone (optarg);
           break;
         case 'o':
           OUTPUT = gdl_string_clone (optarg);
           break;
         case 'f':
           FEATURE = gdl_string_clone (optarg);
           break;
         case 'l':
           FEATURE_ID = gdl_string_clone (optarg);
           break;
         case 'u':
           UPSTREAM_W = atol (optarg);
           break;
         case 'd':
           DOWNSTREAM_W = atol (optarg);
           break;
         case 'p':
           UPSTREAM_N = (size_t)atoi (optarg);
           break;
         case 'w':
           DOWNSTREAM_N = (size_t)atoi (optarg);
           break;
         case 's':
           INSIDE = (size_t)atoi (optarg);
           break;
         case 'j':
           COL_FROM = atoi (optarg);
           break;
         case 'k':
           COL_TO = atoi (optarg);
           break;
         case 't':
           COL_THRESH = (double)atof (optarg);
           break;
         case 'e':
           EXTRACT_C = gdl_string_clone (optarg);
           break;
         case '?':
           GDL_ERROR_VAL ("Unknown arguments", GDL_EINVAL, -1);
         default:
           GDL_ERROR_VAL ("Bad arguments", GDL_EINVAL, -1);
         }
     }
   /* Print any remaining command line arguments (not options). 
   if (optind < argc)
     {
       printf ("non-option ARGV-elements: ");
       while (optind < argc)
         printf ("%s ", argv[optind++]);
       putchar ('\n');
     }*/
}

static int
check_argument (void)
{
	if (INPUT == 0)
	{
		GDL_ERROR_VAL ("No input file defined",
		               GDL_FAILURE,
		               1);
	}
	if (OUTPUT == 0)
	{
		GDL_ERROR_VAL ("No output stem defined",
		               GDL_FAILURE,
		               1);
	}
	if (FEATURE == 0)
	{
		GDL_ERROR_VAL ("No feature file defined",
		               GDL_FAILURE,
		               1);
	}
	
	
	return GDL_SUCCESS;
}

static int
help (void)
{
	print_header (PROGRAM);
	print_header (PROGRAM);
	printf ("\n");
    printf ("%s computes tiling signal profiles in cis-region of genomic features\n",  PROGRAM);
    printf ("\n");
    printf ("*These options set a flag:\n");
    printf ("\n");
    printf ("--help\tDisplay a brief help on program usage\n");
    printf ("--verbose\tOutput message on standard output to see what the program is doing\n");
    printf ("\n");
    printf ("--no-overlap\tIgnore overlapping cis-regions\n");
    printf ("--square\tOutput also the (signal) sum of square\n");
    printf ("\n");
    printf ("*These options require arguments:\n");
    printf ("\n");
    printf ("--input or -i\tThe input tiling dataset file (.db, as converted by NMtdb)\n");
    printf ("--output or -o\tThe stem name of the output file(s)\n");
    printf ("--feature or -f\tThe path to the feature file in gff format (version 3)\n");
    printf ("--feature-id or -l\tThe path to the feature file type and/or ids (see manual)\n");
	printf ("--up-window or -u\tThe window size (in bp) of the upstream region\n");    
    printf ("--down-window or -d\tThe window size (in bp) of the dowstream region\n");
    printf ("--up-nclass or -p\tThe number of bins into the upstream region will be split\n");
	printf ("--down-nclass or -w\tThe number of bins into the downstream region will be split\n");
    printf ("--inside or -s\tThe number of bins used to split the feature region\n");
    printf ("--col-from or -j\tThe first column (from 0) from which threshold has to be computed\n");
    printf ("--col-to or -k\tThe last column (from 0) from which threshold has to be computed\n");
    printf ("--col-thresh or -t\tThe minimal avg. value to consider the feature\n");   
}

static int
get_bin (const unsigned char strand,
         const long start,
         const long end,
         const long position,
         const long upstream_w, 
         const size_t upstream_n,
         const size_t inside_n,
         const long downstream_w, 
         const size_t downstream_n)
{
	if (strand == '+' || strand == '?')
	{
		//printf("%d %d %d\n", start, end, position);
		if (position < start)
		{
			double t = position-(start-upstream_w);
			if (t < 0) return -1;
			t /= upstream_w;
			t *= upstream_n;
			return (t == (double)upstream_n) ? upstream_n-1 : floor(t);
		}
		else if (position > end)
		{
			double t = position-end;
			if (t > (double)downstream_w) return -1;
			t /= downstream_w;
			t *= downstream_n;
			return upstream_n + inside_n + ((t == (double)downstream_n) ? downstream_n-1 : floor(t));
		}
		else
		{
			double t = position-start;
			t /= (end-start+1);
			t *= inside_n;
			return upstream_n + ((t == (double)inside_n) ? inside_n-1 : floor(t));
		}
	}
	else
	{
		if (position > end)
		{
			double t = (end+upstream_w)-position;
			if (t < 0) return -1;
			t /= upstream_w;
			t *= upstream_n;
			return (t == (double)upstream_n) ? upstream_n-1 : floor(t);
		}
		else if (position < start)
		{
			double t = start-position;
			if (t > (double)downstream_w) return -1;
			t /= downstream_w;
			t *= downstream_n;
			return upstream_n + inside_n + ((t == (double)downstream_n) ? downstream_n-1 : floor(t));
		}
		else
		{
			double t = end-position;
			t /= (end-start+1);
			t *= inside_n;
			return upstream_n + ((t == (double)inside_n) ? inside_n-1 : floor(t));
		}
	} 
	return -1;
}

void
push_signal (double ** signal,
             const int bin,
             const gdl_tiling_chromosome * chrom_nucleo,
             const size_t p,
             const size_t square_flag)
{
	size_t i,j;
	if (bin==-1) return;
	for(i = 0; i < chrom_nucleo->narray; i++)
	{
		signal[bin][i] += GDL_TILING_CHROM_GET (chrom_nucleo, p, i);
	}
	j=i;
	if (square_flag)
	{
		double t;
		for(i = 0; i < chrom_nucleo->narray; i++, j++)
		{
			t = GDL_TILING_CHROM_GET (chrom_nucleo, p, i);
			signal[bin][j] += t*t;
		}	
	}
	signal[bin][j] += 1;
}

static int
print_cis_signal (FILE * stream,
			         double ** signal, 
			         const size_t signal_size,
			         const size_t ncis,
			         const long upstream_w, 
			         const size_t upstream_n,
			         const size_t inside_n,
			         const long inside_t,
			         const long downstream_w, 
			         const size_t downstream_n,
			         const size_t square_flag)
{
	size_t i, j, k;
	long pos = 0;
	
	long delta = upstream_w / upstream_n;
	for(j = i = 0; i < upstream_n; i++, j++)
	{
		fprintf (stream, "u %d %d %d %d %d", j, i, ncis, pos, pos+delta);
		pos += delta; 
		for(k = 0; k < signal_size; k++)
		{
			fprintf (stream, " %g", signal[j][k]);
		}
		if (square_flag)
		{
			for(; k < 2*signal_size; k++)
			{
				fprintf (stream, " %g", signal[j][k]);
			}	
		}
		fprintf (stream, " %d %d\n",	(int)floor(signal[j][k]), (int)floor(signal[j][k+1]));
	}
	delta = (inside_t / ncis) / inside_n;
	for(i = 0; i < inside_n; i++, j++)
	{
		fprintf (stream, "i %d %d %d %d %d", j, i, ncis, pos, pos+delta);
		pos += delta; 
		for(k = 0; k < signal_size; k++)
		{
			fprintf (stream, " %g", signal[j][k]);
		}
		if (square_flag)
		{
			for(; k < 2*signal_size; k++)
			{
				fprintf (stream, " %g", signal[j][k]);
			}	
		}
		fprintf (stream, " %d %d\n",	(int)floor(signal[j][k]), (int)floor(signal[j][k+1]));
	}
	delta = downstream_w / downstream_n;
	for(i = 0; i < downstream_n; i++, j++)
	{
		fprintf (stream, "d %d %d %d %d %d", j, i, ncis, pos, pos+delta);
		pos += delta; 
		for(k = 0; k < signal_size; k++)
		{
			fprintf (stream, " %g", signal[j][k]);
		}
		if (square_flag)
		{
			for(; k < 2*signal_size; k++)
			{
				fprintf (stream, " %g", signal[j][k]);
			}	
		}
		fprintf (stream, " %d %d\n",	(int)floor(signal[j][k]), (int)floor(signal[j][k+1]));
	}
}

typedef struct
{
	size_t size;
	gdl_string ** outputs;
	size_t * bin_from;
	size_t * bin_to;
	int * col_from;
	int * col_to;
	double * col_thresh;
} _extract_config;

static _extract_config *
_extract_config_read (const gdl_string * filename)
{
	FILE * stream;	
	size_t i,j,n,b;
	gdl_string * tok, * line = 0;
	_extract_config * extc;
	
	extc = GDL_CALLOC (_extract_config, 1);
	
	stream = gdl_fileopen (filename, "r");
	while (gdl_getline (&line, &n, stream) != -1)
	{
		gdl_string_free (line);
		line=0;
		(extc->size)++;
	}
   gdl_fileclose (filename, stream);
   
   stream = gdl_fileopen (filename, "r");
	
	extc->outputs   = GDL_MALLOC (gdl_string *, extc->size);
	extc->bin_from  = GDL_MALLOC (size_t, extc->size);
	extc->bin_to    = GDL_MALLOC (size_t, extc->size);
	extc->col_from  = GDL_MALLOC (int, extc->size);
	extc->col_to    = GDL_MALLOC (int, extc->size);
	extc->col_thresh = GDL_CALLOC (double, extc->size);
	b=0;
   line=0;
   while (gdl_getline (&line, &n, stream) != -1)
	{ 
		i = j = 0;
		extc->outputs[b] = gdl_string_next_token (line, n, &i, &j);
		tok = gdl_string_next_token (line, n, &i, &j);
		extc->bin_from[b] = (size_t) atoi (tok);
		gdl_string_free (tok);
		tok = gdl_string_next_token (line, n, &i, &j);
		extc->bin_to[b] = (size_t) atoi (tok);
		gdl_string_free (tok);
		tok = gdl_string_next_token (line, n, &i, &j);
		extc->col_from[b] =  atoi (tok);
		gdl_string_free (tok);
		tok = gdl_string_next_token (line, n, &i, &j);
		extc->col_to[b] =  atoi (tok);
		gdl_string_free (tok);
		tok = gdl_string_next_token (line, n, &i, &j);
		extc->col_thresh[b] =  (double)atof (tok);
		gdl_string_free (tok);		
		gdl_string_free (line);line=0;
		b++;
	}
	gdl_fileclose (filename, stream);
	
	return extc;
}

int
main (int argc, char *argv[])
{
  int status;
  
  parse_argument (argc, argv);
  
  if (help_flag)
  {
  	  exit (help());
  }
  
  status = check_argument ();
  
  if (status == GDL_SUCCESS)
  {
  		int status; 
  		gdl_string * file;
  		FILE * stream, * matstream;
  		gdl_runtime * runtime;
  		gdl_hashtable * selected_features = 0;
  		_extract_config * extc = 0;
  		
  		runtime = runtime_setup();
  	
	 fprintf(stderr, "--\n");
	 fprintf(stderr, " NMt2featcis\n");
	 fprintf(stderr, "--\n");
	 
	 gdl_tiling_data * nucleo;
	 
	 stream  = gdl_fileopen (INPUT, "r");
	 nucleo  = gdl_tiling_data_fread (stream);
	 gdl_fileclose (INPUT, stream);
	 
	 if (!nucleo)
	 	 GDL_ERROR_VAL ("Unable to read data file", GDL_FAILURE, GDL_FAILURE);
	
	// Read the feature
	gdl_gfeatures * feature = gdl_gfeatures_alloc (gdl_gfeatures_gff3, FEATURE);
	
	if (FEATURE_ID)
		selected_features = read_selected_features (FEATURE_ID);
		
	if (EXTRACT_C)
		extc = _extract_config_read (EXTRACT_C);
		
	const size_t NCHROM = gdl_gfeatures_size (feature);
	size_t i,j,c,f,p,p_min,p_max,push,ncis=0;
	double ** cis_signal;
	long inside_t = 0;
		
	const size_t NA = gdl_tiling_data_get_dim (nucleo);	
	
	if (square_flag)
	{
		cis_signal = GDL_MATRIX_ALLOC (double, UPSTREAM_N+INSIDE+DOWNSTREAM_N, 2*NA+2);
	}
	else
	{	
		cis_signal = GDL_MATRIX_ALLOC (double, UPSTREAM_N+INSIDE+DOWNSTREAM_N, NA+2);
	}
	
//	if (!no_within_flag)
//	{
//		file   = gdl_string_sprintf ("%s.within", OUTPUT);
//		stream = gdl_fileopen (file, "w");
//		gdl_string_free (file);
//	}
	
	for (c = 0; c < NCHROM; c++)
	{
		gdl_gfeatures_landmark * chrom_feature = gdl_gfeatures_get (feature, c);
		const int chrom_nucidx = gdl_tiling_data_chrom_lookup (nucleo, chrom_feature->seqid);
		if (chrom_nucidx == -1)
		{
			fprintf (stderr, "Chromosome (landmark) %s is not present in the tiling data set\n", chrom_feature->seqid);
			continue;
		}
		const gdl_tiling_chromosome * chrom_nucleo = nucleo->chroms[chrom_nucidx];
		const size_t NPROBE = chrom_nucleo->nprobe;
		for(f = 0; f < chrom_feature->size; f++)
		{
			gdl_gfeatures_gff3_record * rec = (gdl_gfeatures_gff3_record *) chrom_feature->records[f];
			if (is_a_selected_feature (feature, rec, selected_features))
			{
				push=1;
				if (rec->strand=='+' || rec->strand=='?')
				{
					for(p_min = 0;p_min < NPROBE && chrom_nucleo->positions[p_min] < rec->start - UPSTREAM_W;p_min++);
					for(p_max = p_min;p_max < NPROBE && chrom_nucleo->positions[p_max] <= rec->end + DOWNSTREAM_W;p_max++);
				}
				else
				{
					for(p_min = 0;p_min < NPROBE && chrom_nucleo->positions[p_min] < rec->start - DOWNSTREAM_W;p_min++);
					for(p_max = p_min;p_max < NPROBE && chrom_nucleo->positions[p_max] <= rec->end + UPSTREAM_W;p_max++);
				}
				double z=0;
				if (COL_FROM != -1 && COL_TO != -1)
				{
					for(p = p_min;p < p_max;p++)
					{
						for(j = COL_FROM; j <= COL_TO; j++)
						{
							z+=GDL_TILING_CHROM_GET (chrom_nucleo, p, j);
						}
					}
				}
				push = (COL_THRESH > 0) ? (z >= COL_THRESH) : (z <= -COL_THRESH);
				if (push && p_max > p_min)
				{
					inside_t += rec->end-rec->start+1;
					ncis++;
					double * ze = 0;
					if (extc)
					{
						ze = GDL_CALLOC (double, extc->size);
					}
					for(p = p_min;p < p_max;p++)
					{
						int p_bin = get_bin (rec->strand, rec->start, rec->end, chrom_nucleo->positions[p], UPSTREAM_W, UPSTREAM_N, INSIDE, DOWNSTREAM_W, DOWNSTREAM_N);
						if (p_bin == -1) continue;
						if (no_overlap_flag)
						{
							size_t ff, overlap=0;
							for(ff = 0; ff < chrom_feature->size; ff++)
							{
								gdl_gfeatures_gff3_record * orec = (gdl_gfeatures_gff3_record *) chrom_feature->records[ff];
								if (chrom_nucleo->positions[p] >= orec->start && chrom_nucleo->positions[p] <= orec->end)
								{
									if (is_a_selected_feature (feature, orec, selected_features))
									{	
										overlap=1;
										break;
									}
								}
								else if (chrom_nucleo->positions[p] < orec->start)
								{
									break;
								}
							}
							if (overlap)
							{
								continue;
							}
						}
						push_signal(cis_signal, p_bin, chrom_nucleo, p, square_flag);
//						if (!no_within_flag)
//						{
//							if (!gdl_hashtable_lookup (rec->attributes, "ID"))
//							{
//								fprintf (stream, "%s %s %c %d %d %d %g\n", chrom_feature->seqid, gdl_hashtable_lookup (rec->parents[0]->attributes, "ID"), rec->strand, rec->start, rec->end, chrom_nucleo->positions[p], GDL_TILING_CHROM_GET (chrom_nucleo, p, 0));
//							}
//							else
//							{
//								fprintf (stream, "%s %s %c %d %d %d %g\n", chrom_feature->seqid, gdl_hashtable_lookup (rec->attributes, "ID"), rec->strand, rec->start, rec->end, chrom_nucleo->positions[p], GDL_TILING_CHROM_GET (chrom_nucleo, p, 0));
//							}
//						}
						if (extc)
						{
							size_t e;
							for(e = 0; e < extc->size; e++)
							{
								if (p_bin >= extc->bin_from[e] && p_bin <= extc->bin_to[e])
								{
									for(j = extc->col_from[e]; j <= extc->col_to[e]; j++)
									{
										ze[e] += GDL_TILING_CHROM_GET (chrom_nucleo, p, j);
									}
								}
							}
						}
					}
					int p0 = get_bin (rec->strand, rec->start, rec->end, chrom_nucleo->positions[p_min], UPSTREAM_W, UPSTREAM_N, INSIDE, DOWNSTREAM_W, DOWNSTREAM_N);
					int p1 = get_bin (rec->strand, rec->start, rec->end, chrom_nucleo->positions[p_max-1], UPSTREAM_W, UPSTREAM_N, INSIDE, DOWNSTREAM_W, DOWNSTREAM_N);
					if (rec->strand == '+' || rec->strand == '?')
					{
						for(; p0 <= p1; p0++)
						{
							if (square_flag)
								cis_signal[p0][2*NA+1]+=1;
							else
								cis_signal[p0][NA+1]+=1;
						}
					}
					else
					{
						for(; p0 >= p1; p0--)
						{
							if (square_flag)
								cis_signal[p0][2*NA+1]+=1;
							else
								cis_signal[p0][NA+1]+=1;
						}
					}
					if (extc)
					{
						size_t e;
						for(e = 0; e < extc->size; e++)
						{
							if (ze[e] >= extc->col_thresh[e])
							{
								FILE * e_stream = gdl_fileopen (extc->outputs[e], "a");
								if (!gdl_hashtable_lookup (rec->attributes, "ID"))
								{
									fprintf (e_stream, "%s %s %c %d %d %g\n", chrom_feature->seqid, gdl_hashtable_lookup (rec->parents[0]->attributes, "ID"), rec->strand, rec->start, rec->end, ze[e]);
								}
								else
								{
									fprintf (e_stream, "%s %s %c %d %d %g\n", chrom_feature->seqid, gdl_hashtable_lookup (rec->attributes, "ID"), rec->strand, rec->start, rec->end, ze[e]);
								}
								gdl_fileclose(extc->outputs[e], e_stream);
							}
						}
						GDL_FREE (ze);
					}					
				}
			}
		}
	}
	
//	if (!no_within_flag)
//	{
//		file   = gdl_string_sprintf ("%s.within", OUTPUT);
//		gdl_fileclose (file, stream);
//		gdl_string_free (file);
//	}

	file   = gdl_string_sprintf ("%s_cis.txt", OUTPUT);
	stream = gdl_fileopen (file, "w");
    print_cis_signal (stream, cis_signal, NA, ncis, UPSTREAM_W, UPSTREAM_N, INSIDE, inside_t, DOWNSTREAM_W, DOWNSTREAM_N, square_flag);
	gdl_fileclose (file, stream);
	
	gdl_tiling_data_free (nucleo);
	
	runtime_setdown (runtime);
	
  }
  exit(0);
}
