/*  
 *  NMtdb.c
 * 
 *  $Author: gael $, $Date: 2008-09-15 12:32:59 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_list.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_runtime.h>

#include "nucleoMiner.h"

#include "const.h"
#include "header.c"
#include "util.c"

#include "tiling_array.h"

static gdl_string * PROGRAM = "NMtdb";

static int help_flag    = 0;
static int verbose_flag = 0;
static int print_flag = 0;
static int substract_flag = 0;

static gdl_string * INPUT     = NULL;
static gdl_string * CONFIG    = NULL;
static gdl_string * OUTPUT    = NULL;
static int FORCE_TILING = -1;
static double LOG_TRANSFORM   = 0; 

static struct option long_options[] =
{
   /* These options set a flag. */
   {"help", no_argument,       &help_flag, 1},
   {"verbose", no_argument,       &verbose_flag, 1},
   {"brief",   no_argument,       &verbose_flag, 0},
   {"print",   no_argument,       &print_flag, 1},
   {"substract",   no_argument,   &substract_flag, 1},
   /* These options don't set a flag.
      We distinguish them by their indices. */
   {"input",   required_argument, 0, 'i'},         // the data file
   {"config",   required_argument, 0, 'c'},        // the configuration file
   {"output",   required_argument, 0, 'o'},        // the output file
   {"force-tiling",   required_argument, 0, 't'},  // force the tiling size
   {"log-transform",   required_argument, 0, 'l'},  // log-transform the data
   {0, 0, 0, 0}
};
 
static int
parse_argument (int argc, char *argv[])
{
	int c;
     
   while (1)
     {
      
       /* getopt_long stores the option index here. */
       int option_index = 0;
 
       c = getopt_long (argc, argv, "i:o:c:t:l:",
                        long_options, &option_index);
 
       /* Detect the end of the options. */
       if (c == -1)
         break;
         
       switch (c)
         {
         case 0:
           /* If this option set a flag, do nothing else now. */
           if (long_options[option_index].flag != 0)
             break;
           printf ("option %s", long_options[option_index].name);
           if (optarg)
             printf (" with arg %s", optarg);
           printf ("\n");
           break;
 
         case 'i':
           INPUT = gdl_string_clone (optarg);
           break;
         case 'c':
           CONFIG = gdl_string_clone (optarg);
           break;  
         case 'o':
           OUTPUT = gdl_string_clone (optarg);
           break;
         case 't':
           FORCE_TILING = atoi (optarg);
           break;
         case 'l':
         	LOG_TRANSFORM = (double) atof(optarg);
         	break;
         case '?':
           GDL_ERROR_VAL ("Unknown arguments", GDL_EINVAL, -1);
         default:
           GDL_ERROR_VAL ("Bad arguments", GDL_EINVAL, -1);
         }
     }
   /* Print any remaining command line arguments (not options). 
   if (optind < argc)
     {
       printf ("non-option ARGV-elements: ");
       while (optind < argc)
         printf ("%s ", argv[optind++]);
       putchar ('\n');
     }*/
}

static int
check_argument (void)
{
	if (INPUT == 0)
	{
		GDL_ERROR_VAL ("No input file defined",
		               GDL_FAILURE,
		               1);
	}
	if (!print_flag && CONFIG == 0)
	{
		GDL_ERROR_VAL ("No config file defined",
		               GDL_FAILURE,
		               1);
	}
	if (!print_flag && OUTPUT == 0)
	{
		GDL_ERROR_VAL ("No output stem defined",
		               GDL_FAILURE,
		               1);	
	}
	
	return GDL_SUCCESS;
}

static int
help (void)
{
   print_header (PROGRAM);
   printf ("\n");
   printf ("*These options set a flag:\n");
   printf ("\n");
   printf ("--help\tDisplay a brief help on program usage\n");
   printf ("--verbose\tOutput message on standard output to see what the program is doing\n");
   printf ("\n");
   printf ("--print\tPrint the dataset stored in the db file\n");
   printf ("--substract\tSubstract signal between two groups (2) - (1)\n");
   printf ("\n");
   printf ("*These options require arguments:\n");
   printf ("\n");
   printf ("--input or -i\tInput file: either the raw data file or the db file\n");
   printf ("--config or -c\tConfiguration file to create the db file from the raw data file\n");
   printf ("--output or -o\tThe output db file\n");
   printf ("--force-tiling or -t\tForce the tiling step to be a multiple of the given value\n");
   printf ("--log-transform or -l\tLog-transform the intensities with the given base (default is no log transformation)\n");
}


int
main (int argc, char *argv[])
{
  int status;
  
  parse_argument (argc, argv);
  
  if (help_flag)
  {
  	  exit (help());
  }
  
  status = check_argument ();
  
  if (status == GDL_SUCCESS)
  {
  		int status; 
  		gdl_string * file;
  		FILE * stream;
  		gdl_runtime * runtime;
  		
  		runtime = runtime_setup();
  	
  		if (!print_flag)
  		{
  	
  	    fprintf(stdout, "--\n");
		 fprintf(stdout, " NMtdb\n");
		 
		 gdl_tiling_data * data;
		 
		 gdl_tiling_data_conf * data_conf;
		 
		 fprintf(stdout, "--\n");
		 fprintf(stdout, " Read configuration file:");
		 fflush(stdout);
		 
		 stream    = gdl_fileopen (CONFIG, "r");
		 data_conf = gdl_tiling_data_conf_fscanf (stream);
		 
		 gdl_fileclose (CONFIG, stream);
		 fprintf(stdout, " [ OK ]\n");
		 
		 if (substract_flag)
		 {
		 	if (strcmp(data_conf->type,gdl_tiling_data_mean->name) && strcmp(data_conf->type,gdl_tiling_data_median->name))
		 	{
		 		GDL_ERROR_VAL ("[type] in configuration file must be set either to [mean] or [median]\n", GDL_FAILURE, GDL_FAILURE);
		 	}
		 }
		 
		 fprintf(stdout, "--\n");fflush(stdout);
		 fprintf(stdout, " Read data file: wait...");fflush(stdout);
		 
		 data = gdl_tiling_data_alloc (data_conf);
		 stream = gdl_fileopen (INPUT, "r");
		 status = gdl_tiling_data_fscanf (stream, data);
		 gdl_fileclose (INPUT, stream);
		 
		 if (status!=GDL_SUCCESS)
		 	 GDL_ERROR_VAL ("Unable to read data file", GDL_FAILURE, GDL_FAILURE);
		 fprintf(stdout, "\b\b\b\b\b\b\b[ OK ]\n");
		 
		 if (LOG_TRANSFORM > 0)
		 {
		 	fprintf(stdout, "--\n");
		 	fprintf(stdout, " Log Transform (base=%g) the intensities:", LOG_TRANSFORM);
		 	fflush(stdout);
		 	size_t c,p,a;
		 	for(c = 0; c < data->nchrom; c++)
		 	{
			 	gdl_tiling_chromosome * data_c = data->chroms[c];
			 	for(p = 0; p < data_c->nprobe; p++)
			 	{
			 		for(a = 0; a < data_c->narray; a++)
			 		{
			 			double x = GDL_TILING_CHROM_GET (data_c, p, a);
			 			x = log(x)/log(LOG_TRANSFORM);
			 			GDL_TILING_CHROM_SET (data_c, p, a, x);
			 		}
			 	}
		 	}
		 	fprintf(stdout, "\b\b\b\b\b\b\b[ OK ]\n");
		 	fflush(stdout);
		 }
		 
		 if (substract_flag)
		 {
		 	if (data->ngroup!=2)
		 	{	
		 		GDL_ERROR_VAL ("There must be at least and at most two groups when using --log-ratio option", GDL_FAILURE, GDL_FAILURE);
		 	}
		 	fprintf(stdout, "--\n");
		 	fprintf(stdout, " Substract (2) - (1)");
		 	fflush(stdout);
		 	size_t c,p;
		 	for(c = 0; c < data->nchrom; c++)
		 	{
			 	gdl_tiling_chromosome * data_c = data->chroms[c];
			 	for(p = 0; p < data_c->nprobe; p++)
			 	{
			 		double x = GDL_TILING_CHROM_GET (data_c, p, 1);
			 		x -= GDL_TILING_CHROM_GET (data_c, p, 0);
			 		data_c->intensities[p] = x;
			 	}
			 	data_c->narray=1;			 	
		 	}
		 	data->narray=1;
		 	data->ngroup=1;
		 	data->narrayByGroup = GDL_CALLOC (size_t, 1);
		 	data->narrayByGroup[0] = 1;
		 	data->arrayByGroupIdx = GDL_MATRIX_ALLOC (size_t, 1, 1);
		 	data->groupId = GDL_CALLOC (size_t, 1);
		 	fprintf(stdout, "\b\b\b\b\b\b\b[ OK ]\n");
		 	fflush(stdout);
		 }
		 
		 if (FORCE_TILING > 0)
		 {
		 	fprintf(stdout, "--\n");fflush(stdout);
		   fprintf(stdout, " Force tiling (%d):", FORCE_TILING);fflush(stdout);
		 	gdl_tiling_data_force_tiling (data, FORCE_TILING);
		 	fprintf(stdout, " [ OK ]\n");fflush(stdout);
		 }
		 
		 fprintf(stdout, "--\n");fflush(stdout);
		 fprintf(stdout, " Save data:");fflush(stdout);
		 file   = gdl_string_sprintf ("%s", OUTPUT);
		 stream = gdl_fileopen (file, "w");
		 gdl_tiling_data_fwrite (stream, data);
		 gdl_fileclose (file, stream);
		 gdl_string_free (file);
		 fprintf(stdout, " [ OK ]\n");fflush(stdout);
		 fprintf(stdout, "--\n");fflush(stdout);
  		}
  		else
  		{
  			gdl_tiling_data * data;
  			
  			stream = gdl_fileopen (INPUT, "r");
  			data = gdl_tiling_data_fread (stream);
		   gdl_fileclose (INPUT, stream);
  			
  			gdl_tiling_data_fprintf (stdout, data);
  			
  			gdl_tiling_data_free (data);
  		}
		
		runtime_setdown (runtime);
  }
  
  exit(0);
}
