/*  
 *  align_nucleo.c
 * 
 *  $Author: gael $, $Date: 2008-10-17 16:07:07 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
 
#include <math.h>
#include <stdio.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_list.h>
#include <gdl/gdl_machine.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_sort_long.h>
#include <gdl/gdl_runtime.h>
#include <gdl/gdl_vector.h>
#include <gdl/gdl_matrix.h>
#include <gdl/gdl_multireg.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_statistics_double.h>
#include <gdl/gdl_genome_coord.h>
#include <gdl/gdl_sequences.h>

#include "tiling_hmm.h"
#include "tiling_nucleo.h"
#include "tiling_array.h"
#include "align_nucleo.h"

align_nucleo_config *
align_nucleo_config_alloc (const gdl_string * filename)
{
	FILE * stream;
	align_nucleo_config * conf = 0;
	
	stream = gdl_fileopen (filename, "r");
	if (!stream)
		return conf;
	
	size_t ti,tj,n;
	gdl_string * line=0, * key, * value;
	
	conf = GDL_CALLOC (align_nucleo_config, 1);
	while(gdl_getline (&line, &n, stream)!=-1)
	{
		if (line[0]=='#') 
			continue;
		ti = tj = 0;
		key   = gdl_string_next_token (line, n, &ti, &tj);
		value = gdl_string_next_token (line, n, &ti, &tj);
		if (!strcmp(key, "ref_hmm_dir"))
			conf->ref_hmm_dir = value;
		if (!strcmp(key, "qry_hmm_dir"))
			conf->qry_hmm_dir = value;
		if (!strcmp(key, "ref_data_file"))
			conf->ref_data_file = value;
		if (!strcmp(key, "qry_data_file"))
			conf->qry_data_file = value;
		if (!strcmp(key, "ref_qry_coord_file"))
			conf->ref_qry_coord_file = value;
		if (!strcmp(key, "ref_stat_file"))
			conf->ref_stat_file = value;
		if (!strcmp(key, "qry_stat_file"))
			conf->qry_stat_file = value;
		if (!strcmp(key, "ref_seq_file"))
			conf->ref_seq_file = value;
		if (!strcmp(key, "qry_seq_file"))
			conf->qry_seq_file = value;
		gdl_string_free (key);
		gdl_string_free (line);line=0;
	}
	gdl_fileclose (filename, stream);
	
	if (conf->ref_hmm_dir==0)
	{
		GDL_ERROR_VAL ("No ref HMM directory defined", GDL_EINVAL, 0);	
	}
	if (conf->qry_hmm_dir==0)
	{
		GDL_ERROR_VAL ("No qry HMM directory defined", GDL_EINVAL, 0);	
	}
	if (conf->ref_data_file==0)
	{
		GDL_ERROR_VAL ("No ref data file defined", GDL_EINVAL, 0);	
	}
	if (conf->qry_data_file==0)
	{
		GDL_ERROR_VAL ("No qry data file defined", GDL_EINVAL, 0);	
	}
	if (conf->ref_qry_coord_file==0)
	{
		GDL_ERROR_VAL ("No ref to qry file defined", GDL_EINVAL, 0);	
	}
	
	stream = gdl_fileopen (conf->ref_data_file, "r");
	conf->ref_data = gdl_tiling_data_fread (stream);
	//printf ("NCHROM = %d\n", conf->ref_data->nchrom);
	gdl_fileclose (conf->ref_data_file, stream);
	if (!conf->ref_data)
	{
		GDL_ERROR_VAL ("Unable to read ref data file", GDL_FAILURE, 0);	
	}
	stream = gdl_fileopen (conf->qry_data_file, "r");
	conf->qry_data = gdl_tiling_data_fread (stream);
	//printf ("NCHROM = %d\n", conf->qry_data->nchrom);
	gdl_fileclose (conf->qry_data_file, stream);
	if (!conf->qry_data)
	{
		GDL_ERROR_VAL ("Unable to read qry data file", GDL_FAILURE, 0);	
	}
	conf->ref_qry_coord = gdl_genome_coord_alloc (conf->ref_qry_coord_file);
	if (!conf->ref_qry_coord)
	{
		GDL_ERROR_VAL ("Unable to read ref to qry file", GDL_FAILURE, 0);	
	}
	
	if (conf->ref_stat_file)
	{
		stream = gdl_fileopen (conf->ref_stat_file, "r");
		conf->ref_stat = gdl_tiling_data_fread (stream);
		gdl_fileclose (conf->ref_stat_file, stream);
	}
	if (conf->qry_stat_file)
	{
		stream = gdl_fileopen (conf->qry_stat_file, "r");
		conf->qry_stat = gdl_tiling_data_fread (stream);
		gdl_fileclose (conf->qry_stat_file, stream);
	}
	if (conf->ref_seq_file)
	{
		gdl_seq_io * seq_io = gdl_seq_io_alloc (gdl_seq_io_type_get ("fasta"));
		gdl_seq_io_open (seq_io, conf->ref_seq_file);
		gdl_seq_io_parse (seq_io);
		conf->ref_seq = gdl_seq_io_close (seq_io);
		gdl_seqdb_id2seq_create (conf->ref_seq);
		gdl_seq_io_free (seq_io);	
	}
	if (conf->qry_seq_file)
	{
		gdl_seq_io * seq_io = gdl_seq_io_alloc (gdl_seq_io_type_get ("fasta"));
		gdl_seq_io_open (seq_io, conf->qry_seq_file);
		gdl_seq_io_parse (seq_io);
		conf->qry_seq = gdl_seq_io_close (seq_io);
		gdl_seqdb_id2seq_create (conf->qry_seq);
		gdl_seq_io_free (seq_io);
	}
	
	return conf;
}

align_nucleo_nucleosome * 
align_nucleo_nucleosome_alloc (align_nucleo_chromosome * chrom, const size_t chunk, const size_t start, const size_t end, const size_t status, const size_t nclass)
{
	align_nucleo_nucleosome * n;
	
	n = GDL_CALLOC (align_nucleo_nucleosome, 1);
	
	n->chrom = chrom;
	n->chunk_idx = chunk;
	n->start_idx = start;
	n->end_idx   = end;
	n->status    = status;
	n->prob      = GDL_CALLOC (double, (nclass-1)/2+1);
	n->ignore    = 'n';
	
	return n;
}

align_nucleo_linker * 
align_nucleo_linker_alloc (const align_nucleo_chromosome * chrom, const size_t chunk, const size_t start, const size_t end, const size_t nclass)
{
	align_nucleo_linker * n;
	
	n = GDL_CALLOC (align_nucleo_linker, 1);
	
	n->chrom = chrom;
	n->chunk_idx = chunk;
	n->start_idx = start;
	n->end_idx   = end;
	n->prob      = GDL_CALLOC (double, (nclass-1)/2+1);
	
	return n;
}

gdl_tiling_nucleo_chromosome *
load_chrom_hmm (const gdl_string * hmm_dir, const gdl_string * chrom, const size_t D, const size_t C)
{
	gdl_string * file;
	FILE * stream;
	gdl_tiling_nucleo_chromosome * chrom_hmm;
	
	file   = gdl_string_sprintf ("%s/%s.nuc", hmm_dir, chrom);
 	stream = gdl_fileopen (file, "r");
 	if (!stream)
 	{
 		GDL_ERROR_VAL (gdl_string_sprintf ("Unable to read ref hmm file for chromosome %s from directory %s", chrom, hmm_dir), GDL_FAILURE, 0);	
 	}
	chrom_hmm = gdl_tiling_nucleo_chromosome_fscanf (stream, D, C);
	gdl_fileclose (file, stream);
	gdl_string_free (file);
	
	return chrom_hmm;	
}

int
align_nucleo_chromosome_anchor_probe (align_nucleo_chromosome * chrom)
{
	size_t i,j,k;
	
	for (j = i = 0; i < chrom->hmm->size; i++)
	{
		const gdl_tiling_nucleo_chunk * chunk = chrom->hmm->chunks[i];
		for(k = 0; k < chunk->size; k++)
		{
			gdl_tiling_nucleo_probe * probe = chunk->probes[k];
			for(; chrom->data->positions[j] < probe->position && j < chrom->data->nprobe; j++);
			if (j < chrom->data->nprobe && probe->position==chrom->data->positions[j])
			{
				probe->link_idx = j++;
			}
			else if (j == chrom->data->nprobe)
			{
				break;
			}
			else
			{
				probe->link_idx = -1;
			}
		}
	}
}

int
align_nucleo_chromosome_fprintf_probe (FILE * stream, const align_nucleo_chromosome * chrom, const gdl_tiling_nucleo_probe * probe)
{
	size_t k;
	
	fprintf (stream, "%s %d", chrom->data->name, probe->position);
	if (probe->idx>=0)
	{
		for(k = 0; k < chrom->data->narray; k++)
			fprintf (stream, " %e", GDL_TILING_CHROM_GET(chrom->data, probe->idx, k));
	}
	else
	{
		for(k = 0; k < chrom->data->narray; k++)
			fprintf (stream, " .");
	}
	for(k = 0; k < chrom->hmm->D; k++)
		fprintf (stream, " %e", probe->smooth[k]);
	for(k = 0; k < chrom->hmm->C; k++)
		fprintf (stream, " %e", probe->prob[k]);
	fprintf (stream, " %d", probe->calling);
}

int
align_nucleo_chromosome_fprintf (FILE * stream, const align_nucleo_chromosome * chrom, const size_t mode)
{
	size_t i,j,k;
	for (i = 0; i < chrom->hmm->size; i++)
	{
		const gdl_tiling_nucleo_chunk * chunk = chrom->hmm->chunks[i];
		for(j = 0; j < chunk->size; j++)
		{
			gdl_tiling_nucleo_probe * probe = chunk->probes[j];
			if (mode==1 && probe->link_idx==-1) continue;
			if (mode==1 && probe->align==0) continue;
			align_nucleo_chromosome_fprintf_probe (stream, chrom, probe);
			if (probe->align)
			{
				fprintf (stream, " ");
				gdl_tiling_nucleo_probe * qry_probe   = (gdl_tiling_nucleo_probe *) probe->align;
				align_nucleo_chromosome * qry_chrom = (align_nucleo_chromosome *) qry_probe->chrom;
				align_nucleo_chromosome_fprintf_probe (stream, qry_chrom, qry_probe);
			}
			fprintf (stream, "\n");
		}
	}	
}

gdl_tiling_nucleo_probe *
nucleo_hmm_cmp_anchor_qry_probe_at_position (align_nucleo_chromosome * qry_chrom, const long qry_pos, const long tiling_size)
{
	size_t i,j;
	gdl_tiling_nucleo_chunk * chunk;
	gdl_tiling_nucleo_probe * probe;
	
	for(i = 0; i < qry_chrom->hmm->size; i++)
	{
		chunk = qry_chrom->hmm->chunks[i];
		if (qry_pos >= chunk->from_pos && qry_pos <= chunk->to_pos)
		{
			for(j = 0; j < chunk->size; j++)
			{
				probe = chunk->probes[j];
				if (probe->position > qry_pos)
				{
					break;
				}
				else if (fabs((double)(probe->position-qry_pos))<tiling_size)
				{
					probe->chrom = qry_chrom;
					return probe;
				}
			}
		}
		else if (chunk->from_pos > qry_pos)
		{
			break;
		} 
	}
	
	return 0;
}

int
align_nucleo_chromosome_anchor_qry (const align_nucleo_config * conf, align_nucleo_chromosome * ref_chrom, const size_t hmm_dim, const size_t tiling_size) 
{
	size_t i,j;
	gdl_tiling_nucleo_chunk * chunk;
	gdl_tiling_nucleo_probe * probe;
	
	const gdl_chunk_coord_cluster * chunk_cluster = gdl_genome_coord_lookup (conf->ref_qry_coord, "ref", ref_chrom->data->name);
	
	if (!chunk_cluster)
	{
		fprintf (stderr, "[ WARNING] nucleoHmmCmp: no landmark %s in ref to qry coordinates\n", ref_chrom->data->name);
		fflush (stderr);
		return GDL_CONTINUE;
	}
	
	ref_chrom->qry_table = gdl_hashtable_alloc (gdl_hash_default, 0);
	
	for(i = 0; i < ref_chrom->hmm->size; i++)
	{
		int ref_chunk=0;
		chunk = ref_chrom->hmm->chunks[i];
		for(j = 0; j < chunk->size; j++)
		{
			probe = chunk->probes[j];
			if (probe->idx >= 0)
			{
				long ref_pos = probe->position;
				ref_chunk    = gdl_chunk_coord_cluster_goto (chunk_cluster, ref_pos, ref_chunk);
				if (ref_chunk == -1) 
				{
					fprintf (stderr, "[ WARNING] nucleoHmmCmp: unable to map position %d on landmark %s (unmapped region)\n", ref_pos, ref_chrom->data->name);
					fflush (stderr);
					ref_chunk=0;
					continue;
				}
				const gdl_string * qry_name = gdl_dictionary_get (chunk_cluster->seqid_dico, "qry", chunk_cluster->chunks[ref_chunk]->qry_id);
				int qry_idx = gdl_tiling_data_chrom_lookup (conf->qry_data, qry_name);
				if (qry_idx == -1)
				{
					fprintf (stderr, "[ WARNING] nucleoHmmCmp: no landmark %s in qry data\n", qry_name);
					fflush (stderr);
					continue;
				}
				//printf ("%s %d %d %s %d ==> ", ref_chrom->data->name, ref_pos, ref_chunk, qry_name, qry_idx);
				align_nucleo_chromosome * qry_chrom;
				//printf ("ANCHOR::QRY NAME %s\n", qry_name);
				if ((qry_chrom=gdl_hashtable_lookup (ref_chrom->qry_table, qry_name))==0)
				{
					qry_chrom = GDL_CALLOC (align_nucleo_chromosome, 1);
					// load the data 
					qry_chrom->data = conf->qry_data->chroms[qry_idx];
				 	// load the hmm results
				 	qry_chrom->hmm = load_chrom_hmm (conf->qry_hmm_dir, qry_chrom->data->name, (hmm_dim==3) ? 1 : 2, hmm_dim);
				 	// anchor the hmm results on the data 
				 	align_nucleo_chromosome_anchor_probe (qry_chrom);
				 	// add the qry_chrom
				 	gdl_hashtable_add (ref_chrom->qry_table, qry_name, qry_chrom, 0);
				}
				long qry_pos = gdl_chunk_coord_cluster_map (chunk_cluster, ref_pos, ref_chunk);
				probe->align = nucleo_hmm_cmp_anchor_qry_probe_at_position (qry_chrom, qry_pos, tiling_size);
				//printf ("%p %d %p", qry_chrom, qry_pos, probe->extra);
				if (probe->align == 0)
				{
					fprintf (stderr, "[ WARNING] nucleoHmmCmp: unable to map position %d on landmark %s (no probe at that position on qry)\n", ref_pos, ref_chrom->data->name);
					fflush (stderr);
				}
				else
				{
					((gdl_tiling_nucleo_probe *)probe->align)->align=probe;
				}
			}
			else if (probe->observed)
			{
				// if probe->observed output that this probe is not included in the comparaion	
			}
		}
	}		
}

align_nucleo_nucleosome *
align_nucleo_chromosome_call_nucleo (align_nucleo_chromosome * chrom,
                                     const gdl_tiling_nucleo_probe * probe,
                                     const size_t chunk_idx,
                                     const size_t probe_idx,
                                     const size_t nucleo_min,
                                     const size_t nclass,
                                     size_t * score,
                                     size_t * status,
                                     size_t * start,
                                     size_t * cidx,
                                     align_nucleo_chromosome ** chrom_ptr)
{
	align_nucleo_nucleosome * nucleosome = 0;
	
	switch (probe->calling)
	{
		case 0: // linker
			if (*score >= nucleo_min 
			    && (*cidx == chunk_idx && probe_idx>0)
			    && (*chrom_ptr == chrom))
			{
			    size_t i,j,k;
				double t=0;
				nucleosome = align_nucleo_nucleosome_alloc (chrom, chunk_idx, *start, probe_idx-1, *status, nclass);
				size_t ps = GDL_MIN (nucleosome->start_idx, nucleosome->end_idx);
				size_t pe = GDL_MAX (nucleosome->start_idx, nucleosome->end_idx);
				for(i = ps; i <= pe; i++)
				{
					gdl_tiling_nucleo_probe * probe_nuc = chrom->hmm->chunks[chunk_idx]->probes[i];
					nucleosome->prob[0] += probe_nuc->prob[0];
					t += probe_nuc->prob[0];
					for(k = 1; k <= (nclass-1)/2; k++)
					{
						t += probe_nuc->prob[2*k-1]+probe_nuc->prob[2*k];
						nucleosome->prob[k] += probe_nuc->prob[2*k-1]+probe_nuc->prob[2*k];
					}
				}
				for(i = 0; i <= (nclass-1)/2; i++)
					nucleosome->prob[i] /= t;
			}
			*score=0;
			break;
		case 1: // not-acetylated delocalized
		case 2: // not-acetylated well-positioned
		case 3: // acetylated delocalized
		case 4: // acetylated well-positioned
			if ((*score)==0)
			{
				*start     = probe_idx;
				*status    = probe->calling;
				*cidx      = chunk_idx;
				*chrom_ptr = chrom;
			}
			else if (*status != probe->calling) // reset it
			{
				*score     = 0;
				*start     = probe_idx;
				*status    = probe->calling;
				*cidx      = chunk_idx;
				*chrom_ptr = chrom;
			}
			(*score)++;
			break;
	}
	
	return nucleosome;
}

align_nucleo_linker *
align_nucleo_chromosome_call_linker (align_nucleo_chromosome * chrom,
                                       const gdl_tiling_nucleo_probe * probe,
                                       const size_t chunk_idx,
                                       const size_t probe_idx,
                                       const size_t nclass,
                                       size_t * score,
                                       size_t * chunk,
                                       size_t * start,
                                       align_nucleo_chromosome ** chrom_ptr)
{
	align_nucleo_linker * linker = 0;
	switch (probe->calling)
	{
		case 0: // linker
			if ((*score)==0)
			{
				*start     = probe_idx;
				*chunk     = chunk_idx;
				*chrom_ptr = chrom;
			}
			else if (*chunk != chunk_idx || *chrom_ptr != chrom)
			{
				*score=0;	
			}
			(*score)++;
			break;
		case 1: // not-acetylated delocalized
		case 2: // not-acetylated well-positioned
		case 3: // acetylated delocalized
		case 4: // acetylated well-positioned
			if (*score && (*chunk==chunk_idx && *chrom_ptr==chrom))
			{
				size_t i,j,k;
				double t=0;
				linker = align_nucleo_linker_alloc (chrom, chunk_idx, *start, probe_idx-1, nclass);
				size_t ps = GDL_MIN (linker->start_idx, linker->end_idx);
				size_t pe = GDL_MAX (linker->start_idx, linker->end_idx);
				//printf ("ps = %d pe = %d (%d (%d) %p, %p)\n", ps, pe, chunk_idx, chrom->hmm->size, chrom->hmm->chunks[chunk_idx]->size, chrom, * chrom_ptr);
				for(i = ps; i <= pe; i++)
				{
					gdl_tiling_nucleo_probe * probe_nuc = chrom->hmm->chunks[chunk_idx]->probes[i];
					linker->prob[0] += probe_nuc->prob[0];
					t += probe_nuc->prob[0];
					for(k = 1; k <= (nclass-1)/2; k++)
					{
						t += probe_nuc->prob[2*k-1]+probe_nuc->prob[2*k];
						linker->prob[k] += probe_nuc->prob[2*k-1]+probe_nuc->prob[2*k];
					}
				}
				for(i = 0; i <= (nclass-1)/2; i++)
					linker->prob[i] /= t;
				*score=0;
			}
			else if (*chunk!=chunk_idx)
			{
				*score=0;	
			}
			else if (*chrom_ptr!=chrom)
			{
				*score=0;	
			}
			break;	
	}
	return linker;
}

void
align_nucleo_chromosome_nucleo_properties (const gdl_tiling_nucleo_chunk * ref_chunk,
                                           align_nucleo_nucleosome * ref_nucleo)
                                           
{           
	size_t j;
	
	ref_nucleo->map=ref_nucleo->obs=0;
	
	if (ref_nucleo->end_idx > ref_nucleo->start_idx)
	{
		for(j = ref_nucleo->start_idx; j <= ref_nucleo->end_idx; j++)
		{
			const gdl_tiling_nucleo_probe * ref_probe = ref_chunk->probes[j];
			if (ref_probe->align!=0)
			{
				(ref_nucleo->map)+=1.;
			}
			if (ref_probe->link_idx>=0)
			{
				(ref_nucleo->obs)+=1.;	
			}
		}
		ref_nucleo->map   /= (ref_nucleo->end_idx-ref_nucleo->start_idx+1);
		ref_nucleo->obs  /= (ref_nucleo->end_idx-ref_nucleo->start_idx+1);
	}
	else
	{
		for(j = ref_nucleo->end_idx; j <= ref_nucleo->start_idx; j++)
		{
			const gdl_tiling_nucleo_probe * ref_probe = ref_chunk->probes[j];
			if (ref_probe->align!=0)
			{
				(ref_nucleo->map)+=1.;
			}
			if (ref_probe->link_idx>=0)
			{
				(ref_nucleo->obs)+=1.;	
			}
		}
		ref_nucleo->map   /= (ref_nucleo->start_idx-ref_nucleo->end_idx+1);
		ref_nucleo->obs  /= (ref_nucleo->start_idx-ref_nucleo->end_idx+1);	
	}
}

void
align_nucleo_chromosome_linker_properties (const gdl_tiling_nucleo_chunk * ref_chunk,
                                           align_nucleo_linker * ref_linker)
                                           
{           
	size_t j;
	
	ref_linker->map=ref_linker->obs=0;
	
	if (ref_linker->end_idx > ref_linker->start_idx)
	{
		for(j = ref_linker->start_idx; j <= ref_linker->end_idx; j++)
		{
			const gdl_tiling_nucleo_probe * ref_probe = ref_chunk->probes[j];
			if (ref_probe->align!=0)
			{
				(ref_linker->map)+=1.;
			}
			if (ref_probe->link_idx>=0)
			{
				(ref_linker->obs)+=1.;	
			}
		}
		ref_linker->map   /= (ref_linker->end_idx-ref_linker->start_idx+1);
		ref_linker->obs  /= (ref_linker->end_idx-ref_linker->start_idx+1);
	}
	else
	{
		for(j = ref_linker->end_idx; j <= ref_linker->start_idx; j++)
		{
			const gdl_tiling_nucleo_probe * ref_probe = ref_chunk->probes[j];
			if (ref_probe->align!=0)
			{
				(ref_linker->map)+=1.;
			}
			if (ref_probe->link_idx>=0)
			{
				(ref_linker->obs)+=1.;	
			}
		}
		ref_linker->map   /= (ref_linker->start_idx-ref_linker->end_idx+1);
		ref_linker->obs  /= (ref_linker->start_idx-ref_linker->end_idx+1);	
	}
}

int
align_nucleo_chromosome_anchor_nucleo (align_nucleo_chromosome * chrom,
                                         const size_t tiling_size,
                                         const size_t nucleo_min,
                                         const double prb_mapped_min,
                                         const size_t nclass)
{
	size_t i,j,jj,k;
	size_t score,status,start,lscore,lstart,lcidx,rncidx,qncidx;
	size_t qry_score,qry_status,qry_start,ref_rm=0,qry_rm=0;
	align_nucleo_nucleosome * ref_nucleo, * qry_nucleo;
	align_nucleo_linker * ref_linker, * qry_linker;
	align_nucleo_chromosome ** ref_chrom_ptr, ** qry_chrom_ptr;
	align_nucleo_chromosome ** lref_chrom_ptr, ** lqry_chrom_ptr;
	gdl_list * ref_nucleosomes = gdl_list_alloc (gdl_list_default);
	gdl_list * qry_nucleosomes = gdl_list_alloc (gdl_list_default);
	gdl_list * ref_linkers = gdl_list_alloc (gdl_list_default);
	gdl_list * qry_linkers = gdl_list_alloc (gdl_list_default);
	
	ref_chrom_ptr = GDL_MALLOC (align_nucleo_chromosome *, 1);
	qry_chrom_ptr = GDL_MALLOC (align_nucleo_chromosome *, 1);
	lref_chrom_ptr = GDL_MALLOC (align_nucleo_chromosome *, 1);
	lqry_chrom_ptr = GDL_MALLOC (align_nucleo_chromosome *, 1);
	
	for (i = 0; i < chrom->hmm->size; i++)
	{
		const gdl_tiling_nucleo_chunk * chunk = chrom->hmm->chunks[i];
		score=0;
		qry_score=0;
		for(j = 0; j < chunk->size; j++)
		{
			gdl_tiling_nucleo_probe * probe = chunk->probes[j];
			ref_nucleo = align_nucleo_chromosome_call_nucleo (chrom, probe, i, j, nucleo_min, nclass, &score, &status, &start, &rncidx, ref_chrom_ptr);
			if (ref_nucleo)
			{
				ref_nucleo->ref_start_pos = chunk->probes[ref_nucleo->start_idx]->position;
				ref_nucleo->ref_end_pos   = chunk->probes[ref_nucleo->end_idx]->position;
				align_nucleo_chromosome_nucleo_properties (chunk, ref_nucleo);
				if (ref_nucleo->map >= prb_mapped_min)
				{
					ref_nucleo->ignore = 'n';
				}
				else
				{
					ref_nucleo->ignore = 'y';
					ref_rm++;
				}
				ref_nucleo->idx = gdl_list_size (ref_nucleosomes);
				gdl_list_push_back (ref_nucleosomes, ref_nucleo, 0);
				// get the next linker
				ref_linker=0;
				for(k = j; k < chunk->size && ref_linker==0; k++)
				{
					ref_linker = align_nucleo_chromosome_call_linker (chrom, chunk->probes[k], i, k, nclass, &lscore, &lcidx, &lstart, lref_chrom_ptr);
				}
				if (ref_linker)
				{
					ref_linker->ref_start_pos = chunk->probes[ref_linker->start_idx]->position;
					ref_linker->ref_end_pos   = chunk->probes[ref_linker->end_idx]->position;
					ref_nucleo->next_linker   = ref_linker;
					ref_linker->idx = gdl_list_size (ref_linkers);
					align_nucleo_chromosome_linker_properties (chunk, ref_linker);
					gdl_list_push_back (ref_linkers, ref_linker, 0);
				}
			}
			if (probe->align)
			{
				gdl_tiling_nucleo_probe * qry_probe = (gdl_tiling_nucleo_probe *) probe->align;
				align_nucleo_chromosome * qry_chrom = (align_nucleo_chromosome *) qry_probe->chrom;
				qry_nucleo = align_nucleo_chromosome_call_nucleo (qry_chrom, qry_probe, qry_probe->chunk_idx, qry_probe->idx, nucleo_min, nclass, &qry_score, &qry_status, &qry_start, &qncidx, qry_chrom_ptr);
				if (qry_nucleo)
				{
					align_nucleo_chromosome_nucleo_properties (qry_chrom->hmm->chunks[qry_nucleo->chunk_idx], qry_nucleo);
					if (qry_nucleo->map >= prb_mapped_min)
					{
						qry_nucleo->ignore = 'n';
					}
					else
					{
						qry_nucleo->ignore = 'y';
						qry_rm++;
					}
					qry_nucleo->idx = gdl_list_size (qry_nucleosomes);
					gdl_list_push_back (qry_nucleosomes, qry_nucleo, 0);
					// take into accout the frame...
					if (qry_probe->idx > qry_start)
					{
						jj = j - (qry_probe->idx-qry_start);
					}
					else
					{
						jj = j - (qry_start-qry_probe->idx);	
					}
					for(k = jj; k < j; k++)
					{
						chunk->probes[k]->extra = qry_nucleo;
					}
					qry_nucleo->ref_start_pos = chunk->probes[jj]->position;
					qry_nucleo->ref_end_pos   = chunk->probes[j-1]->position;
					// get the next linker
					qry_linker=0;
					for(k = j; k < chunk->size && qry_linker==0; k++)
					{
						if (chunk->probes[k]->align)	
						{
							qry_probe  = (gdl_tiling_nucleo_probe *) chunk->probes[k]->align;
							qry_linker = align_nucleo_chromosome_call_linker (qry_probe->chrom, qry_probe, qry_probe->chunk_idx, qry_probe->idx, nclass, &lscore, &lcidx, &lstart, lqry_chrom_ptr);
						}
					}
					if (qry_linker && qry_probe)
					{
						// take into accout the frame...
						if (qry_probe->idx > lstart)
						{
							jj = k - (qry_probe->idx-lstart);
						}
						else
						{
							jj = k - (lstart-qry_probe->idx);	
						}
						qry_linker->ref_start_pos = chunk->probes[jj]->position;
						qry_linker->ref_end_pos   = chunk->probes[k-1]->position;
						qry_nucleo->next_linker   = qry_linker;
						qry_linker->idx = gdl_list_size (qry_linkers);
						align_nucleo_chromosome_linker_properties (((align_nucleo_chromosome *)qry_probe->chrom)->hmm->chunks[qry_linker->chunk_idx], qry_linker);
						gdl_list_push_back (qry_linkers, qry_linker, 0);
					}
				}
			}	
		}
	}
	GDL_FREE (ref_chrom_ptr);
	GDL_FREE (qry_chrom_ptr);
	GDL_FREE (lref_chrom_ptr);
	GDL_FREE (lqry_chrom_ptr);
	if ((chrom->nref_nucleosome=gdl_list_size (ref_nucleosomes))>0)
	{
		gdl_list_itr * itr = gdl_list_iterator_front (ref_nucleosomes);
		i=0;
		chrom->ref_nucleosomes = GDL_CALLOC (align_nucleo_nucleosome *, chrom->nref_nucleosome); 
		do
		{
			chrom->ref_nucleosomes[i++] = gdl_list_iterator_value (itr);
		}
		while(gdl_list_iterator_next (itr));
		gdl_list_iterator_free (itr);
	}
	if ((chrom->nqry_nucleosome=gdl_list_size (qry_nucleosomes))>0)
	{
		gdl_list_itr * itr = gdl_list_iterator_front (qry_nucleosomes);
		i=0;
		chrom->qry_nucleosomes = GDL_CALLOC (align_nucleo_nucleosome *, chrom->nqry_nucleosome); 
		do
		{
			chrom->qry_nucleosomes[i++] = gdl_list_iterator_value (itr);
		}
		while(gdl_list_iterator_next (itr));
		gdl_list_iterator_free (itr);
	}
	if ((chrom->nref_linker=gdl_list_size (ref_linkers))>0)
	{
		gdl_list_itr * itr = gdl_list_iterator_front (ref_linkers);
		i=0;
		chrom->ref_linkers = GDL_CALLOC (align_nucleo_linker *, chrom->nref_linker); 
		do
		{
			chrom->ref_linkers[i++] = gdl_list_iterator_value (itr);
		}
		while(gdl_list_iterator_next (itr));
		gdl_list_iterator_free (itr);
	}
	if ((chrom->nqry_linker=gdl_list_size (qry_linkers))>0)
	{
		gdl_list_itr * itr = gdl_list_iterator_front (qry_linkers);
		i=0;
		chrom->qry_linkers = GDL_CALLOC (align_nucleo_linker *, chrom->nqry_linker); 
		do
		{
			chrom->qry_linkers[i++] = gdl_list_iterator_value (itr);
		}
		while(gdl_list_iterator_next (itr));
		gdl_list_iterator_free (itr);
	}
	fprintf (stderr, "--\n");
	fprintf (stderr, " Nucleosme calling [ map = %1.4f ]\n", prb_mapped_min);
	fprintf (stderr, "\n");
	fprintf (stderr, "  reference %d nucleosomes (%d badly mapped)\n", chrom->nref_nucleosome, ref_rm);
	fprintf (stderr, "  reference %d linkers\n", chrom->nref_linker);
	fprintf (stderr, "  query     %d nucleosomes (%d badly mapped)\n", chrom->nqry_nucleosome, qry_rm);
	fprintf (stderr, "  query     %d linkers\n", chrom->nqry_linker);
	fprintf (stderr, "--\n");
	fflush(stderr);
	gdl_list_free (ref_nucleosomes);
	gdl_list_free (qry_nucleosomes);
}

void
align_nucleo_chromosome_align_neighbour (align_nucleo_chromosome * chrom,
                                           align_nucleo_nucleosome * ref_nucleo_start,
                                           align_nucleo_nucleosome * ref_nucleo_end,
                                           align_nucleo_nucleosome * qry_nucleo_start,
                                           align_nucleo_nucleosome * qry_nucleo_end,
                                           const gdl_boolean include_start,
                                           const gdl_boolean include_end)
{                                          
	size_t j,k,l;
	size_t ref_region_size = ref_nucleo_end->idx-ref_nucleo_start->idx-1+include_start+include_end;
	size_t qry_region_size = qry_nucleo_end->idx-qry_nucleo_start->idx-1+include_start+include_end;
	
	if (ref_region_size==0 || qry_region_size==0)
		return;
	
	align_nucleo_nucleosome * ref_nucleo, * qry_nucleo;
	double ** dx = GDL_MATRIX_ALLOC (double, ref_region_size, qry_region_size);
	double ref_x, qry_x, min, max=0, delta;
	size_t min_r,min_q;
	for(j = 0; j < ref_region_size; j++)
	{
		ref_nucleo   = chrom->ref_nucleosomes[ref_nucleo_start->idx+j+1-include_start];
		ref_x        = 0.5*(ref_nucleo->ref_start_pos+ref_nucleo->ref_end_pos);
		for(k = 0; k < qry_region_size; k++)
		{
			qry_nucleo = chrom->qry_nucleosomes[qry_nucleo_start->idx+k+1-include_start];
			qry_x      = 0.5*(qry_nucleo->ref_start_pos+qry_nucleo->ref_end_pos);
			dx[j][k]    = fabs(qry_x-ref_x);
			if (dx[j][k] > max) max = dx[j][k];
		}
	}
	for(j = 0; j < GDL_MIN (ref_region_size, qry_region_size); j++)
	{
		min = max+1;
		for(k = 0; k < ref_region_size; k++)
		{
			for(l = 0; l < qry_region_size; l++)
			{
				if (dx[k][l] != -1. && dx[k][l] < min)
				{
					min   = dx[k][l];
					min_r = k;
					min_q = l;
				}
			}
		}
		ref_nucleo = chrom->ref_nucleosomes[ref_nucleo_start->idx+min_r+1-include_start];
		qry_nucleo = chrom->qry_nucleosomes[qry_nucleo_start->idx+min_q+1-include_start];
		ref_nucleo->delta = min;
		qry_nucleo->delta = min;
		delta  = (ref_nucleo->ref_end_pos-ref_nucleo->ref_start_pos+1);
		delta += (qry_nucleo->ref_end_pos-qry_nucleo->ref_start_pos+1);
		delta /= 2.;
		if (min <= delta)
		{
			if (qry_nucleo->align != 0)
			{
				printf("ERROR QRY NUC ALREADY ALIGNED !!\n");
				exit(1);	
			}
			ref_nucleo->align = qry_nucleo;
			qry_nucleo->align = ref_nucleo;
			for(k = 0; k < ref_region_size; k++)
				dx[k][min_q] = -1.;
		}
		for(l = 0; l < qry_region_size; l++)
			dx[min_r][l] = -1.;
	}
	GDL_MATRIX_FREE (dx, ref_region_size);
}

align_nucleo_alignment *
align_nucleo_alignment_alloc (align_nucleo_chromosome * chrom)
{
	size_t i;
	gdl_string * key;
	align_nucleo_nucleosome ** pos_rec;
	align_nucleo_nucleosome * ref_nucleo, * qry_nucleo;
	gdl_hashtable * pos_table = gdl_hashtable_alloc (gdl_hash_default, 0);
	gdl_hashtable * qry_chrom_table = gdl_hashtable_alloc (gdl_hash_default, 0);
	
	for(i = 0; i < chrom->nref_nucleosome; i++)
	{
		ref_nucleo = chrom->ref_nucleosomes[i];
		key     = gdl_string_sprintf ("%d", ref_nucleo->ref_start_pos);
		pos_rec = GDL_CALLOC (align_nucleo_nucleosome *, 2);
		pos_rec[0] = ref_nucleo;
		gdl_hashtable_add (pos_table, key, pos_rec, 0);
		gdl_string_free (key);
	}
	for(i = 0; i < chrom->nqry_nucleosome; i++)
	{
		qry_nucleo = chrom->qry_nucleosomes[i];
		key        = gdl_string_sprintf ("%d", qry_nucleo->ref_start_pos);
		if ((pos_rec=gdl_hashtable_lookup (pos_table, key))!=0)
		{
			pos_rec[1] = qry_nucleo;
		}
		else
		{
			pos_rec = GDL_CALLOC (align_nucleo_nucleosome *, 2);
			pos_rec[1] = qry_nucleo;
		    gdl_hashtable_add (pos_table, key, pos_rec, 0);
		}
		gdl_string_free (key);
		gdl_hashtable_add (qry_chrom_table, qry_nucleo->chrom->data->name, qry_nucleo->chrom, 0);
	}
	
	align_nucleo_alignment * align = 0;
	if (gdl_hashtable_size (pos_table))
	{
		align = GDL_CALLOC (align_nucleo_alignment, 1);
		align->size = gdl_hashtable_size (pos_table);
		align->pos  = GDL_CALLOC (long, align->size);
		align->ref  = GDL_CALLOC (align_nucleo_nucleosome *, align->size);
		align->qry  = GDL_CALLOC (align_nucleo_nucleosome *, align->size);
		
		i=0;
		gdl_hashtable_itr * itr = gdl_hashtable_iterator (pos_table);
		do
		{
			key     = gdl_hashtable_iterator_key (itr);
			pos_rec = gdl_hashtable_iterator_value (itr);
			align->pos[i] = atol(key);
			align->ref[i] = pos_rec[0];
			align->qry[i] = pos_rec[1];
			GDL_FREE (pos_rec);
			i++;
		}
		while (gdl_hashtable_iterator_next (itr));
		gdl_hashtable_iterator_free (itr);
		// then, sort the alignement
		size_t * idx = GDL_CALLOC (size_t, align->size);
		
		gdl_sort_long_index (idx, align->pos, 1, align->size);
		
		align_nucleo_alignment * salign = GDL_CALLOC (align_nucleo_alignment, 1);
		salign->size = align->size;
		salign->pos  = GDL_CALLOC (long, align->size);
		salign->ref  = GDL_CALLOC (align_nucleo_nucleosome *, align->size);
		salign->qry  = GDL_CALLOC (align_nucleo_nucleosome *, align->size);
		
		for(i = 0; i < align->size; i++)
		{
			salign->pos[i] = align->pos[idx[i]];
			salign->ref[i] = align->ref[idx[i]];
			salign->qry[i] = align->qry[idx[i]];
			if (salign->ref[i]) salign->ref[i]->seen = 'n';
			if (salign->qry[i]) salign->qry[i]->seen = 'n';
		}
		
		GDL_FREE (idx);
		GDL_FREE (align->pos);
		GDL_FREE (align->ref);
		GDL_FREE (align->qry);
		GDL_FREE (align);
		
		align            = salign;
		align->ref_chrom = chrom;
	}
	gdl_hashtable_free (pos_table);
	
	if ((align->nqry=gdl_hashtable_size (qry_chrom_table))!=0)
	{
		gdl_hashtable_itr * itr = gdl_hashtable_iterator (qry_chrom_table);
		align->qry_chroms = GDL_MALLOC (align_nucleo_chromosome *, align->nqry);
		i=0;
		do
		{
			align->qry_chroms[i] = gdl_hashtable_iterator_value (itr);
			i++;
		}
		while (gdl_hashtable_iterator_next (itr));
		gdl_hashtable_iterator_free (itr);
	}
	gdl_hashtable_free (qry_chrom_table);
	
	return align;
}

align_nucleo_alignment *
align_nucleo_chromosome_align_nucleo (align_nucleo_chromosome * chrom,
                                        const double prb_overlap_min)
{
	size_t i,j,k,l;
	double overlap, qry_nuc_size, ref_nuc_size;
	align_nucleo_nucleosome * ref_nucleo, * qry_nucleo;
	gdl_tiling_nucleo_chunk * ref_chunk, * qry_chunk;
	gdl_tiling_nucleo_probe * ref_probe, * qry_probe;
	void * qry_extra;
	
	for(i = 0; i < chrom->nref_nucleosome; i++)
	{
		ref_nucleo = chrom->ref_nucleosomes[i];
		if (ref_nucleo->ignore == 'y')
			continue;
		ref_chunk  = chrom->hmm->chunks[ref_nucleo->chunk_idx];
		qry_extra  = 0;
		overlap    = 0.;
		for(j = ref_nucleo->start_idx; j <= ref_nucleo->end_idx; j++)
		{
			ref_probe = ref_chunk->probes[j];
			if (ref_probe->extra!=0)
			{
				if (!qry_extra) qry_extra = ref_probe->extra;
				else if (ref_probe->extra == qry_extra)
				{
					overlap+=1.;
				}
				else
				{
					overlap=0.;
					qry_extra=ref_probe->extra;
				}
			}
		}
		ref_nuc_size = (ref_nucleo->end_idx-ref_nucleo->start_idx+1);
		if (overlap > prb_overlap_min*ref_nuc_size)
		{
			qry_nucleo = (align_nucleo_nucleosome *) qry_extra;
			if (qry_nucleo->end_idx > qry_nucleo->start_idx)
			{
				qry_nuc_size = (qry_nucleo->end_idx-qry_nucleo->start_idx+1);
			}
			else
			{
				qry_nuc_size = (qry_nucleo->start_idx-qry_nucleo->end_idx+1);	
			}
			if (overlap > prb_overlap_min*qry_nuc_size && qry_nucleo->align==0)
			{
				if (ref_nucleo->ref_start_pos == 172502)
				{
					qry_nucleo = qry_extra;
				}
				ref_nucleo->align = qry_extra;
				qry_nucleo->align = ref_nucleo;
			}
		}
	}
	
	align_nucleo_nucleosome * ref_nucleo_start, * ref_nucleo_end;
	align_nucleo_nucleosome * qry_nucleo_start, * qry_nucleo_end;
	
	ref_nucleo_start = ref_nucleo_end = 0;
	for(i = 0; i < chrom->nref_nucleosome; i++)
	{
		ref_nucleo = chrom->ref_nucleosomes[i];
		if (ref_nucleo->ignore == 'y')
			continue;
		if (ref_nucleo->align)
		{
			if (ref_nucleo_start) ref_nucleo_end = ref_nucleo;
			else 
			{
				ref_nucleo_start = ref_nucleo;
				if (ref_nucleo_start->idx > 0)
				{
					align_nucleo_chromosome_align_neighbour (chrom, chrom->ref_nucleosomes[0], ref_nucleo, chrom->qry_nucleosomes[0], ref_nucleo->align, 1, 0);
				}
			}
		}
		if (ref_nucleo_end)
		{
			size_t ref_region_size = ref_nucleo_end->idx-ref_nucleo_start->idx-1;
			if (ref_region_size)
			{
				qry_nucleo_start = (align_nucleo_nucleosome *) ref_nucleo_start->align;
				qry_nucleo_end   = (align_nucleo_nucleosome *) ref_nucleo_end->align; 
				size_t qry_region_size = qry_nucleo_end->idx-qry_nucleo_start->idx-1;
				//printf ("REGION %d-->%d #nuc = %d | %d-->%d #nuc = %d\n", ref_nucleo_start->idx, ref_nucleo_end->idx, ref_region_size, qry_nucleo_start->idx, qry_nucleo_end->idx, qry_region_size);
				if (qry_region_size == ref_region_size)
				{
					for(j = 0; j < ref_region_size; j++)
					{
						ref_nucleo        = chrom->ref_nucleosomes[ref_nucleo_start->idx+j+1];
						qry_nucleo        = chrom->qry_nucleosomes[qry_nucleo_start->idx+j+1];
						if (qry_nucleo->align != 0)
						{
							printf("ERROR QRY NUC ALREADY ALIGNED !!\n");
							exit(1);	
						}
						ref_nucleo->align = qry_nucleo;
						qry_nucleo->align = ref_nucleo;
					}
				}
				else if (qry_region_size)
				{
					align_nucleo_chromosome_align_neighbour (chrom, ref_nucleo_start, ref_nucleo_end, qry_nucleo_start, qry_nucleo_end, 0, 0);
				}
			} 
			ref_nucleo_start = ref_nucleo_end;
			ref_nucleo_end   = 0;
		}
	}
	if (ref_nucleo_start->idx < chrom->nref_nucleosome)
	{
		ref_nucleo_end   = chrom->ref_nucleosomes[chrom->nref_nucleosome-1];
		qry_nucleo_start = ref_nucleo_start->align;
		qry_nucleo_end   = chrom->qry_nucleosomes[chrom->nqry_nucleosome-1];
		align_nucleo_chromosome_align_neighbour (chrom, ref_nucleo_start, ref_nucleo_end, qry_nucleo_start, qry_nucleo_end, 0, 1);
	}
	
	return align_nucleo_alignment_alloc (chrom);
}

double
align_nucleo_nucleosome_genotype_prob (const align_nucleo_nucleosome * ref_nucleo, const align_nucleo_nucleosome * qry_nucleo)
{
	size_t i,j;
	double ref_pr, qry_pr;
	
	i = (ref_nucleo->status/2+ref_nucleo->status%2);
	j = (qry_nucleo->status/2+qry_nucleo->status%2);
	
	return (ref_nucleo->prob[i])*(qry_nucleo->prob[j]);
}

// insertion = nuc in ref and not in qry
// deletion  = nuc in qry and not in ref
double
align_nucleo_nucleosome_indel_prob (const align_nucleo_nucleosome * nucleosome, const size_t min_p)
{
	size_t i, np = 0;
	double pr = 0;
	size_t ps = GDL_MIN (nucleosome->start_idx, nucleosome->end_idx);
	size_t pe = GDL_MAX (nucleosome->start_idx, nucleosome->end_idx);
	for(i = ps; i <= pe; i++)
	{
		gdl_tiling_nucleo_probe * probe_nuc = nucleosome->chrom->hmm->chunks[nucleosome->chunk_idx]->probes[i];
		gdl_tiling_nucleo_probe * qry_probe = (gdl_tiling_nucleo_probe *) probe_nuc->align;
		if (qry_probe)
		{
			pr += (1-probe_nuc->prob[0])*qry_probe->prob[0];
			np++;
		}
	}
	return (np > min_p) ? pr/np : 0;
}

// probabilities of the four different states
// 0 = L x L
// 1 = N x L (insertion)
// 2 = L x N (deletion)
// 3 = N x N (aligned)
int
align_nucleo_nucleosome_state_prob (const align_nucleo_nucleosome * ref_nuc,
                                    const align_nucleo_nucleosome * qry_nuc,
                                    double pr[])
{
	const align_nucleo_nucleosome * nucleosome;
	if (ref_nuc)
	{
		nucleosome = ref_nuc;
	}
	else
	{
		nucleosome = qry_nuc;	
	}
	size_t i, np = 0;
	size_t ps = GDL_MIN (nucleosome->start_idx, nucleosome->end_idx);
	size_t pe = GDL_MAX (nucleosome->start_idx, nucleosome->end_idx);
	for(i = 0; i < 4; i++) pr[i] = 0;
	for(i = ps; i <= pe; i++)
	{
		gdl_tiling_nucleo_probe * probe_nuc = nucleosome->chrom->hmm->chunks[nucleosome->chunk_idx]->probes[i];
		gdl_tiling_nucleo_probe * qry_probe = (gdl_tiling_nucleo_probe *) probe_nuc->align;
		if (qry_probe)
		{
			pr[0] += probe_nuc->prob[0]*qry_probe->prob[0];
			if (ref_nuc) 
			{
				pr[1] += (1.0-probe_nuc->prob[0])*qry_probe->prob[0];
				pr[2] += probe_nuc->prob[0]*(1.0-qry_probe->prob[0]);
			}
			else
			{
				pr[2] += (1.0-probe_nuc->prob[0])*qry_probe->prob[0];
				pr[1] += probe_nuc->prob[0]*(1.0-qry_probe->prob[0]);
			}
			pr[3] += (1.0-probe_nuc->prob[0])*(1.0-qry_probe->prob[0]);
			np++;
		}
	}
	if (np)
	{
		for(i = 0; i < 4; i++)
		{
			pr[i] /= np;
		}
	}
	return np;
}

// probabilities of the four different states
// 0 = L x L
// 1 = N x L (insertion)
// 2 = L x N (deletion)
// 3 = N x N (aligned)
int
align_nucleo_linker_state_prob (const align_nucleo_linker * ref_nuc,
                                const align_nucleo_linker * qry_nuc,
                                double pr[])
{
	const align_nucleo_linker * nucleosome;
	if (ref_nuc)
	{
		nucleosome = ref_nuc;
	}
	else
	{
		nucleosome = qry_nuc;	
	}
	size_t i, np = 0;
	size_t ps = GDL_MIN (nucleosome->start_idx, nucleosome->end_idx);
	size_t pe = GDL_MAX (nucleosome->start_idx, nucleosome->end_idx);
	for(i = ps; i <= pe; i++)
	{
		gdl_tiling_nucleo_probe * probe_nuc = nucleosome->chrom->hmm->chunks[nucleosome->chunk_idx]->probes[i];
		gdl_tiling_nucleo_probe * qry_probe = (gdl_tiling_nucleo_probe *) probe_nuc->align;
		if (qry_probe)
		{
			pr[0] += probe_nuc->prob[0]*qry_probe->prob[0];
			if (ref_nuc) 
			{
				pr[1] += (1.0-probe_nuc->prob[0])*qry_probe->prob[0];
				pr[2] += probe_nuc->prob[0]*(1.0-qry_probe->prob[0]);
			}
			else
			{
				pr[2] += (1.0-probe_nuc->prob[0])*qry_probe->prob[0];
				pr[1] += probe_nuc->prob[0]*(1.0-qry_probe->prob[0]);
			}
			pr[3] += (1.0-probe_nuc->prob[0])*(1.0-qry_probe->prob[0]);
			np++;
		}
	}
	if (np)
	{
		for(i = 0; i < 4; i++)
		{
			pr[i] /= np;
		}
	}
	return np;
}

double
align_nucleo_linker_genotype_prob (const align_nucleo_linker * ref_linker, const align_nucleo_linker * qry_linker)
{
	return (ref_linker->prob[0])*(qry_linker->prob[0]);
}

int
align_nucleo_alignment_fprintf (FILE * stream, const align_nucleo_alignment * align)
{
	size_t i, j, ref_allele, qry_allele, geno;
	double sd, * pr = GDL_MALLOC (double, 4);
	align_nucleo_nucleosome * ref_nucleo, * qry_nucleo;
	
	for(i = 0; i < align->size; i++)
	{
		if (align->ref[i] && align->ref[i]->seen != 'y')
		{
			ref_nucleo = align->ref[i];
			ref_nucleo->seen = 'y';
			align_nucleo_nucleosome_state_prob (ref_nucleo, ref_nucleo->align, pr);
			fprintf (stream, "%s %d %c %d %d %d %d %g %g %g", align->ref_chrom->data->name, ref_nucleo->chunk_idx, ref_nucleo->ignore, ref_nucleo->outlier, ref_nucleo->ref_start_pos, ref_nucleo->ref_end_pos, ref_nucleo->status, ref_nucleo->delta, ref_nucleo->indel, 1.0-ref_nucleo->prob[0]);
			if (ref_nucleo->align)
			{
				qry_nucleo = ref_nucleo->align;
				qry_nucleo->seen = 'y';
				fprintf (stream, " %d %c %d %d %d %d %g %g %g", qry_nucleo->chunk_idx, qry_nucleo->ignore, qry_nucleo->outlier, qry_nucleo->ref_start_pos, qry_nucleo->ref_end_pos, qry_nucleo->status, qry_nucleo->delta, qry_nucleo->indel, 1.0-ref_nucleo->prob[0]);
				ref_allele = ref_nucleo->status/2+ref_nucleo->status%2 - 1;
				qry_allele = qry_nucleo->status/2+qry_nucleo->status%2 - 1;
				geno = ref_allele + qry_allele;
				fprintf (stream, " %g", ref_nucleo->pval);
				fprintf (stream, " %g %d", ref_nucleo->gamma, geno);//, align_nucleo_nucleosome_genotype_prob (ref_nucleo, qry_nucleo));
			}
			else
			{
				fprintf (stream, " - - - - - - - - - - -1 %g", align_nucleo_nucleosome_indel_prob (ref_nucleo, 5));
			}
			for(j = 0; j < 4; j++)
			{
				fprintf (stream, " %1.5f", pr[j]); 	
			}
			fprintf (stream, "\n");
		}
		else if (align->qry[i] && align->qry[i]->seen != 'y')
		{
			qry_nucleo = align->qry[i];
			qry_nucleo->seen = 'y';
			align_nucleo_nucleosome_state_prob (qry_nucleo->align, qry_nucleo, pr);
			if (qry_nucleo->align)
			{
				ref_nucleo = qry_nucleo->align;
				ref_nucleo->seen = 'y';
				fprintf (stream, "%s %d %c %d %d %d %d %g %g %g", align->ref_chrom->data->name, ref_nucleo->chunk_idx, ref_nucleo->ignore, ref_nucleo->outlier, ref_nucleo->ref_start_pos, ref_nucleo->ref_end_pos, ref_nucleo->status, ref_nucleo->delta, ref_nucleo->indel, 1.0-ref_nucleo->prob[0]);
				fprintf (stream, " %d %c %d %d %d %d %g %g %g", qry_nucleo->chunk_idx, qry_nucleo->ignore, qry_nucleo->outlier, qry_nucleo->ref_start_pos, qry_nucleo->ref_end_pos, qry_nucleo->status, qry_nucleo->delta, qry_nucleo->indel, 1.0-qry_nucleo->prob[0]);
				ref_allele = ref_nucleo->status/2+ref_nucleo->status%2 - 1;
				qry_allele = qry_nucleo->status/2+qry_nucleo->status%2 - 1;
				geno = ref_allele + qry_allele;
				fprintf (stream, " %g", ref_nucleo->pval);
				fprintf (stream, " %g %d", ref_nucleo->gamma, geno);//, align_nucleo_nucleosome_genotype_prob (ref_nucleo, qry_nucleo));
			}
			else
			{
				fprintf (stream, "%s - - - - - - - - -", align->ref_chrom->data->name);
				fprintf (stream, " %d %c %d %d %d %d %g %g %g - -1 %g", qry_nucleo->chunk_idx, qry_nucleo->ignore, qry_nucleo->outlier, qry_nucleo->ref_start_pos, qry_nucleo->ref_end_pos, qry_nucleo->status, qry_nucleo->delta, qry_nucleo->indel, 1.0-qry_nucleo->prob[0], align_nucleo_nucleosome_indel_prob (qry_nucleo, 5));
			}
			for(j = 0; j < 4; j++)
			{
				fprintf (stream, " %1.5f", pr[j]);
			}
			fprintf (stream, "\n");
		}
	}
	for(i = 0; i < align->size; i++)
	{
		if (align->ref[i]) align->ref[i]->seen = 'n';
		if (align->qry[i]) align->qry[i]->seen = 'n';
	}
	GDL_FREE (pr);
}

void
compute_nucleo_pval_average (const align_nucleo_config     * conf, 
                              const align_nucleo_chromosome * ref_chrom,
                              align_nucleo_nucleosome * ref_nucleo,
                              align_nucleo_nucleosome * qry_nucleo)
{
	size_t p,n;
	int pidx;
	double * x = GDL_CALLOC (double, 2);
	double d,v;
	if (ref_nucleo)
	{
		const gdl_tiling_nucleo_chunk * chunk = ref_chrom->hmm->chunks[ref_nucleo->chunk_idx];
		d = 0;
		for(n = 0, p = ref_nucleo->start_idx; p <= ref_nucleo->end_idx; p++)
		{
			gdl_tiling_nucleo_probe * probe = chunk->probes[p];
			if ((pidx=probe->link_idx)>=0)
			{
				probe->pval = gdl_tiling_data_get_val (conf->ref_stat, ref_chrom->data->idx, pidx, 0, 0);
				d += probe->pval; 
				n++;
			}
		}
		if (n)
		{
			d /= n;
			ref_nucleo->pval = d;
		}
	}
	if (qry_nucleo)
	{
		const gdl_tiling_nucleo_chunk * chunk = qry_nucleo->chrom->hmm->chunks[qry_nucleo->chunk_idx];
		d = 0;
		size_t p_start = GDL_MIN (qry_nucleo->start_idx, qry_nucleo->end_idx);
		size_t p_end   = GDL_MAX (qry_nucleo->start_idx, qry_nucleo->end_idx);
		for(n = 0, p = p_start; p <= p_end; p++)
		{
			gdl_tiling_nucleo_probe * probe = chunk->probes[p];
			if ((pidx=probe->link_idx)>=0)
			{
				probe->pval = gdl_tiling_data_get_val (conf->qry_stat, qry_nucleo->chrom->data->idx, pidx, 0, 0);
				d += probe->pval;
				n++;
			}
		}
		if (n)
		{
			d /= n;
			qry_nucleo->pval = d;
		}
	}
	GDL_FREE (x);	
}  

size_t 
_anova_linker_size (const align_nucleo_linker * ref_linker,
                    const align_nucleo_linker * qry_linker)
{
	size_t p,rp,qp,rn,qn;
	int pidx;
	const gdl_tiling_nucleo_chunk * ref_chunk = ref_linker->chrom->hmm->chunks[ref_linker->chunk_idx];
	for(rp = rn = 0, p = ref_linker->start_idx; p <= ref_linker->end_idx; p++)
	{
		gdl_tiling_nucleo_probe * probe = ref_chunk->probes[p];
		if ((pidx=probe->link_idx)>=0)
		{
			rp++;
			rn=ref_linker->chrom->data->narray;
		}
	}
	const gdl_tiling_nucleo_chunk * qry_chunk = qry_linker->chrom->hmm->chunks[qry_linker->chunk_idx];
	size_t p_start = GDL_MIN (qry_linker->start_idx, qry_linker->end_idx);
	size_t p_end   = GDL_MAX (qry_linker->start_idx, qry_linker->end_idx);
	for(qn = qp = 0, p = p_start; p <= p_end; p++)
	{
		gdl_tiling_nucleo_probe * probe = qry_chunk->probes[p];
		if ((pidx=probe->link_idx)>=0)
		{
			qp++;
		}
	}

	return (rp+qp)*rn;
}   

size_t 
_anova_linker_design (const align_nucleo_config     * conf,
                      const align_nucleo_linker * ref_linker,
                      const align_nucleo_linker * qry_linker,
                      gdl_matrix * X0,
                      gdl_matrix * X1,
                      gdl_vector * y)
{
	size_t p,g,na,a,i,nm;
	int pidx;
	double yi;
	
	const gdl_tiling_data * ref_data = conf->ref_data;
	const gdl_tiling_nucleo_chunk * ref_chunk = ref_linker->chrom->hmm->chunks[ref_linker->chunk_idx];
	for(i = 0, p = ref_linker->start_idx; p <= ref_linker->end_idx; p++)
	{
		gdl_tiling_nucleo_probe * probe = ref_chunk->probes[p];
		if ((pidx=probe->link_idx)>=0)
		{
			for(g = 0; g < 2; g++) // loop on manip
			{
				na = ref_data->narrayByGroup[g]; 
				for(a = 0; a < na; a++, i++) // loop on replicates
				{
					yi=gdl_tiling_data_get_val (ref_data, ref_linker->chrom->data->idx, pidx, g, a);
					gdl_matrix_set (X0, i, 0, 1.);
					gdl_matrix_set (X0, i, 1, 1.);
					if (!g) 
					{
						gdl_matrix_set (X0, i, 2, 1.);
						gdl_matrix_set (X1, i, 0, 1.);
					}
					gdl_vector_set (y, i, yi);
				}
			}
		}
	}
	const gdl_tiling_data * qry_data = conf->qry_data;
	const gdl_tiling_nucleo_chunk * qry_chunk = qry_linker->chrom->hmm->chunks[qry_linker->chunk_idx];
	size_t p_start = GDL_MIN (qry_linker->start_idx, qry_linker->end_idx);
	size_t p_end   = GDL_MAX (qry_linker->start_idx, qry_linker->end_idx);
	for(p = p_start; p <= p_end; p++)
	{
		gdl_tiling_nucleo_probe * probe = qry_chunk->probes[p];
		if ((pidx=probe->link_idx)>=0)
		{
			for(g = 0; g < 2; g++) // loop on manip
			{
				na = qry_data->narrayByGroup[g]; 
				for(a = 0; a < na; a++, i++) // loop on replicates
				{
					yi=gdl_tiling_data_get_val (qry_data, qry_linker->chrom->data->idx, pidx, g, a);
					gdl_matrix_set (X0, i, 0, 1.);
					if (!g) gdl_matrix_set (X0, i, 2, 1.);
					gdl_vector_set (y, i, yi);
				}
			}
		}
	}
}                             

size_t 
_anova_nucleo_size (const align_nucleo_nucleosome * ref_nucleo,
                    const align_nucleo_nucleosome * qry_nucleo,
                    size_t * np)
{
	size_t p,n,q;
	int pidx;
	const gdl_tiling_nucleo_chunk * ref_chunk = ref_nucleo->chrom->hmm->chunks[ref_nucleo->chunk_idx];
	
	*np=0;
	for(n = 0, p = ref_nucleo->start_idx; p <= ref_nucleo->end_idx; p++)
	{
		gdl_tiling_nucleo_probe * probe = ref_chunk->probes[p];
		probe->eligible = gdl_false;
		if ((pidx=probe->link_idx)>=0 && probe->align)
		{
			n+=ref_nucleo->chrom->data->narray;
			probe->eligible = gdl_true;
			(*np)++;
		}
	}
	const gdl_tiling_nucleo_chunk * qry_chunk = qry_nucleo->chrom->hmm->chunks[qry_nucleo->chunk_idx];
	size_t p_start = GDL_MIN (qry_nucleo->start_idx, qry_nucleo->end_idx);
	size_t p_end   = GDL_MAX (qry_nucleo->start_idx, qry_nucleo->end_idx);
	for(q = 0, p = p_start; p <= p_end; p++)
	{
		gdl_tiling_nucleo_probe * probe = qry_chunk->probes[p];
		probe->eligible = gdl_false;
		if ((pidx=probe->link_idx)>=0 && (probe->align && ((gdl_tiling_nucleo_probe *)probe->align)->eligible))
		{
			probe->eligible = gdl_true;
			n+=qry_nucleo->chrom->data->narray;
			q++;
		}
		else if ((pidx=probe->link_idx)<0 && (probe->align && ((gdl_tiling_nucleo_probe *)probe->align)->eligible))
		{
			((gdl_tiling_nucleo_probe *)probe->align)->eligible=gdl_false;
			if ((*np) > 0) (*np)--;
		}		
	}
	return n;
}   

/**
 * _anova_nucleo_size() MUST HAVE BEEN RUN BEFORE !!
 */
size_t 
_anova_nucleo_design (const align_nucleo_config     * conf,
                      const align_nucleo_nucleosome * ref_nucleo,
                      const align_nucleo_nucleosome * qry_nucleo,
                      gdl_matrix * X0,
                      gdl_matrix * X1,
                      gdl_vector * y,
                      const size_t P,
                      const gdl_boolean no_probe_effect_flag,
                      FILE * stream)
{
	size_t g,na,a,i,np;
	int pidx,p;
	double yi;
	
	const gdl_tiling_data * ref_data = conf->ref_data;
	const gdl_tiling_nucleo_chunk * ref_chunk = ref_nucleo->chrom->hmm->chunks[ref_nucleo->chunk_idx];
	for(np = i = 0, p = ref_nucleo->start_idx; np < P && p <= ref_nucleo->end_idx; p++)
	{
		gdl_tiling_nucleo_probe * probe = ref_chunk->probes[p];
		if ((pidx=probe->link_idx)>=0 && probe->eligible && probe->align)
		{
			np++;
			for(g = 0; g < 2; g++) // loop on manip
			{
				na = ref_data->narrayByGroup[g]; 
				for(a = 0; a < na; a++, i++) // loop on replicates
				{
					yi=gdl_tiling_data_get_val (ref_data, ref_nucleo->chrom->data->idx, pidx, g, a);
					//printf ("%d (%d,%d)\n", i, X0->size1, X0->size2);
					gdl_matrix_set (X0, i, 0, 1.);
					gdl_matrix_set (X0, i, 1, 1.);
					if (!g) 
					{
						gdl_matrix_set (X0, i, 2, 1.);
						gdl_matrix_set (X1, i, 0, 1.);
					}
					if (!no_probe_effect_flag && np < P) gdl_matrix_set (X0, i, 2+np, 1.);
					gdl_vector_set (y, i, yi);
					if (stream) fprintf (stream, "%s %d %d S1 %s %d %g\n", ref_nucleo->chrom->data->name, probe->position, probe->position, (g==0) ? "E1" : "E2", np, yi);
				}
			}
		}
	}
	const gdl_tiling_data * qry_data = conf->qry_data;
	const gdl_tiling_nucleo_chunk * qry_chunk = qry_nucleo->chrom->hmm->chunks[qry_nucleo->chunk_idx];
	if (qry_nucleo->start_idx > qry_nucleo->end_idx)
	{
		for(np = 0, p = qry_nucleo->start_idx; np < P && p >= qry_nucleo->end_idx; p--)
		{
			gdl_tiling_nucleo_probe * probe = qry_chunk->probes[p];
			if ((pidx=probe->link_idx)>=0 && probe->eligible && probe->align && ((gdl_tiling_nucleo_probe *)probe->align)->eligible) 
			{
				np++;
				for(g = 0; g < 2; g++) // loop on manip
				{
					na = qry_data->narrayByGroup[g]; 
					for(a = 0; a < na; a++, i++) // loop on replicates
					{
						yi=gdl_tiling_data_get_val (qry_data, qry_nucleo->chrom->data->idx, pidx, g, a);
						gdl_matrix_set (X0, i, 0, 1.);
						if (!g) gdl_matrix_set (X0, i, 2, 1.);
						if (!no_probe_effect_flag && np < P) gdl_matrix_set (X0, i, 2+np, 1.);
						gdl_vector_set (y, i, yi);
						if (stream) fprintf (stream, "%s %d %d S2 %s %d %g\n", qry_nucleo->chrom->data->name, probe->position, ((gdl_tiling_nucleo_probe *)probe->align)->position, (g==0) ? "E1" : "E2", np, yi);
					}
				}
			}
		}
	}
	else
	{
		for(np = 0, p = qry_nucleo->start_idx;  np < P && p <= qry_nucleo->end_idx; p++)
		{
			gdl_tiling_nucleo_probe * probe = qry_chunk->probes[p];
			if ((pidx=probe->link_idx)>=0 && probe->eligible && probe->align && ((gdl_tiling_nucleo_probe *)probe->align)->eligible)
			{
				np++;
				for(g = 0; g < 2; g++) // loop on manip
				{
					na = qry_data->narrayByGroup[g]; 
					for(a = 0; a < na; a++, i++) // loop on replicates
					{
						yi=gdl_tiling_data_get_val (qry_data, qry_nucleo->chrom->data->idx, pidx, g, a);
						gdl_matrix_set (X0, i, 0, 1.);
						if (!g) gdl_matrix_set (X0, i, 2, 1.);
						if (!no_probe_effect_flag && np < P) gdl_matrix_set (X0, i, 2+np, 1.);
						gdl_vector_set (y, i, yi);
						if (stream) fprintf (stream, "%s %d %d S2 %s %d %g\n", qry_nucleo->chrom->data->name, probe->position, ((gdl_tiling_nucleo_probe *)probe->align)->position, (g==0) ? "E1" : "E2", np, yi);
					}
				}
			}
		}
	}
}  

size_t 
_anova_nucleo_mrk_design (const align_nucleo_config     * conf,
	                      const align_nucleo_nucleosome * ref_nucleo,
	                      const align_nucleo_nucleosome * qry_nucleo,
	                      gdl_matrix * X0,
	                      gdl_matrix * X1,
	                      gdl_vector * y,
	                      const size_t P)
{
	size_t p,g,na,a,i,np;
	int pidx;
	double yi;
	
	const gdl_tiling_data * ref_data = conf->ref_data;
	const gdl_tiling_nucleo_chunk * ref_chunk = ref_nucleo->chrom->hmm->chunks[ref_nucleo->chunk_idx];
	for(np = i = 0, p = ref_nucleo->start_idx; p <= ref_nucleo->end_idx; p++)
	{
		gdl_tiling_nucleo_probe * probe = ref_chunk->probes[p];
		if ((pidx=probe->link_idx)>=0 && probe->align)
		{
			np++;
			for(g = 0; g < 2; g++) // loop on manip
			{
				na = ref_data->narrayByGroup[g]; 
				for(a = 0; a < na; a++, i++) // loop on replicates
				{
					yi=gdl_tiling_data_get_val (ref_data, ref_nucleo->chrom->data->idx, pidx, g, a);
					gdl_matrix_set (X0, i, 0, 1.); // g. mean
					gdl_matrix_set (X0, i, 1, 1.); // strain effect
					if (!g) 
					{
						gdl_matrix_set (X1, i, 0, 1.); // exp. effect
					}
					if (np < P) gdl_matrix_set (X0, i, 1+np, 1.); // probe effect
					gdl_vector_set (y, i, yi);
				}
			}
		}
	}
	const gdl_tiling_data * qry_data = conf->qry_data;
	const gdl_tiling_nucleo_chunk * qry_chunk = qry_nucleo->chrom->hmm->chunks[qry_nucleo->chunk_idx];
	size_t p_start = GDL_MIN (qry_nucleo->start_idx, qry_nucleo->end_idx);
	size_t p_end   = GDL_MAX (qry_nucleo->start_idx, qry_nucleo->end_idx);
	for(np = 0, p = p_start; p <= p_end; p++)
	{
		gdl_tiling_nucleo_probe * probe = qry_chunk->probes[p];
		if ((pidx=probe->link_idx)>=0 && probe->align)
		{
			np++;
			for(g = 0; g < 2; g++) // loop on manip
			{
				na = qry_data->narrayByGroup[g]; 
				for(a = 0; a < na; a++, i++) // loop on replicates
				{
					yi=gdl_tiling_data_get_val (qry_data, qry_nucleo->chrom->data->idx, pidx, g, a);
					gdl_matrix_set (X0, i, 0, 1.);
					if (!g) gdl_matrix_set (X1, i, 0, 1.);
					if (np < P) gdl_matrix_set (X0, i, 1+np, 1.);
					gdl_vector_set (y, i, yi);
				}
			}
		}
	}
}  

void
compute_linker_pval (const align_nucleo_config     * conf, 
                      const align_nucleo_chromosome * ref_chrom,
                      align_nucleo_linker * ref_linker,
                      align_nucleo_linker * qry_linker,
                      const double obs_min)
{
	if (ref_linker->obs < obs_min)
	{
		qry_linker->pval = ref_linker->pval = -1;
		return;
	}
	if (qry_linker->obs < obs_min)
	{
		qry_linker->pval = ref_linker->pval = -1;
		return;	
	}
	if (ref_linker && qry_linker)
	{
		size_t i, j, N, M;
		gdl_multireg_linear_workspace * W;
		gdl_matrix * X0, * X1;
		gdl_vector * y;
		
		// init
		N  = _anova_linker_size (ref_linker, qry_linker);
		M  = 3;
		X0 = gdl_matrix_alloc (N, M);
		X1 = gdl_matrix_alloc (N, 1);
		y  = gdl_vector_alloc (N);
		W  = gdl_multireg_linear_workspace_alloc (N, M+1);
		
		_anova_linker_design (conf, ref_linker, qry_linker, X0, X1, y);

		gdl_multireg_linear_eval (W, X0, X1, y);
		
		qry_linker->pval = gdl_multireg_linear_eval_pvalue(W);
		ref_linker->pval = qry_linker->pval;

		// clean
		gdl_vector_free (y);
		gdl_matrix_free (X1);
		gdl_matrix_free (X0);
		gdl_multireg_linear_workspace_free (W);
	}
	else if (ref_linker)
	{
		ref_linker->pval = -1;	
	}
	else if (qry_linker)
	{
		qry_linker->pval = -1;
	}
}          

void
compute_nucleo_pval (const align_nucleo_config     * conf, 
                      const align_nucleo_chromosome * ref_chrom,
                      align_nucleo_nucleosome * ref_nucleo,
                      align_nucleo_nucleosome * qry_nucleo,
                      const double obs_min,
                      const size_t prb_min,
                      const size_t permut,
                      const gdl_boolean no_probe_effect_flag, 
                      const gdl_rng * rng,
                      FILE * stream)
{
	if (conf->ref_stat && conf->qry_stat)
	{
		compute_nucleo_pval_average (conf, ref_chrom, ref_nucleo, qry_nucleo);
	}
	else if (ref_nucleo && qry_nucleo)
	{
		if (ref_nucleo->obs < obs_min)
		{
			qry_nucleo->pval = ref_nucleo->pval = -1;
			return;
		}
		if (qry_nucleo->obs < obs_min)
		{
			qry_nucleo->pval = ref_nucleo->pval = -1;
			return;	
		}
		size_t i, j, N, M, P;
		gdl_multireg_linear_workspace * W;
		gdl_matrix * X0, * X1;
		gdl_vector * y;
		
		// init
		N = _anova_nucleo_size (ref_nucleo, qry_nucleo, &P);
		if (!P || P < prb_min)
		{
			qry_nucleo->pval = ref_nucleo->pval = -1;
			return;	
		}
		M = 3;
		if (no_probe_effect_flag)
			X0 = gdl_matrix_alloc (N, M);
		else
			X0 = gdl_matrix_alloc (N, M+P-1);
		X1 = gdl_matrix_alloc (N, 1);
		y  = gdl_vector_alloc (N);
		if (no_probe_effect_flag)
			W  = gdl_multireg_linear_workspace_alloc (N, M+1);
		else
			W  = gdl_multireg_linear_workspace_alloc (N, M+P);
		
		if (stream) 
		{
			fprintf(stream, "<nucleosome_alignment chrom=\"%s\" ref_start=\"%d\" ref_end=\"%d\" qry_start=\"%d\" qry_end=\"%d\" shared-probes=\"%d\">\n",ref_nucleo->chrom->data->name,ref_nucleo->ref_start_pos,ref_nucleo->ref_end_pos,qry_nucleo->ref_start_pos,qry_nucleo->ref_end_pos, P);
			fprintf(stream, "<data>\n");
		}
		
		_anova_nucleo_design (conf, ref_nucleo, qry_nucleo, X0, X1, y, P, no_probe_effect_flag, stream);
		
		if (permut)
		{
			gdl_multireg_linear_eval_permut (W, X0, X1, y, permut, rng);
		}
		else
		{
			//printf ("EVAL X0 [%d, %d] ", X0->size1, X0->size2);
			gdl_multireg_linear_eval (W, X0, X1, y);
			//printf (" : OK\n");
			//fflush(stdout);
		}
		
		qry_nucleo->pval  = gdl_multireg_linear_eval_pvalue(W);
		ref_nucleo->pval  = qry_nucleo->pval;
		qry_nucleo->gamma = gdl_vector_get (gdl_multireg_linear_coefficients(W), X0->size2);
		ref_nucleo->gamma = qry_nucleo->gamma;
		
		if (stream) 
		{
			fprintf(stream, "</data>\n");
			fprintf(stream, "<p-value target=\"interaction\">%e</p-value>\n",qry_nucleo->pval);
			fprintf(stream, "</nucleosome_alignment>\n");
			fflush(stream);
		}
		
		// clean
		gdl_vector_free (y);
		gdl_matrix_free (X1);
		gdl_matrix_free (X0);
		gdl_multireg_linear_workspace_free (W);
		
//		M  = 2;
//		X0 = gdl_matrix_alloc (N, M+P-1);
//		X1 = gdl_matrix_alloc (N, 1);
//		y  = gdl_vector_alloc (N);
//		W  = gdl_multireg_linear_workspace_alloc (N, M+P);
//		
//		_anova_nucleo_mrk_design (conf, ref_nucleo, qry_nucleo, X0, X1, y, P);
//		
////		for(i = 0; i < X0->size1; i++)
////		{
////			printf ("%g %1.0f", gdl_vector_get (y, i), gdl_matrix_get (X1, i, 0));
////			for(j = 1; j <  X0->size2; j++)
////			{
////				printf (" %1.0f", gdl_matrix_get (X0, i, j));
////			}
////			printf ("\n");
////		}
//		
//		if (permut)
//		{
//			gdl_multireg_linear_eval_permut (W, X0, X1, y, permut, rng);
//		}
//		else
//		{
//			//printf ("EVAL X0 [%d, %d] ", X0->size1, X0->size2);
//			gdl_multireg_linear_eval (W, X0, X1, y);
//			//printf (" : OK\n");
//			//fflush(stdout);
//		}
//		
//		qry_nucleo->mrk_pval  = gdl_multireg_linear_eval_pvalue (W);
//		ref_nucleo->mrk_pval  = qry_nucleo->mrk_pval;
//		qry_nucleo->mrk_gamma = gdl_vector_get (gdl_multireg_linear_coefficients(W), X0->size2);
//		ref_nucleo->mrk_gamma = qry_nucleo->mrk_gamma;
//		
//		//printf ("PVAL = %g GAMMA = %g\n", qry_nucleo->mrk_pval, qry_nucleo->mrk_gamma);
//		// clean
//		gdl_vector_free (y);
//		gdl_matrix_free (X1);
//		gdl_matrix_free (X0);
//		gdl_multireg_linear_workspace_free (W);
		
	}
	else if (ref_nucleo)
	{
		ref_nucleo->pval = -1;	
	}
	else if (qry_nucleo)
	{
		qry_nucleo->pval = -1;
	}
}  

void
align_nucleo_alignment_pval (const align_nucleo_config    * conf, 
                             const align_nucleo_chromosome * chrom,
                             align_nucleo_alignment * align,
                             const double obs_min,
                             const size_t prb_min,
                             const size_t permut,
                             const gdl_boolean no_probe_effect_flag,
                             const gdl_rng * rng)
{
	size_t i;
	align_nucleo_nucleosome * ref_nucleo, * qry_nucleo;

	for(i = 0; i < align->size; i++)
	{
		if (align->ref[i] && align->ref[i]->seen != 'y' && align->ref[i]->ignore != 'y')
		{
			ref_nucleo = align->ref[i];
			ref_nucleo->seen = 'y';
			if (ref_nucleo->align)
			{
				qry_nucleo = ref_nucleo->align;
				qry_nucleo->seen = 'y';
				compute_nucleo_pval (conf, chrom, ref_nucleo, qry_nucleo, obs_min, prb_min, permut, no_probe_effect_flag, rng, 0);
			}
			else
			{
				compute_nucleo_pval (conf, chrom, ref_nucleo, 0, obs_min, prb_min, permut, no_probe_effect_flag, rng, 0);	
			}
		}
		else if (align->qry[i] && align->qry[i]->seen != 'y' && align->qry[i]->ignore != 'y')
		{
			qry_nucleo = align->qry[i];
			qry_nucleo->seen = 'y';
			if (qry_nucleo->align)
			{
				ref_nucleo = qry_nucleo->align;
				ref_nucleo->seen = 'y';
				compute_nucleo_pval (conf, chrom, ref_nucleo, qry_nucleo, obs_min, prb_min, permut, no_probe_effect_flag, rng, 0);
			}
			else
			{
				compute_nucleo_pval (conf, chrom, 0, qry_nucleo, obs_min, prb_min, permut, no_probe_effect_flag, rng, 0);	
			}
		}
	}
	for(i = 0; i < align->size; i++)
	{
		if (align->ref[i]) align->ref[i]->seen = 'n';
		if (align->qry[i]) align->qry[i]->seen = 'n';
	}	
} 

void
align_nucleo_alignment_anova_extract (const align_nucleo_config    * conf, 
		                             const align_nucleo_chromosome * chrom,
		                             align_nucleo_alignment * align,
		                             const double obs_min,
		                             const size_t prb_min,
		                             const gdl_boolean no_probe_effect_flag,
		                             const gdl_rng * rng,
		                             const long start,
		                             const long end,
		                             FILE * stream)
{
	size_t i;
	align_nucleo_nucleosome * ref_nucleo, * qry_nucleo;
	
	for(i = 0; i < align->size; i++)
	{
		if (!align->ref[i])
			continue;
		if ((align->ref[i]->ref_start_pos >= start && align->ref[i]->ref_start_pos <= end)
		    || (align->ref[i]->ref_end_pos >= start && align->ref[i]->ref_end_pos <= end)
		    && (align->ref[i]->ignore != 'y')
		    && (align->ref[i]->align != 0))
		{
			ref_nucleo = align->ref[i];
			qry_nucleo = ref_nucleo->align;
			compute_nucleo_pval (conf, chrom, ref_nucleo, qry_nucleo, obs_min, prb_min, 0, no_probe_effect_flag, rng, stream);
			fflush (stream);
		}
	}
} 


void
align_nucleo_linker_alignment_pval (const align_nucleo_config    * conf, 
                                     const align_nucleo_chromosome * chrom,
                                     align_nucleo_alignment * align,
                                     const long size_min,
                                     const double obs_min)
{
	size_t i;
	align_nucleo_linker * ref_linker, * qry_linker;
	align_nucleo_nucleosome * ref_nucleo, * qry_nucleo;
	
	for(i = 0; i < align->size; i++)
	{
		if (align->ref[i] && align->ref[i]->seen != 'y' && align->ref[i]->ignore != 'y')
		{
			ref_nucleo = align->ref[i];
			ref_nucleo->seen = 'y';
			if (ref_nucleo->align)
			{
				qry_nucleo = ref_nucleo->align;
				qry_nucleo->seen = 'y';
				// is the next one well aligned ?
				if (i < align->size - 1)
				{
					if (((align->ref[i+1] && align->ref[i+1]->seen != 'y') && align->ref[i+1]->align)
					   || ((align->qry[i+1] && align->qry[i+1]->seen != 'y') && align->qry[i+1]->align))
					{
						ref_linker = ref_nucleo->next_linker;
						qry_linker = qry_nucleo->next_linker;
						if (ref_linker->ref_end_pos-ref_linker->ref_start_pos+1 >= size_min)
						{
							compute_linker_pval (conf, chrom, ref_linker, qry_linker, obs_min);
						}
						else
						{
							ref_linker->pval = qry_linker->pval = -1;	
						}
					}
				}				
			}
		}
		else if (align->qry[i] && align->qry[i]->seen != 'y' && align->qry[i]->ignore != 'y')
		{
			qry_nucleo = align->qry[i];
			qry_nucleo->seen = 'y';
			if (qry_nucleo->align)
			{
				ref_nucleo = qry_nucleo->align;
				ref_nucleo->seen = 'y';
				if (i < align->size - 1)
				{
					if (((align->ref[i+1] && align->ref[i+1]->seen != 'y') && align->ref[i+1]->align)
					   || ((align->qry[i+1] && align->qry[i+1]->seen != 'y') && align->qry[i+1]->align))
					{
						ref_linker = ref_nucleo->next_linker;
						qry_linker = qry_nucleo->next_linker;
						if (ref_linker->ref_end_pos-ref_linker->ref_start_pos+1 >= size_min)
						{
							compute_linker_pval (conf, chrom, ref_linker, qry_linker, obs_min);
						}
						else
						{
							ref_linker->pval = qry_linker->pval = -1;	
						}
					}
				}
			}
		}
	}
	for(i = 0; i < align->size; i++)
	{
		if (align->ref[i]) align->ref[i]->seen = 'n';
		if (align->qry[i]) align->qry[i]->seen = 'n';
	}	
}                                           

int
align_nucleo_linker_alignment_fprintf (FILE * stream, const align_nucleo_alignment * align)
{
	size_t i,j;
	double * pr = GDL_MALLOC (double, 4);
	align_nucleo_linker * ref_linker, * qry_linker;
	align_nucleo_nucleosome * ref_nucleo, * qry_nucleo;
	
	for(i = 0; i < align->size; i++)
	{
		if (align->ref[i] && align->ref[i]->seen != 'y' && align->ref[i]->ignore != 'y')
		{
			ref_nucleo = align->ref[i];
			ref_nucleo->seen = 'y';
			if (ref_nucleo->align)
			{
				qry_nucleo = ref_nucleo->align;
				qry_nucleo->seen = 'y';
				// is the next one well aligned ?
				if (i < align->size - 1)
				{
					if (((align->ref[i+1] && align->ref[i+1]->seen != 'y') && align->ref[i+1]->align)
					   || ((align->qry[i+1] && align->qry[i+1]->seen != 'y') && align->qry[i+1]->align))
					{
						ref_linker = ref_nucleo->next_linker;
						qry_linker = qry_nucleo->next_linker;
						if (ref_linker == 0 || qry_linker == 0)
							continue;
						align_nucleo_linker_state_prob (ref_linker, qry_linker, pr);
						fprintf (stream, "%s %d %d %d %g", ref_nucleo->chrom->data->name, ref_linker->ref_start_pos, ref_linker->ref_end_pos, ref_linker->outlier, ref_linker->prob[0]);
						fprintf (stream, " %d %d %d %g", qry_linker->ref_start_pos, qry_linker->ref_end_pos, qry_linker->outlier, qry_linker->prob[0]);
						fprintf (stream, " %g %g", ref_linker->pval, align_nucleo_linker_genotype_prob (ref_linker, qry_linker));
					}
					for(j = 0; j < 4; j++)
					{
						fprintf (stream, " %1.5f", pr[j]);
					}
					fprintf (stream, "\n");
				}
			}
		}
		else if (align->qry[i] && align->qry[i]->seen != 'y' && align->qry[i]->ignore != 'y')
		{
			qry_nucleo = align->qry[i];
			qry_nucleo->seen = 'y';
			if (qry_nucleo->align)
			{
				ref_nucleo = qry_nucleo->align;
				ref_nucleo->seen = 'y';
				if (i < align->size - 1)
				{
					if (((align->ref[i+1] && align->ref[i+1]->seen != 'y') && align->ref[i+1]->align)
					   || ((align->qry[i+1] && align->qry[i+1]->seen != 'y') && align->qry[i+1]->align))
					{
						ref_linker = ref_nucleo->next_linker;
						qry_linker = qry_nucleo->next_linker;
						if (ref_linker == 0 || qry_linker == 0)
							continue;
						align_nucleo_linker_state_prob (ref_linker, qry_linker, pr);
						fprintf (stream, "%s %d %d %d %g", ref_nucleo->chrom->data->name, ref_linker->ref_start_pos, ref_linker->ref_end_pos, ref_linker->outlier, ref_linker->prob[0]);
						fprintf (stream, " %d %d %d %g", qry_linker->ref_start_pos, qry_linker->ref_end_pos, qry_linker->outlier, qry_linker->prob[0]);
						fprintf (stream, " %g %g", qry_linker->pval, align_nucleo_linker_genotype_prob (ref_linker, qry_linker));
						for(j = 0; j < 4; j++)
						{
							fprintf (stream, " %1.5f", pr[j]);
						}
						fprintf (stream, "\n");
					}
				}
			}
		}
	}
	for(i = 0; i < align->size; i++)
	{
		if (align->ref[i]) align->ref[i]->seen = 'n';
		if (align->qry[i]) align->qry[i]->seen = 'n';
	}
	GDL_FREE (pr);
}

align_nucleo_linker *
align_nucleo_linker_fread (FILE * stream,
                            const align_nucleo_config * config,
                            const align_nucleo_alignment * align,
                            const unsigned char kind)
{
	if (stream)
	{
		int status;
		int chrom_idx;
		unsigned char has;
		align_nucleo_linker * nuc;
		
		nuc = GDL_CALLOC (align_nucleo_linker, 1);	
		
		status = fread (&(nuc->idx), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->chunk_idx), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->start_idx), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->end_idx), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->ref_start_pos), sizeof(long), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->ref_end_pos), sizeof(long), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->pval), sizeof(double), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->map), sizeof(double), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->obs), sizeof(double), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->outlier), sizeof(gdl_boolean), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		nuc->prob = GDL_CALLOC (double, (config->hmm_nclass-1)/2+1);
		status = fread (nuc->prob, sizeof(double), (config->hmm_nclass-1)/2+1, stream);
		GDL_FREAD_STATUS (status, (config->hmm_nclass-1)/2+1);
		if (kind == 'q')
		{
			status = fread (&chrom_idx, sizeof(int), 1, stream);
			GDL_FREAD_STATUS (status, 1);
			nuc->chrom = align->qry_chroms[chrom_idx];			
		}
		else
		{
			nuc->chrom = align->ref_chrom;
		}
		
		return nuc;
	}
	return 0;
}

align_nucleo_nucleosome *
align_nucleo_nucleosome_fread (FILE * stream,
                               const align_nucleo_config * config,
                               const align_nucleo_alignment * align,
                               const unsigned char kind)
{
	if (stream && align)
	{
		int status;
		int chrom_idx;
		unsigned char has;
		align_nucleo_nucleosome * nuc;
		
		nuc = GDL_CALLOC (align_nucleo_nucleosome, 1);
		
		status = fread (&(nuc->idx), sizeof(int), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->ignore), sizeof(unsigned char), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->chunk_idx), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->start_idx), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->end_idx), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->status), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->ref_start_pos), sizeof(long), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->ref_end_pos), sizeof(long), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->pval), sizeof(double), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->map), sizeof(double), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->obs), sizeof(double), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->delta), sizeof(double), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->outlier), sizeof(gdl_boolean), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->indel), sizeof(double), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->gamma), sizeof(double), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		nuc->prob = GDL_CALLOC (double, (config->hmm_nclass-1)/2+1);
		status = fread (nuc->prob, sizeof(double), (config->hmm_nclass-1)/2+1, stream);
		GDL_FREAD_STATUS (status, (config->hmm_nclass-1)/2+1);
		status = fread (&(nuc->align_idx), sizeof(int), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		if (kind == 'q')
		{
			status = fread (&chrom_idx, sizeof(int), 1, stream);
			GDL_FREAD_STATUS (status, 1);
			nuc->chrom = align->qry_chroms[chrom_idx];
		}
		else
		{
			nuc->chrom = align->ref_chrom;
		}
		status = fread (&has, sizeof(unsigned char), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		if (has == 'y')
		{
			nuc->next_linker = align_nucleo_linker_fread (stream, config, align, kind);
			GDL_FREAD_STATUS (nuc->next_linker!=0, 1);
		}
		
		return nuc;
	}
	return 0;
}

align_nucleo_alignment *
align_nucleo_alignment_fread (FILE * stream,
                              const align_nucleo_config * config)
{
	if (stream && config)
	{
		size_t i, j, qx, qcx, qpx, rcx, rpx, nmapped;
		int status, cx;
		unsigned char has;
		gdl_string * name;
		align_nucleo_alignment  * align;
		
		// read the ref chromosome
		name = gdl_string_fread (stream);
		GDL_FREAD_STATUS (name!=0, 1);
		cx = gdl_tiling_data_chrom_lookup (config->ref_data, name);
		gdl_string_free (name);
		if (cx < 0)
		{
			return 0;
		}
		align            = GDL_CALLOC (align_nucleo_alignment, 1);
		align->ref_chrom = GDL_CALLOC (align_nucleo_chromosome, 1);
		align->ref_chrom->data = config->ref_data->chroms[cx];
		align->ref_chrom->hmm  = load_chrom_hmm (config->ref_hmm_dir, align->ref_chrom->data->name, (config->hmm_nclass) ? 1 : 2, config->hmm_nclass);
		align_nucleo_chromosome_anchor_probe (align->ref_chrom);
		// read the qry chromosomes
		status = fread (&(align->nqry), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		align->qry_chroms = GDL_CALLOC (align_nucleo_chromosome *, align->nqry);
		for(i = 0; i < align->nqry; i++)
		{
			// read the i^th qry chromosome
			name = gdl_string_fread (stream);
			GDL_FREAD_STATUS (name!=0, 1);
			cx = gdl_tiling_data_chrom_lookup (config->qry_data, name);
			if (cx < 0)
			{
				return 0;
			}
			align->qry_chroms[i] = GDL_CALLOC (align_nucleo_chromosome, 1);
			align->qry_chroms[i]->data = config->qry_data->chroms[cx];
			align->qry_chroms[i]->hmm  = load_chrom_hmm (config->qry_hmm_dir, align->qry_chroms[i]->data->name, (config->hmm_nclass) ? 1 : 2, config->hmm_nclass);
			align_nucleo_chromosome_anchor_probe (align->qry_chroms[i]);
			gdl_string_free (name);
		}
		status = fread (&nmapped, sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		for(i = 0; i < nmapped; i++)
		{
			align_nucleo_chromosome * qry_chrom;
			gdl_tiling_nucleo_chunk * ref_chunk, * qry_chunk;
			gdl_tiling_nucleo_probe * ref_probe, * qry_probe; 
			
			status = fread (&rcx, sizeof(size_t), 1, stream);
			GDL_FREAD_STATUS (status, 1);
			
			ref_chunk = align->ref_chrom->hmm->chunks[rcx];
			
			status = fread (&rpx, sizeof(size_t), 1, stream);
			GDL_FREAD_STATUS (status, 1);
			
			ref_probe = ref_chunk->probes[rpx];
			
			status = fread (&qx, sizeof(size_t), 1, stream);
			GDL_FREAD_STATUS (status, 1);	
			
			qry_chrom = align->qry_chroms[qx];
			
			status = fread (&qcx, sizeof(size_t), 1, stream);
			GDL_FREAD_STATUS (status, 1);
			
			qry_chunk = qry_chrom->hmm->chunks[qcx];
			
			status = fread (&qpx, sizeof(size_t), 1, stream);
			GDL_FREAD_STATUS (status, 1);
			
			qry_probe = qry_chunk->probes[qpx];
			
			ref_probe->align = qry_probe;
			qry_probe->align = ref_probe;
			qry_probe->chrom = qry_chrom;
		}
		// read the nucleosomes
		status = fread (&(align->size), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		align->pos = GDL_MALLOC (long, align->size);
		align->ref = GDL_CALLOC (align_nucleo_nucleosome *, align->size);
		align->qry = GDL_CALLOC (align_nucleo_nucleosome *, align->size);
		
		status = fread (align->pos, sizeof(long), align->size, stream);
		GDL_FREAD_STATUS (status, align->size);
		for(i = 0; i < align->size; i++)
		{
			status = fread (&has, sizeof(unsigned char), 1, stream);
			GDL_FREAD_STATUS (status, 1);
			if (has == 'y')
			{
				align->ref[i] = align_nucleo_nucleosome_fread (stream, config, align, 'r');
				GDL_FREAD_STATUS (align->ref[i]!=0, 1);
			}
			status = fread (&has, sizeof(unsigned char), 1, stream);
			GDL_FREAD_STATUS (status, 1);
			if (has == 'y')
			{
				align->qry[i] = align_nucleo_nucleosome_fread (stream, config, align, 'q');
				GDL_FREAD_STATUS (align->qry[i]!=0, 1);
			}
		}
		for(i = 0; i < align->size; i++)
		{
			if (align->ref[i] && align->ref[i]->align_idx>=0)
			{
				align->ref[i]->align = align->qry[align->ref[i]->align_idx];
			}
			if (align->qry[i] && align->qry[i]->align_idx>=0)
			{
				align->qry[i]->align = align->ref[align->qry[i]->align_idx];
			}
		}
		
		return align;
	}
	return 0;
}

int
align_nucleo_linker_fwrite (FILE * stream,
                            const align_nucleo_config * config,
                            const align_nucleo_alignment * align,
                            const int chrom_idx,
                            const align_nucleo_linker * nuc)
{
	if (stream)
	{
		int status;
		unsigned char has;
		
		status = fwrite (&(nuc->idx), sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->chunk_idx), sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->start_idx), sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->end_idx), sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->ref_start_pos), sizeof(long), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->ref_end_pos), sizeof(long), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->pval), sizeof(double), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->map), sizeof(double), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->obs), sizeof(double), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->outlier), sizeof(gdl_boolean), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (nuc->prob, sizeof(double), (config->hmm_nclass-1)/2+1, stream);
		GDL_FWRITE_STATUS (status, (config->hmm_nclass-1)/2+1);
		if (chrom_idx >= 0)
		{
			status = fwrite (&chrom_idx, sizeof(int), 1, stream);
			GDL_FWRITE_STATUS (status, 1);
		}
		
		return GDL_SUCCESS;
	}
	return GDL_EINVAL;
}

int
align_nucleo_nucleosome_fwrite (FILE * stream,
                                const align_nucleo_config * config,
                                const align_nucleo_alignment * align,
                                const int chrom_idx,
                                const align_nucleo_nucleosome * nuc)
{
	if (stream)
	{
		int status, nox=-1;
		unsigned char has;
		
		status = fwrite (&(nuc->idx), sizeof(int), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->ignore), sizeof(unsigned char), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->chunk_idx), sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->start_idx), sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->end_idx), sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->status), sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->ref_start_pos), sizeof(long), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->ref_end_pos), sizeof(long), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->pval), sizeof(double), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->map), sizeof(double), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->obs), sizeof(double), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->delta), sizeof(double), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->outlier), sizeof(gdl_boolean), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->indel), sizeof(double), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->gamma), sizeof(double), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (nuc->prob, sizeof(double), (config->hmm_nclass-1)/2+1, stream);
		GDL_FWRITE_STATUS (status, (config->hmm_nclass-1)/2+1);
		if (nuc->align)
		{
			status = fwrite (&(nuc->align->idx), sizeof(int), 1, stream);
			GDL_FWRITE_STATUS (status, 1);
		}
		else
		{
			status = fwrite (&nox, sizeof(int), 1, stream);
			GDL_FWRITE_STATUS (status, 1);
		}
		if (chrom_idx >= 0)
		{
			status = fwrite (&chrom_idx, sizeof(int), 1, stream);
			GDL_FWRITE_STATUS (status, 1);
		}
		has = (nuc->next_linker) ? 'y' : 'n';
		status = fwrite (&has, sizeof(unsigned char), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		if (has == 'y')
		{
			status = align_nucleo_linker_fwrite (stream, config, align, chrom_idx, nuc->next_linker);
			GDL_FWRITE_STATUS (status, GDL_SUCCESS);
		}		
		return GDL_SUCCESS;
	}
	return GDL_EINVAL;
}

int
align_nucleo_alignment_fwrite (FILE * stream,
                               const align_nucleo_config * config,
                               const align_nucleo_alignment * align)
{
	if ((stream && align) && config)
	{
		size_t i,j,k,nmapped;
		int status, chrom_idx;
		unsigned char has;
		gdl_string * name;
		
		status = gdl_string_fwrite (stream, align->ref_chrom->data->name);
		GDL_FWRITE_STATUS (status, GDL_SUCCESS);
		status = fwrite (&(align->nqry), sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		for(i = 0; i < align->nqry; i++)
		{
			status = gdl_string_fwrite (stream, align->qry_chroms[i]->data->name);
			GDL_FWRITE_STATUS (status, GDL_SUCCESS);
		}
		// save the probe ref<->qry mapping
		align_nucleo_chromosome * qry_chrom;
		gdl_tiling_nucleo_chunk * ref_chunk, * qry_chunk;
		gdl_tiling_nucleo_probe * ref_probe, * qry_probe; 
		
		for(nmapped = i = 0; i < align->ref_chrom->hmm->size; i++)
		{
			ref_chunk = align->ref_chrom->hmm->chunks[i];
			for(j = 0; j < ref_chunk->size; j++)
			{
				ref_probe = ref_chunk->probes[j];
				if (ref_probe->idx >= 0 && ref_probe->align)
				{
					qry_probe = ref_probe->align;
					qry_chrom = qry_probe->chrom;
					for(k = 0; k < align->nqry; k++)
					{
						if (!strcmp(qry_chrom->data->name,align->qry_chroms[k]->data->name))
							break;
					}
					if (k < align->nqry) nmapped++;
				}
			}
		}
		status = fwrite (&nmapped, sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		for(i = 0; i < align->ref_chrom->hmm->size; i++)
		{
			ref_chunk = align->ref_chrom->hmm->chunks[i];
			for(j = 0; j < ref_chunk->size; j++)
			{
				ref_probe = ref_chunk->probes[j];
				if (ref_probe->idx >= 0 && ref_probe->align)
				{
					qry_probe = ref_probe->align;
					qry_chrom = qry_probe->chrom;
					for(k = 0; k < align->nqry; k++)
					{
						if (!strcmp(qry_chrom->data->name,align->qry_chroms[k]->data->name))
							break;
					}
					if (k < align->nqry)
					{
						status = fwrite (&i, sizeof(size_t), 1, stream);
						GDL_FWRITE_STATUS (status, 1);
						status = fwrite (&j, sizeof(size_t), 1, stream);
						GDL_FWRITE_STATUS (status, 1);
						status = fwrite (&k, sizeof(size_t), 1, stream);
						GDL_FWRITE_STATUS (status, 1);
						status = fwrite (&(qry_probe->chunk_idx), sizeof(size_t), 1, stream);
						GDL_FWRITE_STATUS (status, 1);
						status = fwrite (&(qry_probe->idx), sizeof(size_t), 1, stream);
						GDL_FWRITE_STATUS (status, 1);
					}
				}
			}
		}	
		// read the nucleosomes
		status = fwrite (&(align->size), sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (align->pos, sizeof(long), align->size, stream);
		GDL_FWRITE_STATUS (status, align->size);
		for(i = 0; i < align->size; i++)
		{
			if (align->ref[i])
			{
				align->ref[i]->idx = i;
			}
			if (align->qry[i])
			{
				align->qry[i]->idx = i;
			}	
		}
		for(i = 0; i < align->size; i++)
		{
			has = (align->ref[i]) ? 'y' : 'n';
			status = fwrite (&has, sizeof(unsigned char), 1, stream);
			GDL_FWRITE_STATUS (status, 1);
			if (has == 'y')
			{
				status = align_nucleo_nucleosome_fwrite (stream, config, align, -1, align->ref[i]);
				GDL_FWRITE_STATUS (status, GDL_SUCCESS);
			}
			has = (align->qry[i]) ? 'y' : 'n';
			status = fwrite (&has, sizeof(unsigned char), 1, stream);
			GDL_FWRITE_STATUS (status, 1);
			if (has == 'y')
			{
				chrom_idx=-1;
				for(j = 0; j < align->nqry; j++)
				{
					if (align->qry[i]->chrom == align->qry_chroms[j])
					{
						chrom_idx=j;
						break;
					} 
				}
				status = align_nucleo_nucleosome_fwrite (stream, config, align, chrom_idx, align->qry[i]);
				GDL_FWRITE_STATUS (status, GDL_SUCCESS);
			}
		}
		
		return GDL_SUCCESS;
	}
 	return GDL_EINVAL;
}

void
align_nucleo_linker_free (align_nucleo_linker * n)
{
	if (n)
	{
		GDL_FREE (n->prob);
	   GDL_FREE (n);
	}	
}

void
align_nucleo_nucleosome_free (align_nucleo_nucleosome * n)
{
	if (n)
	{
		GDL_FREE (n->prob);
	   align_nucleo_linker_free (n->next_linker);
		GDL_FREE (n);
	}	
}

void
align_nucleo_alignment_free (align_nucleo_alignment * a)
{
	if (a)
	{
		size_t i;
		GDL_FREE (a->pos);
		for(i = 0; i < a->size; i++)
		{
			align_nucleo_nucleosome_free (a->ref[i]);
			align_nucleo_nucleosome_free (a->qry[i]);		
		}
		GDL_FREE (a->ref);
		GDL_FREE (a->qry);
		for(i = 0; i < a->nqry; i++)
		{
			gdl_tiling_nucleo_chromosome_free (a->qry_chroms[i]->hmm);			
		}
		GDL_FREE (a->qry_chroms);
		gdl_tiling_nucleo_chromosome_free (a->ref_chrom->hmm);
		//GDL_FREE (a->ref_chrom);
		GDL_FREE (a);
	}	
}

static void
_align_nucleo_linker_alignment_filter (align_nucleo_nucleosome * nucleos[], const size_t N)
{
	size_t i, nl=0;
	double iq, q1, q2, * nx, * lx;
	
	nx = GDL_MALLOC (double, N);
	lx = GDL_MALLOC (double, N);
	
	for(i = 0; i < N; i++)
	{
		nx[i] = nucleos[i]->ref_end_pos-nucleos[i]->ref_start_pos+1;
		if (nucleos[i]->next_linker)
		{
			align_nucleo_linker * linker = nucleos[i]->next_linker;
			lx[nl++] = linker->ref_end_pos-linker->ref_start_pos+1;
		}
	}
	
	gdl_sort (nx, 1, N);
	q1 = gdl_stats_quantile_from_sorted_data (nx, 1, N, 0.25);
	q2 = gdl_stats_quantile_from_sorted_data (nx, 1, N, 0.75);
	iq = q2-q1;
	
	for(i = 0; i < N; i++)
	{
		q1 = nucleos[i]->ref_end_pos-nucleos[i]->ref_start_pos+1;
		if (q1 > q2 + 3*iq)
		{
			nucleos[i]->outlier = gdl_true;
			nucleos[i]->ignore  = 'y';
		} 	
	}
	
	gdl_sort (lx, 1, nl);
	q1 = gdl_stats_quantile_from_sorted_data (lx, 1, nl, 0.25);
	q2 = gdl_stats_quantile_from_sorted_data (lx, 1, nl, 0.75);
	iq = q2-q1;
	
	for(i = 0; i < N; i++)
	{
		if (nucleos[i]->next_linker)
		{
			align_nucleo_linker * linker = nucleos[i]->next_linker;
			q1 = linker->ref_end_pos-linker->ref_start_pos+1;
			if (q1 > q2 + 3*iq)
			{
				linker->outlier = gdl_true;
			} 	
		} 	
	}
	
	GDL_FREE (nx);
	GDL_FREE (lx);
} 

int
align_nucleo_linker_alignment_filter (align_nucleo_alignment * align)
{
	size_t i,j,k,nref,nqry;
	int l;
	double score;
	align_nucleo_linker * ref_linker, * qry_linker;
	align_nucleo_nucleosome * ref_nucleo, * qry_nucleo;
	align_nucleo_nucleosome ** ref_nucleos, ** qry_nucleos;
	
	nref=nqry=0;
	for(i = 0; i < align->size; i++)
	{
		if (align->ref[i]) nref++;
		if (align->qry[i]) nqry++;
	}
	ref_nucleos = GDL_MALLOC (align_nucleo_nucleosome *, nref);
	qry_nucleos = GDL_MALLOC (align_nucleo_nucleosome *, nqry);
	for(i = j = k = 0; i < align->size; i++)
	{
		if (align->ref[i]) ref_nucleos[j++]=align->ref[i];
		if (align->qry[i]) qry_nucleos[k++]=align->qry[i];
	}
	
	_align_nucleo_linker_alignment_filter (ref_nucleos, nref);
	_align_nucleo_linker_alignment_filter (qry_nucleos, nqry);
	
	// now check if some qry nulceosomes overlap with outlier ref nucleosome
	for(i = 0; i < align->size; i++)
	{
		if (align->ref[i] && align->ref[i]->outlier)
		{
			if (align->ref[i]->align)
			{
				align->ref[i]->align->outlier = gdl_true;
				align->ref[i]->align->ignore  = 'y';
			}
			const long start_pos = align->ref[i]->ref_start_pos;
			const long end_pos   = align->ref[i]->ref_end_pos;
			for(l = i-1; l >= 0; l--)
			{
				if (!align->qry[l]) continue;
				if (align->qry[l]->ref_end_pos < start_pos) break;
				score = align->qry[l]->ref_end_pos - start_pos;
				score /= (align->qry[l]->ref_end_pos - align->qry[l]->ref_start_pos);
				if (score >= 0.5)
				{
					align->qry[l]->outlier = gdl_true;
					align->qry[l]->ignore  = 'y';
				}
			}
			for(l = i+1; l < align->size; l++)
			{
				if (!align->qry[l]) continue;
				if (align->qry[l]->ref_start_pos > end_pos) break;
				if (align->qry[l]->ref_end_pos < end_pos)
				{
					score = 1;
				}
				else
				{
					score = end_pos - align->qry[l]->ref_start_pos;
					score /= (align->qry[l]->ref_end_pos - align->qry[l]->ref_start_pos);
				}
				if (score >= 0.5)
				{
					align->qry[l]->outlier = gdl_true;
					align->qry[l]->ignore  = 'y';
				}
			}
		}
	}
	// now check if some ref nulceosomes overlap with outlier qry nucleosome
	for(i = 0; i < align->size; i++)
	{
		if (align->qry[i] && align->qry[i]->outlier)
		{
			if (align->qry[i]->align)
			{
				align->qry[i]->align->outlier = gdl_true;
				align->qry[i]->align->ignore  = 'y';
			}
			const long start_pos = align->qry[i]->ref_start_pos;
			const long end_pos   = align->qry[i]->ref_end_pos;
			for(l = i-1; l >= 0; l--)
			{
				if (!align->ref[l]) continue;
				if (align->ref[l]->ref_end_pos < start_pos) break;
				score = align->ref[l]->ref_end_pos - start_pos;
				score /= (align->ref[l]->ref_end_pos - align->ref[l]->ref_start_pos);
				if (score >= 0.5)
				{
					align->ref[l]->outlier = gdl_true;
					align->ref[l]->ignore  = 'y';
				}
			}
			for(l = i+1; l < align->size; l++)
			{
				if (!align->ref[l]) continue;
				if (align->ref[l]->ref_start_pos > end_pos) break;
				if (align->ref[l]->ref_end_pos < end_pos)
				{
					score = 1;
				}
				else
				{
					score = end_pos - align->ref[l]->ref_start_pos;
					score /= (align->ref[l]->ref_end_pos - align->ref[l]->ref_start_pos);
				}
				if (score >= 0.5)
				{
					align->ref[l]->outlier = gdl_true;
					align->ref[l]->ignore  = 'y';
				}
			}
		}
	}
	// now check if some qry nulceosomes overlap with outlier ref linker
	for(i = 0; i < align->size; i++)
	{
		if (align->ref[i] && align->ref[i]->next_linker && align->ref[i]->next_linker->outlier)
		{
			const long start_pos = align->ref[i]->next_linker->ref_start_pos;
			const long end_pos   = align->ref[i]->next_linker->ref_end_pos;
			for(l = i-1; l >= 0; l--)
			{
				if (!align->qry[l]) continue;
				if (align->qry[l]->ref_end_pos < start_pos) break;
				score = align->qry[l]->ref_end_pos - start_pos;
				score /= (align->qry[l]->ref_end_pos - align->qry[l]->ref_start_pos);
				if (score >= 0.5)
				{
					align->qry[l]->outlier = gdl_true;
					align->qry[l]->ignore  = 'y';
				}
			}
			for(l = i+1; l < align->size; l++)
			{
				if (!align->qry[l]) continue;
				if (align->qry[l]->ref_start_pos > end_pos) break;
				if (align->qry[l]->ref_end_pos < end_pos)
				{
					score = 1;
				}
				else
				{
					score = end_pos - align->qry[l]->ref_start_pos;
					score /= (align->qry[l]->ref_end_pos - align->qry[l]->ref_start_pos);
				}
				if (score >= 0.5)
				{
					align->qry[l]->outlier = gdl_true;
					align->qry[l]->ignore  = 'y';
				}
			}
		}
	}
	// now check if some ref nulceosomes overlap with outlier qry linkers
	for(i = 0; i < align->size; i++)
	{
		if (align->qry[i] && align->qry[i]->next_linker && align->qry[i]->next_linker->outlier)
		{
			const long start_pos = align->qry[i]->next_linker->ref_start_pos;
			const long end_pos   = align->qry[i]->next_linker->ref_end_pos;
			for(l = i-1; l >= 0; l--)
			{
				if (!align->ref[l]) continue;
				if (align->ref[l]->ref_end_pos < start_pos) break;
				score = align->ref[l]->ref_end_pos - start_pos;
				score /= (align->ref[l]->ref_end_pos - align->ref[l]->ref_start_pos);
				if (score >= 0.5)
				{
					align->ref[l]->outlier = gdl_true;
					align->ref[l]->ignore  = 'y';
				}
			}
			for(l = i+1; l < align->size; l++)
			{
				if (!align->ref[l]) continue;
				if (align->ref[l]->ref_start_pos > end_pos) break;
				if (align->ref[l]->ref_end_pos < end_pos)
				{
					score = 1;
				}
				else
				{
					score = end_pos - align->ref[l]->ref_start_pos;
					score /= (align->ref[l]->ref_end_pos - align->ref[l]->ref_start_pos);
				}
				if (score >= 0.5)
				{
					align->ref[l]->outlier = gdl_true;
					align->ref[l]->ignore  = 'y';
				}
			}
		}
	}
	
	GDL_FREE (ref_nucleos);
	GDL_FREE (qry_nucleos);
}

void
_distance_to_proximal_nucleosome_on_qry (align_nucleo_alignment * align, const size_t i, double * x_left, double * x_right)
{
	int j;
	double up,down;
	align_nucleo_nucleosome * qry_nucleo;
	
	for(j = i+1; j < align->size; j++)
	{
		if (align->ref[j] && align->ref[j]->align)
		{
			qry_nucleo = align->ref[j]->align;
			down = 0.5*(qry_nucleo->ref_start_pos+qry_nucleo->ref_end_pos);
			break;
		}
		else if (align->qry[j])
		{
			qry_nucleo = align->qry[j];
			down = 0.5*(qry_nucleo->ref_start_pos+qry_nucleo->ref_end_pos);
			break;
		}	
	}
	up = GDL_POSINF;
	for(j = i-1; j >= 0; j--)
	{
		if (align->ref[j] && align->ref[j]->align)
		{
			qry_nucleo = align->ref[j]->align;
			up = 0.5*(qry_nucleo->ref_start_pos+qry_nucleo->ref_end_pos);
			break;
		}
		else if (align->qry[j])
		{
			qry_nucleo = align->qry[j];
			up = 0.5*(qry_nucleo->ref_start_pos+qry_nucleo->ref_end_pos);
			break;
		}	
	}
	*x_left =up;
	*x_right=down;
}

void
_distance_to_proximal_nucleosome_on_ref (align_nucleo_alignment * align, const size_t i, double * x_left, double * x_right)
{
	int j;
	double up,down;
	align_nucleo_nucleosome * qry_nucleo;
	
	for(j = i+1; j < align->size; j++)
	{
		if (align->qry[j] && align->qry[j]->align)
		{
			qry_nucleo = align->qry[j]->align;
			down = 0.5*(qry_nucleo->ref_start_pos+qry_nucleo->ref_end_pos);
			break;
		}
		else if (align->ref[j])
		{
			qry_nucleo = align->ref[j];
			down = 0.5*(qry_nucleo->ref_start_pos+qry_nucleo->ref_end_pos);
			break;
		}	
	}
	up = GDL_POSINF;
	for(j = i-1; j >= 0; j--)
	{
		if (align->qry[j] && align->qry[j]->align)
		{
			qry_nucleo = align->qry[j]->align;
			up = 0.5*(qry_nucleo->ref_start_pos+qry_nucleo->ref_end_pos);
			break;
		}
		else if (align->ref[j])
		{
			qry_nucleo = align->ref[j];
			up = 0.5*(qry_nucleo->ref_start_pos+qry_nucleo->ref_end_pos);
			break;
		}	
	}
	*x_left =up;
	*x_right=down;
}

int
align_nucleo_linker_alignment_statistics (const align_nucleo_config * conf, align_nucleo_alignment * align)
{
	size_t i;
	int j, cstart=0, cend=0;
	long del_size, ins_size;
	double delta,up,down;
	align_nucleo_nucleosome * ref_nucleo, * qry_nucleo;
	const gdl_chunk_coord_cluster * chunk_cluster = gdl_genome_coord_lookup (conf->ref_qry_coord, "ref", align->ref_chrom->data->name);
	
	for(i = 0; i < align->size; i++)
	{
		if (align->ref[i] && align->ref[i]->seen != 'y' && align->ref[i]->ignore != 'y')
		{
			ref_nucleo = align->ref[i];
			ref_nucleo->seen = 'y';
			delta  = 0.5*(ref_nucleo->ref_start_pos+ref_nucleo->ref_end_pos);
			if (ref_nucleo->align)
			{
				qry_nucleo = ref_nucleo->align;
				qry_nucleo->seen = 'y';
				// compute delta
				delta -= 0.5*(qry_nucleo->ref_start_pos+qry_nucleo->ref_end_pos);
				qry_nucleo->delta = -delta;
				// compute the deletion score for that nucleosome
				cstart = gdl_chunk_coord_cluster_goto (chunk_cluster, qry_nucleo->ref_start_pos, 0);
				cend   = gdl_chunk_coord_cluster_goto (chunk_cluster, qry_nucleo->ref_end_pos, cstart);
				del_size=ins_size=0;
				gdl_chunk_coord_cluster_indel_size (chunk_cluster, cstart, cend, qry_nucleo->ref_start_pos, qry_nucleo->ref_end_pos, &del_size, &ins_size);
				//printf ("(QRY)INS_SIZE (%d,%d %d,%d) = %d\n", qry_nucleo->ref_start_pos, cstart, qry_nucleo->ref_end_pos, cend, del_size);
				qry_nucleo->indel = ((double)del_size)/(qry_nucleo->ref_end_pos-qry_nucleo->ref_start_pos+1);
			}
			else
			{
				// find the proximal nucleosome on query
				_distance_to_proximal_nucleosome_on_qry (align, i, &up, &down);
				//delta -= GDL_MIN (up, down);
				if (fabs(delta-up) < fabs(delta-down))
				{
					delta -= up;
				}
				else
				{
					delta -= down;
				}
				// compute the insertion score for that nucleosome
			}
			ref_nucleo->delta = delta;
			cstart = gdl_chunk_coord_cluster_goto (chunk_cluster, ref_nucleo->ref_start_pos, 0);
			cend   = gdl_chunk_coord_cluster_goto (chunk_cluster, ref_nucleo->ref_end_pos, cstart);
			del_size=ins_size=0;
			gdl_chunk_coord_cluster_indel_size (chunk_cluster, cstart, cend, ref_nucleo->ref_start_pos, ref_nucleo->ref_end_pos, &del_size, &ins_size);
			//printf ("(REF)INS_SIZE (%d,%d %d,%d) = %d\n", ref_nucleo->ref_start_pos, cstart, ref_nucleo->ref_end_pos, cend, ins_size);
			ref_nucleo->indel = ((double)ins_size)/(ref_nucleo->ref_end_pos-ref_nucleo->ref_start_pos+1);
		}
		else if (align->qry[i] && align->qry[i]->seen != 'y' && align->qry[i]->ignore != 'y')
		{
			qry_nucleo = align->qry[i];
			qry_nucleo->seen = 'y';
			delta = -0.5*(qry_nucleo->ref_start_pos+qry_nucleo->ref_end_pos);
			if (qry_nucleo->align && qry_nucleo->align->ignore != 'y')
			{
				ref_nucleo = qry_nucleo->align;
				ref_nucleo->seen = 'y';
				delta += 0.5*(ref_nucleo->ref_start_pos+ref_nucleo->ref_end_pos);
				ref_nucleo->delta = -delta;
				cstart = gdl_chunk_coord_cluster_goto (chunk_cluster, ref_nucleo->ref_start_pos, 0);
				cend   = gdl_chunk_coord_cluster_goto (chunk_cluster, ref_nucleo->ref_end_pos, cstart);
				del_size=ins_size=0;
				gdl_chunk_coord_cluster_indel_size (chunk_cluster, cstart, cend, ref_nucleo->ref_start_pos, ref_nucleo->ref_end_pos, &del_size, &ins_size);
				//printf ("(REF)INS_SIZE (%d,%d %d,%d) = %d\n", ref_nucleo->ref_start_pos, cstart, ref_nucleo->ref_end_pos, cend, ins_size);
				ref_nucleo->indel = ((double)ins_size)/(ref_nucleo->ref_end_pos-ref_nucleo->ref_start_pos+1);
			}
			else
			{
				// find the proximal nucleosome on reference
				_distance_to_proximal_nucleosome_on_ref (align, i, &up, &down);
				if (fabs(delta+up) < fabs(delta+down))
				{
					delta += up;
				}
				else
				{
					delta += down;
				}
			}
			qry_nucleo->delta = delta;
			// compute the deletion score for that nucleosome
			cstart = gdl_chunk_coord_cluster_goto (chunk_cluster, qry_nucleo->ref_start_pos, 0);
			cend   = gdl_chunk_coord_cluster_goto (chunk_cluster, qry_nucleo->ref_end_pos, cstart);
			del_size=ins_size=0;
			gdl_chunk_coord_cluster_indel_size (chunk_cluster, cstart, cend, qry_nucleo->ref_start_pos, qry_nucleo->ref_end_pos, &del_size, &ins_size);
			//printf ("(QRY)INS_SIZE (%d,%d %d,%d) = %d\n", qry_nucleo->ref_start_pos, cstart, qry_nucleo->ref_end_pos, cend, del_size);
			qry_nucleo->indel = ((double)del_size)/(qry_nucleo->ref_end_pos-qry_nucleo->ref_start_pos+1);
		}
	}
	for(i = 0; i < align->size; i++)
	{
		if (align->ref[i]) align->ref[i]->seen = 'n';
		if (align->qry[i]) align->qry[i]->seen = 'n';
	}	
}

void
align_nucleo_alignment_ref_flanking_qry (align_nucleo_alignment * align, const size_t i, align_nucleo_nucleosome ** ref_left, align_nucleo_nucleosome ** ref_right)
{
	int j;
	
	for(j = i+1; j < align->size; j++)
	{
		if (align->ref[j] && align->ref[j]->align)
		{
			*ref_right = align->ref[j]->align;
			break;
		}
		else if (align->qry[j])
		{
			*ref_right = align->qry[j];
			break;
		}	
	}
	for(j = i-1; j >= 0; j--)
	{
		if (align->ref[j] && align->ref[j]->align)
		{
			*ref_left = align->ref[j]->align;
			break;
		}
		else if (align->qry[j])
		{
			*ref_left = align->qry[j];
			break;
		}	
	}
}

void
align_nucleo_alignment_qry_flanking_ref (align_nucleo_alignment * align, const size_t i, align_nucleo_nucleosome ** qry_left, align_nucleo_nucleosome ** qry_right)
{
	int j;
	
	for(j = i+1; j < align->size; j++)
	{
		if (align->qry[j] && align->qry[j]->align)
		{
			*qry_right = align->qry[j]->align;
			break;
		}
		else if (align->ref[j])
		{
			*qry_right = align->ref[j];
			break;
		}	
	}
	for(j = i-1; j >= 0; j--)
	{
		if (align->qry[j] && align->qry[j]->align)
		{
			*qry_left = align->qry[j]->align;
			break;
		}
		else if (align->ref[j])
		{
			*qry_left = align->ref[j];
			break;
		}	
	}
}
