/*  
 *  align_nucleo.h
 * 
 *  $Author: gael $, $Date: 2008-10-17 10:34:08 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
 
#ifndef __GDL_ALIGN_NUCLEO_H__
#define __GDL_ALIGN_NUCLEO_H__

#include <math.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_randist.h>
#include <gdl/gdl_string.h>
#include <gdl/gdl_vector.h>
#include <gdl/gdl_matrix.h>
#include <gdl/gdl_multireg.h>
#include <gdl/gdl_genome_coord.h>
#include <gdl/gdl_sequences.h>

#include "tiling_hmm.h"
#include "tiling_nucleo.h"
#include "tiling_array.h"

__BEGIN_DECLS

typedef struct
{
	// ref files
	gdl_string * ref_data_file;
	gdl_string * ref_hmm_dir;
	gdl_string * ref_stat_file;
	gdl_string * ref_qry_coord_file;
	gdl_string * ref_seq_file;
	// qry files
	gdl_string * qry_data_file;
	gdl_string * qry_stat_file;
	gdl_string * qry_hmm_dir;
	gdl_string * qry_seq_file;
	// other
	size_t hmm_nclass;
	// objects
	gdl_tiling_data * ref_data;
	gdl_tiling_data * qry_data;
	gdl_tiling_data * ref_stat;
	gdl_tiling_data * qry_stat;
	gdl_seqdb * ref_seq;
	gdl_seqdb * qry_seq;
	gdl_genome_coord * ref_qry_coord;
} align_nucleo_config;

typedef struct _align_nucleo_chromosome align_nucleo_chromosome;

typedef struct
{
	size_t idx;
	unsigned char seen;
	unsigned char _seen;
	size_t chunk_idx;
	size_t start_idx;
	size_t end_idx;
	long ref_start_pos;
	long ref_end_pos;
	const align_nucleo_chromosome * chrom;
	double pval;
	double * prob;
	double map;
	double obs;
	gdl_boolean outlier;
} align_nucleo_linker;

typedef struct _align_nucleo_nucleosome align_nucleo_nucleosome;

struct _align_nucleo_nucleosome
{
	int idx;
	unsigned char ignore;
	unsigned char seen;
	unsigned char _seen;
	size_t chunk_idx;
	size_t start_idx;
	size_t end_idx;
	size_t status;
	int align_idx;
	long ref_start_pos;
	long ref_end_pos;
	align_nucleo_chromosome * chrom;
	align_nucleo_nucleosome * align;
	double * prob;
	double delta;
	double pval;
	double map;
	double obs;
	double indel;
	double gamma;
	double mrk_pval;
	double mrk_gamma;
	gdl_boolean outlier;
	align_nucleo_linker * next_linker;
};

typedef struct
{
	size_t size;
	size_t nqry;
	long * pos;
	align_nucleo_chromosome  * ref_chrom;
	align_nucleo_chromosome ** qry_chroms;
	align_nucleo_nucleosome ** ref;
	align_nucleo_nucleosome ** qry;
} align_nucleo_alignment;

struct _align_nucleo_chromosome
{
	gdl_tiling_chromosome        * data;
	gdl_tiling_nucleo_chromosome * hmm;
	gdl_hashtable * qry_table;
	size_t nref_nucleosome;
	size_t nqry_nucleosome;
	align_nucleo_nucleosome ** ref_nucleosomes;
	align_nucleo_nucleosome ** qry_nucleosomes;
	size_t nref_linker;
	size_t nqry_linker;
	align_nucleo_linker ** ref_linkers;
	align_nucleo_linker ** qry_linkers;
};

align_nucleo_config * align_nucleo_config_alloc (const gdl_string * filename);
align_nucleo_nucleosome * align_nucleo_nucleosome_alloc (align_nucleo_chromosome * chrom, const size_t chunk, const size_t start, const size_t end, const size_t status, const size_t nclass);
align_nucleo_linker * align_nucleo_linker_alloc (const align_nucleo_chromosome * chrom, const size_t chunk, const size_t start, const size_t end, const size_t nclass);
gdl_tiling_nucleo_chromosome * load_chrom_hmm (const gdl_string * hmm_dir, const gdl_string * chrom, const size_t D, const size_t C);
int align_nucleo_chromosome_anchor_probe (align_nucleo_chromosome * chrom);
int align_nucleo_chromosome_fprintf_probe (FILE * stream, const align_nucleo_chromosome * chrom, const gdl_tiling_nucleo_probe * probe);
int align_nucleo_chromosome_fprintf (FILE * stream, const align_nucleo_chromosome * chrom, const size_t mode);
gdl_tiling_nucleo_probe * nucleo_hmm_cmp_anchor_qry_probe_at_position (align_nucleo_chromosome * qry_chrom, const long qry_pos, const long tiling_size);
int align_nucleo_chromosome_anchor_qry (const align_nucleo_config * conf, align_nucleo_chromosome * ref_chrom, const size_t hmm_dim, const size_t tiling_size); 
align_nucleo_nucleosome * align_nucleo_chromosome_call_nucleo (align_nucleo_chromosome * chrom, const gdl_tiling_nucleo_probe * probe, const size_t chunk_idx, const size_t probe_idx, const size_t nucleo_min, const size_t nclass, size_t * score, size_t * status, size_t * start, size_t * cidx, align_nucleo_chromosome ** chrom_ptr);
align_nucleo_linker * align_nucleo_chromosome_call_linker (align_nucleo_chromosome * chrom, const gdl_tiling_nucleo_probe * probe, const size_t chunk_idx, const size_t probe_idx, const size_t nclass, size_t * score, size_t * chunk, size_t * start, align_nucleo_chromosome ** chrom_ptr);
void align_nucleo_chromosome_nucleo_properties (const gdl_tiling_nucleo_chunk * ref_chunk, align_nucleo_nucleosome * ref_nucleo);
void align_nucleo_chromosome_linker_properties (const gdl_tiling_nucleo_chunk * ref_chunk, align_nucleo_linker * ref_linker);
int align_nucleo_chromosome_anchor_nucleo (align_nucleo_chromosome * chrom, const size_t tiling_size, const size_t nucleo_min, const double prb_mapped_min, const size_t nclass);
void align_nucleo_chromosome_align_neighbour (align_nucleo_chromosome * chrom, align_nucleo_nucleosome * ref_nucleo_start, align_nucleo_nucleosome * ref_nucleo_end, align_nucleo_nucleosome * qry_nucleo_start, align_nucleo_nucleosome * qry_nucleo_end, const gdl_boolean include_start, const gdl_boolean include_end);
align_nucleo_alignment * align_nucleo_alignment_alloc (align_nucleo_chromosome * chrom);
align_nucleo_alignment * align_nucleo_chromosome_align_nucleo (align_nucleo_chromosome * chrom, const double prb_overlap_min);
double align_nucleo_nucleosome_genotype_prob (const align_nucleo_nucleosome * ref_nucleo, const align_nucleo_nucleosome * qry_nucleo);
double align_nucleo_nucleosome_indel_prob (const align_nucleo_nucleosome * nucleosome, const size_t min_p);
double align_nucleo_linker_genotype_prob (const align_nucleo_linker * ref_linker, const align_nucleo_linker * qry_linker);
int align_nucleo_alignment_fprintf (FILE * stream, const align_nucleo_alignment * align);
void align_nucleo_alignment_pvalue (const align_nucleo_config * conf, const align_nucleo_chromosome * chrom, align_nucleo_alignment * align, const double obs_min, const size_t permut, const gdl_rng * rng);
void align_nucleo_linker_alignment_pvalue  (const align_nucleo_config * conf, const align_nucleo_chromosome * chrom, align_nucleo_alignment * align, const long size_min, const double obs_min);
int align_nucleo_linker_alignment_fprintf (FILE * stream, const align_nucleo_alignment * align);
int align_nucleo_linker_alignment_filter (align_nucleo_alignment * align);
int align_nucleo_linker_alignment_statistics (const align_nucleo_config * conf, align_nucleo_alignment * align);

void align_nucleo_alignment_anova_extract (const align_nucleo_config    * conf, 
		                             const align_nucleo_chromosome * chrom,
		                             align_nucleo_alignment * align,
		                             const double obs_min,
		                             const size_t prb_min,
		                             const gdl_boolean no_probe_effect_flag,
		                             const gdl_rng * rng,
		                             const long start,
		                             const long end,
		                             FILE * stream);

void align_nucleo_alignment_pval (const align_nucleo_config    * conf, 
                             const align_nucleo_chromosome * chrom,
                             align_nucleo_alignment * align,
                             const double obs_min,
                             const size_t prb_min,
                             const size_t permut,
                             const gdl_boolean no_probe_effect_flag,
                             const gdl_rng * rng);

align_nucleo_linker * align_nucleo_linker_fread (FILE * stream, const align_nucleo_config * config, const align_nucleo_alignment * align, const unsigned char kind);
align_nucleo_nucleosome * align_nucleo_nucleosome_fread (FILE * stream, const align_nucleo_config * config, const align_nucleo_alignment * align, const unsigned char kind);
align_nucleo_alignment * align_nucleo_alignment_fread (FILE * stream, const align_nucleo_config * config);
int align_nucleo_alignment_frwite (FILE * stream, const align_nucleo_config * config, const align_nucleo_alignment * a);

__END_DECLS

#endif
