/*  
 *  align_nucleo2.c
 * 
 *  $Author: gael $, $Date: 2008-10-17 16:07:07 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
 
#include <math.h>
#include <stdio.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_list.h>
#include <gdl/gdl_machine.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_sort_long.h>
#include <gdl/gdl_runtime.h>
#include <gdl/gdl_vector.h>
#include <gdl/gdl_matrix.h>
#include <gdl/gdl_multireg.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_statistics_double.h>
#include <gdl/gdl_genome_coord.h>
#include <gdl/gdl_sequences.h>

#include "tiling_hmm.h"
#include "tiling_nucleo.h"
#include "tiling_array.h"
#include "align_nucleo2.h"

/*==============================================================================================
 *
 * CONFIG + WORKSPACE 
 * 
 * =============================================================================================*/

nm_align_nucleo_config *
nm_align_nucleo_config_alloc (const gdl_string * filename)
{
	FILE * stream;
	nm_align_nucleo_config * conf = 0;
	
	stream = gdl_fileopen (filename, "r");
	if (!stream)
		return conf;
	
	size_t ti,tj,n;
	gdl_string * line=0, * key, * value;
	
	conf = GDL_CALLOC (nm_align_nucleo_config, 1);
	while(gdl_getline (&line, &n, stream)!=-1)
	{
		if (line[0]=='#') 
			continue;
		ti = tj = 0;
		key   = gdl_string_next_token (line, n, &ti, &tj);
		value = gdl_string_next_token (line, n, &ti, &tj);
		if (!strcmp(key, "ref_hmm_dir"))
			conf->ref_hmm_dir = value;
		if (!strcmp(key, "qry_hmm_dir"))
			conf->qry_hmm_dir = value;
		if (!strcmp(key, "ref_qry_coord_file"))
			conf->ref_qry_coord_file = value;
		if (!strcmp(key, "ref_seq_file"))
			conf->ref_seq_file = value;
		if (!strcmp(key, "qry_seq_file"))
			conf->qry_seq_file = value;
		if (!strcmp(key, "ref_landmark"))
		{
			conf->chroms = gdl_string_split (value, ",", &(conf->nchrom));
		}
		gdl_string_free (key);
		gdl_string_free (line);line=0;
	}
	gdl_fileclose (filename, stream);
	
	if (conf->ref_hmm_dir==0)
	{
		GDL_ERROR_VAL ("No ref HMM directory defined", GDL_EINVAL, 0);	
	}
	if (conf->qry_hmm_dir==0)
	{
		GDL_ERROR_VAL ("No qry HMM directory defined", GDL_EINVAL, 0);	
	}
	if (conf->ref_qry_coord_file==0)
	{
		GDL_ERROR_VAL ("No ref to qry file defined", GDL_EINVAL, 0);	
	}
	
	conf->ref_qry_coord = gdl_genome_coord_alloc (conf->ref_qry_coord_file);
	if (!conf->ref_qry_coord)
	{
		GDL_ERROR_VAL ("Unable to read ref to qry file", GDL_FAILURE, 0);	
	}
	if (conf->ref_seq_file)
	{
		gdl_seq_io * seq_io = gdl_seq_io_alloc (gdl_seq_io_type_get ("fasta"));
		gdl_seq_io_open (seq_io, conf->ref_seq_file);
		gdl_seq_io_parse (seq_io);
		conf->ref_seq = gdl_seq_io_close (seq_io);
		gdl_seqdb_id2seq_create (conf->ref_seq);
		gdl_seq_io_free (seq_io);	
	}
	if (conf->qry_seq_file)
	{
		gdl_seq_io * seq_io = gdl_seq_io_alloc (gdl_seq_io_type_get ("fasta"));
		gdl_seq_io_open (seq_io, conf->qry_seq_file);
		gdl_seq_io_parse (seq_io);
		conf->qry_seq = gdl_seq_io_close (seq_io);
		gdl_seqdb_id2seq_create (conf->qry_seq);
		gdl_seq_io_free (seq_io);
	}
	
	return conf;
}

static gdl_tiling_nucleo_chromosome *
load_chrom_hmm (const gdl_string * hmm_dir, const gdl_string * chrom, const size_t D, const size_t C)
{
	gdl_string * file;
	FILE * stream;
	gdl_tiling_nucleo_chromosome * chrom_hmm;
	
	file   = gdl_string_sprintf ("%s/%s.nuc", hmm_dir, chrom);
 	stream = gdl_fileopen (file, "r");
 	if (!stream)
 	{
 		GDL_ERROR_VAL (gdl_string_sprintf ("Unable to read ref hmm file for chromosome %s from directory %s", chrom, hmm_dir), GDL_FAILURE, 0);	
 	}
	chrom_hmm = gdl_tiling_nucleo_chromosome_fscanf (stream, D, C);
	gdl_fileclose (file, stream);
	gdl_string_free (file);
	
	return chrom_hmm;	
}

nm_align_nucleo_workspace * 
nm_align_nucleo_workspace_alloc (nm_align_nucleo_config * c)
{
	size_t i;
	nm_align_nucleo_workspace * w;
	
	w = GDL_CALLOC (nm_align_nucleo_workspace, 1);
	
	w->config = c;
	w->nref_chrom = c->nchrom;
	w->ref_chroms = GDL_MALLOC (nm_align_nucleo_chromosome *, c->nchrom);
	for(i = 0; i < w->nref_chrom; i++)
	{
		w->ref_chroms[i]      = nm_align_nucleo_chromosome_alloc (c->chroms[i]);
		w->ref_chroms[i]->hmm = load_chrom_hmm (c->ref_hmm_dir, w->ref_chroms[i]->name, 1, 3); 	
	}
	w->_qry_dico = gdl_hashtable_alloc (gdl_hash_default, 0);
	
	return w;
}

void
nm_align_nucleo_workspace_free (nm_align_nucleo_workspace * w)
{
	if (w)
	{
		size_t i;
		for(i = 0; i < w->nref_chrom; i++)
		{
			nm_align_nucleo_chromosome_free (w->ref_chroms[i]);
			nm_align_nucleo_alignment_free (w->chrom_aligned[i]);	
		}
		GDL_FREE (w->ref_chroms);
		GDL_FREE (w->chrom_aligned);
		for(i = 0; i < w->nqry_chrom; i++)
		{
			nm_align_nucleo_chromosome_free (w->qry_chroms[i]);
		}
		GDL_FREE (w->qry_chroms);
		gdl_hashtable_free (w->_qry_dico);
		GDL_FREE (w);
	}
}

int
nm_align_nucleo_workspace_anchor_qry (nm_align_nucleo_workspace * workspace,
                                      nm_align_nucleo_chromosome * ref_chrom,
                                      const size_t hmm_dim,
                                      const size_t tiling_size,
                                      FILE * logger) 
{
	size_t i,j;
	gdl_tiling_nucleo_chunk * chunk;
	gdl_tiling_nucleo_probe * probe;
	const nm_align_nucleo_config * conf = workspace->config;
	
	const gdl_chunk_coord_cluster * chunk_cluster = gdl_genome_coord_lookup (conf->ref_qry_coord, "ref", ref_chrom->name);
	
	if (!chunk_cluster)
	{
		fprintf (stderr, "[ WARNING ] No landmark %s in ref to qry coordinates\n", ref_chrom->name);
		fflush (stderr);
		return GDL_CONTINUE;
	}
	
	for(i = 0; i < ref_chrom->hmm->size; i++)
	{
		int ref_chunk=0;
		chunk = ref_chrom->hmm->chunks[i];
		for(j = 0; j < chunk->size; j++)
		{
			probe = chunk->probes[j];
			if (probe->idx >= 0)
			{
				long ref_pos = probe->position;
				ref_chunk    = gdl_chunk_coord_cluster_goto (chunk_cluster, ref_pos, ref_chunk);
				if (ref_chunk == -1) 
				{
					if (logger)
					{
						fprintf (logger, "[ WARNING] Unable to map position %d on landmark %s (unmapped region)\n", ref_pos, ref_chrom->name);
						fflush (logger);
					}
					ref_chunk=0;
					continue;
				}
				const gdl_string * qry_name = gdl_dictionary_get (chunk_cluster->seqid_dico, "qry", chunk_cluster->chunks[ref_chunk]->qry_id);
				nm_align_nucleo_chromosome * qry_chrom;
				if ((qry_chrom = gdl_hashtable_lookup (workspace->_qry_dico, qry_name))==0)
				{
					qry_chrom = nm_align_nucleo_chromosome_alloc (qry_name);
					// load the hmm results
				 	qry_chrom->hmm = load_chrom_hmm (conf->qry_hmm_dir, qry_name, (hmm_dim==3) ? 1 : 2, hmm_dim);
				 	// add the qry_chrom
				 	gdl_hashtable_add (workspace->_qry_dico, qry_name, qry_chrom, 0);
				}
				//probe-> = gdl_chunk_coord_cluster_map (chunk_cluster, ref_pos, ref_chunk)
			}
		}
	}		
}

int 
nm_align_nucleo_chromosome_map_qry (nm_align_nucleo_workspace * w, FILE * logger)
{
	size_t c,i,j;
	const nm_align_nucleo_config * conf = w->config;
	
	for(c = 0; c < w->nqry_chrom; c++)
	{
		nm_align_nucleo_chromosome * ref_chrom = 0;
		nm_align_nucleo_chromosome * qry_chrom = w->qry_chroms[c];
		const gdl_chunk_coord_cluster * chunk_cluster = gdl_genome_coord_lookup (conf->ref_qry_coord, "qry", qry_chrom->name);
		if (!chunk_cluster)
		{
			fprintf (stderr, "[ WARNING] No landmark %s in ref to qry coordinates\n", qry_chrom->name);
			fflush (stderr);
			return GDL_CONTINUE;
		}
		for(i = 0; i < qry_chrom->n_nucleosome; i++)
		{
			nm_align_nucleo_nucleosome * qry_nuc = qry_chrom->nucleosomes[i];
			long qry_start = qry_nuc->qry_start_pos;
			long qry_end   = qry_nuc->qry_end_pos;
			int qry_start_chunk=0, qry_end_chunk=0;
			
			qry_start_chunk = gdl_chunk_coord_cluster_goto (chunk_cluster, qry_start, qry_start_chunk);
			if (qry_start_chunk == -1) 
			{
				if (logger)
				{
					fprintf (logger, "[ WARNING] Unable to map position %d on landmark %s (unmapped region)\n", qry_start, qry_chrom->name);
					fflush (logger);
				}
				continue;
			}
			qry_end_chunk = gdl_chunk_coord_cluster_goto (chunk_cluster, qry_end, qry_end_chunk);
			if (qry_end_chunk == -1) 
			{
				if (logger)
				{
					fprintf (logger, "[ WARNING] Unable to map position %d on landmark %s (unmapped region)\n", qry_end, qry_chrom->name);
					fflush (logger);
				}
				continue;
			}
//			const gdl_string * ref_name = gdl_dictionary_get (chunk_cluster->seqid_dico, "ref", chunk_cluster->chunks[qry_start_chunk]->ref_id);
//			if (qry_start_chunk != qry_end_chunk)
//			{
//				if (logger)
//				{
//					fprintf (logger, "[ WARNING] Query nucleosome lies between two distinct chunks\n");
//					fflush (logger);
//				}
//				if (!strcmp(ref_name, "chrIV"))
//				{
//					printf ("[ WARNING] Query nucleosome lies between two distinct chunks\n");
//					qry_nuc->ref_start_pos = gdl_chunk_coord_cluster_map (chunk_cluster, qry_start, qry_start_chunk);
//            		qry_nuc->ref_end_pos   = gdl_chunk_coord_cluster_map (chunk_cluster, qry_end, qry_end_chunk);
//            		printf (" ==> %d %d\n", qry_nuc->ref_start_pos, qry_nuc->ref_end_pos);
//				}
//				continue;
//			}
			qry_nuc->ref_start_pos = gdl_chunk_coord_cluster_map (chunk_cluster, qry_start, qry_start_chunk);
            qry_nuc->ref_end_pos   = gdl_chunk_coord_cluster_map (chunk_cluster, qry_end, qry_end_chunk);
            if (qry_nuc->ref_start_pos > qry_nuc->ref_end_pos)
            {
            	qry_start = qry_nuc->ref_start_pos;
            	qry_nuc->ref_start_pos = qry_nuc->ref_end_pos;
            	qry_nuc->ref_end_pos   = qry_start;
            }
            const gdl_string * ref_name = gdl_dictionary_get (chunk_cluster->seqid_dico, "ref", chunk_cluster->chunks[qry_start_chunk]->ref_id);
            if (!ref_chrom || strcmp(ref_chrom->name, ref_name))
            {
            	for(j = 0; j < w->nref_chrom; j++)
            	{
            		if (!strcmp(w->ref_chroms[j]->name, ref_name))
            		{
            			break;
            		}	
            	}
            	if (j == w->nref_chrom)
            	{
            		GDL_WARNING ("Unexpected error: reference index (name) is out of bound", GDL_EINVAL);
            		continue;
            	}
            	ref_chrom = w->ref_chroms[j];
            }
            gdl_list_push_back (ref_chrom->_mapped_nucleosomes, qry_nuc, 0);
		}
	}		
}

void 
nm_align_nucleo_workspace_init_qry (nm_align_nucleo_workspace * w)
{
	if ((w->nqry_chrom=gdl_hashtable_size (w->_qry_dico))!=0)
	{
		size_t j, i = 0;
		gdl_hashtable_itr * itr;
		
		w->qry_chroms = GDL_MALLOC (nm_align_nucleo_chromosome *, w->nqry_chrom);
		
		itr = gdl_hashtable_iterator (w->_qry_dico);
		
		do
		{
			w->qry_chroms[i++] = gdl_hashtable_iterator_value (itr);
		}
		while (gdl_hashtable_iterator_next (itr));
		
		gdl_hashtable_iterator_free (itr);
		
	}	
}

/*==============================================================================================
 *
 *  CHROMOSOME 
 * 
 * =============================================================================================*/
 
nm_align_nucleo_chromosome *
nm_align_nucleo_chromosome_alloc (const gdl_string * name)
{
	nm_align_nucleo_chromosome * c;
	
	c = GDL_CALLOC (nm_align_nucleo_chromosome, 1);
	
	c->name = gdl_string_clone (name);
	
	c->_mapped_nucleosomes = gdl_list_alloc (gdl_list_default); 
	
	return c;
}

void
nm_align_nucleo_chromosome_free (nm_align_nucleo_chromosome * c)
{
	if (c)
	{
		size_t i;
		gdl_string_free (c->name);
		gdl_tiling_nucleo_chromosome_free (c->hmm);
		for(i = 0; i < c->n_nucleosome; i++)
		{
			nm_align_nucleo_nucleosome_free (c->nucleosomes[i]);
		}
		GDL_FREE (c->nucleosomes);
		for(i = 0; i < c->n_linker; i++)
		{
			nm_align_nucleo_linker_free (c->linkers[i]);
		}
		GDL_FREE (c->linkers);
		gdl_list_free (c->_mapped_nucleosomes);
		GDL_FREE (c);
	}	
}

nm_align_nucleo_chromosome *
nm_align_nucleo_chromosome_fread (FILE * stream)
{
	if (stream)
	{
		int i, status;
		size_t n_mapped;
		gdl_string * name;
		nm_align_nucleo_chromosome * c;
		
		name = gdl_string_fread (stream);
		GDL_FREAD_STATUS (name != 0, 1);
		
		c = nm_align_nucleo_chromosome_alloc (name);
		
		gdl_string_free (name);
		
		status = fread (&(c->n_nucleosome), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		
		c->nucleosomes = GDL_MALLOC (nm_align_nucleo_nucleosome *, c->n_nucleosome);
		
		for(i = 0; i < c->n_nucleosome; i++)
		{
			c->nucleosomes[i] = nm_align_nucleo_nucleosome_fread (stream);
		}
		
		status = fread (&(n_mapped), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		
		if (n_mapped)
		{
			for(i = 0; i < n_mapped; i++)
			{
				nm_align_nucleo_nucleosome * qry_nuc = nm_align_nucleo_nucleosome_fread (stream);
				if (qry_nuc->align_idx >= 0)
				{
					c->nucleosomes[qry_nuc->align_idx]->align = qry_nuc;
					qry_nuc->align = c->nucleosomes[qry_nuc->align_idx];
				}
				gdl_list_push_back (c->_mapped_nucleosomes, qry_nuc, 0);
			}	
		}
		
		return c;
	}
	
	return 0;
}

int
nm_align_nucleo_chromosome_fwrite (FILE * stream, const nm_align_nucleo_chromosome * c)
{
	if (stream && c)
	{
		int i, status;
		size_t n_mapped;
		
		status = gdl_string_fwrite (stream, c->name);
		GDL_FWRITE_STATUS (status, GDL_SUCCESS);
		status = fwrite (&(c->n_nucleosome), sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		for(i = 0; i < c->n_nucleosome; i++)
		{
			status = nm_align_nucleo_nucleosome_fwrite (stream, c->nucleosomes[i]);
			GDL_FWRITE_STATUS (status, GDL_SUCCESS);
		}
		n_mapped = gdl_list_size (c->_mapped_nucleosomes);
		status = fwrite (&(n_mapped), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		if (n_mapped)
		{
			gdl_list_itr * itr = gdl_list_iterator_front (c->_mapped_nucleosomes);
			do
			{
				nm_align_nucleo_nucleosome * qry_nuc = (nm_align_nucleo_nucleosome *) gdl_list_iterator_value (itr);
				status = nm_align_nucleo_nucleosome_fwrite (stream, qry_nuc);
				GDL_FWRITE_STATUS (status, GDL_SUCCESS);
			}
			while(gdl_list_iterator_next (itr));
			gdl_list_iterator_free (itr);
		}
		
		return GDL_SUCCESS;
	}
	
	return GDL_EINVAL;
}

nm_align_nucleo_nucleosome * 
nm_align_nucleo_nucleosome_alloc (nm_align_nucleo_chromosome * chrom, const size_t chunk, const size_t start, const size_t end, const size_t status, const size_t nclass)
{
	nm_align_nucleo_nucleosome * n;
	
	n = GDL_CALLOC (nm_align_nucleo_nucleosome, 1);
	
	n->chrom = chrom;
	n->chunk_idx = chunk;
	n->start_idx = start;
	n->end_idx   = end;
	n->status    = status;
	n->prob      = GDL_CALLOC (double, (nclass-1)/2+1);
	n->ignore    = 'n';
	
	return n;
}

nm_align_nucleo_linker * 
nm_align_nucleo_linker_alloc (const nm_align_nucleo_chromosome * chrom, const size_t chunk, const size_t start, const size_t end, const size_t nclass)
{
	nm_align_nucleo_linker * n;
	
	n = GDL_CALLOC (nm_align_nucleo_linker, 1);
	
	n->chrom = chrom;
	n->chunk_idx = chunk;
	n->start_idx = start;
	n->end_idx   = end;
	n->prob      = GDL_CALLOC (double, (nclass-1)/2+1);
	
	return n;
}

nm_align_nucleo_nucleosome *
nm_align_nucleo_chromosome_call_nucleo (nm_align_nucleo_chromosome * chrom,
                                     const gdl_tiling_nucleo_probe * probe,
                                     const size_t chunk_idx,
                                     const size_t probe_idx,
                                     const size_t nucleo_min,
                                     const size_t nclass,
                                     size_t * score,
                                     size_t * status,
                                     size_t * start,
                                     size_t * cidx,
                                     nm_align_nucleo_chromosome ** chrom_ptr)
{
	nm_align_nucleo_nucleosome * nucleosome = 0;
	
	switch (probe->calling)
	{
		case 0: // linker
			if (*score >= nucleo_min 
			    && (*cidx == chunk_idx && probe_idx>0)
			    && (*chrom_ptr == chrom))
			{
			    size_t i,j,k;
				double t=0;
				nucleosome = nm_align_nucleo_nucleosome_alloc (chrom, chunk_idx, *start, probe_idx-1, *status, nclass);
				size_t ps = GDL_MIN (nucleosome->start_idx, nucleosome->end_idx);
				size_t pe = GDL_MAX (nucleosome->start_idx, nucleosome->end_idx);
				for(i = ps; i <= pe; i++)
				{
					gdl_tiling_nucleo_probe * probe_nuc = chrom->hmm->chunks[chunk_idx]->probes[i];
					nucleosome->prob[0] += probe_nuc->prob[0];
					t += probe_nuc->prob[0];
					for(k = 1; k <= (nclass-1)/2; k++)
					{
						t += probe_nuc->prob[2*k-1]+probe_nuc->prob[2*k];
						nucleosome->prob[k] += probe_nuc->prob[2*k-1]+probe_nuc->prob[2*k];
					}
				}
				for(i = 0; i <= (nclass-1)/2; i++)
					nucleosome->prob[i] /= t;
			}
			*score=0;
			break;
		case 1: // not-acetylated delocalized
		case 2: // not-acetylated well-positioned
		case 3: // acetylated delocalized
		case 4: // acetylated well-positioned
			if ((*score)==0)
			{
				*start     = probe_idx;
				*status    = probe->calling;
				*cidx      = chunk_idx;
				*chrom_ptr = chrom;
			}
			else if (*status != probe->calling) // reset it
			{
				*score     = 0;
				*start     = probe_idx;
				*status    = probe->calling;
				*cidx      = chunk_idx;
				*chrom_ptr = chrom;
			}
			(*score)++;
			break;
	}
	
	return nucleosome;
}

nm_align_nucleo_linker *
nm_align_nucleo_chromosome_call_linker (nm_align_nucleo_chromosome * chrom,
                                       const gdl_tiling_nucleo_probe * probe,
                                       const size_t chunk_idx,
                                       const size_t probe_idx,
                                       const size_t nclass,
                                       size_t * score,
                                       size_t * chunk,
                                       size_t * start,
                                       nm_align_nucleo_chromosome ** chrom_ptr)
{
	nm_align_nucleo_linker * linker = 0;
	switch (probe->calling)
	{
		case 0: // linker
			if ((*score)==0)
			{
				*start     = probe_idx;
				*chunk     = chunk_idx;
				*chrom_ptr = chrom;
			}
			else if (*chunk != chunk_idx || *chrom_ptr != chrom)
			{
				*score=0;	
			}
			(*score)++;
			break;
		case 1: // not-acetylated delocalized
		case 2: // not-acetylated well-positioned
		case 3: // acetylated delocalized
		case 4: // acetylated well-positioned
			if (*score && (*chunk==chunk_idx && *chrom_ptr==chrom))
			{
				size_t i,j,k;
				double t=0;
				linker = nm_align_nucleo_linker_alloc (chrom, chunk_idx, *start, probe_idx-1, nclass);
				size_t ps = GDL_MIN (linker->start_idx, linker->end_idx);
				size_t pe = GDL_MAX (linker->start_idx, linker->end_idx);
				//printf ("ps = %d pe = %d (%d (%d) %p, %p)\n", ps, pe, chunk_idx, chrom->hmm->size, chrom->hmm->chunks[chunk_idx]->size, chrom, * chrom_ptr);
				for(i = ps; i <= pe; i++)
				{
					gdl_tiling_nucleo_probe * probe_nuc = chrom->hmm->chunks[chunk_idx]->probes[i];
					linker->prob[0] += probe_nuc->prob[0];
					t += probe_nuc->prob[0];
					for(k = 1; k <= (nclass-1)/2; k++)
					{
						t += probe_nuc->prob[2*k-1]+probe_nuc->prob[2*k];
						linker->prob[k] += probe_nuc->prob[2*k-1]+probe_nuc->prob[2*k];
					}
				}
				for(i = 0; i <= (nclass-1)/2; i++)
					linker->prob[i] /= t;
				*score=0;
			}
			else if (*chunk!=chunk_idx)
			{
				*score=0;	
			}
			else if (*chrom_ptr!=chrom)
			{
				*score=0;	
			}
			break;	
	}
	return linker;
}

int
nm_align_nucleo_chromosome_anchor_nucleo (nm_align_nucleo_chromosome * chrom,
                                         const size_t tiling_size,
                                         const size_t nucleo_min,
                                         const size_t nclass,
                                         const size_t query_flag)
{
	size_t i,j,jj,k;
	size_t score,status,start,lscore,lstart,lcidx,rncidx,qncidx;
	size_t qry_score,qry_status,qry_start,ref_rm=0,qry_rm=0;
	nm_align_nucleo_nucleosome * ref_nucleo;
	nm_align_nucleo_linker * ref_linker;
	nm_align_nucleo_chromosome ** ref_chrom_ptr;
	nm_align_nucleo_chromosome ** lref_chrom_ptr;
	gdl_list * ref_nucleosomes = gdl_list_alloc (gdl_list_default);
	gdl_list * ref_linkers = gdl_list_alloc (gdl_list_default);
	
	ref_chrom_ptr = GDL_MALLOC (nm_align_nucleo_chromosome *, 1);
	lref_chrom_ptr = GDL_MALLOC (nm_align_nucleo_chromosome *, 1);
	
	for (i = 0; i < chrom->hmm->size; i++)
	{
		const gdl_tiling_nucleo_chunk * chunk = chrom->hmm->chunks[i];
		score=0;
		for(j = 0; j < chunk->size; j++)
		{
			gdl_tiling_nucleo_probe * probe = chunk->probes[j];
			ref_nucleo = nm_align_nucleo_chromosome_call_nucleo (chrom, probe, i, j, nucleo_min, nclass, &score, &status, &start, &rncidx, ref_chrom_ptr);
			if (ref_nucleo)
			{
				ref_nucleo->ref_start_pos = chunk->probes[ref_nucleo->start_idx]->position;
				ref_nucleo->ref_end_pos   = chunk->probes[ref_nucleo->end_idx]->position;
				
				if (query_flag)
				{
					ref_nucleo->qry_start_pos = chunk->probes[ref_nucleo->start_idx]->position;
					ref_nucleo->qry_end_pos   = chunk->probes[ref_nucleo->end_idx]->position;
					ref_nucleo->qry_seq_name  = gdl_string_clone (chrom->name);
				}
				
				ref_nucleo->ignore = 'n';
				ref_nucleo->idx = gdl_list_size (ref_nucleosomes);
				gdl_list_push_back (ref_nucleosomes, ref_nucleo, 0);
				// get the next linker
				ref_linker=0;
				for(k = j; k < chunk->size && ref_linker==0; k++)
				{
					ref_linker = nm_align_nucleo_chromosome_call_linker (chrom, chunk->probes[k], i, k, nclass, &lscore, &lcidx, &lstart, lref_chrom_ptr);
				}
				if (ref_linker)
				{
					ref_linker->ref_start_pos = chunk->probes[ref_linker->start_idx]->position;
					ref_linker->ref_end_pos   = chunk->probes[ref_linker->end_idx]->position;
					ref_nucleo->next_linker   = ref_linker;
					ref_linker->idx = gdl_list_size (ref_linkers);
					gdl_list_push_back (ref_linkers, ref_linker, 0);
				}
			}	
		}
	}
	GDL_FREE (ref_chrom_ptr);
	GDL_FREE (lref_chrom_ptr);
	if ((chrom->n_nucleosome=gdl_list_size (ref_nucleosomes))>0)
	{
		gdl_list_itr * itr = gdl_list_iterator_front (ref_nucleosomes);
		i=0;
		chrom->nucleosomes = GDL_CALLOC (nm_align_nucleo_nucleosome *, chrom->n_nucleosome); 
		do
		{
			chrom->nucleosomes[i++] = gdl_list_iterator_value (itr);
		}
		while(gdl_list_iterator_next (itr));
		gdl_list_iterator_free (itr);
	}
	if ((chrom->n_linker=gdl_list_size (ref_linkers))>0)
	{
		gdl_list_itr * itr = gdl_list_iterator_front (ref_linkers);
		i=0;
		chrom->linkers = GDL_CALLOC (nm_align_nucleo_linker *, chrom->n_linker); 
		do
		{
			chrom->linkers[i++] = gdl_list_iterator_value (itr);
		}
		while(gdl_list_iterator_next (itr));
		gdl_list_iterator_free (itr);
	}
	fprintf (stderr, "--\n");
	fprintf (stderr, " Nucleosme calling\n");
	fprintf (stderr, "\n");
	fprintf (stderr, "  Found %d nucleosomes\n", chrom->n_nucleosome);
	fprintf (stderr, "  Found %d linkers\n", chrom->n_linker);
	fprintf (stderr, "--\n");
	fflush(stderr);
	gdl_list_free (ref_nucleosomes);
}

nm_align_nucleo_nucleosome *
nm_align_nucleo_nucleosome_fread (FILE * stream)
{
	if (stream)
	{
		int status;
		int chrom_idx;
		unsigned char has;
		nm_align_nucleo_nucleosome * nuc;
		
		nuc = GDL_CALLOC (nm_align_nucleo_nucleosome, 1);
		
		status = fread (&(nuc->idx), sizeof(int), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->ignore), sizeof(unsigned char), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->chunk_idx), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->start_idx), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->end_idx), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->align_idx), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->status), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->ref_start_pos), sizeof(long), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->ref_end_pos), sizeof(long), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->pval), sizeof(double), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->map), sizeof(double), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->obs), sizeof(double), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->delta), sizeof(double), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->outlier), sizeof(gdl_boolean), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->indel), sizeof(double), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&(nuc->gamma), sizeof(double), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		nuc->prob = GDL_CALLOC (double, 3);
		status = fread (nuc->prob, sizeof(double), 3, stream);
		GDL_FREAD_STATUS (status, 3);
		
		nuc->qry_seq_name = gdl_string_fread (stream);
		GDL_FREAD_STATUS (nuc->qry_seq_name != 0, 1);
		
		if (strcmp(nuc->qry_seq_name, "NULL"))
		{
			status = fread (&(nuc->qry_start_pos), sizeof(long), 1, stream);
			GDL_FREAD_STATUS (status, 1);
			status = fread (&(nuc->qry_end_pos), sizeof(long), 1, stream);
			GDL_FREAD_STATUS (status, 1);
		}
		else
		{
			gdl_string_free (nuc->qry_seq_name);
			nuc->qry_seq_name = NULL;	
		}
		
		return nuc;
	}
	return 0;
}

int
nm_align_nucleo_nucleosome_fwrite (FILE * stream, const nm_align_nucleo_nucleosome * nuc)
{
	if (stream)
	{
		int status, nox=-1;
		unsigned char has;
		
		status = fwrite (&(nuc->idx), sizeof(int), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->ignore), sizeof(unsigned char), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->chunk_idx), sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->start_idx), sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->end_idx), sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->align_idx), sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->status), sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->ref_start_pos), sizeof(long), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->ref_end_pos), sizeof(long), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->pval), sizeof(double), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->map), sizeof(double), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->obs), sizeof(double), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->delta), sizeof(double), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->outlier), sizeof(gdl_boolean), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->indel), sizeof(double), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&(nuc->gamma), sizeof(double), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (nuc->prob, sizeof(double), 3, stream);
		GDL_FWRITE_STATUS (status, 3);
		
		if (nuc->qry_seq_name)
		{
			status = gdl_string_fwrite (stream, nuc->qry_seq_name);
			GDL_FWRITE_STATUS (status, GDL_SUCCESS);
			status = fwrite (&(nuc->qry_start_pos), sizeof(long), 1, stream);
			GDL_FWRITE_STATUS (status, 1);
			status = fwrite (&(nuc->qry_end_pos), sizeof(long), 1, stream);
			GDL_FWRITE_STATUS (status, 1);
		}
		else
		{
			status = gdl_string_fwrite (stream, "NULL");
			GDL_FWRITE_STATUS (status, GDL_SUCCESS);
		}
				
		return GDL_SUCCESS;
	}
	return GDL_EINVAL;
}

void
nm_align_nucleo_linker_free (nm_align_nucleo_linker * n)
{
	if (n)
	{
		GDL_FREE (n->prob);
	   GDL_FREE (n);
	}	
}

void
nm_align_nucleo_nucleosome_free (nm_align_nucleo_nucleosome * n)
{
	if (n)
	{
		GDL_FREE (n->prob);
	   nm_align_nucleo_linker_free (n->next_linker);
		GDL_FREE (n);
	}	
}

void
nm_align_nucleo_alignment_free (nm_align_nucleo_alignment * a)
{
	if (a)
	{
		size_t i;
		GDL_FREE (a->pos);
		for(i = 0; i < a->size; i++)
		{
			nm_align_nucleo_nucleosome_free (a->ref[i]);
			nm_align_nucleo_nucleosome_free (a->qry[i]);		
		}
		GDL_FREE (a->ref);
		GDL_FREE (a->qry);
		for(i = 0; i < a->nqry; i++)
		{
			gdl_tiling_nucleo_chromosome_free (a->qry_chroms[i]->hmm);			
		}
		GDL_FREE (a->qry_chroms);
		gdl_tiling_nucleo_chromosome_free (a->ref_chrom->hmm);
		//GDL_FREE (a->ref_chrom);
		GDL_FREE (a);
	}	
}
