/*  
 *  align_nucleo.h
 * 
 *  $Author: gael $, $Date: 2008-10-17 10:34:08 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
 
#ifndef __GDL_ALIGN_NUCLEO2_H__
#define __GDL_ALIGN_NUCLEO2_H__

#include <math.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_randist.h>
#include <gdl/gdl_string.h>
#include <gdl/gdl_vector.h>
#include <gdl/gdl_matrix.h>
#include <gdl/gdl_multireg.h>
#include <gdl/gdl_genome_coord.h>
#include <gdl/gdl_sequences.h>

#include "tiling_hmm.h"
#include "tiling_nucleo.h"
#include "tiling_array.h"

__BEGIN_DECLS

typedef struct
{
	// ref landmarks
	size_t nchrom;
	gdl_string ** chroms;
	// ref files
	gdl_string * ref_hmm_dir;
	gdl_string * ref_qry_coord_file;
	gdl_string * ref_seq_file;
	// qry files
	gdl_string * qry_hmm_dir;
	gdl_string * qry_seq_file;
	// objects
	gdl_seqdb * ref_seq;
	gdl_seqdb * qry_seq;
	gdl_genome_coord * ref_qry_coord;
} nm_align_nucleo_config;

nm_align_nucleo_config * nm_align_nucleo_config_alloc (const gdl_string * filename);

typedef struct _nm_align_nucleo_chromosome nm_align_nucleo_chromosome;

typedef struct
{
	size_t idx;
	unsigned char seen;
	unsigned char _seen;
	size_t chunk_idx;
	size_t start_idx;
	size_t end_idx;
	long ref_start_pos;
	long ref_end_pos;
	const nm_align_nucleo_chromosome * chrom;
	double pval;
	double * prob;
	double map;
	double obs;
	gdl_boolean outlier;
} nm_align_nucleo_linker;

typedef struct _nm_align_nucleo_nucleosome nm_align_nucleo_nucleosome;

struct _nm_align_nucleo_nucleosome
{
	int idx;
	unsigned char ignore;
	unsigned char seen;
	unsigned char _seen;
	size_t chunk_idx;
	size_t start_idx;
	size_t end_idx;
	size_t status;
	int align_idx;
	long ref_start_pos;
	long ref_end_pos;
	long qry_start_pos;
	long qry_end_pos;
	gdl_string * qry_seq_name;
	nm_align_nucleo_chromosome * chrom;
	nm_align_nucleo_nucleosome * align;
	double * prob;
	double delta;
	double pval;
	double map;
	double obs;
	double indel;
	double gamma;
	double mrk_pval;
	double mrk_gamma;
	gdl_boolean outlier;
	nm_align_nucleo_linker * next_linker;
};

typedef struct
{
	size_t size;
	size_t nqry;
	long * pos;
	nm_align_nucleo_chromosome  * ref_chrom;
	nm_align_nucleo_chromosome ** qry_chroms;
	nm_align_nucleo_nucleosome ** ref;
	nm_align_nucleo_nucleosome ** qry;
} nm_align_nucleo_alignment;

struct _nm_align_nucleo_chromosome
{
	gdl_string * name;
	gdl_tiling_nucleo_chromosome * hmm;
	size_t n_nucleosome;
	nm_align_nucleo_nucleosome ** nucleosomes;
	size_t n_linker;
	nm_align_nucleo_linker ** linkers;
	// utilities
	gdl_list * _mapped_nucleosomes;
};

typedef struct
{
	nm_align_nucleo_config * config; 
	size_t nref_chrom;
	size_t nqry_chrom;
	nm_align_nucleo_alignment  ** chrom_aligned;
	nm_align_nucleo_chromosome ** ref_chroms;
	nm_align_nucleo_chromosome ** qry_chroms;
	// utilities
	gdl_hashtable * _qry_dico;
} nm_align_nucleo_workspace;

nm_align_nucleo_workspace * nm_align_nucleo_workspace_alloc (nm_align_nucleo_config * c);
void nm_align_nucleo_workspace_free (nm_align_nucleo_workspace * w);
int nm_align_nucleo_workspace_anchor_qry (nm_align_nucleo_workspace * w, nm_align_nucleo_chromosome * ref_chrom, const size_t hmm_dim, const size_t tiling_size, FILE * logger);
void nm_align_nucleo_workspace_init_qry (nm_align_nucleo_workspace * w);
int nm_align_nucleo_chromosome_map_qry (nm_align_nucleo_workspace * w, FILE * logger);

nm_align_nucleo_chromosome * nm_align_nucleo_chromosome_alloc (const gdl_string * name);
void nm_align_nucleo_chromosome_free (nm_align_nucleo_chromosome * c);
nm_align_nucleo_chromosome * nm_align_nucleo_chromosome_fread (FILE * stream);

nm_align_nucleo_nucleosome * nm_align_nucleo_nucleosome_alloc (nm_align_nucleo_chromosome * chrom, const size_t chunk, const size_t start, const size_t end, const size_t status, const size_t nclass);
void nm_align_nucleo_nucleosome_free (nm_align_nucleo_nucleosome * n);
nm_align_nucleo_linker * nm_align_nucleo_linker_alloc (const nm_align_nucleo_chromosome * chrom, const size_t chunk, const size_t start, const size_t end, const size_t nclass);
void nm_align_nucleo_linker_free (nm_align_nucleo_linker * l);
//gdl_tiling_nucleo_chromosome * load_chrom_hmm (const gdl_string * hmm_dir, const gdl_string * chrom, const size_t D, const size_t C);
int nm_align_nucleo_chromosome_anchor_probe (nm_align_nucleo_chromosome * chrom);
int nm_align_nucleo_chromosome_fprintf_probe (FILE * stream, const nm_align_nucleo_chromosome * chrom, const gdl_tiling_nucleo_probe * probe);
int nm_align_nucleo_chromosome_fprintf (FILE * stream, const nm_align_nucleo_chromosome * chrom, const size_t mode);

nm_align_nucleo_nucleosome * nm_align_nucleo_nucleosome_fread (FILE * stream);
int nm_align_nucleo_nucleosome_fwrite (FILE * stream, const nm_align_nucleo_nucleosome * n);

nm_align_nucleo_nucleosome * nm_align_nucleo_chromosome_call_nucleo (nm_align_nucleo_chromosome * chrom, const gdl_tiling_nucleo_probe * probe, const size_t chunk_idx, const size_t probe_idx, const size_t nucleo_min, const size_t nclass, size_t * score, size_t * status, size_t * start, size_t * cidx, nm_align_nucleo_chromosome ** chrom_ptr);
nm_align_nucleo_linker * nm_align_nucleo_chromosome_call_linker (nm_align_nucleo_chromosome * chrom, const gdl_tiling_nucleo_probe * probe, const size_t chunk_idx, const size_t probe_idx, const size_t nclass, size_t * score, size_t * chunk, size_t * start, nm_align_nucleo_chromosome ** chrom_ptr);
void nm_align_nucleo_chromosome_nucleo_properties (const gdl_tiling_nucleo_chunk * ref_chunk, nm_align_nucleo_nucleosome * ref_nucleo);
void nm_align_nucleo_chromosome_linker_properties (const gdl_tiling_nucleo_chunk * ref_chunk, nm_align_nucleo_linker * ref_linker);
int nm_align_nucleo_chromosome_anchor_nucleo (nm_align_nucleo_chromosome * chrom, const size_t tiling_size, const size_t nucleo_min, const size_t nclass, const size_t query_flag);
void nm_align_nucleo_chromosome_align_neighbour (nm_align_nucleo_chromosome * chrom, nm_align_nucleo_nucleosome * ref_nucleo_start, nm_align_nucleo_nucleosome * ref_nucleo_end, nm_align_nucleo_nucleosome * qry_nucleo_start, nm_align_nucleo_nucleosome * qry_nucleo_end, const gdl_boolean include_start, const gdl_boolean include_end);
nm_align_nucleo_alignment * nm_align_nucleo_alignment_alloc (nm_align_nucleo_chromosome * chrom);
void nm_align_nucleo_alignment_free (nm_align_nucleo_alignment * a);
nm_align_nucleo_alignment * nm_align_nucleo_chromosome_perform (nm_align_nucleo_chromosome * chrom, const double overlap_min);
double nm_align_nucleo_nucleosome_genotype_prob (const nm_align_nucleo_nucleosome * ref_nucleo, const nm_align_nucleo_nucleosome * qry_nucleo);
double nm_align_nucleo_nucleosome_indel_prob (const nm_align_nucleo_nucleosome * nucleosome, const size_t min_p);
double nm_align_nucleo_linker_genotype_prob (const nm_align_nucleo_linker * ref_linker, const nm_align_nucleo_linker * qry_linker);
int nm_align_nucleo_alignment_fprintf (FILE * stream, const nm_align_nucleo_alignment * align);
void nm_align_nucleo_alignment_pvalue (const nm_align_nucleo_config * conf, const nm_align_nucleo_chromosome * chrom, nm_align_nucleo_alignment * align, const double obs_min, const size_t permut, const gdl_rng * rng);
void nm_align_nucleo_linker_alignment_pvalue  (const nm_align_nucleo_config * conf, const nm_align_nucleo_chromosome * chrom, nm_align_nucleo_alignment * align, const long size_min, const double obs_min);
int nm_align_nucleo_linker_alignment_fprintf (FILE * stream, const nm_align_nucleo_alignment * align);
int nm_align_nucleo_linker_alignment_filter (nm_align_nucleo_alignment * align);
int nm_align_nucleo_linker_alignment_statistics (const nm_align_nucleo_config * conf, nm_align_nucleo_alignment * align);

void nm_align_nucleo_alignment_anova_extract (const nm_align_nucleo_config    * conf, 
		                             const nm_align_nucleo_chromosome * chrom,
		                             nm_align_nucleo_alignment * align,
		                             const double obs_min,
		                             const size_t prb_min,
		                             const gdl_boolean no_probe_effect_flag,
		                             const gdl_rng * rng,
		                             const long start,
		                             const long end,
		                             FILE * stream);

void nm_align_nucleo_alignment_pval (const nm_align_nucleo_config    * conf, 
                             const nm_align_nucleo_chromosome * chrom,
                             nm_align_nucleo_alignment * align,
                             const double obs_min,
                             const size_t prb_min,
                             const size_t permut,
                             const gdl_boolean no_probe_effect_flag,
                             const gdl_rng * rng);

nm_align_nucleo_alignment * nm_align_nucleo_alignment_fread (FILE * stream, const nm_align_nucleo_config * config);
int nm_align_nucleo_alignment_frwite (FILE * stream, const nm_align_nucleo_config * config, const nm_align_nucleo_alignment * a);

__END_DECLS

#endif
