/*  
 *  tiling_array.c
 * 
 *  $Author: gael $, $Date: 2008-09-15 12:32:59 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
 
#include <math.h>

#include <gdl/gdl_list.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_statistics_double.h>

#include "tiling_array.h"

gdl_tiling_chromosome *
gdl_tiling_chromosome_alloc (const gdl_string * name, const size_t nprobe, const size_t narray)
{
	gdl_tiling_chromosome * c;
	
	c = GDL_CALLOC (gdl_tiling_chromosome, 1);
	
	c->name   = gdl_string_clone (name);
	c->nprobe = nprobe;
	c->narray = narray;
	c->positions   = GDL_CALLOC (long, nprobe);
	c->intensities = GDL_CALLOC (double, nprobe*narray);
	
	return c;
}

void
gdl_tiling_chromosome_free (gdl_tiling_chromosome * c)
{
	if (c)
	{
		GDL_FREE (c->name);
		GDL_FREE (c->positions);
		GDL_FREE (c->intensities);
		GDL_FREE (c);
	}
}

double
gdl_tiling_chromosome_get (gdl_tiling_chromosome * c, const size_t p, const size_t a)
{
	
}

gdl_tiling_chromosome *
gdl_tiling_chromosome_fread (FILE * stream)
{
	if (stream)
	{
		int status;
		gdl_string * name;
		size_t nprobe, narray;
		gdl_tiling_chromosome * c;
		
		name = gdl_string_fread (stream);
		GDL_FREAD_STATUS (name!=0, 1);
		status = fread (&nprobe, sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&narray, sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		
		c = gdl_tiling_chromosome_alloc (name, nprobe, narray);
		
		status = fread (c->positions, sizeof(long), nprobe, stream);
		GDL_FREAD_STATUS (status, nprobe);
		status = fread (c->intensities, sizeof(double), nprobe*narray, stream);
		GDL_FREAD_STATUS (status, nprobe*narray);
		
		gdl_string_free (name);
		
		return c;
	}
	return 0;
}

int
gdl_tiling_chromosome_fwrite (FILE * stream, const gdl_tiling_chromosome * c)
{
	if (stream && c)
	{
		int status;
		
		status = gdl_string_fwrite (stream, c->name);
		GDL_FWRITE_STATUS (status, GDL_SUCCESS);
		status = fwrite (&c->nprobe, sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&c->narray, sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (c->positions, sizeof(long), c->nprobe, stream);
		GDL_FWRITE_STATUS (status, c->nprobe);
		status = fwrite (c->intensities, sizeof(double), c->nprobe*c->narray, stream);
		GDL_FWRITE_STATUS (status, c->nprobe*c->narray);
		
		return GDL_SUCCESS;
	}
	return GDL_EINVAL;
}

gdl_tiling_data *
gdl_tiling_data_alloc (const gdl_tiling_data_conf * conf)
{
	const size_t narray    = conf->narray;
	const size_t * groupId = conf->groupId;
	const gdl_tiling_data_type * T;
	size_t i, * group;
	gdl_tiling_data * d;
	
	d = GDL_CALLOC (gdl_tiling_data, 1);
	
	if (!strcmp(conf->type, gdl_tiling_data_full->name))
		T = gdl_tiling_data_full;
	else if (!strcmp(conf->type, gdl_tiling_data_mean->name))
		T = gdl_tiling_data_mean;
	else if (!strcmp(conf->type, gdl_tiling_data_median->name))
		T = gdl_tiling_data_median;
	else
		GDL_ERROR_VAL ("Unvalid gdl_tiling_data type name in gdl_tiling_data_conf", GDL_EINVAL, 0);
	
	d->T       = T;
	d->name    = gdl_string_clone (conf->name);
	d->narray  = narray;
	d->groupId = GDL_MALLOC (size_t, narray);
	memcpy (d->groupId, groupId, sizeof(size_t)*narray);
	group = GDL_CALLOC (size_t, narray);
	for(i=0;i<narray;i++) 
		group[groupId[i]]++;
	for(i=0;i<narray && group[i]!=0;i++);
	d->ngroup  = i;
	d->narrayByGroup = GDL_CALLOC (size_t, d->ngroup);
	memcpy (d->narrayByGroup, group, sizeof(size_t)*d->ngroup);
	d->arrayByGroupIdx = GDL_MALLOC (size_t *, d->ngroup);
	for(i=0;i<d->ngroup;i++)
	{
		group[i]=0;
		d->arrayByGroupIdx[i]=GDL_MALLOC (size_t, d->narrayByGroup[i]);
	}
	for(i=0;i<narray;i++)
		d->arrayByGroupIdx[groupId[i]][group[groupId[i]]++]=i;
	
	return d;	
}

gdl_tiling_data *
gdl_tiling_data_alloc2 (const size_t narray)
{
	size_t i;
	gdl_tiling_data * d;
	
	d = GDL_CALLOC (gdl_tiling_data, 1);
	
	d->T       = gdl_tiling_data_full;
	d->name    = gdl_string_sprintf ("%s", "data");
	d->narray  = narray;
	d->ngroup  = narray;
	d->groupId = GDL_MALLOC (size_t, narray);
	d->narrayByGroup = GDL_CALLOC (size_t, narray);
	d->arrayByGroupIdx = GDL_MALLOC (size_t *, narray);
	for(i=0;i<narray;i++) 
	{
		d->groupId[i]=i;
		d->narrayByGroup[i]=1;
		d->arrayByGroupIdx[i]=GDL_MALLOC (size_t, 1);
		d->arrayByGroupIdx[i][0]=i;
	}	
	return d;	
}

void
gdl_tiling_data_free (gdl_tiling_data * d)
{
	if (d)
	{
		size_t i;
		if (d->nchrom)
		{
			for(i=0;i<d->nchrom;i++)
				gdl_tiling_chromosome_free (d->chroms[i]);
			GDL_FREE (d->chroms);
		}
		for(i=0;i<d->ngroup;i++)
			GDL_FREE (d->arrayByGroupIdx[i]);
		GDL_FREE (d->arrayByGroupIdx);
		GDL_FREE (d->narrayByGroup);
		GDL_FREE (d->groupId);
		gdl_string_free (d->name);
		GDL_FREE (d);
	}	
}

gdl_tiling_data_conf *
gdl_tiling_data_conf_fscanf (FILE * stream)
{
	if (stream)
	{	
		size_t i,j,k,n,s;
		gdl_string * line, * tok;
		gdl_tiling_data_conf * c;
		
		c = GDL_CALLOC (gdl_tiling_data_conf, 1);
		
		line=0;
		while(gdl_getline (&line, &n, stream)!=-1)
		{
			for(s=0;s<n && isspace(line[s]); s++);
			if (line[s]!='#')
			{
				if (!strncmp(line+s,"narray",6))
				{
					for(;s<n && line[s]!='='; s++);
					for(s++;s<n && isspace(line[s]); s++);
					c->narray  = (size_t)atoi(line+s);
					c->groupId = GDL_CALLOC (size_t, c->narray);
				}
				else if (!strncmp(line+s,"groupId",7))
				{
					for(;s<n && line[s]!='='; s++);
					for(s++;s<n && isspace(line[s]); s++);
					for(i=j=k=0;i<c->narray;i++)
					{
						tok=gdl_string_next_token (line+s, n, &j, &k);
						c->groupId[i]=(size_t)atoi(tok);
						c->groupId[i]--;
						gdl_string_free (tok);
					}
				}
				else if (!strncmp(line+s,"type",4))
				{
					for(;s<n && line[s]!='='; s++);
					for(s++;s<n && isspace(line[s]); s++);
					c->type = gdl_string_clone (line+s);
				}
				else if (!strncmp(line+s,"name",4))
				{
					for(;s<n && line[s]!='='; s++);
					for(s++;s<n && isspace(line[s]); s++);
					c->name = gdl_string_clone (line+s);
				}
			}
			gdl_string_free (line);line=0;
		}
		
		return c;
	}
	return 0;
}

size_t
gdl_tiling_data_get_dim (const gdl_tiling_data * d)
{
	if (d->T == gdl_tiling_data_full)
		return d->narray;
	else if (d->T == gdl_tiling_data_mean)
		return d->ngroup;
   else if (d->T == gdl_tiling_data_median)
		return d->ngroup;
	return 0;
}

static int
gdl_tiling_data_fscanf_line (const gdl_tiling_data * d, gdl_tiling_chromosome * c, const size_t p, const gdl_string * line, const size_t n)
{
	size_t i,j,k,g;
	gdl_string * tok;
	
	i=j=0;
	tok = gdl_string_next_token (line, n, &i, &j); // chromosome name
	gdl_string_free (tok);
	tok = gdl_string_next_token (line, n, &i, &j); // probe position
	c->positions[p]=atol(tok);
	gdl_string_free (tok);
	if (d->T == gdl_tiling_data_full)
	{
		
		for(k=0;k<d->narray;k++)
		{
			tok = gdl_string_next_token (line, n, &i, &j);
			GDL_TILING_CHROM_SET(c,p,k,(double)atof(tok));
			gdl_string_free (tok);
		}
		return GDL_SUCCESS;
	}
	else if (  d->T == gdl_tiling_data_mean
	        || d->T == gdl_tiling_data_median)
	{
		double * tmp = GDL_MALLOC (double, d->narray);
		for(k=0;k<d->narray;k++)
		{
			tok = gdl_string_next_token (line, n, &i, &j);
			tmp[k]=(double)atof(tok);
			gdl_string_free (tok);
		}
		for(g=0;g<d->ngroup;g++)
		{
			double x=0;
			if (d->T == gdl_tiling_data_mean)
			{
				for(k=0;k<d->narrayByGroup[g];k++)
				{
					x+=tmp[d->arrayByGroupIdx[g][k]];
				}
				x /= d->narrayByGroup[g];
			}
			else if (d->T == gdl_tiling_data_median)
			{
				if (d->narrayByGroup[g]==1)
					x=tmp[d->arrayByGroupIdx[g][0]];
				else if (d->narrayByGroup[g]==2)
					x=0.5*(tmp[d->arrayByGroupIdx[g][0]]+tmp[d->arrayByGroupIdx[g][1]]);
				else
				{
					const size_t n = d->narrayByGroup[g];
					double * y = GDL_MALLOC (double, n);
  					for(k=0;k<n;k++)
  						y[k]=tmp[d->arrayByGroupIdx[g][k]];
  					gdl_sort (y, 1, n);	
  					x=gdl_stats_median_from_sorted_data (y, 1, n);
  					GDL_FREE (y);
				}
			}
			GDL_TILING_CHROM_SET(c,p,g,x);
		}
		GDL_FREE (tmp);
		return GDL_SUCCESS;
	}
	else
		return GDL_FAILURE;
}

int
gdl_tiling_data_fscanf (FILE * stream, gdl_tiling_data * d)
{
	if (stream && d)
	{
		int status;
		size_t i,j,k,n,s,p,nprobe,narray;
		long offset;
		gdl_string * line, * tok, * cur;
		gdl_list * chrom_list;
		gdl_tiling_chromosome * chrom;
		
		chrom_list = gdl_list_alloc (gdl_list_default);
		// get the number of arrays (this depends on the type
		// of the data set)
		narray = gdl_tiling_data_get_dim (d);
		// ignore header
		line=0;gdl_getline (&line, &n, stream);
		offset=ftell (stream);
		status=(gdl_getline (&line, &n, stream)!=-1) ? GDL_CONTINUE : GDL_FAILURE;
		do
		{
			i=j=0;cur=gdl_string_next_token (line, n, &i, &j);
			nprobe=1;
			
			while(gdl_getline (&line, &n, stream)!=-1)
			{
				i=j=0;
				for(i=0;i<n && !isspace(line[i]);i++);
				if (i != strlen(cur) || strncmp(cur,line,i)) break;
				nprobe++;
			}
			//fprintf(stdout, "Chromosome [ %s ] contains [ %d ] tiled probes\n", cur, nprobe);
			
			chrom = gdl_tiling_chromosome_alloc (cur, nprobe, narray);
			// push the pointer into the buffer
			gdl_list_push_back (chrom_list, chrom, 0);
			 
			gdl_string_free(cur);
			
			// come back at the first line for this chromosome
			fseek (stream, offset, SEEK_SET);
			
			for(p=0;p<nprobe && status==GDL_CONTINUE;p++)
			{
				gdl_getline (&line, &n, stream);
				gdl_tiling_data_fscanf_line (d, chrom, p, line, n);
			}
			
			offset=ftell (stream);
			
			status=(gdl_getline (&line, &n, stream)!=-1) ? GDL_CONTINUE : GDL_FAILURE;
		
		} while(status==GDL_CONTINUE);
		
		size_t nchrom = gdl_list_size (chrom_list);
		if (nchrom)
		{
			gdl_list_itr * itr = gdl_list_iterator_front (chrom_list);
			i=0;
			d->nprobe = 0;
			d->nchrom = nchrom;
			d->chroms = GDL_MALLOC (gdl_tiling_chromosome *, nchrom);
			do
			{
				d->chroms[i]=gdl_list_iterator_value(itr);
				d->nprobe+=d->chroms[i]->nprobe;
				i++;
			}
			while(gdl_list_iterator_next(itr));
			gdl_list_iterator_free (itr);
			
			status = GDL_SUCCESS;
		}
		else
		{
			status = GDL_FAILURE;	
		}
		gdl_list_free (chrom_list);
		
		return status;
	}
	return GDL_EINVAL;
}

int
gdl_tiling_data_fprintf (FILE * stream, const gdl_tiling_data * d)
{
	if (stream && d)
	{
		size_t a,c,p,narray=gdl_tiling_data_get_dim (d);
		
		fprintf (stream, "chromosome position");
		for(a=0;a<narray;a++) fprintf (stream, " A%d", a+1);
		fprintf (stream,"\n");
		for(c=0;c<d->nchrom;c++)
		{
			gdl_tiling_chromosome * chrom = d->chroms[c];
			for(p=0;p<chrom->nprobe;p++)
			{
				fprintf (stream, "%s %d", chrom->name, chrom->positions[p]);
				for(a=0;a<narray;a++) fprintf (stream, " %e", GDL_TILING_CHROM_GET(chrom,p,a));
				fprintf (stream,"\n");
			}	
		}
		
		return GDL_SUCCESS;
	}
	return GDL_EINVAL;	
}

gdl_tiling_data *
gdl_tiling_data_fread (FILE * stream)
{
	if (stream)
	{
		int status;
		size_t i, nchrom;
	   const gdl_tiling_data_type * T;
	   gdl_string * type, * name;
	   size_t narray, * groupId;
	   gdl_tiling_data_conf conf; 
		gdl_tiling_data * d;
		
		status = fread (&narray, sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		groupId = GDL_MALLOC (size_t, narray);
		status = fread (groupId, sizeof(size_t), narray, stream);
		GDL_FREAD_STATUS (status, narray);
		type = gdl_string_fread (stream);
		GDL_FREAD_STATUS (type!=0, 1);
		name = gdl_string_fread (stream);
		GDL_FREAD_STATUS (name!=0, 1);	
		
		conf.type    = type;
		conf.name    = name;
		conf.groupId = groupId;
		conf.narray  = narray;
		
		d = gdl_tiling_data_alloc (&conf);
		
		status = fread (&nchrom, sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		
		if (nchrom)
		{
			d->nprobe=0;
			d->nchrom=nchrom;
			d->chroms=GDL_MALLOC (gdl_tiling_chromosome *, nchrom);
			for(i=0;i<nchrom;i++)
			{
				d->chroms[i]=gdl_tiling_chromosome_fread(stream);
				d->chroms[i]->idx=i;
				d->nprobe += d->chroms[i]->nprobe;
				GDL_FREAD_STATUS (d->chroms[i]!=0, 1);
			}
		}
		
		GDL_FREE (groupId);
		gdl_string_free (name);
		gdl_string_free (type);
		
		return d;
	}
	return 0;	
}

int
gdl_tiling_data_fwrite (FILE * stream, const gdl_tiling_data * d)
{
	if (stream && d)
	{
		int status;
		size_t i;
	   
		status = fwrite (&d->narray, sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (d->groupId, sizeof(size_t), d->narray, stream);
		GDL_FWRITE_STATUS (status, d->narray);
		status = gdl_string_fwrite (stream, d->T->name);
		GDL_FWRITE_STATUS (status, GDL_SUCCESS);
		status = gdl_string_fwrite (stream, d->name);
		GDL_FWRITE_STATUS (status, GDL_SUCCESS);
		status = fwrite (&d->nchrom, sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		for(i=0;i<d->nchrom;i++)
		{
			status=gdl_tiling_chromosome_fwrite(stream, d->chroms[i]);
			GDL_FWRITE_STATUS (status, GDL_SUCCESS);
		}
		return GDL_SUCCESS;
	}
	return GDL_EINVAL;	
}

long
gdl_tiling_data_get_pos (const gdl_tiling_data * d, const size_t c, const size_t i)
{
	return d->chroms[c]->positions[i];	
}

double
gdl_tiling_data_get_val (const gdl_tiling_data * d, const size_t c, const size_t i, const size_t g, const size_t a)
{
	return GDL_TILING_CHROM_GET (d->chroms[c], i, d->arrayByGroupIdx[g][a]);
}

gdl_boolean
gdl_tiling_data_has_missing (const gdl_tiling_data * d, const size_t c, const size_t i)
{
	size_t a;
	for(a = 0; a < d->chroms[c]->narray; a++)
		if (gdl_isinf (GDL_TILING_CHROM_GET (d->chroms[c], i, a)))
			break;
	return (a == d->chroms[c]->narray) ? gdl_false : gdl_true;
}

int
gdl_tiling_data_chrom_lookup (const gdl_tiling_data * d, const gdl_string * name)
{
	size_t c;
	for(c = 0; c < d->nchrom; c++)
	{
		if (!strcmp(d->chroms[c]->name, name))
			return c;	
	}
	return -1;
}

double
gdl_tiling_data_get_mean_val (const gdl_tiling_data * d, const size_t c, const size_t i, const size_t g)
{
	
}

void
gdl_tiling_data_force_tiling (gdl_tiling_data * d, const long tiling)
{
	size_t c,p;
	long prev,offset,til,q,r;
	
	for(c = 0; c < d->nchrom; c++)
	{
		gdl_tiling_chromosome * chrom = d->chroms[c];
		prev = chrom->positions[0];
		for (p = 1; p < chrom->nprobe; p++)
		{
			til = chrom->positions[p]-prev;
			if (til < 0) continue;
			if (til%tiling!=0)
			{
				q=til/tiling;
				r=til-q*tiling;
				if (r > tiling/2)
					chrom->positions[p] = prev+(q+1)*tiling;
				else
					chrom->positions[p] = prev+q*tiling;
			}
			prev = chrom->positions[p];
		}
	}
}

double
gdl_tiling_data_sigma_shrinkage_estimator (const gdl_tiling_data * data, double * Bs)
{
	size_t c,i,j,k,v,I=0,nmiss=0;
	double x_ijk, x_ij, s_ij, s_i, s=0, S=0, B, mu, var;
	
	v=0;
	for(j=0;j<data->ngroup;j++)
	{
		v+=(data->narrayByGroup[j]-1);
	}
	for(c = 0; c < data->nchrom; c++)
	 {
	 	 for(i = 0 ; i < data->chroms[c]->nprobe; i++)
	 	 {
	 	 	 if (gdl_tiling_data_has_missing (data, c, i))
	 	 	 {
	 	 	 	nmiss++;
	 	 	 	continue;
	 	 	 }
	 	 	 s_i = 0;
	 	 	 for(j = 0; j < data->ngroup; j++)
	 	 	 {
	 	 	 	const size_t K_j = data->narrayByGroup[j];
	 	 	 	x_ij=s_ij=0;
	 	 	 	for(k = 0; k < K_j; k++)
	 	 	 	{ 
	 	 	 		x_ijk = gdl_tiling_data_get_val (data, c, i, j, k);
	 	 	 		x_ij += x_ijk;
	 	 	 		s_ij += x_ijk*x_ijk;
	 	 	 	}
	 	 	 	x_ij /= K_j;
		 	 	s_ij -= K_j*x_ij*x_ij;
		 	 	s_i  += s_ij;
	 	 	 }
	 	 	 s_i /= v;
	 	 	 s   += s_i;
	 	 	 S   += s_i*s_i;
	 	 }
	 	 I+=data->chroms[c]->nprobe;
	 }
	 s /= (I-nmiss);
	 S -= (I-nmiss)*s*s;
	 // shrinkage factor
	 *Bs = (2.0/v)*(I-1)/((1+2.0/v)*I)+2*s*s*(I-1)/((1+2.0/v)*v*S);
	 
	 return s;
}

gdl_tiling_ttest *
gdl_tiling_ttest_alloc (const gdl_tiling_data * data)
{
	size_t c,i,j,k,v,I=0;
	double x_ijk, x_ij, s_ij, s_i, s=0, S=0, B;
	gdl_tiling_ttest * t;
	
	t = GDL_CALLOC (gdl_tiling_ttest, 1);
	t->ngroup = data->ngroup;
	t->nchrom = data->nchrom;
	t->nprobe = data->nprobe;
	t->K      = GDL_MALLOC (size_t, t->ngroup);
	for(j=0;j<t->ngroup;j++)
	{
		t->K[j]=data->narrayByGroup[j];
		t->nu +=(t->K[j]-1);
	}
	t->mu     = GDL_CALLOC (double, t->ngroup*t->nprobe);
	t->var    = GDL_CALLOC (double, t->nprobe);
	
	v = t->nu;
	for(c = 0; c < data->nchrom; c++)
	 {
	 	 for(i = 0 ; i < data->chroms[c]->nprobe; i++)
	 	 {
	 	 	 if (gdl_tiling_data_has_missing (data, c, i))
	 	 	 {
	 	 	 	t->nmiss++;
	 	 	 	continue;
	 	 	 }
	 	 	 s_i = 0;
	 	 	 for(j = 0; j < data->ngroup; j++)
	 	 	 {
	 	 	 	const size_t K_j = data->narrayByGroup[j];
	 	 	 	x_ij=s_ij=0;
	 	 	 	for(k = 0; k < K_j; k++)
	 	 	 	{ 
	 	 	 		x_ijk = gdl_tiling_data_get_val (data, c, i, j, k);
	 	 	 		x_ij += x_ijk;
	 	 	 		s_ij += x_ijk*x_ijk;
	 	 	 	}
	 	 	 	x_ij /= K_j;
		 	 	s_ij -= K_j*x_ij*x_ij;
		 	 	s_i  += s_ij;
		 	 	t->mu[(I+i)*t->ngroup+j]=x_ij;
	 	 	 }
	 	 	 s_i /= v;
	 	 	 s   += s_i;
	 	 	 S   += s_i*s_i;
	 	 	 t->var[I+i]=s_i;
	 	 }
	 	 I+=data->chroms[c]->nprobe;
	 }
	 s /= (I-t->nmiss);
	 S -= (I-t->nmiss)*s*s;
	 // shrinkage factor
	 t->shrink = (2.0/v)*(I-1)/((1+2.0/v)*I)+2*s*s*(I-1)/((1+2.0/v)*v*S);
	 t->tvar   = s;
	 for(j=0;j<t->ngroup;j++)
	 {
	 	t->factor += 1.0/t->K[j];
	 }
	 t->factor = sqrt(t->factor);
	 t->data   = data;
	 
	 return t;
}

void
gdl_tiling_ttest_free (gdl_tiling_ttest * t)
{
	if (t)
	{
		GDL_FREE (t->K);
		GDL_FREE (t->mu);
		GDL_FREE (t->var);
		GDL_FREE (t);
	}
}

double
gdl_tiling_ttest_get (const gdl_tiling_ttest * t, const size_t c, const size_t p)
{
	if (gdl_tiling_data_has_missing (t->data, c, p))
		return 0;
	size_t I = 0;
	size_t cc;
	double t_i, s_i;
	
	for(cc = 0; cc < c && cc < t->data->nchrom; cc++)
		I+=t->data->chroms[c]->nprobe;
	
	t_i  = t->mu[(I+p)*t->ngroup];
 	t_i -= t->mu[(I+p)*t->ngroup+1];
 	s_i  = sqrt((1-t->shrink)*t->var[I+p]+t->shrink*t->tvar);
 	s_i *= t->factor;
 	
 	return t_i/s_i;
}

int
gdl_tiling_ttest_fprintf (FILE * stream, const gdl_tiling_data * data, const gdl_tiling_ttest * t)
{
	if (stream && t)
	{
		size_t c,i,j,k,I=0;
		double t_i, s_i, factor=0;
		
		for(j=0;j<t->ngroup;j++) factor += 1.0/t->K[j];
		factor = sqrt(factor);
		
		fprintf (stream, "chromosome\tposition\tscore\n");
		for(c = 0; c < data->nchrom; c++)
		 {
		 	 for(i = 0 ; i < data->chroms[c]->nprobe; i++)
		 	 {
		 	 	 if (gdl_tiling_data_has_missing (data, c, i))
		 	 	 	continue;
		 	 	 t_i  = t->mu[(I+i)*t->ngroup];
		 	 	 t_i -= t->mu[(I+i)*t->ngroup+1];
		 	 	 s_i  = sqrt((1-t->shrink)*t->var[I+i]+t->shrink*t->tvar);
		 	 	 s_i *= factor;
		 	 	 fprintf (stream, "%s\t%d\t%e\n", data->chroms[c]->name, data->chroms[c]->positions[i], t_i/s_i);
		 	 }
		 	 I+=data->chroms[c]->nprobe;
		 }
		
		return GDL_SUCCESS;
	}
	return GDL_EINVAL;
}
	
static const gdl_tiling_data_type _gdl_tiling_data_full = 
{
	"full"
};

const gdl_tiling_data_type * gdl_tiling_data_full = &_gdl_tiling_data_full;

static const gdl_tiling_data_type _gdl_tiling_data_mean = 
{
	"mean"
};

const gdl_tiling_data_type * gdl_tiling_data_mean = &_gdl_tiling_data_mean;

static const gdl_tiling_data_type _gdl_tiling_data_median = 
{
	"median"
};

const gdl_tiling_data_type * gdl_tiling_data_median = &_gdl_tiling_data_median;
