/*  
 *  tiling_array.h
 * 
 *  $Author: gael $, $Date: 2008-09-15 12:32:59 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
 
#ifndef __GDL_TILING_ARRAY_H__
#define __GDL_TILING_ARRAY_H__

#include <math.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_randist.h>
#include <gdl/gdl_string.h>

__BEGIN_DECLS

#define GDL_TILING_CHROM_GET(c,p,a)((c)->intensities[(p)*((c)->narray)+(a)])
#define GDL_TILING_CHROM_SET(c,p,a,x)((c)->intensities[(p)*((c)->narray)+(a)]=(x))

typedef struct
{
	size_t idx;
	gdl_string * name;
	size_t nprobe;
	size_t narray;
	long * positions;
	double * intensities;
} gdl_tiling_chromosome;

gdl_tiling_chromosome * gdl_tiling_chromosome_alloc (const gdl_string * name, const size_t nprobe, const size_t narray);
void gdl_tiling_chromosome_free (gdl_tiling_chromosome * c);
gdl_tiling_chromosome * gdl_tiling_chromosome_fread (FILE * stream);
int gdl_tiling_chromosome_fwrite (FILE * stream, const gdl_tiling_chromosome * c);

typedef struct
{
	gdl_string * name;
} gdl_tiling_data_type;

typedef struct
{
	size_t narray;
	size_t * groupId;
	gdl_string * type;
	gdl_string * name;
} gdl_tiling_data_conf;

gdl_tiling_data_conf * gdl_tiling_data_conf_fscanf (FILE * stream);

typedef struct
{
	const gdl_tiling_data_type * T;
	gdl_string * name;        // the name of the data set
	size_t nchrom;            // the number of chromosomes
	size_t narray;            // the number of tiling arrays
	size_t ngroup;            // the number of groups
	size_t nprobe;            // the total number of probes
	size_t * narrayByGroup;   // the number of arrays per group
	size_t ** arrayByGroupIdx; // the index of the array by group in the data set
	size_t * groupId;         // the group labels for each array
	gdl_tiling_chromosome ** chroms; // the data by chromosomes
} gdl_tiling_data;

gdl_tiling_data * gdl_tiling_data_alloc (const gdl_tiling_data_conf * conf);
gdl_tiling_data * gdl_tiling_data_alloc2 (const size_t ncolumn);
void gdl_tiling_data_free (gdl_tiling_data * d);
int gdl_tiling_data_fscanf (FILE * stream, gdl_tiling_data * d);
int gdl_tiling_data_fprintf (FILE * stream, const gdl_tiling_data * d);
gdl_tiling_data * gdl_tiling_data_fread (FILE * stream);
int gdl_tiling_data_fwrite (FILE * stream, const gdl_tiling_data * d);

long gdl_tiling_data_get_pos (const gdl_tiling_data * d, const size_t c, const size_t i);
double gdl_tiling_data_get_val (const gdl_tiling_data * d, const size_t c, const size_t i, const size_t g, const size_t a);
double gdl_tiling_data_get_mean_val (const gdl_tiling_data * d, const size_t c, const size_t i, const size_t g);
size_t gdl_tiling_data_get_dim (const gdl_tiling_data * d);
gdl_boolean gdl_tiling_data_has_missing (const gdl_tiling_data * d, const size_t c, const size_t i);
void gdl_tiling_data_force_tiling (gdl_tiling_data * d, const long tiling);
int gdl_tiling_data_chrom_lookup (const gdl_tiling_data * d, const gdl_string * name); 
double gdl_tiling_data_sigma_shrinkage_estimator (const gdl_tiling_data * data, double * Bs);


typedef struct
{
	size_t ngroup;  // number of conditions
	size_t nchrom;  // number of chromosomes
	size_t nprobe;  // number of probes
	size_t nmiss;   // number of probes with missing info 
	size_t nu;      // total number of df
	size_t * K;     // number of replicates per condition
	double * mu;    // probe means 
	double * var;   // probe variances
	double shrink;	 // empirical Bayes shrinkage estimator 
	double tvar;
	double factor;
	const gdl_tiling_data * data;
} gdl_tiling_ttest;

gdl_tiling_ttest * gdl_tiling_ttest_alloc (const gdl_tiling_data * d);
void gdl_tiling_ttest_free (gdl_tiling_ttest * t);
int gdl_tiling_ttest_fprintf (FILE * stream, const gdl_tiling_data * data, const gdl_tiling_ttest * t);

GDL_VAR const gdl_tiling_data_type * gdl_tiling_data_full;
GDL_VAR const gdl_tiling_data_type * gdl_tiling_data_mean;
GDL_VAR const gdl_tiling_data_type * gdl_tiling_data_median;

__END_DECLS

#endif
