/*  
 *  tiling_feature.c
 * 
 *  $Author: gael $, $Date: 2008-09-15 12:32:58 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
 
#include <math.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_machine.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_string.h>
#include <gdl/gdl_hash.h>

#include "tiling_feature.h"

gdl_tiling_gene_expcmp *
gdl_tiling_gene_expcmp_alloc (const gdl_string * file)
{
	FILE * stream;
	size_t ti,tj,n;
	gdl_string * line=0, * tok;
	gdl_hashtable * gdb = 0;
	
	stream = gdl_fileopen (file, "r");
	
	gdb = gdl_hashtable_alloc (gdl_hash_default, 0);
	// skip header
	gdl_getline (&line, &n, stream);
	gdl_string_free (line);line=0;
	while(gdl_getline (&line, &n, stream)!=-1)
	{
		ti = tj = 0;
		gdl_tiling_gene_expcmp_entry * ge = GDL_CALLOC (gdl_tiling_gene_expcmp_entry, 1);
		ge->gene_id = gdl_string_next_token (line, n, &ti, &tj);
		tok = gdl_string_next_token (line, n, &ti, &tj);
		if (!strcmp(tok, "NA"))
			ge->fold = GDL_NAN;
		else
			ge->fold = (double)atof(tok);
		gdl_string_free (tok);
		tok   = gdl_string_next_token (line, n, &ti, &tj);
		if (!strcmp(tok, "NA"))
			ge->statistic = GDL_NAN;
		else
			ge->statistic = (double)atof(tok);
		gdl_string_free (tok);
		tok   = gdl_string_next_token (line, n, &ti, &tj);
		if (!strcmp(tok, "NA"))
			ge->pvalue = GDL_NAN;
		else
			ge->pvalue = (double)atof(tok);
		gdl_string_free (tok);
		tok   = gdl_string_next_token (line, n, &ti, &tj);
		if (!strcmp(tok, "NA"))
			ge->is_significant = 0;
		else
			ge->is_significant = (tok[0]=='T' || tok[0]=='1') ? 1 : 0;
		gdl_string_free (tok);
		if (tj != n)
		{
			ge->comment = gdl_string_next_token (line, n, &ti, &tj);
		}
		gdl_hashtable_add (gdb, ge->gene_id, ge, 0);
		gdl_string_free (line);line=0;
	}
	gdl_fileclose (file, stream);
	
	return gdb;	
}

void
gdl_tiling_gene_expcmp_free (gdl_tiling_gene_expcmp * gdb)
{
	if (gdb)
	{
		if (gdl_hashtable_size (gdb))
		{
			gdl_hashtable_itr * itr = gdl_hashtable_iterator (gdb);
			do
			{
				gdl_tiling_gene_expcmp_entry * ge = gdl_hashtable_iterator_value (itr);
				GDL_FREE (ge->gene_id);
				GDL_FREE (ge->comment);
				GDL_FREE (ge);
			}
			while(gdl_hashtable_iterator_next (itr));
			gdl_hashtable_iterator_free (itr);
		}
		gdl_hashtable_free (gdb); 
	}
}

gdl_tiling_gene_expcmp_entry *
gdl_tiling_gene_expcmp_lookup (gdl_tiling_gene_expcmp * gdb, const gdl_string * gene_id)
{
	return gdl_hashtable_lookup (gdb, gene_id);
}
