/*  
 *  tiling_hmm.c
 * 
 *  $Author: gael $, $Date: 2008-09-15 12:32:59 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
 
#include <math.h>

#include <gdl/gdl_machine.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_statistics_double.h>

#include "tiling_hmm.h"

gdl_tiling_hmm_seq *
gdl_tiling_hmm_seq_alloc1 (const gdl_tiling_chromosome * data, const size_t from, const size_t to)
{
	size_t i;
	gdl_tiling_hmm_seq * o;
	
	o = GDL_MALLOC (gdl_tiling_hmm_seq, 1);
	
	o->data = data;
	o->size = to-from+1;
	o->nobs = o->size;
	o->t    = GDL_MALLOC (int, o->size);
	for(i = 0; i < o->size; i++) o->t[i]=from+i;
	
	return o;	
}

gdl_tiling_hmm_seq *
gdl_tiling_hmm_seq_alloc2 (const gdl_tiling_chromosome * data, const size_t from, const long physical_step, const size_t probe_size)
{
	size_t i,j;
	long pos;
	gdl_tiling_hmm_seq * o;
	
	o = GDL_CALLOC (gdl_tiling_hmm_seq, 1);
	
	o->data = data;
	o->size = probe_size;
	o->t    = GDL_CALLOC (int, o->size);
	o->t[0] = from;
	j   = from+1;
	pos = data->positions[from];
	o->nobs = 1;
	for(i=1;i<o->size;i++)
	{
		pos += physical_step;
		if ((j < o->data->nprobe) && (data->positions[j] >  pos)) o->t[i]=-1;
		else if ((j < o->data->nprobe) && (data->positions[j] <  pos))
		{
			GDL_FREE (o->t);
			GDL_FREE (o);
			return 0;
		} 
		else if (j < o->data->nprobe) o->t[i]=j++;
		else o->t[i] = -1;
		if (o->t[i] != -1) (o->nobs)++;
	}
	
	return o;	
}

gdl_tiling_hmm_seq *
gdl_tiling_hmm_seq_alloc3 (const gdl_tiling_chromosome * data, const long from, const long tiling_size, const size_t window_size, size_t * p_cur)
{
	const size_t P = data->nprobe;
	size_t i,j,k,p_from;
	long pos;
	gdl_tiling_hmm_seq * o;
	
	for(i = *p_cur; i < P && data->positions[i] < from; i++);
	if (i == P || (data->positions[P-1]-from+1) < (window_size-1)*tiling_size)
		return 0;

	o       = GDL_CALLOC (gdl_tiling_hmm_seq, 1);
	o->data = data;
	o->size = window_size;
	o->t    = GDL_MALLOC (int, o->size);
	for(k = 0; k < o->size; k++) o->t[k] = -1;
	
	if (data->positions[i] == from)
	{
		o->t[0] = i;
		o->nobs = 1;
		j = i+1;
		(*p_cur)++;
	}
	else
		j = i;
	
	pos = from;
	
	for(i = 1; i < o->size && j < P; i++)
	{
		// in case where there is two probes located 
		// at the same position
		if (data->positions[j] == pos) {i--;j++;continue;}
		pos += tiling_size;
		if ((j < P) && (data->positions[j] >  pos)) o->t[i]=-1;
		else if ((j < P) && (data->positions[j] <  pos))
		{
			// Unexpected error
			GDL_FREE (o->t);
			GDL_FREE (o);
			return 0;
		} 
		else if (j < P)
		{
			o->t[i] = j++;
		   (o->nobs)++;
		}
	}
	
	return o;	
}

gdl_boolean
gdl_tiling_hmm_seq_missing (const gdl_tiling_hmm_seq * o, const size_t t)
{
	if (o)
		return (o->t[t]==-1) ? gdl_true : gdl_false;
	else
		return gdl_true;
}

size_t
gdl_tiling_hmm_seq_size (const gdl_tiling_hmm_seq * o)
{
	if (o)
		return o->size;
	else
		return 0;
}

size_t
gdl_tiling_hmm_seq_dim (const gdl_tiling_hmm_seq * o)
{
	if (o)
		return o->data->narray;
	else
		return 0;
}

double
gdl_tiling_hmm_seq_get (const gdl_tiling_hmm_seq * o, const size_t t, const size_t d)
{
	if (o)
		return (o->t[t]!=-1) ? GDL_TILING_CHROM_GET (o->data, o->t[t], d) : 0;
	else
		return 0;
}

double *
gdl_tiling_hmm_seq_vector (const gdl_tiling_hmm_seq * o, const size_t d)
{
	if (o)
	{
		size_t i,j;
		double * x;
		x = GDL_MALLOC (double, o->nobs);
		for(j = i = 0; i < o->nobs && j < o->size; j++)
		{
			if (o->t[j]==-1)
				continue;
			x[i++] = GDL_TILING_CHROM_GET (o->data, o->t[j], d);
			if (gdl_isnan (x[i-1])) printf (">>ERROR %d %d %d %g\n", i-1, j, o->t[j], GDL_TILING_CHROM_GET (o->data, o->t[j], d));
		}		
		return x;
	}
	return 0;	
}

void
gdl_tiling_hmm_seq_free (gdl_tiling_hmm_seq * o)
{
	if (o)
	{
		GDL_FREE (o->t);
		GDL_FREE (o);	
	}
}

static int
gdl_tiling_hmm_initforward (gdl_tiling_hmm * hmm, const gdl_tiling_hmm_seq * seq)
{
	// hmm parameters
	const size_t K = hmm->nstate;
	double * scale = hmm->scale;
	double ** alpha = hmm->alpha;
	double *log_p  = &hmm->log_p;
	//
	size_t k;
	double b_0, c_0;
	// initialization
	scale[0]=0;
	for(k=0;k<K;k++)
	{
		b_0=(*hmm->states[k]->out_b)(seq, 0, hmm, hmm->states[k]->gauss_class);
		alpha[0][k]=hmm->states[k]->pi*b_0;
		scale[0]+=alpha[0][k];
	}
	if (scale[0] >= GDL_SQRT_DBL_EPSILON)
	{
      c_0 = 1 / scale[0];
      for (k=0;k<K;k++) alpha[0][k] *= c_0;
      *log_p = -log (c_0);
      return GDL_SUCCESS;
   }
   else
   {
   	// first obseration cannot be generated by the HMM	
      *log_p = +1;
      return GDL_EUNDRFLW;
   }
}

static double
gdl_tiling_hmm_stepforward (gdl_tiling_hmm_state * state, double * alpha_t, const double b_t)
{
	size_t n,id;
	double s_t=0;
	
	if (b_t < GDL_SQRT_DBL_EPSILON)
    return 0.;
	
	for(s_t=0,n=0;n<state->in_states;n++)
	{
		id   = state->in_id[n];
		s_t += alpha_t[id]*state->in_a[n];
	}
	
	s_t*=b_t;
	
	return s_t;
}

int
gdl_tiling_hmm_forward (gdl_tiling_hmm * hmm, const gdl_tiling_hmm_seq * seq)
{
	// model global parameters
	const size_t T = gdl_tiling_hmm_seq_size(seq);
	const size_t N =  hmm->nstate;
	double * scale = hmm->scale;
	double ** alpha = hmm->alpha;
	double *log_p   = &hmm->log_p;
	// 
	size_t t,n,l,id;
	double b_t,s_t,c_t;
	
	// initialization
	if (gdl_tiling_hmm_initforward (hmm, seq)!=GDL_SUCCESS)
	{
		*log_p==+1;
		return GDL_EUNDRFLW;
	}
	// induction
	for(t=1;t<T;t++)
	{
		scale[t]=0;
		for(n=0;n<N;n++)
		{
			b_t=(*hmm->states[n]->out_b)(seq, t, hmm, hmm->states[n]->gauss_class);
			alpha[t][n]=gdl_tiling_hmm_stepforward (hmm->states[n], alpha[t-1], b_t);
			scale[t]+=alpha[t][n];
		}
		if (scale[t] < GDL_SQRT_DBL_EPSILON)
		{
        /* Obs. can't be generated by hmm */
        *log_p = +1.0;
        break;
      }
      c_t = 1/scale[t];
      for(n=0;n<N;n++) alpha[t][n]*=c_t;
      *log_p -= log (c_t);
	}
	
	return (*log_p==+1) ? GDL_EUNDRFLW : GDL_SUCCESS;
}

int
gdl_tiling_hmm_backward (gdl_tiling_hmm * hmm, const gdl_tiling_hmm_seq * seq)
{
	// model global parameters
	const size_t T = gdl_tiling_hmm_seq_size(seq);
	const size_t N =  hmm->nstate;
	double * scale = hmm->scale;
	double ** beta  = hmm->beta;
	// 
	int t;
	size_t n,l,id;
	double b_t,s_t,c_t;
	
	// initialization
	for(n=0;n<N;n++) beta[T-1][n]=1.0;
	// induction
	for(t = T-2; t >= 0; t--)
	{
		for (n = 0; n < N; n++)
	   {
	   	gdl_tiling_hmm_state * state = hmm->states[n];
	   	s_t = 0.0;
	   	for (l = 0; l < state->out_states; l++)
	   	{  	
	          id   = state->out_id[l];
        	    b_t  = (*hmm->states[n]->out_b)(seq, t+1, hmm, hmm->states[id]->gauss_class);
             s_t += state->out_a[l] * b_t * beta[t+1][id];
          }
	       /* updating beta[t] */
	       beta[t][n] = s_t / scale[t+1];
      }
	}
	
	return GDL_SUCCESS;
}

typedef struct
{
	/* Pi */
	double * pi_num;
	double pi_denom;
	/* B */
	double ** b_num_x;
	double ** b_num_xx;
	double ** b_denom;
	/* A */
	double * a_denom;
	double ** a_num;
} gdl_tiling_hmm_store;

static gdl_tiling_hmm_store *
gdl_tiling_hmm_store_alloc (const gdl_tiling_hmm * hmm)
{
	const size_t N = hmm->nstate;
	const size_t D = hmm->ndim;
	size_t i;
	gdl_tiling_hmm_store * s;
	
	s = GDL_CALLOC (gdl_tiling_hmm_store, 1);
	
	/* Pi */
	s->pi_num   = GDL_CALLOC (double, N);
	/* B */
	s->b_denom  = GDL_MATRIX_ALLOC (double, N, D);
	s->b_num_x  = GDL_MATRIX_ALLOC (double, N, D);
	s->b_num_xx = GDL_MATRIX_ALLOC (double, N, D);
	/* A */
	s->a_denom  = GDL_CALLOC (double, N);
	s->a_num    = GDL_MALLOC (double *, N);
	for(i = 0; i < N; i++)
	{
		s->a_num[i] = GDL_CALLOC (double, hmm->states[i]->out_states);
	}
	
	return s;
}

static void
gdl_tiling_hmm_store_reset (gdl_tiling_hmm_store * s, const gdl_tiling_hmm * hmm)
{
	const size_t N = hmm->nstate;
	const size_t D = hmm->ndim;
	size_t i,j;
	
	for(i = 0; i < N; i++)
	{
		/* Pi */
		s->pi_denom =0;
		s->pi_num[i]=0;
		/* B */
		for(j = 0; j < D; j++)
		{
			s->b_denom[i][j] =0;
			s->b_num_x[i][j] =0;
			s->b_num_xx[i][j]=0;
		}
		/* A */
		s->a_denom[i]=0;
		for(j = 0; j < hmm->states[i]->out_states; j++)
		{
			s->a_num[i][j]=0;
		}
	}		
}

static void
gdl_tiling_hmm_store_free (gdl_tiling_hmm_store * s, const gdl_tiling_hmm * hmm)
{
	if (s)
	{
		const size_t N = hmm->nstate;
		const size_t D = hmm->ndim;
		size_t i;
		
		/* Pi */
		GDL_FREE (s->pi_num);
		/* B */
		GDL_MATRIX_FREE (s->b_denom, N);
		GDL_MATRIX_FREE (s->b_num_x, N);
		GDL_MATRIX_FREE (s->b_num_xx, N);
		/* A */
		GDL_FREE (s->a_denom);
		GDL_MATRIX_FREE (s->a_num, N);
		GDL_FREE (s);
	}	
}

static int
gdl_tiling_hmm_baum_welch_update (gdl_tiling_hmm * hmm, const gdl_tiling_hmm_store * store)
{
	const size_t C  = hmm->ngauss_class;
	const size_t D  = hmm->ndim;
   const size_t N  = hmm->nstate;
	int res = -1;
   int d, h, i, j, m, l, j_id, positive;
   double factor, p_i, xbar, * mu, * var, * obs, * switcher;
   int hist, col, size, reachable;
   gdl_tiling_hmm_state * state;
  
   if (hmm->update_b)
   {
	   mu  = GDL_CALLOC (double, C);
	   var = GDL_CALLOC (double, C);
	   obs = GDL_CALLOC (double, C);
   }
   
	for (i = 0; i < N; i++)
	{
		state = hmm->states[i];
      reachable = 1;
      /* Pi */
      if (hmm->update_a && state->has_pi)
      {
         state->pi = store->pi_num[i] / store->pi_denom;
         //printf ("STATE %d pi = %g\n", i, state->pi);
      }
      /* A */
      if (!state->out_fix && hmm->update_a)
      {
	      /* note: denom. might be 0; never reached state? */
	      p_i = 0.0;
	      if (store->a_denom[i] < GDL_SQRT_DBL_EPSILON)
	      {
	      	for (h = 0; h < state->in_states; h++)
	      	{
	         	p_i += state->in_a[h];
	      	}
		      if (p_i == 0)
		      {
		        // output some warnings...
		        reachable = 0;
		      }
		      factor = 0.0;
			}
		   else
		   {
		      factor = (1 / store->a_denom[i]);
		   }
			positive = 0;
			
		   for (j = 0; j < state->out_states; j++)
	      {
		      /* TEST: denom. < numerator */
		      if ((store->a_denom[i] - store->a_num[i][j]) <= -GDL_SQRT_DBL_EPSILON)
		      {
		      	// WARNING...
		      }
		      
		      state->out_a[j] = store->a_num[i][j] * factor;
		      
		      //printf ("STATE %d out_a[%d] = %g\n", i, j, state->out_a[j]);
		      
		      if (store->a_num[i][j] >= GDL_SQRT_DBL_EPSILON)
		      {
		         positive = 1;
		      }
		      /* important: also update in_a  */
		      l = 0;
		      j_id = state->out_id[j];
		      while (l < hmm->states[j_id]->in_states)
		      {
		        if (hmm->states[j_id]->in_id[l] == i)
		          break;
		        else
		          l++;
		      }
		      if (l == hmm->states[j_id]->in_states)
		      {
		        // ERROR !
		      }
	      	hmm->states[j_id]->in_a[l] = state->out_a[j];
	      }
    	}
    	
		/* if fix, continue to next state */
		if (state->b_fix || !hmm->update_b)
			continue;
		/* B */
 		for (d = 0; d < D; d++)
		{
		  /* If the denominator is very small, we have not seen many emissions
			  in this state with this history.
			  We are conservative and just skip them.
		   */
		   if (store->b_denom[i][d] < GDL_SQRT_DBL_EPSILON)
			  continue;
			
			//printf ("STORE %d %d %d %g %g\n", i, d, state->gauss_class[d], store->b_num_x[i][d], store->b_denom[i][d]);
	 		
			mu[state->gauss_class[d]]  += store->b_num_x[i][d];
			var[state->gauss_class[d]] += store->b_num_xx[i][d];
			obs[state->gauss_class[d]] += store->b_denom[i][d];
		 }
	}

	if (hmm->update_b)
	{
		double tot_var  = 0;
		size_t ntot_var = 0;
		for(i = 0; i < C; i++)
		{
			//printf ("OUT_B %d %d mu %g var %g\n", i, d, mu[i], var[i]);
	 		if (obs[i] < GDL_SQRT_DBL_EPSILON)
	 			continue;
	 		else 
	 			factor = (1 / obs[i]);
	 		mu[i]  *= factor;
	 		var[i] *= factor;
	 		var[i] -= mu[i]*mu[i];
	 		if (var[i] > 0.)
	 		{
	 			tot_var  += obs[i]*var[i];
	 			ntot_var += obs[i];
	 		}
		}
		tot_var /= ntot_var;
		for(i = 0; i < C; i++)
		{
			if (var[i] <= tot_var*GDL_SQRT_DBL_EPSILON)
			{
				var[i] = tot_var;
			}
		}
		switcher = hmm->mu;
		hmm->mu  = mu;
		mu = switcher;
		switcher = hmm->var;
		hmm->var = var;
		var = switcher;
		
		GDL_FREE (mu);
		GDL_FREE (var);
		GDL_FREE (obs);
	}
	
	return GDL_SUCCESS;
}

static int
gdl_tiling_hmm_baum_welch_one_step (gdl_tiling_hmm * hmm, const gdl_tiling_hmm_seq * seq, gdl_tiling_hmm_store * store)
{
  const size_t T  = gdl_tiling_hmm_seq_size (seq);
  const size_t D  = gdl_tiling_hmm_seq_dim (seq);
  const size_t N  = hmm->nstate;
  double * scale  = hmm->scale;
  double ** alpha = hmm->alpha;
  double ** beta  = hmm->beta;
  double * log_p  = &hmm->log_p;
  int res = -1;
  int k, i, j, t, j_id, valid, d;
  double otd,b_t,gamma;
  gdl_tiling_hmm_state * state;
  
  valid = 0;
  
  if (gdl_tiling_hmm_forward (hmm, seq) != GDL_SUCCESS)
  {
     return GDL_FAILURE;
  }
  if (*log_p != +1) /* Tiling sequence can be generated */
  {        
      valid = 1;
      if (gdl_tiling_hmm_backward (hmm, seq) != GDL_SUCCESS)
      {
        return GDL_FAILURE;
      }
      /* loop over all states */
      for (i = 0; i < N; i++)
      {
        state = hmm->states[i];
        /* Pi */
        if (state->has_pi)
        {
	        gamma = alpha[0][i] * beta[0][i];
	        store->pi_num[i] += gamma;
		     store->pi_denom  += gamma;
		  }
		  for (t = 0; t < T - 1; t++)
		  {
		  	 if (gdl_tiling_hmm_seq_missing (seq, t)) continue;
		  	 
		  	 gamma = alpha[t][i] * beta[t][i];
		  	 /* B */
		  	 if (!state->b_fix)
          {
          	 for(d = 0; d < D; d++)
	        	 {
	        	 	 otd = gdl_tiling_hmm_seq_get (seq, t, d);
		          store->b_num_x[i][d]  += gamma*otd;
		          store->b_num_xx[i][d] += gamma*otd*otd;
		          store->b_denom[i][d]  += gamma;
	        	 }   
          }
          /* A */
          if (!state->out_fix)
          {
	          store->a_denom[i] += gamma;
	          for (j = 0; j < state->out_states; j++)
	          {
	            j_id = state->out_id[j];
	            b_t  = (hmm->states[j_id]->out_b)(seq, t+1, hmm, hmm->states[j_id]->gauss_class);
	            store->a_num[i][j] += (alpha[t][i] * state->out_a[j] * b_t * beta[t + 1][j_id] * (1.0 / scale[t + 1]));
	          }
          }
        }
        t=T-1;
        /* B: last iteration for t==T-1 */
        if (!state->b_fix && !gdl_tiling_hmm_seq_missing (seq, t))
        {
	        gamma = alpha[t][i] * beta[t][i];
           for(d = 0; d < D; d++)
        	  {
        	 	 otd = gdl_tiling_hmm_seq_get (seq, t, d);
	          store->b_num_x[i][d]  += gamma*otd;
	          store->b_num_xx[i][d] += gamma*otd*otd;
	          store->b_denom[i][d]  += gamma;
        	  }
        }
      }
      
      gdl_tiling_hmm_baum_welch_update (hmm, store);
  } 
  else
  {
    GDL_WARNING ("Tiling sequence can't be built from HMM!", GDL_FAILURE);
  }

	return GDL_SUCCESS;
}

int
gdl_tiling_hmm_baum_welch (gdl_tiling_hmm * hmm, const gdl_tiling_hmm_seq * seq, int max_step, double likelihood_delta)
{
  int n, k, valid;
  double * log_p = &hmm->log_p;
  double log_p_old, log_p_k, diff;
  gdl_tiling_hmm_store * store;
  int status;
  char *str;

  /* local store for all iterations */
  store = gdl_tiling_hmm_store_alloc (hmm);
  /* allocate also alpha, beta and scale buffers */
  gdl_tiling_hmm_util_alloc (hmm, seq->size);
  
  log_p_old = -GDL_DBL_MAX;
  n = 1;

  /* main loop Baum-Welch-Alg. */
  while (n <= max_step)
  {
    status = gdl_tiling_hmm_baum_welch_one_step (hmm, seq, store);
    
    if (*log_p == +1)
    {
      fprintf (stderr, "gdl_tiling_hmm_baum_welch stopped: the tiling sequence cannot be built from the HMM!\n");
      status = GDL_EUNDRFLW;
      break;
    }
    
    //printf ("%d LOG_P %g\n", n, *log_p);

    diff = *log_p - log_p_old;
    /* error in convergence ? */
    if (diff < -GDL_SQRT_DBL_EPSILON)
    {
      status = GDL_ENOPROG;
    }
    else if (*log_p > GDL_SQRT_DBL_EPSILON)
    {
      status = GDL_ERANGE;
    }
	 	
    /* stop iterations? */
    if (diff < fabs ((double) likelihood_delta * (*log_p)))
    {
      //printf("Convergence after %d steps\n", n); 
		status = GDL_SUCCESS;
		break;
    }
    else
    {
      /* for next iteration */
      log_p_old = *log_p;
      gdl_tiling_hmm_store_reset (store, hmm);  /* sets all fields to zero */
      n++;
    }
  }
  
  gdl_tiling_hmm_store_free (store, hmm);
  
  gdl_tiling_hmm_forward (hmm, seq);
  
  gdl_tiling_hmm_util_free (hmm, seq->size);
  
  return status;
}

int
gdl_tiling_hmm_forward_backward (gdl_tiling_hmm * hmm, const gdl_tiling_hmm_seq * seq, double *prob[])
{
  const size_t T  = gdl_tiling_hmm_seq_size (seq);
  const size_t D  = gdl_tiling_hmm_seq_dim (seq);
  const size_t C  = hmm->nclass;
  const size_t N  = hmm->nstate;
  double * log_p  = &hmm->log_p;
  int res = -1;
  int k, i, j, t, j_id, valid, c;
  double gamma;
  gdl_tiling_hmm_state * state;
  double * norm;
  
  /* allocate alpha, beta and scale buffers */
  gdl_tiling_hmm_util_alloc (hmm, seq->size);
  double * scale  = hmm->scale;
  double ** alpha = hmm->alpha;
  double ** beta  = hmm->beta;
  
  valid = 0;
  if (gdl_tiling_hmm_forward (hmm, seq) != GDL_SUCCESS)
  {
  	  gdl_tiling_hmm_util_free (hmm, seq->size);
     return GDL_FAILURE;
  }
  if (*log_p != +1) /* Tiling sequence can be generated */
  {        
      valid = 1;
      if (gdl_tiling_hmm_backward (hmm, seq) != GDL_SUCCESS)
      {
        gdl_tiling_hmm_util_free (hmm, seq->size);
        return GDL_FAILURE;
      }
      
      /* Clean probs */
      for (t = 0; t < T; t++)
      	for(c = 0; c < C; c++)
      			prob[t][c] = 0;
      			
      norm = GDL_CALLOC (double, T);
      
      /* loop over all states */
      for (i = 0; i < N; i++)
      {
        state = hmm->states[i];
        /* Pi */
        if (state->has_pi)
        {
        	  gamma = alpha[0][i]*beta[0][i];
        	  prob[0][state->class] += gamma;
        	  norm[0] += gamma;
        }
		  for (t = 1; t < T; t++)
		  {
		  	 //if (gdl_tiling_hmm_seq_missing (seq, t)) continue;
		  	 prob[t][state->class] += alpha[t][i] * beta[t][i];
        	 norm[t] += alpha[t][i] * beta[t][i];
		  }
      }
      /* normalized */
      for (t = 0; t < T; t++)
      	if (norm[t]) // here should be a better check for tiny values...
      		for(c = 0; c < C; c++)
      			prob[t][c] /= norm[t];
      
      GDL_FREE (norm);			
  } 
  else
  {
  	 gdl_tiling_hmm_util_free (hmm, seq->size);
    GDL_WARNING ("Tiling sequence can't be built from HMM!", GDL_FAILURE);
  }
  
  gdl_tiling_hmm_util_free (hmm, seq->size); 

  return GDL_SUCCESS;
}

int
gdl_tiling_hmm_smooth (const gdl_tiling_hmm * hmm, const double gamma[], double x[])
{
	const size_t C  = hmm->nclass;
   const size_t D  = hmm->ndim;
	size_t c,d,g;
	
	for(d = 0; d < D; d++)
	{
		x[d]=0;
		for(c = 0; c < C; c++)
		{
			g = hmm->class2gauss[c][d];
			x[d] += gamma[c]*hmm->mu[g];
		}
	}
	
	return GDL_SUCCESS;
}

typedef struct 
{
  double **log_in_a;
  double **log_b;
  double *phi;
  double *phi_new;
  int **psi;
  int *topo_order;
  int topo_order_length;
} gdl_tiling_hmm_viterbi_store;

static gdl_tiling_hmm_viterbi_store *
gdl_tiling_hmm_viterbi_store_alloc (const gdl_tiling_hmm * hmm, const size_t len)
{
  const size_t N = hmm->nstate;
  size_t j;
  gdl_tiling_hmm_viterbi_store * v;
  
  
  v = GDL_CALLOC (gdl_tiling_hmm_viterbi_store, 1);

  /* Allocate the log_in_a's -> individal lenghts */
  v->log_in_a = GDL_MALLOC (double *, N);
  for (j = 0; j < N; j++)
    v->log_in_a[j] = GDL_CALLOC (double, hmm->states[j]->in_states);
  v->log_b = GDL_MATRIX_ALLOC (double, N, len);
  v->phi   = GDL_CALLOC (double, N);
  v->phi_new = GDL_CALLOC (double, N);
  v->psi = GDL_MATRIX_ALLOC (int, len, N);
  v->topo_order_length = 0;
  v->topo_order = GDL_CALLOC (int, N);

  return (v);
}     

static void
gdl_tiling_hmm_viterbi_store_free (gdl_tiling_hmm_viterbi_store * v, const size_t N, const size_t T)
{
	if (v)
	{
		GDL_MATRIX_FREE (v->log_in_a, N);
		GDL_MATRIX_FREE (v->log_b, N);
  		GDL_FREE (v->phi);
  		GDL_FREE (v->phi_new);
  		GDL_MATRIX_FREE(v->psi, T);
  		GDL_FREE (v->topo_order);
		GDL_FREE (v);
	}	
}

static void
gdl_tiling_hmm_viterbi_precompute (const gdl_tiling_hmm * hmm, const gdl_tiling_hmm_seq * seq, gdl_tiling_hmm_viterbi_store * v)
{
  const size_t N = hmm->nstate;
  const size_t T = seq->size; 
  double b_t;
  int i, j, t;
  /* Precomputing the log(a_ij) */
  for (j = 0; j < N; j++)
    for (i = 0; i < hmm->states[j]->in_states; i++)
      if (hmm->states[j]->in_a[i] < GDL_DBL_EPSILON)
        v->log_in_a[j][i] = +1; /* Not used any further in the calculations */
      else
        v->log_in_a[j][i] = log (hmm->states[j]->in_a[i]);
  /* Precomputing the log(bj(ot)) */
  for (j = 0; j < N; j++)
    for (t = 0; t < T; t++)
    {
 	    b_t = (*hmm->states[j]->out_b)(seq, t, hmm, hmm->states[j]->gauss_class);
 	    if (b_t <= GDL_DBL_EPSILON)
          v->log_b[j][t] = +1;
       else
          v->log_b[j][t] = log (b_t);
    }
}

int *
gdl_tiling_hmm_viterbi (const gdl_tiling_hmm * hmm, const gdl_tiling_hmm_seq * seq)
{
  const size_t T  = gdl_tiling_hmm_seq_size (seq);
  const size_t D  = gdl_tiling_hmm_seq_dim (seq);
  const size_t N  = hmm->nstate;
  double log_p;
  int *state_seq = NULL;
  int t, j, i, k, St;
  double value, max_value;
  gdl_tiling_hmm_viterbi_store * v;
  
  int len_path;
  int cur_len_path = 0; /* the current length of the viterbi path */ 
  int lastemState;
  int state_seq_index;

  len_path = T;
  
  v = gdl_tiling_hmm_viterbi_store_alloc (hmm, T);
  
  /* allocating state_seq array */
  state_seq = GDL_MALLOC (int, len_path);
  /* initialization of state_seq with -1 */
  for (i = 0; i < len_path; i++)
     state_seq[i] = -1;
  
  /* Precomputing the log(a_ij) and log(bj(ot)) */
  gdl_tiling_hmm_viterbi_precompute (hmm, seq, v);

  /* Initialization, that is t = 0 */
  for (j = 0; j < N; j++)
  {
    if (hmm->states[j]->pi == 0.0 || v->log_b[j][0] == +1)
      v->phi[j] = +1;
    else
      v->phi[j] = log (hmm->states[j]->pi) + v->log_b[j][0];
  }
  /* t > 0 */
  for (t = 1; t < T; t++)
  {
    for (j = 0; j < N; j++)
    {
		/** initialization of phi, psi **/
      v->phi_new[j] = +1;
      v->psi[t][j]  = -1;
    }
    for (k = 0; k < N; k++)
    {
      /* Determine the maximum */
      /* max_phi = phi[i] + log_in_a[j][i] ... */
      St = k;
      max_value     = -GDL_DBL_MAX;
      v->psi[t][St] = -1;
      for (i = 0; i < hmm->states[St]->in_states; i++)
      {
         if (v->phi[hmm->states[St]->in_id[i]] != +1 && v->log_in_a[St][i] != +1)
      	{
	         value = v->phi[hmm->states[St]->in_id[i]] + v->log_in_a[St][i];
	         if (value > max_value)
	         {
	           max_value = value;
	           v->psi[t][St] = hmm->states[St]->in_id[i];
	         }
      	}
      }

      /* No maximum found (that is, state never reached)
         or the output O[t] = 0.0: */
      if (max_value == -GDL_DBL_MAX || v->log_b[St][t] == +1)
      {
      	v->phi_new[St] = +1;
      }
      else
      	v->phi_new[St] = max_value + v->log_b[St][t];
    }                           /* complete time step for emitting states */
	 /* First now replace the old phi with the new phi */
    for (j = 0; j < N; j++)
    {
       v->phi[j] = v->phi_new[j];
    }
  } 
  /* Termination */
  state_seq_index = len_path - 1;
  max_value = -GDL_DBL_MAX;
  state_seq[state_seq_index] = -1;
  for (j = 0; j < N; j++)
  {
    if (v->phi[j] != +1 && v->phi[j] > max_value)
    {
      max_value = v->phi[j];
      state_seq[state_seq_index] = j;
    }
  }  
  if (max_value == -GDL_DBL_MAX)
  {
    log_p = +1;
  }
  else
  {
  		log_p = max_value;
      lastemState = state_seq[len_path - 1];
      for (t = T - 2, i = len_path - 2; t >= 0; t--)
      {
        state_seq[i--] = v->psi[t + 1][lastemState];
        lastemState = v->psi[t + 1][lastemState];
      }  
   }
    
  /* Free the memory space */
  gdl_tiling_hmm_viterbi_store_free (v, N, T);
  
  for(t = 0; t < T; t++)
     if (state_seq[t]!=-1)
	     state_seq[t] = hmm->states[state_seq[t]]->class;
 
  return (state_seq);
}

int
gdl_tiling_hmm_init (gdl_tiling_hmm * hmm, const gdl_tiling_hmm_seq * seq, const int gauss_path[])
{
	const size_t D = hmm->ndim;
	const size_t N = hmm->nstate;
	const size_t C = hmm->ngauss_class;
	const size_t T = seq->size;
	int status;
	size_t i,j,j_id,l,t,s,d,c,n=0,min,*nc;
	double x;
	
	for (c = 0; c < C; c++)
		hmm->mu[c]=hmm->var[c]=0;
	
	if ((gauss_path==0 || gauss_path[0]==-1) && hmm->T->init_b)
	{
		status=(hmm->T->init_b)(hmm, seq, 0);
	}
	else if (gauss_path!=0 && gauss_path[0]!=-1)
   // First compute the mean/var estimates
   // for each class of state
	{
		nc = GDL_CALLOC (size_t, C);
		for(t = 0; t < T; t++)
		{
			if (gdl_tiling_hmm_seq_missing (seq, t))
				continue;
			for(d = 0; d < D; d++)
			{
				c = gauss_path[t];
				nc[c]++;
				x = gdl_tiling_hmm_seq_get (seq, t, d);
				hmm->mu[c]+=x;
				hmm->var[c]+=x*x;
			}
			n++;
		}
		if (n < 3)
		{
			GDL_FREE (nc);
			return GDL_EINVAL;
		}
		
		size_t nok = 0;
		double tot_var  = 0;
		double tot_mu   = 0;
		for (c = 0; c < C; c++)
		{
			if (nc[c] > 2)
			{
				hmm->mu[c]  /= nc[c];
				hmm->var[c] /= nc[c];
				hmm->var[c] -= hmm->mu[c]*hmm->mu[c];
				nok += nc[c];
				tot_var += nc[c]*hmm->var[c];
				tot_mu  += nc[c]*hmm->mu[c]; 
			}
		}
		tot_var /= nok;
		tot_mu  /= nok;
		for (c = 0; c < C; c++)
		{
			if (nc[c] < 3) 
			{
				hmm->mu[c]  = tot_mu;
			   hmm->var[c] = tot_var;
			}
		}
		GDL_FREE (nc);
		
		status = GDL_SUCCESS;
	}
	else
	{
		return GDL_FAILURE;	
	}
	// Then compute the transition probabilities
	// First init paramaters with uniform prob.
	size_t npi=0;
	for(i = 0; i < N; i++)
	{
		if (hmm->states[i]->has_pi) npi++;
		if (!hmm->states[i]->out_fix)
		{
			x = 1.0/hmm->states[i]->out_states;
			for(j = 0; j < hmm->states[i]->out_states; j++)
			{
				hmm->states[i]->out_a[j] = x;
				j_id = hmm->states[i]->out_id[j];
				l=0;
		      while (l < hmm->states[j_id]->in_states)
		      {
		        if (hmm->states[j_id]->in_id[l] == i) break;
		        else l++;
		      }
		      if (l < hmm->states[j_id]->in_states)
		      {
		      	hmm->states[j_id]->in_a[l] = hmm->states[i]->out_a[j];
		      }
		      else {/* ERROR !*/}
			}
		}
	}
	x = 1.0/npi;
	for(i = 0; i < N; i++) 
		if (hmm->states[i]->has_pi)
			hmm->states[i]->pi=x;
	
	//hmm->update_b = gdl_false;
	//gdl_tiling_hmm_baum_welch (hmm, seq, 1, 1);
	//hmm->update_b = gdl_true;
	
	return status;
}

int
gdl_tiling_hmm_util_alloc (gdl_tiling_hmm * hmm, const size_t T)
{
	hmm->alpha = GDL_MATRIX_ALLOC (double, T, hmm->nstate);
	hmm->beta  = GDL_MATRIX_ALLOC (double, T, hmm->nstate);
	hmm->scale = GDL_CALLOC (double, T);
}

int
gdl_tiling_hmm_util_free (gdl_tiling_hmm * hmm, const size_t T)
{
	GDL_MATRIX_FREE (hmm->alpha, T);hmm->alpha=0;
	GDL_MATRIX_FREE (hmm->beta, T);hmm->beta=0;
	GDL_FREE (hmm->scale);hmm->scale=0;
}

gdl_tiling_hmm_state *
gdl_tiling_hmm_state_alloc (const size_t class, const size_t gauss_class[], const size_t ndim, const size_t in_states, const size_t out_states, double (*out_b)(const gdl_tiling_hmm_seq * o, size_t t, const gdl_tiling_hmm * hmm, size_t * class))
{
	size_t i;
	gdl_tiling_hmm_state * s;
	
	s = GDL_CALLOC (gdl_tiling_hmm_state, 1);
	
	s->out_b = out_b;
	s->gauss_class = GDL_MALLOC (size_t, ndim);
	memcpy (s->gauss_class, gauss_class, sizeof(size_t)*ndim);
	s->class = class;
	s->out_states = out_states;
	s->out_id     = GDL_CALLOC (size_t, out_states);
	s->out_a      = GDL_CALLOC (double, out_states);
	for(i=0;i<out_states;i++) s->out_a[i] = 1.0/out_states;
	s->in_states  = in_states;
	s->in_id      = GDL_CALLOC (size_t, in_states);
	s->in_a       = GDL_CALLOC (double, in_states);
	for(i=0;i<in_states;i++) s->in_a[i] = 1.0/in_states;
	
	return s;	
}

void
gdl_tiling_hmm_state_free (gdl_tiling_hmm_state * state)
{
	if (state)
	{
		GDL_FREE (state->gauss_class);
		GDL_FREE (state->out_id);
		GDL_FREE (state->out_a);
		GDL_FREE (state->in_id);
		GDL_FREE (state->in_a);
		GDL_FREE (state);	
	}	
}

gdl_tiling_hmm_state *
gdl_tiling_hmm_state_fread (FILE * stream, const size_t ndim, double (*out_b)(const gdl_tiling_hmm_seq * o, size_t t, const gdl_tiling_hmm * hmm, size_t * class))
{
	if (stream)
	{
		int status;
		size_t class, * gauss_class, in_states, out_states;
		gdl_tiling_hmm_state * state;

		gauss_class = GDL_CALLOC (size_t, ndim);

		status = fread(&class, sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread(gauss_class, sizeof(size_t), ndim, stream);
		GDL_FREAD_STATUS (status, ndim);
		status = fread(&in_states, sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread(&out_states, sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		
		state = gdl_tiling_hmm_state_alloc (class, gauss_class, ndim, in_states, out_states, out_b); 
  		
  		GDL_FREE (gauss_class);
  		
  		status = fread(&state->has_pi, sizeof(gdl_boolean), 1, stream);
		GDL_FREAD_STATUS (status, 1);
  		status = fread(&state->pi, sizeof(double), 1, stream);
		GDL_FREAD_STATUS (status, 1);
  		status = fread(state->out_id, sizeof(size_t), out_states, stream);
		GDL_FREAD_STATUS (status, out_states);
		status = fread(state->in_id, sizeof(size_t), in_states, stream);
		GDL_FREAD_STATUS (status, in_states);
 		status = fread(state->out_a, sizeof(double), out_states, stream);
		GDL_FREAD_STATUS (status, out_states);
		status = fread(state->in_a, sizeof(double), in_states, stream);
		GDL_FREAD_STATUS (status, in_states);
 		status = fread(&state->b_fix, sizeof(gdl_boolean), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread(&state->out_fix, sizeof(gdl_boolean), 1, stream);
		GDL_FREAD_STATUS (status, 1);
  		status = fread(&state->in_fix, sizeof(gdl_boolean), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		
		return state;
	}
	return 0;	
}

int
gdl_tiling_hmm_state_fwrite (FILE * stream, const gdl_tiling_hmm_state * state, const size_t ndim)
{
	if (stream && state)
	{
		int status;
		
		status = fwrite(&state->class, sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite(state->gauss_class, sizeof(size_t), ndim, stream);
		GDL_FWRITE_STATUS (status, ndim);
		status = fwrite(&state->in_states, sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite(&state->out_states, sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite(&state->has_pi, sizeof(gdl_boolean), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite(&state->pi, sizeof(double), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
  		status = fwrite(state->out_id, sizeof(size_t), state->out_states, stream);
		GDL_FWRITE_STATUS (status, state->out_states);
		status = fwrite(state->in_id, sizeof(size_t), state->in_states, stream);
		GDL_FWRITE_STATUS (status, state->in_states);
 		status = fwrite(state->out_a, sizeof(double), state->out_states, stream);
		GDL_FWRITE_STATUS (status, state->out_states);
		status = fwrite(state->in_a, sizeof(double), state->in_states, stream);
		GDL_FWRITE_STATUS (status, state->in_states);
 		status = fwrite(&state->b_fix, sizeof(gdl_boolean), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite(&state->out_fix, sizeof(gdl_boolean), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
  		status = fwrite(&state->in_fix, sizeof(gdl_boolean), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		
		return GDL_SUCCESS;
	}
	return GDL_EINVAL;
}

/**
 * Allocate a vector of pointers on the state of the HMM
 * as defined by the incidence matrix A[N][N], so that A(i,j)=1 if there
 * is a transition from state i to state j, and the matrix of 
 * classes C[N][D] which gives the D-tuple class of each state, and
 * the emission function out_b for each state.
 */
gdl_tiling_hmm_state **
gdl_tiling_hmm_state_matrix_alloc (size_t * A[], size_t C[], size_t * GC[], const size_t N, const size_t D, double (*out_b)(const gdl_tiling_hmm_seq * o, size_t t, const gdl_tiling_hmm * hmm, size_t * class))
{
	size_t i,j,k,in_states,out_states,not_in_fix=0;
	double z;
	gdl_tiling_hmm_state ** states;
	gdl_tiling_hmm_state * state;
	
	states = GDL_MALLOC (gdl_tiling_hmm_state *, N);
	
	for(i = 0; i < N; i++)
	{
		in_states=out_states=0;
		for(j = 0; j < N; j++) out_states+=A[i][j];
		for(j = 0; j < N; j++) in_states +=A[j][i];
		state = gdl_tiling_hmm_state_alloc (C[i], GC[i], D, in_states, out_states, out_b);
		z = 1.0/out_states;
		if (out_states)
			for(k = j = 0; j < N; j++)
			{
				if (A[i][j])
				{
					state->out_id[k]=j;
					state->out_a[k]=z;
					k++;
				}
			}
		if (in_states)
			for(k = j = 0; j < N; j++)
				if (A[j][i]) 
					state->in_id[k++]=j;
		state->in_fix  = (in_states==1);
		not_in_fix    += (in_states>1);
		state->out_fix = (out_states==1);
		
		states[i] = state;
	}
	
	return states;
}

gdl_tiling_hmm *
gdl_tiling_hmm_alloc (gdl_tiling_hmm_type * T)
{
	gdl_tiling_hmm * hmm;
	
	hmm = GDL_CALLOC (gdl_tiling_hmm, 1);
	
	(T->alloc)(T, hmm);
	
	hmm->T = T;
	
	hmm->update_a = gdl_true;
	hmm->update_b = gdl_true;
	
	return hmm;
}

void
gdl_tiling_hmm_free (gdl_tiling_hmm * hmm)
{
	if (hmm)
	{
		size_t i;
		GDL_FREE (hmm->mu);
		GDL_FREE (hmm->var);
		gdl_tiling_hmm_type_free (hmm->T);
		for(i = 0; i < hmm->nstate; i++)
		{
			gdl_tiling_hmm_state_free (hmm->states[i]);
		}
		GDL_FREE(hmm);
	}
}

static int
gdl_tiling_hmm_nucleosome_alloc (const void * vT, gdl_tiling_hmm * hmm)
{
	gdl_tiling_hmm_type * T = (gdl_tiling_hmm_type *) vT;
	
	/***
	 * So, following Yuan et al., we have 3 classes here:
	 *  1 - Linker (L)
	 *  2 - Well-positioned nucleosomes (N) 
	 *  3 - Delocalised nucleosomes (D)
	 * 
	 * For the first, class we have a single state.
	 * For the second one, we have 'nucleo_nprobe_max' states and only states
	 * from 'nucleo_nprobe_min' to 'nucleo_nprobe_max' can output on the linker state
	 * For the last one, we have 'nucleo_nprobe_max'+1 states and only the last state
	 * can output on the linker state (or loop on himself)
	 * So the total number of states S will be S=2*('nucleo_nprobe_max'+1)
	 * 
	 * If 'ndim' == 2 (means that we are looking jointly to several tiled genomes)
	 * then we have S^2 states...
	 */
	const size_t D  = T->ndim;
	const size_t N  = 1+T->nucleo_nprobe_max+T->nucleo_deloc_max;
	 
	 if (D==1)
	 {
		const size_t C  = 3;
		const size_t GC = 2;
		size_t i, j, * class, ** gauss_class, ** A;
		gdl_tiling_hmm_state * state;
	 
		 hmm->nclass       = C;
		 hmm->ngauss_class = GC;
		 hmm->ndim         = D;
		 hmm->class2gauss  = GDL_MATRIX_ALLOC (size_t, C, D);
		 hmm->mu           = GDL_CALLOC (double, GC);
		 hmm->var          = GDL_CALLOC (double, GC);
		 
		 // Define the design matrix of the HMM
		 A = GDL_MATRIX_ALLOC (size_t, N, N);
		 A[0][0]=1; /* linker with itself */
		 A[0][1]=1; /* linker to well-positionned nucleosome */
		 A[0][T->nucleo_nprobe_max+1]=1; /* linker to delocalised nucleosome */
		 for(j=1;j<=T->nucleo_nprobe_max;j++)
		 {
		 	if (j<T->nucleo_nprobe_max)  A[j][j+1]=1; /* well-positionned nucleosome to the next one */
		 	if (j>=T->nucleo_nprobe_min) A[j][0]=1;   /* terminal well-positionned nucleosome to the linker */
		 }
		 for(;j<N-1;j++) A[j][j+1]=1; /* delocalised nucleosome to the next one */
		 A[N-1][0]=1; /* terminal delocalised nucleosome to the linker */
		 A[N-1][N-1]=1; /* terminal delocalised nucleosome to itself */
	 	// Set the class of the states
	 	class       = GDL_CALLOC (size_t, N);
	 	gauss_class = GDL_MATRIX_ALLOC (size_t, N, 1);
	 	hmm->class2gauss[0][0] = 0; // linker class as its own gauss class
	 	hmm->class2gauss[1][0] = 1; // well-positionned and fuzzy clases 
	 	hmm->class2gauss[2][0] = 1; // share the same gauss glass
	 	for(j=1;j<=T->nucleo_nprobe_max;j++) {class[j]=2;gauss_class[j][0]=1;}
	 	for(;j<N;j++) {class[j]=1;gauss_class[j][0]=1;}
		 
		 hmm->nstate = N;
		 hmm->states = gdl_tiling_hmm_state_matrix_alloc (A, class, gauss_class, N, 1, T->out_b);
		 // set pi
		 hmm->states[0]->has_pi = gdl_true;
		 hmm->states[0]->pi = 0.5;
		 hmm->states[1]->has_pi = gdl_true; 
		 hmm->states[1]->pi = 0.25;
		 hmm->states[T->nucleo_nprobe_max+1]->has_pi = gdl_true;
		 hmm->states[T->nucleo_nprobe_max+1]->pi = 0.25;		 
	 }
	 else if (D==2 && !T->heterogeneous)
	 {
	 	const size_t C  = 5;
		const size_t GC = 2;
		size_t i, j, * class, ** gauss_class, ** A;
		gdl_tiling_hmm_state * state;
	 
		 hmm->nclass       = C;
		 hmm->ngauss_class = GC;
		 hmm->ndim         = D;
		 hmm->class2gauss  = GDL_MATRIX_ALLOC (size_t, C, D);
		 hmm->mu           = GDL_CALLOC (double, GC);
		 hmm->var          = GDL_CALLOC (double, GC);
	 	
	 	 // Define the design matrix of the HMM
		 A = GDL_MATRIX_ALLOC (size_t, 2*N-1, 2*N-1);
		 A[0][0]=1; /* linker x linker with itself */
		 A[0][1]=1; /* linker x linker to well-positionned nucleosome x linker*/
		 A[0][T->nucleo_nprobe_max+1]=1;
		 A[0][N]=1; /* linker x linker to  well-positionned nucleosome x well-positionned nucleosome*/
		 for(j=1;j<=T->nucleo_nprobe_max;j++)
		 {
		 	if (j<T->nucleo_nprobe_max) 
		 	{
		 		A[j][j+1]=1; /* well-positionned nucleosome to the next one */
		 		A[N+j-1][N+j]=1;
		 	}
		 	if (j>=T->nucleo_nprobe_min) 
		 	{
		 		A[j][0]=1;   /* terminal well-positionned nucleosome to the linker */
		 		A[N+j-1][0]=1;
		 	}
		 }
		 for(;j<N-1;j++)
		 {
		 	  A[j][j+1]=1; /* delocalised nucleosome to the next one */
		 	  A[N+j-1][N+j]=1;
		 }
		 A[N-1][0]=1; /* terminal delocalised nucleosome to the linker */
		 A[2*N-2][0]=1;
		 A[N-1][N-1]=1; /* terminal delocalised nucleosome to itself */
		 A[2*N-2][2*N-2]=1;
		 // Set the class of the states
	 	 class       = GDL_CALLOC (size_t, 2*N-1);
	 	 gauss_class = GDL_MATRIX_ALLOC (size_t, 2*N-1, 2);
	 	 hmm->class2gauss[0][0] = 0;
	 	 hmm->class2gauss[0][1] = 0;
	 	 hmm->class2gauss[1][0] = 1; 
	 	 hmm->class2gauss[1][1] = 0;  
	 	 hmm->class2gauss[2][0] = 1;
	 	 hmm->class2gauss[2][1] = 0;  
	 	 hmm->class2gauss[3][0] = 1;
	 	 hmm->class2gauss[3][1] = 1;
	 	 hmm->class2gauss[4][0] = 1;
	 	 hmm->class2gauss[4][1] = 1;
	 	 for(j=1;j<=T->nucleo_nprobe_max;j++)
	 	 {
	 	 	class[j]=2;
	 	 	class[N+j-1]=4;
	 	 	gauss_class[j][0]=1;
	 	 	gauss_class[j][1]=0;
	 	 	gauss_class[N+j-1][0]=1;
	 	 	gauss_class[N+j-1][1]=1;
	 	 }
	 	 for(;j<N;j++)
	 	 {
	 	 	class[j]=1;
	 	 	class[N+j-1]=3;
	 	 	gauss_class[j][0]=1;
	 	 	gauss_class[j][1]=0;
	 	 	gauss_class[N+j-1][0]=1;
	 	 	gauss_class[N+j-1][1]=1;
	 	 }
		 hmm->nstate = 2*N-1;
		 hmm->states = gdl_tiling_hmm_state_matrix_alloc (A, class, gauss_class, 2*N-1, 2, T->out_b);
		 // set pi
		 hmm->states[0]->has_pi = gdl_true;
		 hmm->states[0]->pi = 0.5;
		 hmm->states[1]->has_pi = gdl_true; 
		 hmm->states[1]->pi = 0.125;
		 hmm->states[T->nucleo_nprobe_max+1]->has_pi = gdl_true;
		 hmm->states[T->nucleo_nprobe_max+1]->pi = 0.125;
		 hmm->states[N]->has_pi = gdl_true;
		 hmm->states[N]->pi = 0.125;
		 hmm->states[N+T->nucleo_nprobe_max]->has_pi = gdl_true;
		 hmm->states[N+T->nucleo_nprobe_max]->pi = 0.125;
	 }
	 else if (D==2 && T->heterogeneous)
	 {
	 	const size_t C  = 5;
		const size_t GC = 4;
		size_t i, j, * class, ** gauss_class, ** A;
		gdl_tiling_hmm_state * state;
	 
		 hmm->nclass       = C;
		 hmm->ngauss_class = GC;
		 hmm->ndim         = D;
		 hmm->class2gauss  = GDL_MATRIX_ALLOC (size_t, C, D);
		 hmm->mu           = GDL_CALLOC (double, GC);
		 hmm->var          = GDL_CALLOC (double, GC);
	 	
	 	 // Define the design matrix of the HMM
		 A = GDL_MATRIX_ALLOC (size_t, 2*N-1, 2*N-1);
		 A[0][0]=1; /* linker x linker with itself */
		 A[0][1]=1; /* linker x linker to well-positionned nucleosome x linker*/
		 A[0][T->nucleo_nprobe_max+1]=1;
		 A[0][N]=1; /* linker x linker to  well-positionned nucleosome x well-positionned nucleosome*/
		 for(j=1;j<=T->nucleo_nprobe_max;j++)
		 {
		 	if (j<T->nucleo_nprobe_max) 
		 	{
		 		A[j][j+1]=1; /* well-positionned nucleosome to the next one */
		 		A[N+j-1][N+j]=1;
		 	}
		 	if (j>=T->nucleo_nprobe_min) 
		 	{
		 		A[j][0]=1;   /* terminal well-positionned nucleosome to the linker */
		 		A[N+j-1][0]=1;
		 	}
		 }
		 for(;j<N-1;j++)
		 {
		 	  A[j][j+1]=1; /* delocalised nucleosome to the next one */
		 	  A[N+j-1][N+j]=1;
		 }
		 A[N-1][0]=1; /* terminal delocalised nucleosome to the linker */
		 A[2*N-2][0]=1;
		 A[N-1][N-1]=1; /* terminal delocalised nucleosome to itself */
		 A[2*N-2][2*N-2]=1;
		 // Set the class of the states
	 	 class       = GDL_CALLOC (size_t, 2*N-1);
	 	 gauss_class = GDL_MATRIX_ALLOC (size_t, 2*N-1, 2);
	 	 hmm->class2gauss[0][0] = 0;
	 	 hmm->class2gauss[0][1] = 2;
	 	 hmm->class2gauss[1][0] = 1; 
	 	 hmm->class2gauss[1][1] = 3;  
	 	 hmm->class2gauss[2][0] = 1;
	 	 hmm->class2gauss[2][1] = 3;  
	 	 hmm->class2gauss[3][0] = 1;
	 	 hmm->class2gauss[3][1] = 2;
	 	 hmm->class2gauss[4][0] = 1;
	 	 hmm->class2gauss[4][1] = 2;
	 	 gauss_class[0][0]=0;
	 	 gauss_class[0][1]=2;
	 	 for(j=1;j<=T->nucleo_nprobe_max;j++)
	 	 {
	 	 	class[j]=2;
	 	 	class[N+j-1]=4;
	 	 	gauss_class[j][0]=1;
	 	 	gauss_class[j][1]=3;
	 	 	gauss_class[N+j-1][0]=1;
	 	 	gauss_class[N+j-1][1]=3;
	 	 }
	 	 for(;j<N;j++)
	 	 {
	 	 	class[j]=1;
	 	 	class[N+j-1]=3;
	 	 	gauss_class[j][0]=1;
	 	 	gauss_class[j][1]=2;
	 	 	gauss_class[N+j-1][0]=1;
	 	 	gauss_class[N+j-1][1]=2;
	 	 }
		 hmm->nstate = 2*N-1;
		 hmm->states = gdl_tiling_hmm_state_matrix_alloc (A, class, gauss_class, 2*N-1, 2, T->out_b);
		 // set pi
		 hmm->states[0]->has_pi = gdl_true;
		 hmm->states[0]->pi = 0.5;
		 hmm->states[1]->has_pi = gdl_true; 
		 hmm->states[1]->pi = 0.125;
		 hmm->states[T->nucleo_nprobe_max+1]->has_pi = gdl_true;
		 hmm->states[T->nucleo_nprobe_max+1]->pi = 0.125;
		 hmm->states[N]->has_pi = gdl_true;
		 hmm->states[N]->pi = 0.125;
		 hmm->states[N+T->nucleo_nprobe_max]->has_pi = gdl_true;
		 hmm->states[N+T->nucleo_nprobe_max]->pi = 0.125;
	 }
	 else
	 {
	 	 // error
	 }
	
	 return GDL_SUCCESS;	
}

gdl_tiling_hmm_type *
gdl_tiling_hmm_type_alloc (const gdl_tiling_hmm_type * T)
{
	gdl_tiling_hmm_type * t;
	
	t = GDL_MALLOC (gdl_tiling_hmm_type, 1);
	
	memcpy (t, T, sizeof(gdl_tiling_hmm_type));
	
	t->name = gdl_string_clone (T->name);
	
	return t;	
} 

void
gdl_tiling_hmm_type_free (gdl_tiling_hmm_type * T)
{
	if (T)
	{
		gdl_string_free (T->name);
		GDL_FREE (T);	
	}	
}

static double
gdl_tiling_hmm_nucleosome_out_b (const gdl_tiling_hmm_seq * seq, size_t t, const gdl_tiling_hmm * hmm, size_t * class)
{
	if (!gdl_tiling_hmm_seq_missing (seq, t))
	{ 
		const size_t D = hmm->ndim;
		size_t c,d;
		double x,z,pr = 1;
		double mu,var_1,log_pr=0;
		
		for(d = 0; d < D; d++)
		{
			c     = class[d];
			mu    = hmm->mu[c];
			var_1 = 1.0 / hmm->var[c];
			x = gdl_tiling_hmm_seq_get (seq, t, d);
			x -= mu;
			log_pr += log(var_1) - x*x*var_1;
		}
		
		return exp(0.5*log_pr);
	}
	
	return 1;
}

static int
gdl_tiling_hmm_nucleosome_init_b (gdl_tiling_hmm * hmm, const gdl_tiling_hmm_seq * seq, const gdl_rng * rng)
{
	const size_t D = hmm->ndim;
	const size_t N = hmm->nstate;
	const size_t C = hmm->ngauss_class;
	const size_t T = seq->size;
	size_t i,j,j_id,l,t,s,d,c,n=0,min,**nc;
	double ** X, mu0, sig0, mu1, sig1;
	double factor;
	
	X = GDL_MALLOC (double *, D);
	
	for (d = 0; d < D; d++)
	{
		X[d] = gdl_tiling_hmm_seq_vector (seq, d);
	}
	for (c = 0; c < C; c++)
	{
		hmm->mu[c] = hmm->var[c] = 0;
	}
	if (hmm->T->heterogeneous)
	{
		const size_t ncd = C / D;
		size_t c0 = 0;
		for (d = 0; d < D; d++)
		{
			size_t c1 = (c0 + ncd)/2;
			if (c0 != 0 && c0 % 2 == 0) c1++;
			if (seq->nobs > 5)
			{
				n    = seq->nobs/2;
				gdl_sort (X[d], 1, seq->nobs);
				mu0  = gdl_stats_mean (X[d], 1, n);
				sig0 = gdl_stats_variance (X[d], 1, seq->nobs);
				mu1  = gdl_stats_mean (X[d]+n, 1, seq->nobs-n);
				for (c = c0; c < c1; c++)
				{
					hmm->mu[c]  += (0.85*mu0+0.15*mu1);
					hmm->var[c] += sig0;
				}
				for (c = c1; c < (c0 + ncd); c++)
				{
					hmm->mu[c]  += (0.15*mu0+0.85*mu1);
					hmm->var[c] += sig0;
				}
			}
			else
			{
				n    = seq->nobs;
				mu0  = gdl_stats_mean (X[d], 1, n);
				sig0 = gdl_stats_variance_with_fixed_mean (X[d], 1, n, mu0);
				for (c = c0; c < (c0 + ncd); c++)
				{
					hmm->mu[c]  += mu0;
					hmm->var[c] += sig0;
				}
			}
			c0 += ncd;
		}
		factor = 1.0;
	}
	else
	{
		for (d = 0; d < D; d++)
		{
			if (seq->nobs > 5)
			{
				n    = seq->nobs/2;
				gdl_sort (X[d], 1, seq->nobs);
				mu0  = gdl_stats_mean (X[d], 1, n);
				sig0 = gdl_stats_variance (X[d], 1, seq->nobs);
				mu1  = gdl_stats_mean (X[d]+n, 1, seq->nobs-n);
				for (c = 0; c < C/2+C%2; c++)
				{
					hmm->mu[c]  += (0.85*mu0+0.15*mu1);
					hmm->var[c] += sig0;
				}
				for (c = C/2+C%2; c < C; c++)
				{
					hmm->mu[c]  += (0.15*mu0+0.85*mu1);
					hmm->var[c] += sig0;
				}
			}
			else
			{
				n    = seq->nobs;
				mu0  = gdl_stats_mean (X[d], 1, n);
				sig0 = gdl_stats_variance_with_fixed_mean (X[d], 1, n, mu0);
				for (c = 0; c < C; c++)
				{
					hmm->mu[c]  += mu0;
					hmm->var[c] += sig0;
				}
			}		
		}
		factor = 1.0/D;
	}
	for (c = 0; c < C; c++)
	{
		hmm->mu[c]  *= factor;
		hmm->var[c] *= factor;
		//printf ("INIT(%d %d) %d %g %g\n", seq->nobs, hmm->T->heterogeneous, c, hmm->mu[c], sqrt(hmm->var[c]));
	}
	
	GDL_MATRIX_FREE (X, D);
	
	return GDL_SUCCESS;
	
}

gdl_tiling_hmm_type *
gdl_tiling_hmm_type_fread (FILE * stream)
{
	if (stream)
	{
		int status;
		gdl_string * name;
		gdl_tiling_hmm_type * T;
		
		name = gdl_string_fread (stream);
		GDL_FREAD_STATUS (name!=0, 1);
		
		if (!strcmp(name, gdl_tiling_hmm_nucleosome->name))
		{
			T = gdl_tiling_hmm_type_alloc (gdl_tiling_hmm_nucleosome);
			status = fread (&T->nucleo_nprobe_min, sizeof(size_t), 1, stream);
			GDL_FREAD_STATUS (status, 1);
			status = fread (&T->nucleo_nprobe_max, sizeof(size_t), 1, stream);
			GDL_FREAD_STATUS (status, 1);
			status = fread (&T->nucleo_deloc_max, sizeof(size_t), 1, stream);
			GDL_FREAD_STATUS (status, 1);
			status = fread (&T->ndim, sizeof(size_t), 1, stream);
			GDL_FREAD_STATUS (status, 1);
		}
		else
		{
			GDL_FREAD_STATUS (1==0, 1);
		}
		
		return T;
	}
	return 0;
}

int
gdl_tiling_hmm_type_fwrite (FILE * stream, const gdl_tiling_hmm_type * T)
{
	if (stream && T)
	{
		int status;
		
		status = gdl_string_fwrite (stream, T->name);
		GDL_FWRITE_STATUS (status, GDL_SUCCESS);
		
		if (!strcmp(T->name, gdl_tiling_hmm_nucleosome->name))
		{
			status = fwrite (&T->nucleo_nprobe_min, sizeof(size_t), 1, stream);
			GDL_FWRITE_STATUS (status, 1);
			status = fwrite (&T->nucleo_nprobe_max, sizeof(size_t), 1, stream);
			GDL_FWRITE_STATUS (status, 1);
			status = fwrite (&T->nucleo_deloc_max, sizeof(size_t), 1, stream);
			GDL_FWRITE_STATUS (status, 1);
			status = fwrite (&T->ndim, sizeof(size_t), 1, stream);
			GDL_FWRITE_STATUS (status, 1);
		}
		
		return GDL_SUCCESS;		
	}
	return GDL_EINVAL;
}

gdl_tiling_hmm *
gdl_tiling_hmm_fread (FILE * stream)
{
	if (stream)
	{
		size_t i;
		int status;
		gdl_tiling_hmm_type * T;
		gdl_tiling_hmm * hmm;
		
		T = gdl_tiling_hmm_type_fread (stream);
		
		hmm = gdl_tiling_hmm_alloc (T);
		
		status = fread (&hmm->nstate, sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&hmm->nclass, sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&hmm->ngauss_class, sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (&hmm->ndim, sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread (hmm->mu, sizeof(double), hmm->ngauss_class, stream);
		GDL_FREAD_STATUS (status, hmm->ngauss_class);
		status = fread (hmm->var, sizeof(double), hmm->ngauss_class, stream);
		GDL_FREAD_STATUS (status, hmm->ngauss_class);	
		for(i = 0; i < hmm->nstate; i++)
		{
			gdl_tiling_hmm_state_free (hmm->states[i]);
			hmm->states[i] = gdl_tiling_hmm_state_fread (stream, hmm->ndim, T->out_b);
			GDL_FREAD_STATUS (hmm->states[i]!=0, 1);	
		}
		
		return hmm;
	}
	return 0;	
}


int
gdl_tiling_hmm_fwrite (FILE * stream, const gdl_tiling_hmm * hmm)
{
	if (stream && hmm)
	{
		size_t i;
		int status;
		
		status = gdl_tiling_hmm_type_fwrite (stream, hmm->T);
		GDL_FWRITE_STATUS (status, GDL_SUCCESS);
		status = fwrite (&hmm->nstate, sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&hmm->nclass, sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&hmm->ngauss_class, sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&hmm->ndim, sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (hmm->mu, sizeof(double), hmm->ngauss_class, stream);
		GDL_FWRITE_STATUS (status, hmm->ngauss_class);
		status = fwrite (hmm->var, sizeof(double), hmm->ngauss_class, stream);
		GDL_FWRITE_STATUS (status, hmm->ngauss_class);	
		for(i = 0; i < hmm->nstate; i++)
		{
			status = gdl_tiling_hmm_state_fwrite (stream, hmm->states[i], hmm->ndim);
			GDL_FWRITE_STATUS (status, GDL_SUCCESS);
		}
		
		return GDL_SUCCESS;
	}
	
	return GDL_EINVAL;
}

gdl_tiling_hmm_backup *
gdl_tiling_hmm_backup_alloc (const gdl_tiling_hmm * hmm)
{
	if (!hmm) return 0;
	
	size_t i,j,io,ii,ip;	
	gdl_tiling_hmm_backup * b;
	
	b = GDL_CALLOC (gdl_tiling_hmm_backup, 1);
	
	b->ndim = hmm->ndim;
	b->ngauss_class = hmm->ngauss_class;
	
	b->mu  = GDL_MALLOC (double, b->ngauss_class);
	b->var = GDL_MALLOC (double, b->ngauss_class);
	
	memcpy (b->mu, hmm->mu, sizeof(double)*b->ngauss_class);
	memcpy (b->var, hmm->var, sizeof(double)*b->ngauss_class);	
	// now, get nin and nout
	for(i = 0; i < hmm->nstate; i++)
	{
		if (!hmm->states[i]->out_fix) // keep only out toward superior states
			for(j = 0; j < hmm->states[i]->out_states; j++)
				if (hmm->states[i]->out_id[j] >= i)
					(b->nout)++;
		if (!hmm->states[i]->in_fix) // keep only in from strictly superior states
			for(j = 0; j < hmm->states[i]->in_states; j++)
				if (hmm->states[i]->in_id[j] > i)
					(b->nin)++;
		if (hmm->states[i]->has_pi)
			(b->npi)++;
		
	}
	b->out_a  = GDL_CALLOC (double, b->nout);
	b->in_a   = GDL_CALLOC (double, b->nin);
	b->pi     = GDL_CALLOC (double, b->npi);
	b->out_id = GDL_CALLOC (size_t, b->nout);
	b->in_id  = GDL_CALLOC (size_t, b->nin);
	for(io = ii = ip= i = 0; i < hmm->nstate; i++)
	{
		if (!hmm->states[i]->out_fix)
		{
			for(j = 0; j < hmm->states[i]->out_states; j++)
			{
				if (hmm->states[i]->out_id[j] >= i)
				{
					b->out_a[io]    = hmm->states[i]->out_a[j];
					b->out_id[io++] = i;
				}
			}
		}
		if (!hmm->states[i]->in_fix)
		{
			for(j = 0; j < hmm->states[i]->in_states; j++)
			{
				if (hmm->states[i]->in_id[j] > i)
				{
					b->in_a[ii]   = hmm->states[i]->in_a[j];
					b->in_a[ii++] = i;
				}
			}
		}
		if (hmm->states[i]->has_pi)
			b->pi[ip++] = hmm->states[i]->pi;
	}
	
	return b;
}

void
gdl_tiling_hmm_backup_free (gdl_tiling_hmm_backup * b)
{
	if (b)
	{
		GDL_FREE (b->mu);
		GDL_FREE (b->var);
		GDL_FREE (b->out_a);
		GDL_FREE (b->in_a);
		GDL_FREE (b->pi);
		GDL_FREE (b);
	}	
}

int
gdl_tiling_hmm_backup_force (gdl_tiling_hmm_backup * b, const gdl_tiling_hmm * hmm)
{
	size_t i,j,io,ii,ip;
	
	memcpy (b->mu, hmm->mu, sizeof(double)*b->ngauss_class);
	memcpy (b->var, hmm->var, sizeof(double)*b->ngauss_class);	
	for(io = ii = ip = i = 0; i < hmm->nstate; i++)
	{
		if (!hmm->states[i]->out_fix)
		{
			for(j = 0; j < hmm->states[i]->out_states; j++)
				if (hmm->states[i]->out_id[j] >= i)
					b->out_a[io++]=hmm->states[i]->out_a[j];
		}
		if (!hmm->states[i]->in_fix)
		{
			for(j = 0; j < hmm->states[i]->in_states; j++)
				if (hmm->states[i]->in_id[j] > i)
					b->in_a[ii++]=hmm->states[i]->in_a[j];
		}
		if (hmm->states[i]->has_pi)
			b->pi[ip++] = hmm->states[i]->pi;
	}
	return GDL_SUCCESS;
}

int
gdl_tiling_hmm_backup_restore (gdl_tiling_hmm_backup * b, gdl_tiling_hmm * hmm)
{
	if (b && hmm)
	{
		size_t i,j,io,ii,ip,k;
		
		memcpy (hmm->mu, b->mu, sizeof(double)*b->ngauss_class);
		memcpy (hmm->var, b->var, sizeof(double)*b->ngauss_class);	
		for(io = ii = ip = i = 0; i < hmm->nstate; i++)
		{
			if (!hmm->states[i]->out_fix)
			{
				for(j = 0; j < hmm->states[i]->out_states; j++)
				{
					if (hmm->states[i]->out_id[j] >= i)
					{
						hmm->states[i]->out_a[j]=b->out_a[io++];
					}
					else
					{
						k=0;
						for(k = 0; k < hmm->states[hmm->states[i]->out_id[j]]->in_states; k++)
							if (hmm->states[hmm->states[i]->out_id[j]]->in_id[k]==i)
								break;
						hmm->states[i]->out_a[j]=hmm->states[hmm->states[i]->out_id[j]]->in_a[k];
					}
				}
			}		
			if (!hmm->states[i]->in_fix)
			{
				for(j = 0; j < hmm->states[i]->in_states; j++)
				{
					if (hmm->states[i]->in_id[j] > i)
					{
						hmm->states[i]->in_a[j]=b->in_a[ii++];
					}
					else
					{
						k=0;
						for(k = 0; k < hmm->states[hmm->states[i]->in_id[j]]->out_states; k++)
							if (hmm->states[hmm->states[i]->in_id[j]]->out_id[k]==i)
								break;
						hmm->states[i]->in_a[j]=hmm->states[hmm->states[i]->in_id[j]]->out_a[k];
					}
				}
			}
			if (hmm->states[i]->has_pi)
				hmm->states[i]->pi = b->pi[ip++];
		}
		
		return GDL_SUCCESS;
	}
	return GDL_EINVAL;
}

int
gdl_tiling_hmm_backup_fread (FILE * stream, gdl_tiling_hmm_backup * b)
{
	if (stream)
	{
		int status;
		size_t i;
		
		status = fread (b->mu, sizeof(double), b->ngauss_class, stream);
		GDL_FREAD_STATUS (status, b->ngauss_class);
		status = fread (b->var, sizeof(double), b->ngauss_class, stream);
		GDL_FREAD_STATUS (status, b->ngauss_class);
		status = fread (b->out_a, sizeof(double), b->nout, stream);
		GDL_FREAD_STATUS (status, b->nout);
		status = fread (b->in_a, sizeof(double), b->nin, stream);
		GDL_FREAD_STATUS (status, b->nin);
		status = fread (b->pi, sizeof(double), b->npi, stream);
		GDL_FREAD_STATUS (status, b->npi);
		
		return GDL_SUCCESS;
	}
	return GDL_EINVAL;
}

int
gdl_tiling_hmm_backup_fprintf_header (FILE * stream, const gdl_tiling_hmm_backup * b)
{
	size_t i;
	for(i = 0; i < b->ngauss_class; i++)
		fprintf(stream, " mu(%d) var(%d)", i, i);	
	for(i = 0; i < b->nin; i++)
		fprintf(stream, " pr(in,%d)", b->in_id[i]);
	for(i = 0; i < b->nout; i++)
		fprintf(stream, " pr(out,%d)", b->out_id[i]);
	for(i = 0; i < b->npi; i++)
		fprintf(stream, " pr(0,%d)", i);	
	
}

int
gdl_tiling_hmm_backup_fprintf_empty (FILE * stream, const gdl_tiling_hmm_backup * b)
{
	size_t i;
	for(i = 0; i < b->ngauss_class; i++)
		fprintf(stream, " - -", i);	
	for(i = 0; i < b->nin; i++)
		fprintf(stream, " -", b->in_id[i]);
	for(i = 0; i < b->nout; i++)
		fprintf(stream, " -", b->out_id[i]);
	for(i = 0; i < b->npi; i++)
		fprintf(stream, " -", i);	
	
}

int
gdl_tiling_hmm_backup_fprintf (FILE * stream, const gdl_tiling_hmm_backup * b)
{
	size_t i;
	for(i = 0; i < b->ngauss_class; i++)
		fprintf(stream, " %g %g", b->mu[i], b->var[i]);	
	for(i = 0; i < b->nin; i++)
		fprintf(stream, " %1.10f", b->in_a[i]);
	for(i = 0; i < b->nout; i++)
		fprintf(stream, " %1.10f", b->out_a[i]);
	for(i = 0; i < b->npi; i++)
		fprintf(stream, " %1.10f", b->pi[i]);	
}

int
gdl_tiling_hmm_backup_fwrite (FILE * stream, const gdl_tiling_hmm_backup * b)
{
	if (stream && b)
	{
		int status;
		size_t i;
		
		status = fwrite (b->mu, sizeof(double), b->ngauss_class, stream);
		GDL_FWRITE_STATUS (status, b->ngauss_class);
		status = fwrite (b->var, sizeof(double), b->ngauss_class, stream);
		GDL_FWRITE_STATUS (status, b->ngauss_class);
		status = fwrite (b->out_a, sizeof(double), b->nout, stream);
		GDL_FWRITE_STATUS (status, b->nout);
		status = fwrite (b->in_a, sizeof(double), b->nin, stream);
		GDL_FWRITE_STATUS (status, b->nin);
		status = fwrite (b->pi, sizeof(double), b->npi, stream);
		GDL_FWRITE_STATUS (status, b->npi);
		
		return GDL_SUCCESS;
	}
	return GDL_EINVAL;
}

static const gdl_tiling_hmm_type _gdl_tiling_hmm_nucleosome =
{
	0,
	"hmm_nucleosome",
	&gdl_tiling_hmm_nucleosome_alloc,
	&gdl_tiling_hmm_nucleosome_out_b,
	6,
	8,
	9,
	1,
	&gdl_tiling_hmm_nucleosome_init_b,
	0,
	0
};

const gdl_tiling_hmm_type * gdl_tiling_hmm_nucleosome = &_gdl_tiling_hmm_nucleosome;
