/*  
 *  tiling_hmm.h
 * 
 *  $Author: gael $, $Date: 2008-09-15 12:32:59 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
 
#ifndef __GDL_TILING_HMM_H__
#define __GDL_TILING_HMM_H__

#include <math.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_randist.h>
#include <gdl/gdl_string.h>

#include "tiling_array.h"

__BEGIN_DECLS

typedef struct _gdl_tiling_hmm gdl_tiling_hmm;

typedef struct
{
	size_t size;
	size_t nobs;
	int * t;
	const gdl_tiling_chromosome * data;
} gdl_tiling_hmm_seq;

gdl_tiling_hmm_seq * gdl_tiling_hmm_seq_alloc1 (const gdl_tiling_chromosome * data, const size_t from, const size_t to);
gdl_tiling_hmm_seq * gdl_tiling_hmm_seq_alloc2 (const gdl_tiling_chromosome * data, const size_t from, const long tiling_size, const size_t probe_size);
gdl_tiling_hmm_seq * gdl_tiling_hmm_seq_alloc3 (const gdl_tiling_chromosome * data, const long from, const long tiling_size, const size_t window_size, size_t * p_cur);
void gdl_tiling_hmm_seq_free (gdl_tiling_hmm_seq * o);
size_t gdl_tiling_hmm_seq_size (const gdl_tiling_hmm_seq * o);
size_t gdl_tiling_hmm_seq_dim (const gdl_tiling_hmm_seq * o);
gdl_boolean gdl_tiling_hmm_seq_missing (const gdl_tiling_hmm_seq * o, size_t t);
double gdl_tiling_hmm_seq_get (const gdl_tiling_hmm_seq * o, const size_t t, const size_t d);
double * gdl_tiling_hmm_seq_vector (const gdl_tiling_hmm_seq * o, const size_t d);

typedef struct
{
  /** Classes of the sate */
  size_t class;
  /** Gaussian components of the sate */
  size_t * gauss_class;
  /** Has initial prob */
  gdl_boolean has_pi;
  /** Initial probability */
  double pi;
  /** Output proba */
  double (*out_b)(const gdl_tiling_hmm_seq * o, size_t t, const gdl_tiling_hmm * hmm, size_t * class);
  /** IDs of the following states */
  size_t * out_id;
  /** IDs of the previous states */
  size_t * in_id;
  /** transition probs to successor states */
  double * out_a;
  /** transition probs from predecessor states */
  double * in_a;
  /** Number of successor states */
  size_t out_states;
  /** Number of precursor states */
  size_t in_states;
  /** state emission prob. is fixed */
  gdl_boolean b_fix;
  /** state transition probs to successor are fixed */
  gdl_boolean out_fix;
  /** state transition probs to predecessor are fixed */
  gdl_boolean in_fix;
} gdl_tiling_hmm_state;

gdl_tiling_hmm_state * gdl_tiling_hmm_state_alloc (const size_t class, const size_t gauss_class[], const size_t ndim, const size_t in_states, const size_t out_states, double (*out_b)(const gdl_tiling_hmm_seq * o, size_t t, const gdl_tiling_hmm * hmm, size_t * class));
void gdl_tiling_hmm_state_free (gdl_tiling_hmm_state * state);
gdl_tiling_hmm_state * gdl_tiling_hmm_state_fread (FILE * stream, const size_t ndim, double (*out_b)(const gdl_tiling_hmm_seq * o, size_t t, const gdl_tiling_hmm * hmm, size_t * class));
int gdl_tiling_hmm_state_fwrite (FILE * stream, const gdl_tiling_hmm_state * state, const size_t ndim);

typedef struct
{
	gdl_boolean heterogeneous;
	gdl_string * name;
	int (*alloc)(const void * type, gdl_tiling_hmm * hmm);
	double (*out_b)(const gdl_tiling_hmm_seq * o, size_t t, const gdl_tiling_hmm * hmm, size_t * class);
	size_t nucleo_nprobe_min;
	size_t nucleo_nprobe_max;
	size_t nucleo_deloc_max;
	size_t ndim;
	int (*init_b)(gdl_tiling_hmm * hmm, const gdl_tiling_hmm_seq * seq, const gdl_rng * rng);
	int (*init_a)(gdl_tiling_hmm * hmm, const gdl_tiling_hmm_seq * seq, const gdl_rng * rng);
	int (*init_pi)(gdl_tiling_hmm * hmm, const gdl_tiling_hmm_seq * seq, const gdl_rng * rng);
} gdl_tiling_hmm_type;

gdl_tiling_hmm_type * gdl_tiling_hmm_type_alloc (const gdl_tiling_hmm_type * T);
void gdl_tiling_hmm_type_free (gdl_tiling_hmm_type * T);
gdl_tiling_hmm_type * gdl_tiling_hmm_type_fread (FILE * stream);
int gdl_tiling_hmm_type_fwrite (FILE * stream, const gdl_tiling_hmm_type * T);

struct _gdl_tiling_hmm 
{
	size_t nstate;
	size_t nclass;
	size_t ngauss_class;
	size_t **class2gauss;
	size_t ndim;
	double * mu;
	double * var;
	double ** alpha;
	double ** beta;
	double * scale;
	double log_p;
	gdl_tiling_hmm_type * T;
	gdl_tiling_hmm_state ** states;
	gdl_boolean update_b;
	gdl_boolean update_a;
};

gdl_tiling_hmm * gdl_tiling_hmm_alloc (gdl_tiling_hmm_type * T);
void gdl_tiling_hmm_free (gdl_tiling_hmm * hmm);
int gdl_tiling_hmm_util_alloc (gdl_tiling_hmm * hmm, const size_t T);
int gdl_tiling_hmm_util_free (gdl_tiling_hmm * hmm, const size_t T);
int gdl_tiling_hmm_forward (gdl_tiling_hmm * hmm, const gdl_tiling_hmm_seq * seq);
int gdl_tiling_hmm_backward (gdl_tiling_hmm * hmm, const gdl_tiling_hmm_seq * seq);
int gdl_tiling_hmm_baum_welch (gdl_tiling_hmm * hmm, const gdl_tiling_hmm_seq * seq, int max_step, double likelihood_delta);
int gdl_tiling_hmm_init (gdl_tiling_hmm * hmm, const gdl_tiling_hmm_seq * seq, const int gauss_path[]);
int * gdl_tiling_hmm_viterbi (const gdl_tiling_hmm * hmm, const gdl_tiling_hmm_seq * seq);
int gdl_tiling_hmm_forward_backward (gdl_tiling_hmm * hmm, const gdl_tiling_hmm_seq * seq, double *prob[]);
int gdl_tiling_hmm_smooth (const gdl_tiling_hmm * hmm, const double gamma[], double x[]);

gdl_tiling_hmm * gdl_tiling_hmm_fread (FILE * stream);
int gdl_tiling_hmm_fwrite (FILE * stream, const gdl_tiling_hmm * hmm);

typedef struct
{
	size_t ndim;
	size_t ngauss_class;
	size_t nout;
	size_t nin;
	size_t npi;
	size_t * out_id;
	size_t * in_id;
	double * mu;
	double * var;
	double * out_a;
	double * in_a;
	double * pi;
} gdl_tiling_hmm_backup;

gdl_tiling_hmm_backup * gdl_tiling_hmm_backup_alloc (const gdl_tiling_hmm * hmm);
void gdl_tiling_hmm_backup_free (gdl_tiling_hmm_backup * b);
int gdl_tiling_hmm_backup_force (gdl_tiling_hmm_backup * b, const gdl_tiling_hmm * hmm);
int gdl_tiling_hmm_backup_restore (gdl_tiling_hmm_backup * b, gdl_tiling_hmm * hmm);
int gdl_tiling_hmm_backup_fread (FILE * stream, gdl_tiling_hmm_backup * b);
int gdl_tiling_hmm_backup_fwrite (FILE * stream, const gdl_tiling_hmm_backup * b);
int gdl_tiling_hmm_backup_fprintf_header (FILE * stream, const gdl_tiling_hmm_backup * b);
int gdl_tiling_hmm_backup_fprintf_empty (FILE * stream, const gdl_tiling_hmm_backup * b);
int gdl_tiling_hmm_backup_fprintf (FILE * stream, const gdl_tiling_hmm_backup * b);

GDL_VAR const gdl_tiling_hmm_type * gdl_tiling_hmm_nucleosome;

__END_DECLS

#endif
