/*  
 *  tiling_nucleo.c
 * 
 *  $Author: gael $, $Date: 2008-09-15 12:32:58 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
 
#include <math.h>
#include <stdio.h>

#include <gdl/gdl_machine.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_list.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_statistics_double.h>

#include "tiling_hmm.h"
#include "tiling_nucleo.h"

/*************************
 * 
 * 
 * Private functions
 * 
 *  
 *************************/
//
// Add something to control the way the results are pushed into
// 0 - mean
// 1 - wmean
// 2 - best
// 
static int 
gdl_tiling_nucleo_hmm_window_push (gdl_tiling_nucleo_hmm * hmm)
{
	size_t p;
	
	gdl_tiling_hmm_backup_force (hmm->wbackup, hmm->whmm);
	for(p = 0; p < hmm->wsize; p++)
	{
		gdl_tiling_nucleo_hmm_probe_push (hmm, hmm->wprobes[p], hmm->wbackup, hmm->_wprobs[p], 1);
	}
}

static int 
gdl_tiling_nucleo_hmm_window_dump (FILE * stream, const gdl_tiling_nucleo_hmm * hmm, const gdl_boolean observed, const size_t mode)
{
	size_t i,j,p;
	gdl_tiling_nucleo_hmm_probe * tmp;
	gdl_tiling_nucleo_hmm_probe * probe = hmm->wprobes[0];
	
	if (!mode)
	{
		// txt output
		fprintf (stream, "%s %d %d", hmm->chrom->name, hmm->_pos_cur, observed);
		if (!probe->npush)
		{
			// print missing result for this probe
			for(i = 0; i < hmm->whmm->ngauss_class; i++)
				for(j = 0; j < hmm->whmm->ndim; j++)
					fprintf (stream, " - -");
			for(i = 0; i < hmm->nout+hmm->nin+hmm->npi; i++)
				fprintf (stream, " -");
		}
		else
		{
			// print + scale + reset
			for(i = 0; i < hmm->G; i++)
			{
				fprintf (stream, " %e %e", probe->mu[i]/probe->scale, probe->var[i]/probe->scale);
				probe->mu[i]=probe->var[i]=0;
			}
			for(i = 0; i < hmm->nout; i++)
			{
				fprintf (stream, " %1.5f", probe->out_a[i]/probe->scale);
				probe->out_a[i]=0;
			}
			for(i = 0; i < hmm->nin; i++)
			{
				fprintf (stream, " %1.5f", probe->in_a[i]/probe->scale);
				probe->in_a[i]=0;
			}
			for(i = 0; i < hmm->npi; i++)
			{
				fprintf (stream, " %1.5f", probe->pi[i]/probe->scale);
				probe->pi[i]=0;
			}
			probe->scale=0;
			probe->npush=0;
		}
		fprintf (stream, "\n");
	}
	else
	{
		// binary output
		int status;
		unsigned char missing = 'n';
		status = fwrite (&hmm->_pos_cur, sizeof(long), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite (&observed, sizeof(gdl_boolean), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		if (!probe->npush)
		{
			missing = 'y';
			status = fwrite (&missing, sizeof(unsigned char), 1, stream);
			GDL_FWRITE_STATUS (status, 1);
		}
		else
		{
			status = fwrite (&missing, sizeof(unsigned char), 1, stream);
			GDL_FWRITE_STATUS (status, 1);
			// dump the probs
			for(i = 0; i < hmm->whmm->nclass; i++)
				probe->gamma[i] /= probe->scale;
			status = fwrite (probe->gamma, sizeof(double), hmm->whmm->nclass, stream);
			GDL_FWRITE_STATUS (status, hmm->whmm->nclass);
			for(i = 0; i < hmm->whmm->nclass; i++)
				probe->gamma[i] = 0;
			// this is a modified piece of code from hmm_backup_fread to speed
			// up the dumping/reseting process of 'probes'...
			// if any changes appear in hmm_backup_fread function
			// this should be reported here, otherwise
			// the following restore function won't work :-(
			for(i = 0; i < hmm->G; i++)
			{
				probe->mu[i]  /= probe->scale;
				probe->var[i] /= probe->scale;
			}
			status = fwrite (probe->mu, sizeof(double), hmm->G, stream);
			GDL_FWRITE_STATUS (status, hmm->G);
			status = fwrite (probe->var, sizeof(double), hmm->G, stream);
			GDL_FWRITE_STATUS (status, hmm->G);
			for(i = 0; i < hmm->G; i++)
			{
				probe->mu[i]=probe->var[i]=0;
			}	
			for(i = 0; i < hmm->nout; i++)
				probe->out_a[i]/=probe->scale;
			status = fwrite (probe->out_a, sizeof(double), hmm->nout, stream);
			GDL_FWRITE_STATUS (status, hmm->nout);
			for(i = 0; i < hmm->nout; i++)
				probe->out_a[i]=0;
			for(i = 0; i < hmm->nin; i++)
				probe->in_a[i]/=probe->scale;	
			status = fwrite (probe->in_a, sizeof(double), hmm->nin, stream);
			GDL_FWRITE_STATUS (status, hmm->nin);
			for(i = 0; i < hmm->nin; i++)
				probe->in_a[i]=0;
			for(i = 0; i < hmm->npi; i++)
				probe->pi[i]/=probe->scale;
			status = fwrite (probe->pi, sizeof(double), hmm->npi, stream);
			GDL_FWRITE_STATUS (status, hmm->npi);
			for(i = 0; i < hmm->npi; i++)
				probe->pi[i]=0;
			probe->scale=0;
			probe->npush=0;
		}
	}
	fflush(stream);
	// shift
	for(p = 1; p < hmm->wsize; p++)
		hmm->wprobes[p-1]=hmm->wprobes[p];
	hmm->wprobes[p-1]=probe;
}

static int 
gdl_tiling_nucleo_hmm_full_dump (FILE * stream, gdl_tiling_nucleo_hmm * hmm, const gdl_tiling_hmm_seq * seq, const size_t mode)
{
	size_t t;
	
	for(t = 0; t < seq->size; t++)
   {
		hmm->_pos_cur += hmm->tsize;
		gdl_tiling_nucleo_hmm_window_dump (stream, hmm, gdl_tiling_hmm_seq_missing (seq, t), mode);
	}
}

static int 
gdl_tiling_nucleo_hmm_window_restore (FILE * stream, gdl_tiling_nucleo_hmm * hmm)
{
	int status;
	unsigned char missing;
	
	status = fread (&hmm->_pos_cur, sizeof(long), 1, stream);
	GDL_FREAD_STATUS (status, 1);
	status = fread (&hmm->_pos_observed, sizeof(gdl_boolean), 1, stream);
	GDL_FREAD_STATUS (status, 1);
	status = fread (&hmm->_pos_missing, sizeof(unsigned char), 1, stream);
	GDL_FREAD_STATUS (status, 1);
	if (hmm->_pos_missing=='n')
	{
		status = fread (hmm->_wprobs[0], sizeof(double), hmm->whmm->nclass, stream);
		GDL_FREAD_STATUS (status, hmm->whmm->nclass);
		status = gdl_tiling_hmm_backup_fread (stream, hmm->wbackup);
		GDL_FREAD_STATUS (status, GDL_SUCCESS);
		gdl_tiling_hmm_backup_restore (hmm->wbackup, hmm->whmm);
		//gdl_tiling_hmm_backup_fprintf (stdout, hmm->wbackup);
		//fprintf(stdout, "\n");
		//fflush(stdout);
	}
	
	return GDL_SUCCESS;
}

static int
gdl_tiling_nucleo_hmm_window_init (gdl_tiling_nucleo_hmm       * hmm,
		                             const gdl_tiling_hmm_seq    * window,
		                             const gdl_tiling_chromosome * atlas)
{
	long from = hmm->_pos_cur;
	long to   = from + hmm->psize - hmm->tsize;
	
	if (atlas && hmm->D==1)
	{
		const size_t P = atlas->nprobe;
		size_t p,pto,pfrom;
		
		p = (hmm->_atlas_cur == -1) ? 0 : hmm->_atlas_cur;
		for(;p < P && atlas->positions[p] != from; p++)
		pfrom=p;
		if (p == P)
		{
			if (hmm->_logger) fprintf(hmm->_logger, "gdl_tiling_nucleo_hmm: [ WARNING ] Unable to find starting probe position %d into the ATLAS\n", from);
			hmm->_gpath[0]=-1;
		}
		else
		{
			hmm->_atlas_cur=pfrom;
			for(;p < P && atlas->positions[p]!=to; p++);
			pto=p;
			if (pto == P)
			{
				if (hmm->_logger) fprintf(hmm->_logger, "gdl_tiling_nucleo_hmm: [ WARNING ] Unable to find ending probe position %d into the ATLAS\n", to);
				hmm->_gpath[0]=-1;
			}
			else if (window->size != pto-pfrom+1)
			{
				if (hmm->_logger) fprintf(hmm->_logger, "gdl_tiling_nucleo_hmm: [ WARNING ] The tiling step [ %d ] is different comparing to the ATLAS\n", hmm->tsize);
				hmm->_gpath[0]=-1;
			} 
			else
			{
				for (p = pfrom;p <= pto; p++)
				{
					int call = (int)GDL_TILING_CHROM_GET(atlas, p, 0);
					hmm->_gpath[p-pfrom] = (!call) ? 0 : 1;
				}
			}
		}
	}
	else
	{
		hmm->_gpath[0]=-1;
	}
	
	return gdl_tiling_hmm_init (hmm->whmm, window, hmm->_gpath);
}
/*************************
 * 
 * 
 * Public functions
 * 
 *  
 *************************/
 
gdl_tiling_nucleo_hmm_probe *
gdl_tiling_nucleo_hmm_probe_alloc (const gdl_tiling_nucleo_hmm * hmm)
{
	gdl_tiling_nucleo_hmm_probe * p;
	
	p = GDL_CALLOC (gdl_tiling_nucleo_hmm_probe, 1);
	
	p->mu  = GDL_CALLOC (double, hmm->G);
	p->var = GDL_CALLOC (double, hmm->G);
	p->out_a = GDL_CALLOC (double, hmm->nout);
	p->in_a  = GDL_CALLOC (double, hmm->nin);
	p->pi    = GDL_CALLOC (double, hmm->npi);
	p->gamma = GDL_CALLOC (double, hmm->whmm->nclass);
	
	return p;
}

void
gdl_tiling_nucleo_hmm_probe_free (gdl_tiling_nucleo_hmm_probe * p, const gdl_tiling_nucleo_hmm * hmm)
{
	if (p)
	{
		GDL_FREE (p->mu);
		GDL_FREE (p->var);
		GDL_FREE (p->in_a);
		GDL_FREE (p->out_a);
		GDL_FREE (p->pi);
		GDL_FREE (p->gamma);
		GDL_FREE (p);
	}
}

int
gdl_tiling_nucleo_hmm_probe_push (const gdl_tiling_nucleo_hmm * hmm, gdl_tiling_nucleo_hmm_probe * p, const gdl_tiling_hmm_backup * b, const double gamma[], const double w)
{
	size_t i,j;
	
	for(i = 0; i < b->ngauss_class; i++)
	{		
		p->mu[i]  += w*b->mu[i];
		p->var[i] += w*b->var[i];
	}
	for(i = 0; i < b->nin; i++)
		p->in_a[i] += w*b->in_a[i];
	for(i = 0; i < b->nout; i++)
		p->out_a[i] += w*b->out_a[i];
	for(i = 0; i < b->npi; i++)
		p->pi[i] += w*b->pi[i];
	
	for(i = 0; i < hmm->whmm->nclass; i++)
		p->gamma[i] += w*gamma[i];
	
	p->scale+=w;
	(p->npush)++;
	 	
	return GDL_SUCCESS;
}	

gdl_tiling_nucleo_hmm *
gdl_tiling_nucleo_hmm_alloc (const size_t ndim, const size_t nucleo_min, const size_t nucleo_max)
{
	gdl_tiling_nucleo_hmm * h;
	
	h = GDL_CALLOC (gdl_tiling_nucleo_hmm, 1);
	
	h->D = ndim;
	h->nucleo_min = nucleo_min;
	h->nucleo_max = nucleo_max;
	
	return h;
}

void
gdl_tiling_nucleo_hmm_free (gdl_tiling_nucleo_hmm * h)
{
	if (h)
	{
		size_t p;
		for(p = 0; p < h->wsize; p++)
			gdl_tiling_nucleo_hmm_probe_free (h->wprobes[p], h);
		GDL_FREE (h->wprobes);
		gdl_tiling_hmm_free (h->whmm);
		gdl_tiling_hmm_backup_free (h->wbackup);
		GDL_FREE (h->_gpath);
		GDL_FREE (h);	
	}
}

FILE *
gdl_tiling_nucleo_hmm_set_logger (gdl_tiling_nucleo_hmm * h, FILE * logger)
{
	FILE * old = h->_logger;
	h->_logger = logger;
	return old;	
}

int
gdl_tiling_nucleo_hmm_chrom_slide_init (gdl_tiling_nucleo_hmm * hmm, const gdl_tiling_chromosome * chrom, const size_t window_size, const size_t tiling)
{
	size_t p;
	// Clean wprobes
	if (hmm->wprobes)
		for(p = 0; p < hmm->wsize; p++)
			gdl_tiling_nucleo_hmm_probe_free (hmm->wprobes[p], hmm);
	GDL_FREE (hmm->wprobes);
	if (hmm->_wprobs)
		GDL_MATRIX_FREE (hmm->_wprobs, hmm->wsize);
	// Clean and init the window HMM
	gdl_tiling_hmm_free (hmm->whmm);
	gdl_tiling_hmm_backup_free (hmm->wbackup);
	gdl_tiling_hmm_type * T = gdl_tiling_hmm_type_alloc (gdl_tiling_hmm_nucleosome);
	T->ndim                 = hmm->D;
	T->nucleo_nprobe_min    = hmm->nucleo_min;
	T->nucleo_nprobe_max    = hmm->nucleo_max;
	T->nucleo_deloc_max     = hmm->nucleo_max+1;
	T->heterogeneous        =  hmm->heterogeneous;
	hmm->whmm    = gdl_tiling_hmm_alloc (T);
	hmm->wbackup = gdl_tiling_hmm_backup_alloc (hmm->whmm);
	hmm->G      = hmm->wbackup->ngauss_class;
	hmm->nin    = hmm->wbackup->nin;
	hmm->nout   = hmm->wbackup->nout;
	hmm->npi    = hmm->wbackup->npi;
	hmm->out_id = hmm->wbackup->out_id;
	hmm->in_id  = hmm->wbackup->in_id;
	// Init window parameters
	hmm->wsize  = window_size;
	hmm->tsize  = tiling;
	hmm->psize  = (hmm->wsize-1)*hmm->tsize;
	hmm->chrom  = chrom;
	// Init wprobes
	hmm->wprobes = GDL_MALLOC (gdl_tiling_nucleo_hmm_probe *, hmm->wsize);
	for(p = 0; p < hmm->wsize; p++)
			hmm->wprobes[p] = gdl_tiling_nucleo_hmm_probe_alloc (hmm);
	// Init wprobs
	hmm->_wprobs = GDL_MATRIX_ALLOC (double,hmm->wsize,hmm->whmm->nclass);
	// Some other stuffs...
	hmm->_p_cur     = 0;
	hmm->_pos_cur   = chrom->positions[0];
	hmm->_last_pos  = chrom->positions[chrom->nprobe-1];
	hmm->_atlas_cur = -1;
	GDL_FREE (hmm->_gpath);
	hmm->_gpath = GDL_CALLOC (int, hmm->wsize);
	// return the number of windows
	return (hmm->_last_pos-hmm->_pos_cur)/hmm->tsize;
}

int
gdl_tiling_nucleo_hmm_chrom_full_init (gdl_tiling_nucleo_hmm * hmm, const gdl_tiling_chromosome * chrom, const size_t tiling)
{
	size_t p;
	// Clean wprobes
	if (hmm->wprobes)
		for(p = 0; p < hmm->wsize; p++)
			gdl_tiling_nucleo_hmm_probe_free (hmm->wprobes[p], hmm);
	GDL_FREE (hmm->wprobes);
	if (hmm->_wprobs)
		GDL_MATRIX_FREE (hmm->_wprobs, hmm->wsize);
	// Clean and init the window HMM
	gdl_tiling_hmm_free (hmm->whmm);
	gdl_tiling_hmm_backup_free (hmm->wbackup);
	gdl_tiling_hmm_type * T = gdl_tiling_hmm_type_alloc (gdl_tiling_hmm_nucleosome);
	T->ndim                 = hmm->D;
	T->nucleo_nprobe_min    = hmm->nucleo_min;
	T->nucleo_nprobe_max    = hmm->nucleo_max;
	T->nucleo_deloc_max     = hmm->nucleo_max+1;
	T->heterogeneous        =  hmm->heterogeneous;
	hmm->whmm    = gdl_tiling_hmm_alloc (T);
	hmm->wbackup = gdl_tiling_hmm_backup_alloc (hmm->whmm);
	hmm->G      = hmm->wbackup->ngauss_class;
	hmm->nin    = hmm->wbackup->nin;
	hmm->nout   = hmm->wbackup->nout;
	hmm->npi    = hmm->wbackup->npi;
	hmm->out_id = hmm->wbackup->out_id;
	hmm->in_id  = hmm->wbackup->in_id;
	// Some other stuffs...
	hmm->_p_cur     = 0;
	hmm->_pos_cur   = chrom->positions[0];
	hmm->_last_pos  = chrom->positions[chrom->nprobe-1];
	hmm->_atlas_cur = -1;
	// Init window parameters
	hmm->wsize  = (hmm->_last_pos-hmm->_pos_cur)/tiling;
	hmm->tsize  = tiling;
	hmm->psize  = (hmm->wsize-1)*hmm->tsize;
	hmm->chrom  = chrom;
	// Init wprobes
	hmm->wprobes = GDL_MALLOC (gdl_tiling_nucleo_hmm_probe *, hmm->wsize);
	for(p = 0; p < hmm->wsize; p++)
			hmm->wprobes[p] = gdl_tiling_nucleo_hmm_probe_alloc (hmm);
	// Init wprobs
	hmm->_wprobs = GDL_MATRIX_ALLOC (double,hmm->wsize,hmm->whmm->nclass);
	
	GDL_FREE (hmm->_gpath);
	hmm->_gpath = GDL_CALLOC (int, hmm->wsize);
	// return the number of windows
	return 1;
}

int
gdl_tiling_nucleo_hmm_chrom_slide (gdl_tiling_nucleo_hmm * hmm,
                                   const gdl_tiling_chromosome * atlas,
                                   const size_t iter_max,
                                   const double epsilon,
                                   const double missing_thresh,
                                   FILE * stream,
                                   const size_t omode)

{
	if (hmm->_pos_cur > hmm->_last_pos) return GDL_SUCCESS;
	
	const size_t P = hmm->psize;
	const size_t T = hmm->tsize;
	const size_t W = hmm->wsize;
	int status;
	
   gdl_tiling_hmm_seq * window = gdl_tiling_hmm_seq_alloc3 (hmm->chrom, hmm->_pos_cur, T, W, &hmm->_p_cur);
	if (window && ((double)window->nobs) >= missing_thresh*((double)window->size))
   {
      status = gdl_tiling_nucleo_hmm_window_init (hmm, window, atlas);
	   if (status == GDL_SUCCESS)
	   {
	 	   status = gdl_tiling_hmm_baum_welch (hmm->whmm, window, iter_max, epsilon);
	 	   if (status != GDL_EUNDRFLW)
	 	   {
	 	   	gdl_tiling_hmm_forward_backward (hmm->whmm, window, hmm->_wprobs);
	 	   	gdl_tiling_nucleo_hmm_window_push (hmm);
	 	   }
	   }  
 	}
 	gdl_tiling_nucleo_hmm_window_dump (stream, hmm, gdl_tiling_hmm_seq_missing(window, 0), omode);
 	gdl_tiling_hmm_seq_free (window);
 	
	hmm->_pos_cur += T;
	if (hmm->_pos_cur > hmm->_last_pos) return GDL_SUCCESS;
	else return GDL_CONTINUE;
}

int
gdl_tiling_nucleo_hmm_chrom_full (gdl_tiling_nucleo_hmm * hmm,
                                  const gdl_tiling_chromosome * atlas,
                                  const size_t iter_max,
                                  const double epsilon,
                                  const double missing_thresh,
                                  FILE * stream,
                                  const size_t omode)

{
	if (hmm->_pos_cur > hmm->_last_pos) return GDL_SUCCESS;
	
	const size_t P = hmm->psize;
	const size_t T = hmm->tsize;
	const size_t W = hmm->wsize;
	int status;
	
   gdl_tiling_hmm_seq * window = gdl_tiling_hmm_seq_alloc3 (hmm->chrom, hmm->_pos_cur, T, W, &hmm->_p_cur);
	if (window && ((double)window->nobs) >= missing_thresh*((double)window->size))
   {
      status = gdl_tiling_nucleo_hmm_window_init (hmm, window, atlas);
	   if (status == GDL_SUCCESS)
	   {
	 	   status = gdl_tiling_hmm_baum_welch (hmm->whmm, window, iter_max, epsilon);
	 	   gdl_tiling_hmm_forward_backward (hmm->whmm, window, hmm->_wprobs);
	 	   gdl_tiling_nucleo_hmm_window_push (hmm);
	 	}  
 	}
 	gdl_tiling_nucleo_hmm_full_dump (stream, hmm, window, omode);
 	gdl_tiling_hmm_seq_free (window);
 	
	return GDL_SUCCESS;
}

/****************************************************
 * 
 * Whole-chromosome Viterbi algorithm
 * 
 *****************************************************/
 
/**
 * Here, we loop on the chromosome results and split the chromosome
 * into as many chunks as there are missing probe result (at a given probe)
 * or if the outlier_rm=true as there are probe results obviously out of 
 * the range...
 */
int
gdl_tiling_nucleo_hmm_chrom_to_chunks (gdl_tiling_nucleo_hmm * hmm, const chunk_min, const size_t remove_min, const gdl_boolean outlier_rm)
{
	const size_t T = hmm->T;
	size_t i,g,d,t;
	size_t nchunk = 0;
	gdl_tiling_nucleo_hmm_chunk * chunk;
	unsigned char * states;
	gdl_list * chunk_buffer;
	long offset;
	
	states = GDL_CALLOC (unsigned char, T);
	if (outlier_rm)
	{
		const size_t G = hmm->G;
		const size_t D = hmm->D;
		size_t nout=0;
		double m0,e,*x,*sx;
		
		offset = ftell (hmm->_stream);
		x = GDL_CALLOC (double, T);
		for(t = 0; t < T; t++)
		{
			// restore hmm parameter at probe t
			gdl_tiling_nucleo_hmm_window_restore (hmm->_stream, hmm);
			if (hmm->_pos_missing=='y')
				states[t]='m';
			else
				states[t]='y';
			// compute the mean/var of the delta between linker and nucleosome
			m0 = hmm->wbackup->mu[0];
			for(g = 1; g < G; g++)
			{
			   e = hmm->wbackup->mu[g]-m0;
			   x[t] += e;
			}
			x[t] /= (G-1);
		}
		// get outliers
		{
			double q1, q2, iq;
			
			sx = GDL_MALLOC (double, T);
			memcpy (sx, x, sizeof (double)*T);
			gdl_sort (sx, 1, T);
			q1 = gdl_stats_quantile_from_sorted_data (sx, 1, T, 0.25);
			q2 = gdl_stats_quantile_from_sorted_data (sx, 1, T, 0.75);
			iq = q2-q1;
			for (t = 0; t < T; t++)
			{
				if (states[t]=='m') continue;
				//if (x[t] > q2 + 3.5*iq || x[t] < q1 - 3.5*iq)
				if (x[t] < q1 - 3*iq)
				{
					states[t]='o';
					nout++;
				}
			}
			GDL_FREE (sx);
		}
		GDL_FREE (x);
		// goto the beginning of the file
		fseek (hmm->_stream, offset, SEEK_SET);
	}
	else
	{
		offset = ftell (hmm->_stream);
		for(t = 0; t < T; t++)
		{
			// restore hmm parameter at probe t
			gdl_tiling_nucleo_hmm_window_restore (hmm->_stream, hmm);
			if (hmm->_pos_missing=='y')
				states[t]='m';
			else
				states[t]='y';
		}
		// goto the beginning of the file
		fseek (hmm->_stream, offset, SEEK_SET);
	}
	double * pi;
	size_t t0,ts,te,nn;
	int tr=-1;
	long pos0,poss,pose;
	t0=0;
	
	chunk_buffer = gdl_list_alloc (gdl_list_default);
	offset = ftell (hmm->_stream);
	
	while(t0 < T)
	{
		for(t = t0; t < T && states[t] != 'y'; t++)
		{
			// restore hmm parameter at probe t
			gdl_tiling_nucleo_hmm_window_restore (hmm->_stream, hmm);
			if (t==t0) pos0 = hmm->_pos_cur;
		}
		tr = ts = t;
		if ((ts > t0 && ts-t0 > remove_min) && hmm->_logger)
		{
			fprintf (hmm->_logger, ">Remove chunk|chr=%s|from=%d|to=%d|\n",hmm->chrom->name,pos0,pos0+(ts-t0-1)*hmm->tsize);
			for(i = t0; i < ts; i++)
			{ 
				if (i>t0 && (i-t0)%50==0) fprintf (hmm->_logger, "\n"); 
				fprintf (hmm->_logger, "%c", states[i]);
			}
			fprintf (hmm->_logger,"\n");
		}
		else if (ts > t0 && ts-t0 <= remove_min)
		{
			tr = ts;
			ts = t0;
			poss = pos0;
		}
		
		if (ts == T) break;
		
		pi = GDL_CALLOC (double, hmm->npi);
		for(nn = 0, t = tr; t < T && states[t] == 'y'; t++, nn++)
		{
			gdl_tiling_nucleo_hmm_window_restore (hmm->_stream, hmm);
			if (t == ts) poss = hmm->_pos_cur;
			for(i = 0; i < hmm->npi; i++)
				pi[i] += hmm->wbackup->pi[i];
		}
		te = t-1;
		pose = hmm->_pos_cur;
		for(i = 0; i < hmm->npi; i++) pi[i] /= nn;
		
		if (te-ts+1 > chunk_min)
		{
			chunk = gdl_tiling_nucleo_hmm_chunk_alloc (ts, te, pi, hmm->npi);
			chunk->from_pos = poss;
			chunk->to_pos   = pose;
			gdl_list_push_back (chunk_buffer, chunk, 0);
			if (hmm->_logger)
				fprintf(hmm->_logger, ">Keep chunk|chr=%s|from=%d|to=%d|\n",hmm->chrom->name,chunk->from_pos,chunk->to_pos);
		}
		else
		{
			if (hmm->_logger)
				fprintf(hmm->_logger, ">Discard chunk|chr=%s|from=%d|to=%d|\n",hmm->chrom->name,poss,pose);
		}	
		GDL_FREE (pi);
		t0 = t;
	}
	GDL_FREE (states);
	// goto the beginning of the file
	fseek (hmm->_stream, offset, SEEK_SET);
	
	if ((nchunk=gdl_list_size (chunk_buffer)))
	{
		gdl_list_itr * itr = gdl_list_iterator_front (chunk_buffer);
		hmm->_nchunk = nchunk;
		hmm->_chunks = GDL_CALLOC (gdl_tiling_nucleo_hmm_chunk *, nchunk);
		i=0;
		do
		{
			hmm->_chunks[i++] = gdl_list_iterator_value (itr);
		}
		while(gdl_list_iterator_next (itr));
		gdl_list_iterator_free (itr);
	}
	gdl_list_free (chunk_buffer);
	
	// now, clusterize possible contiguous chunks...
	for(i = 1; i < nchunk; i++)
	{
		if (hmm->_chunks[i-1]->to+1==hmm->_chunks[i]->from)
		{
			if (hmm->_logger)
				fprintf(hmm->_logger, ">Merge chunks %d and %d into a single chunk\n", i, i+1);
			hmm->_chunks[i-1]->to     = hmm->_chunks[i]->to;
			hmm->_chunks[i-1]->to_pos = hmm->_chunks[i]->to_pos;
			hmm->_chunks[i-1]->size  += hmm->_chunks[i]->size;
			for(t = 0; t < hmm->npi; t++)
				hmm->_chunks[i-1]->pi[t] = 0.5*(hmm->_chunks[i-1]->pi[t]+hmm->_chunks[i]->pi[t]);
			gdl_tiling_nucleo_hmm_chunk_free (hmm->_chunks[i]);
			for(t = i; t < nchunk-1; t++)
				hmm->_chunks[t]=hmm->_chunks[t+1];
			nchunk--;
			i--;	
		}
	}
	hmm->_nchunk = nchunk;
	// END
}

int
gdl_tiling_nucleo_hmm_chrom_viterbi_start (gdl_tiling_nucleo_hmm * hmm, const gdl_tiling_chromosome * chrom, const size_t tiling, const size_t chunk_min, const size_t remove_min, const gdl_boolean outlier_rm, const gdl_string * wdir)
{
	hmm->chrom = chrom;
	// Clean and init the window HMM
	gdl_tiling_hmm_free (hmm->whmm);
	gdl_tiling_hmm_backup_free (hmm->wbackup);
	gdl_tiling_hmm_type * T = gdl_tiling_hmm_type_alloc (gdl_tiling_hmm_nucleosome);
	T->ndim                 = hmm->D;
	T->nucleo_nprobe_min    = hmm->nucleo_min;
	T->nucleo_nprobe_max    = hmm->nucleo_max;
	T->nucleo_deloc_max     = hmm->nucleo_max+1;
	T->heterogeneous        =  hmm->heterogeneous;
	hmm->whmm    = gdl_tiling_hmm_alloc (T);
	hmm->N       = hmm->whmm->nstate;
	hmm->wbackup = gdl_tiling_hmm_backup_alloc (hmm->whmm);
	hmm->G       = hmm->wbackup->ngauss_class;
	hmm->nin     = hmm->wbackup->nin;
	hmm->nout    = hmm->wbackup->nout;
	hmm->npi     = hmm->wbackup->npi;
	hmm->out_id  = hmm->wbackup->out_id;
	hmm->in_id   = hmm->wbackup->in_id;
	// Some other stuffs...
	hmm->tsize      = tiling;
	hmm->_p_cur     = 0;
	hmm->_pos_cur   = chrom->positions[0];
	hmm->_last_pos  = chrom->positions[chrom->nprobe-1];
	hmm->T          = (hmm->_last_pos-hmm->_pos_cur)/hmm->tsize;
	hmm->_wprobs    = GDL_MATRIX_ALLOC (double, 1, hmm->whmm->nclass);
	// open the stream
	gdl_string * file = gdl_string_sprintf ("%s/%s.nhw", wdir, chrom->name);
	hmm->_stream = gdl_fileopen (file, "r");
	hmm->_offset = ftell (hmm->_stream);
	gdl_string_free (file);
	// get the chunks and the pi's per chunk
	gdl_tiling_nucleo_hmm_chrom_to_chunks (hmm, chunk_min, remove_min, outlier_rm);
	return GDL_SUCCESS;
}

int
gdl_tiling_nucleo_hmm_chrom_viterbi_end (gdl_tiling_nucleo_hmm * hmm, const gdl_tiling_chromosome * chrom, const gdl_string * wdir)
{
	hmm->chrom = 0;
	// Clean and init the window HMM
	gdl_tiling_hmm_free (hmm->whmm);hmm->whmm=0;
	gdl_tiling_hmm_backup_free (hmm->wbackup);hmm->wbackup=0;
	hmm->out_id  = 0;
	hmm->in_id   = 0;
	// close the stream
	gdl_string * file = gdl_string_sprintf ("%s/%s.nhw", wdir, chrom->name);
	gdl_fileclose (file, hmm->_stream);
	gdl_string_free (file);
	// Free the chunks...
	size_t c;
	for(c = 0; c < hmm->_nchunk; c++)
		gdl_tiling_nucleo_hmm_chunk_free (hmm->_chunks[c]);
	GDL_FREE (hmm->_chunks);hmm->_chunks=0;
	return GDL_SUCCESS;
}


typedef struct 
{
  double **log_in_a;
  double *log_b;
  double *phi;
  double *phi_new;
  int **psi;
} gdl_tiling_nucleo_hmm_chrom_viterbi_store;

static gdl_tiling_nucleo_hmm_chrom_viterbi_store *
gdl_tiling_nucleo_hmm_chrom_viterbi_store_alloc (const gdl_tiling_nucleo_hmm * hmm, const size_t T)
{
  const size_t N = hmm->N;
  size_t j;
  gdl_tiling_nucleo_hmm_chrom_viterbi_store * v;
  
  v = GDL_CALLOC (gdl_tiling_nucleo_hmm_chrom_viterbi_store, 1);

  /* Allocate the log_in_a's -> individal lenghts */
  v->log_in_a = GDL_MALLOC (double *, N);
  for (j = 0; j < N; j++)
    v->log_in_a[j] = GDL_CALLOC (double, hmm->whmm->states[j]->in_states);
  v->log_b         = GDL_CALLOC (double, N);
  v->phi           = GDL_CALLOC (double, N);
  v->phi_new       = GDL_CALLOC (double, N);
  v->psi           = GDL_MATRIX_ALLOC (int, T, N);
  
  return (v);
}     

static void
gdl_tiling_nucleo_hmm_chrom_viterbi_store_free (gdl_tiling_nucleo_hmm_chrom_viterbi_store * v, const size_t N, const size_t T)
{
	if (v)
	{
		GDL_MATRIX_FREE (v->log_in_a, N);
		GDL_FREE (v->log_b);
  		GDL_FREE (v->phi);
  		GDL_FREE (v->phi_new);
  		GDL_MATRIX_FREE(v->psi, T);
		GDL_FREE (v);
	}	
}

static int
gdl_tiling_nucleo_hmm_chrom_viterbi_precompute (gdl_tiling_nucleo_hmm * hmm, gdl_tiling_nucleo_hmm_chunk * chunk, const size_t t, gdl_tiling_nucleo_hmm_chrom_viterbi_store * v)
{
  const size_t N = hmm->N; 
  double b_t;
  int i, j;
  
  if (hmm->_pos_missing == 'n')
  {
	  /* Precomputing the log(a_ij) */
	  for (j = 0; j < N; j++)
	  {
	  	for (i = 0; i < hmm->whmm->states[j]->in_states; i++)
	  	{
	  		if (hmm->whmm->states[j]->in_a[i] < GDL_DBL_EPSILON)
	        v->log_in_a[j][i] = +1; /* Not used any further in the calculations */
	      else
	        v->log_in_a[j][i] = log (hmm->whmm->states[j]->in_a[i]);
	  	}
	  }      
	  /* Precomputing the log(bj(ot)) */
	  for (j = 0; j < N; j++)
	  {
	    b_t = (*hmm->whmm->states[j]->out_b)(chunk->seq, t, hmm->whmm, hmm->whmm->states[j]->gauss_class);
	    if (gdl_isnan (b_t))
	    	 v->log_b[j] = +1;
	    else if (b_t <= GDL_DBL_EPSILON)
	       v->log_b[j] = 0;//+1;
	    else
	   	 v->log_b[j] = log (b_t);
	  }
	  return GDL_SUCCESS;
  }
  else
  		return GDL_CONTINUE;
} 
/****
 * First, get the chromosome chunks = f(missing, outlier(mu_nuc-mu_linker))
 * On each chunk, run the viterbi algorithm
 * Then output the results...
 */

int
gdl_tiling_nucleo_hmm_chrom_viterbi_perform (gdl_tiling_nucleo_hmm * hmm)
{
  const size_t D  = hmm->D;
  const size_t N  = hmm->N;
  const size_t C  = hmm->_nchunk;
  size_t c;
  double log_p;
  int t, j, i, k, St;
  double value, max_value;
  gdl_tiling_nucleo_hmm_chrom_viterbi_store * v;
  
  if (hmm->_logger)
  {
  	  fprintf (hmm->_logger, ">Viterbi|chr=%s|nchunk=%d|\n",hmm->chrom->name,C);
  	  fflush(hmm->_logger);
  }
  // loop on the chunks
  for(c = 0; c < C; c++)
  {
	  int len_path;
	  int cur_len_path = 0; /* the current length of the viterbi path */ 
	  int lastemState;
	  int state_seq_index;
	  int * state_seq;
	  gdl_tiling_nucleo_hmm_chunk * chunk;
	  
	  chunk    = hmm->_chunks[c];
	  len_path = chunk->size;
	  
	  if (hmm->_logger)
	  {
		 fprintf (hmm->_logger, " Chunk %d|from=%d|to=%d|size=%d\n", c+1, chunk->from_pos, chunk->to_pos, chunk->size);
		 fflush(hmm->_logger);
	  }
	  v = gdl_tiling_nucleo_hmm_chrom_viterbi_store_alloc (hmm, len_path);
	  /* init the chunk sequence */
	  gdl_tiling_nucleo_hmm_chunk_start_hmm (hmm, chunk);
	  
	  /* allocating state_seq array */
	  state_seq = GDL_MALLOC (int, len_path);
	  /* initialization of state_seq with -1 */
	  for (i = 0; i < len_path; i++) state_seq[i] = -1;
	  /* Initialization, that is t = 0 */
	  /* Precomputing the log(a_ij) and log(bj(ot)) */
          /* the hmm has been correclty positioned with gdl_tiling_nucleo_hmm_chunk_start_hmm (hmm, chunk) */
	  gdl_tiling_nucleo_hmm_chrom_viterbi_precompute (hmm, chunk, 0, v);
	  for (j = 0; j < N; j++)
	  {
	  	 if (hmm->whmm->states[j]->pi == 0.0 || v->log_b[j] == +1)
	      v->phi[j] = +1;
	    else
	      v->phi[j] = log (hmm->whmm->states[j]->pi) + v->log_b[j];
	  }
	  /* t > 0 */
	  const size_t T = chunk->size;
  	  for (t = 1; t < T; t++)
	  {
              // load whmm at position t...
              gdl_tiling_nucleo_hmm_window_restore (hmm->_stream, hmm);
	      gdl_tiling_nucleo_hmm_chrom_viterbi_precompute (hmm, chunk, t, v);
	    for (j = 0; j < N; j++)
	    {
			/** initialization of phi, psi **/
	      v->phi_new[j] = +1;
	      v->psi[t][j]  = -1;
	    }
	    for (k = 0; k < N; k++)
	    {
	    	/* Determine the maximum */
	      /* max_phi = phi[i] + log_in_a[j][i] ... */
	      St = k;
	      max_value     = -GDL_DBL_MAX;
	      v->psi[t][St] = -1;
	      for (i = 0; i < hmm->whmm->states[St]->in_states; i++)
	      {
	      	if (v->phi[hmm->whmm->states[St]->in_id[i]] != +1 && v->log_in_a[St][i] != +1)
	      	{
		         value = v->phi[hmm->whmm->states[St]->in_id[i]] + v->log_in_a[St][i];
		         if (value > max_value)
		         {
		           max_value = value;
		           v->psi[t][St] = hmm->whmm->states[St]->in_id[i];
	            }
	      	}
	      }
	      /* No maximum found (that is, state never reached)
	         or the output O[t] = 0.0: */
              if (max_value == -GDL_DBL_MAX || v->log_b[St] == +1)
	      {
	      	v->phi_new[St] = +1;
	      }
	      else
	      {
	      	v->phi_new[St] = max_value + v->log_b[St];
	      }
	    } /* complete time step for emitting states */
		 /* First now replace the old phi with the new phi */
	    for (j = 0; j < N; j++)
	    {
	    	 v->phi[j] = v->phi_new[j];	    	 
	    }
	  } 
	  /* Termination */
	  state_seq_index = len_path - 1;
	  max_value       = -GDL_DBL_MAX;
	  state_seq[state_seq_index] = -1;
	  for (j = 0; j < N; j++)
	  {
	    if (v->phi[j] != +1 && v->phi[j] > max_value)
	    {
	      max_value = v->phi[j];
	      state_seq[state_seq_index] = j;
	    }
	  }  
	  if (max_value == -GDL_DBL_MAX)
	  {
	    log_p = +1;
	  }
	  else
	  {
	      log_p = max_value;
	      lastemState = state_seq[len_path - 1];
	      for (t = T - 2, i = len_path - 2; t >= 0; t--)
	      {
	        state_seq[i--] = v->psi[t + 1][lastemState];
	        lastemState = v->psi[t + 1][lastemState];
	      }  
	   }
	   for(t = 0; t < len_path; t++)
           {
     		if (state_seq[t]!=-1)
                {
	    	   state_seq[t] = hmm->whmm->states[state_seq[t]]->class;
                }
	   }
           chunk->viterbi = state_seq;
	   /* Free the memory space */
	   gdl_tiling_nucleo_hmm_chrom_viterbi_store_free (v, N, T);
	   gdl_tiling_nucleo_hmm_chunk_end_hmm (chunk);
  }
  
  // goto the beginning of the file
  fseek (hmm->_stream, hmm->_offset, SEEK_SET);
 
  return GDL_SUCCESS;
}

int
gdl_tiling_nucleo_hmm_chrom_viterbi_summarize (FILE * stream, gdl_tiling_nucleo_hmm * hmm, const gdl_tiling_chromosome * atlas)
{
	const size_t C = hmm->whmm->nclass;
	const size_t D = hmm->D;
	size_t c,d,p,pa,t,PA=0;
	long cur;
	double *sx;
	gdl_tiling_nucleo_hmm_chunk * chunk;
	
	sx = GDL_CALLOC (double, D);
	
	gdl_tiling_nucleo_hmm_window_restore (hmm->_stream, hmm);
	cur = hmm->_pos_cur;
	if (atlas)
	{
		PA = atlas->nprobe;
		for(pa = 0;pa < PA && atlas->positions[pa] != cur; pa++)
		if (pa == PA)
		{
			if (hmm->_logger) 
				fprintf(hmm->_logger, "gdl_tiling_nucleo_hmm: [ WARNING ] Unable to find probe position %d into the ATLAS\n", cur);
		}
		fprintf (stream, "chromosome probe position virtual ignore chunk calling atlas");
	}
	else
		fprintf (stream, "chromosome probe position virtual ignore chunk calling");
	for(d = 0; d < D; d++) fprintf (stream, " s%d", d+1);
	for(d = 0; d < C; d++) fprintf (stream, " p%d", d+1);
	fprintf (stream, "\n");
	for(p = c = 0; c < hmm->_nchunk; c++)
	{
		chunk = hmm->_chunks[c];
		if (chunk->from_pos > cur)
		{
			while(cur < chunk->from_pos)
			{
				if (atlas)
					if (pa < PA)
						fprintf (stream, "%s %d %ld %d %c - - %d", hmm->chrom->name, p+1, hmm->_pos_cur, hmm->_pos_observed, hmm->_pos_missing, (size_t)atlas->intensities[pa]);
					else
						fprintf (stream, "%s %d %ld %d %c - - -", hmm->chrom->name, p+1, hmm->_pos_cur, hmm->_pos_observed, hmm->_pos_missing);
				else
					fprintf (stream, "%s %d %ld %d %c - -", hmm->chrom->name, p+1, hmm->_pos_cur, hmm->_pos_observed, hmm->_pos_missing);
				for(d = 0; d < D+C; d++) fprintf (stream, " -");
				fprintf (stream, "\n");
				gdl_tiling_nucleo_hmm_window_restore (hmm->_stream, hmm);
				cur = hmm->_pos_cur;
				p++;pa++;
			}
		}
		t = 0;
		while(cur <= chunk->to_pos)
		{
			gdl_tiling_hmm_smooth (hmm->whmm, hmm->_wprobs[0], sx);
			if (atlas)
			{
				if (pa < PA)
					fprintf (stream, "%s %d %ld %d %c %d %d %d", hmm->chrom->name, p+1, hmm->_pos_cur, hmm->_pos_observed, hmm->_pos_missing, c+1, chunk->viterbi[t], (size_t)atlas->intensities[pa]);
				else
					fprintf (stream, "%s %d %ld %d %c %d %d -", hmm->chrom->name, p+1, hmm->_pos_cur, hmm->_pos_observed, hmm->_pos_missing, c+1, chunk->viterbi[t]);
			}
			else
			{
				fprintf (stream, "%s %d %ld %d %c %d %d", hmm->chrom->name, p+1, hmm->_pos_cur, hmm->_pos_observed, hmm->_pos_missing, c+1, chunk->viterbi[t]);
			}
			for(d = 0; d < D; d++)
				fprintf (stream, " %g", sx[d]);
			for(d = 0; d < C; d++)
				fprintf (stream, " %1.3f", hmm->_wprobs[0][d]);
			fprintf (stream, "\n");
			fflush(stream);
			gdl_tiling_nucleo_hmm_window_restore (hmm->_stream, hmm);
			cur = hmm->_pos_cur;
			p++;pa++;
			t++;
		}
	}
	// goto the beginning of the file
   fseek (hmm->_stream, hmm->_offset, SEEK_SET);
   
   GDL_FREE (sx);
}

int
gdl_tiling_nucleo_hmm_chrom_viterbi_fprintf (FILE * stream, gdl_tiling_nucleo_hmm * hmm)
{
	const size_t C = hmm->whmm->nclass;
	const size_t D = hmm->D;
	size_t c,d,p,pa,t,PA=0;
	long cur;
	gdl_tiling_nucleo_hmm_chunk * chunk;
	
	gdl_tiling_nucleo_hmm_window_restore (hmm->_stream, hmm);
	cur = hmm->_pos_cur;
	
	fprintf (stream, "chromosome probe position virtual ignore chunk");
	gdl_tiling_hmm_backup_fprintf_header (stream, hmm->wbackup);
	fprintf (stream, "\n");
	for(p = c = 0; c < hmm->_nchunk; c++)
	{
		chunk = hmm->_chunks[c];
		if (chunk->from_pos > cur)
		{
			while(cur != chunk->from_pos)
			{
				fprintf (stream, "%s %d %ld %d %c -", hmm->chrom->name, p+1, hmm->_pos_cur, hmm->_pos_observed, hmm->_pos_missing);
				gdl_tiling_hmm_backup_fprintf_empty (stream, hmm->wbackup);
				fprintf (stream, "\n");
				gdl_tiling_nucleo_hmm_window_restore (hmm->_stream, hmm);
				cur = hmm->_pos_cur;
				p++;pa++;
			}
		}
		t = 0;
		while(cur <= chunk->to_pos)
		{
			fprintf (stream, "%s %d %ld %d %c %d", hmm->chrom->name, p+1, hmm->_pos_cur, hmm->_pos_observed, hmm->_pos_missing, c+1);
			gdl_tiling_hmm_backup_fprintf (stream, hmm->wbackup);
			fprintf (stream, "\n");
			gdl_tiling_nucleo_hmm_window_restore (hmm->_stream, hmm);
			cur = hmm->_pos_cur;
			p++;pa++;
			t++;
		}
	}
	// goto the beginning of the file
   fseek (hmm->_stream, hmm->_offset, SEEK_SET);
}

gdl_tiling_nucleo_hmm_chunk *
gdl_tiling_nucleo_hmm_chunk_alloc (const size_t from, const size_t to, const double pi[], const size_t npi)
{
	gdl_tiling_nucleo_hmm_chunk * c;
	
	c = GDL_CALLOC (gdl_tiling_nucleo_hmm_chunk, 1);
	
	c->size = to-from+1;
	c->from = from;
	c->to   = to;
	c->pi   = GDL_MALLOC (double, npi);
	memcpy (c->pi, pi, sizeof(double)*npi);
	
	return c;
}

void
gdl_tiling_nucleo_hmm_chunk_free (gdl_tiling_nucleo_hmm_chunk * c)
{
	if (c)
	{
		GDL_FREE (c->pi);
		gdl_tiling_hmm_seq_free (c->seq);
		GDL_FREE (c);
	}
}

int
gdl_tiling_nucleo_hmm_chunk_start_hmm (gdl_tiling_nucleo_hmm * hmm, gdl_tiling_nucleo_hmm_chunk * c)
{
	size_t i,p=0;
        long cur;
	// first create the probe sequence for the HMM
	c->seq = gdl_tiling_hmm_seq_alloc3 (hmm->chrom, c->from_pos, hmm->tsize, c->size, &p);
	// Then set the pi of the HMM
	for(p = i = 0; i < hmm->N; i++)
		if (hmm->whmm->states[i]->has_pi)
			hmm->whmm->states[i]->pi = c->pi[p++];
        // position the hmm at the starting position of the chunk
        if (c->from_pos > cur)
        {
           while(cur < c->from_pos)
           {
              gdl_tiling_nucleo_hmm_window_restore (hmm->_stream, hmm);
              cur = hmm->_pos_cur;
           }
        }
}

int
gdl_tiling_nucleo_hmm_chunk_end_hmm (gdl_tiling_nucleo_hmm_chunk * c)
{
	gdl_tiling_hmm_seq_free (c->seq);
	c->seq=0;
}

/**
 *  HMM final output
 */

gdl_tiling_nucleo_probe *
gdl_tiling_nucleo_probe_alloc (const size_t D, const size_t C)
{
	gdl_tiling_nucleo_probe * p;
	
	p = GDL_CALLOC (gdl_tiling_nucleo_probe, 1);
	
	p->smooth = GDL_CALLOC (double, D);
	p->prob   = GDL_CALLOC (double, C);
	
	return p;
}

void
gdl_tiling_nucleo_probe_free (gdl_tiling_nucleo_probe * p)
{
	if (p)
	{
		GDL_FREE (p->smooth);
		GDL_FREE (p->prob);
		GDL_FREE (p);	
	}
}

gdl_tiling_nucleo_chunk *
gdl_tiling_nucleo_chunk_alloc (const size_t size)
{
	gdl_tiling_nucleo_chunk * c;
	
	c = GDL_CALLOC (gdl_tiling_nucleo_chunk, 1);
	
	c->size = size;
	c->probes = GDL_CALLOC (gdl_tiling_nucleo_probe *, size);
	
	return c;
}

void
gdl_tiling_nucleo_chunk_free (gdl_tiling_nucleo_chunk * c)
{
	if (c)
	{
		size_t i;
		for(i = 0; i < c->size; i++)
			gdl_tiling_nucleo_probe_free (c->probes[i]);
		GDL_FREE (c->probes);
		GDL_FREE (c);
	}
}

void
gdl_tiling_nucleo_chunk_init (gdl_tiling_nucleo_chunk * c)
{
	if (!c)
		return;
	
	c->from     = c->probes[0]->idx;
	c->to       = c->probes[c->size-1]->idx;
	c->from_pos = c->probes[0]->position;
	c->to_pos   = c->probes[c->size-1]->position;
}

gdl_tiling_nucleo_chromosome *
gdl_tiling_nucleo_chromosome_alloc (const gdl_string * name, const size_t D, const size_t C, const size_t size)
{
	gdl_tiling_nucleo_chromosome * c;
	
	c = GDL_CALLOC (gdl_tiling_nucleo_chromosome, 1);
	
	c->name = gdl_string_clone (name);
	c->D = D;
	c->C = C;
	c->size = size;
	c->chunks = GDL_CALLOC (gdl_tiling_nucleo_chunk *, size);
	
	return c;
}

void
gdl_tiling_nucleo_chromosome_free (gdl_tiling_nucleo_chromosome * c)
{
	if (c)
	{
		size_t i;
		for (i = 0; i < c->size; i++)
			gdl_tiling_nucleo_chunk_free (c->chunks[i]);
		GDL_FREE (c->chunks);
		GDL_FREE (c->name);
		GDL_FREE (c);	
	}
}

gdl_tiling_nucleo_chromosome *
gdl_tiling_nucleo_chromosome_fscanf (FILE * stream, const size_t D, const size_t C)
{
	int cur = -1;
	size_t n,ti,tj,d,c,nchunk=0;
	gdl_string * line=0, * tok, *chr=0;
	gdl_tiling_nucleo_chromosome * chrom = 0;
	gdl_tiling_nucleo_probe * probe;
	gdl_list * probes = gdl_list_alloc (gdl_list_default);
	
	// skip the header
	gdl_getline (&line, &n, stream);
	gdl_string_free (line);line=0;
	
	while(gdl_getline (&line, &n, stream)!=-1)
	{
		ti=tj=0;
		// chrom name
		tok=gdl_string_next_token (line, n, &ti, &tj);
		if (!chr) chr = tok;
		else gdl_string_free (tok);
		probe =  gdl_tiling_nucleo_probe_alloc (D, C);
		// probe idx
		tok=gdl_string_next_token (line, n, &ti, &tj);
		probe->idx = atoi (tok);
		gdl_string_free (tok);
		// probe position
		tok=gdl_string_next_token (line, n, &ti, &tj);
		probe->position = atol (tok);
		gdl_string_free (tok);
		// probe observed status
		tok=gdl_string_next_token (line, n, &ti, &tj);
		probe->observed = (gdl_boolean)atoi(tok);
		gdl_string_free (tok);
		// probe missing status
		tok=gdl_string_next_token (line, n, &ti, &tj);
		probe->missing = tok[0];
		gdl_string_free (tok);
		// chunk
		tok = gdl_string_next_token (line, n, &ti, &tj);
		probe->chunk_idx = atoi(tok)-1;
		if (probe->chunk_idx>=0 && probe->chunk_idx != cur)
		{
			nchunk++;
			cur=probe->chunk_idx;
		}
		gdl_string_free (tok);
		// calling
		tok=gdl_string_next_token (line, n, &ti, &tj);
		probe->calling = atoi (tok);
		gdl_string_free (tok);
		// smooth -->D
		for(d = 0; d < D; d++)
		{
			tok=gdl_string_next_token (line, n, &ti, &tj);
			probe->smooth[d] = (double) atof (tok);
			gdl_string_free (tok);
		}
		// prob. -->C
		for(c = 0; c < C; c++)
		{
			tok=gdl_string_next_token (line, n, &ti, &tj);
			probe->prob[c] = (double) atof (tok);
			gdl_string_free (tok);
		}		
		gdl_list_push_back (probes, probe, 0);
		gdl_string_free (line);line=0;
	}
	
	if (gdl_list_size (probes))
	{
		gdl_list_itr * itr;
		chrom = gdl_tiling_nucleo_chromosome_alloc (chr, D, C, nchunk);
		gdl_string_free (chr);
		
		// Count the number if probes per chunk
		size_t * nprobes = GDL_CALLOC (size_t, nchunk);
		itr = gdl_list_iterator_front (probes);
		do
		{
			probe = (gdl_tiling_nucleo_probe *) gdl_list_iterator_value (itr);
			if (probe->chunk_idx>=0)
				nprobes[probe->chunk_idx]++;
		}
		while (gdl_list_iterator_next (itr));
		gdl_list_iterator_free (itr);
		// Init the chunks
		for(c = 0; c < nchunk; c++)
			chrom->chunks[c] = gdl_tiling_nucleo_chunk_alloc (nprobes[c]);	
		GDL_FREE (nprobes);
		nprobes = GDL_CALLOC (size_t, nchunk);
		itr = gdl_list_iterator_front (probes);
		do
		{
			probe = (gdl_tiling_nucleo_probe *) gdl_list_iterator_value (itr);
			if (probe->chunk_idx>=0)
			{
				probe->idx = nprobes[probe->chunk_idx];
				chrom->chunks[probe->chunk_idx]->probes[nprobes[probe->chunk_idx]] = probe;
				nprobes[probe->chunk_idx]++;
			}
			else
			{
				gdl_tiling_nucleo_probe_free (probe);
			}
		}
		while (gdl_list_iterator_next (itr));
		gdl_list_iterator_free (itr);
		GDL_FREE (nprobes);
		for(c = 0; c < nchunk; c++)
	   	gdl_tiling_nucleo_chunk_init (chrom->chunks[c]);	
	}
	
	gdl_list_free (probes);
	
	return chrom;
}

