/*  
 *  tiling_nucleo.h
 * 
 *  $Author: gael $, $Date: 2008-09-15 12:32:58 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
 
#ifndef __GDL_TILING_NUCLEO_H__
#define __GDL_TILING_NUCLEO_H__

#include <math.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_version.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_randist.h>
#include <gdl/gdl_string.h>

#include "tiling_array.h"
#include "tiling_hmm.h"

__BEGIN_DECLS

typedef struct _gdl_tiling_nucleo_hmm gdl_tiling_nucleo_hmm;

typedef struct
{
	size_t npush;
	double * mu;
	double * var;
	double * out_a;
	double * in_a;
	double * pi;
	double * gamma;
	double scale;
} gdl_tiling_nucleo_hmm_probe;

gdl_tiling_nucleo_hmm_probe * gdl_tiling_nucleo_hmm_probe_alloc (const gdl_tiling_nucleo_hmm * hmm);
void gdl_tiling_nucleo_hmm_probe_free (gdl_tiling_nucleo_hmm_probe * p, const gdl_tiling_nucleo_hmm * hmm);
int gdl_tiling_nucleo_hmm_probe_push (const gdl_tiling_nucleo_hmm * hmm, gdl_tiling_nucleo_hmm_probe * p, const gdl_tiling_hmm_backup * b, const double probs[], const double w);

typedef struct
{
	size_t size;
	size_t from;
	size_t to;
	long from_pos;
	long to_pos;
	double * pi;
	int * viterbi;
	gdl_tiling_hmm_seq * seq;
} gdl_tiling_nucleo_hmm_chunk;

gdl_tiling_nucleo_hmm_chunk * gdl_tiling_nucleo_hmm_chunk_alloc (const size_t from, const size_t to, const double pi[], const size_t npi);
void gdl_tiling_nucleo_hmm_chunk_free (gdl_tiling_nucleo_hmm_chunk * c);
int gdl_tiling_nucleo_hmm_chunk_start_hmm (gdl_tiling_nucleo_hmm * hmm, gdl_tiling_nucleo_hmm_chunk * c);
int gdl_tiling_nucleo_hmm_chunk_end_hmm (gdl_tiling_nucleo_hmm_chunk * c);

struct _gdl_tiling_nucleo_hmm
{
	gdl_boolean heterogeneous;
	size_t T;
	size_t D;
	size_t G;
	size_t C;
	size_t N;
	size_t nucleo_min;
	size_t nucleo_max;
	long   psize;
	size_t tsize;
	size_t wsize;
	size_t nout;
	size_t nin;
	size_t npi;
	size_t * out_id;
	size_t * in_id;
	gdl_tiling_hmm        * whmm;
	gdl_tiling_hmm_backup * wbackup;
	gdl_tiling_nucleo_hmm_probe ** wprobes;
	const gdl_tiling_chromosome * chrom;
	// Utilities
	double      **_wprobs;
	int         * _gpath;
	int           _atlas_cur;
	size_t        _p_cur;
	gdl_boolean   _pos_observed;
	unsigned char _pos_missing;
	long          _pos_cur;
	long          _last_pos;
	size_t        _nchunk;
	gdl_tiling_nucleo_hmm_chunk ** _chunks;
	// IO
	long   _offset;
	FILE * _stream;
	FILE * _logger;
};

gdl_tiling_nucleo_hmm * gdl_tiling_nucleo_hmm_alloc (const size_t ndim, const size_t nucleo_min, const size_t nucleo_max);
void gdl_tiling_nucleo_hmm_free (gdl_tiling_nucleo_hmm * h);
FILE * gdl_tiling_nucleo_hmm_set_logger (gdl_tiling_nucleo_hmm * h, FILE * logger);

int gdl_tiling_nucleo_hmm_chrom_slide_init (gdl_tiling_nucleo_hmm * h, const gdl_tiling_chromosome * chrom, const size_t window_size, const size_t tiling);
int gdl_tiling_nucleo_hmm_chrom_slide (gdl_tiling_nucleo_hmm * h,
                                       const gdl_tiling_chromosome * atlas,
                                       const size_t iter_max,
                                       const double epsilon,
                                       const double missing_thresh,
                                       FILE * stream,
                                       const size_t omode);
int gdl_tiling_nucleo_hmm_chrom_full_init (gdl_tiling_nucleo_hmm * h, const gdl_tiling_chromosome * chrom, const size_t tiling);
int gdl_tiling_nucleo_hmm_chrom_full (gdl_tiling_nucleo_hmm * h,
                                       const gdl_tiling_chromosome * atlas,
                                       const size_t iter_max,
                                       const double epsilon,
                                       const double missing_thresh,
                                       FILE * stream,
                                       const size_t omode);                                       

int gdl_tiling_nucleo_hmm_chrom_viterbi_start (gdl_tiling_nucleo_hmm * h, const gdl_tiling_chromosome * chrom, const size_t tiling, const size_t chunk_min, const size_t remove_min, const gdl_boolean outlier_rm, const gdl_string * wdir);
int gdl_tiling_nucleo_hmm_chrom_viterbi_perform (gdl_tiling_nucleo_hmm * hmm);
int gdl_tiling_nucleo_hmm_chrom_viterbi_summarize (FILE * stream, gdl_tiling_nucleo_hmm * hmm, const gdl_tiling_chromosome * atlas);                                       
int gdl_tiling_nucleo_hmm_chrom_viterbi_end (gdl_tiling_nucleo_hmm * h, const gdl_tiling_chromosome * chrom, const gdl_string * wdir);

typedef struct
{
	size_t        idx;
	gdl_boolean   eligible;
	int           link_idx;
	int           chunk_idx;
	long          position;
	gdl_boolean   observed;
	unsigned char missing;
	size_t calling;
	double * smooth;
	double * prob;
	double pval;
	void   * extra;
	void   * align;
	void   * chrom;
} gdl_tiling_nucleo_probe;

gdl_tiling_nucleo_probe * gdl_tiling_nucleo_probe_alloc (const size_t D, const size_t C);
void gdl_tiling_nucleo_probe_free (gdl_tiling_nucleo_probe * p);

typedef struct
{
	size_t size;
	size_t from;
	size_t to;
	long   from_pos;
	long   to_pos;
	gdl_tiling_nucleo_probe ** probes;
} gdl_tiling_nucleo_chunk;

gdl_tiling_nucleo_chunk * gdl_tiling_nucleo_chunk_alloc (const size_t size);
void gdl_tiling_nucleo_chunk_free (gdl_tiling_nucleo_chunk * c);
void gdl_tiling_nucleo_chunk_init (gdl_tiling_nucleo_chunk * c);

typedef struct
{
	size_t D;
	size_t C;
	gdl_string * name;
	size_t size;
	gdl_tiling_nucleo_chunk ** chunks;
} gdl_tiling_nucleo_chromosome;

gdl_tiling_nucleo_chromosome * gdl_tiling_nucleo_chromosome_alloc (const gdl_string * name, const size_t D, const size_t C, const size_t size);
void gdl_tiling_nucleo_chromosome_free (gdl_tiling_nucleo_chromosome * c);

typedef struct
{
	gdl_string * name;
	size_t size;
	gdl_tiling_nucleo_chromosome ** chroms;
} gdl_tiling_nucleo_genome;
                                        
gdl_tiling_nucleo_chromosome * gdl_tiling_nucleo_chromosome_fscanf (FILE * stream, const size_t D, const size_t C);
                                        
__END_DECLS

#endif
