/*  
 *  util.c
 * 
 *  $Author: gael $, $Date: 2008-09-15 12:33:00 $, $Version$
 *
 *  NucleoMiner : a C package for nucleosome mapping and SNEP detection.
 *
 *  Reference:  Nagarajan et al. PLoS Genetics 2010, 6(4): e1000913. Natural Single-Nucleosome Epi-Polymorphisms in Yeast
 * 
 *  Copyright (C) 2010  Jean-Baptiste Veyrieras, Gael Yvert, CNRS/ENS-Lyon, France.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */

#include <gdl/gdl_io.h>
#include <gdl/gdl_runtime.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_gfeatures.h>
#include <gdl/gdl_gfeatures_gff3.h>

#include "tiling_array.h"
#include "tiling_hmm.h"
#include "tiling_nucleo.h"
#include "align_nucleo.h"


static gdl_runtime * 
runtime_setup (void)
{
	gdl_runtime * run;
	FILE * stream;
	
	if (gdl_isfile (runtime_file))
	{
		stream = gdl_fileopen (runtime_file, "r");
		
		run = gdl_runtime_fread (stream);
		
		gdl_fileclose (runtime_file, stream);
		
		if (!run)
		{
			gdl_string * err = gdl_string_sprintf ("Unable to load runtime from %s\n", runtime_file);
			GDL_ERROR_NULL (err, GDL_FAILURE);
			exit (1);
		}
	}
	else
	{
		const gdl_runtime_type * T = gdl_runtime_env_setup ();
		
		run = gdl_runtime_alloc (T);
	}
	
	return run;
}

static int 
runtime_setdown (gdl_runtime * run)
{
	int status;
	FILE * stream;
	
	stream = gdl_fileopen (runtime_file, "w");
		
	status = gdl_runtime_fwrite (stream, run);
	
	if (status != GDL_SUCCESS)
	{
		GDL_ERROR_VAL ("Unable to save the runtime", GDL_FAILURE, GDL_FAILURE);
		exit (1);
	}
		
	gdl_fileclose (runtime_file, stream);
	
	return GDL_SUCCESS;
}


static gdl_boolean
is_a_selected_feature (const gdl_gfeatures * feature,
                       const gdl_gfeatures_gff3_record * rec,
                       gdl_hashtable * selected_feature)
{
	if (!selected_feature)
		return gdl_true;
	
	const gdl_string * rec_type = gdl_dictionary_get (feature->dico, "type", rec->type_d);
	
	if (gdl_hashtable_lookup (selected_feature, ".")) // means that all features are needed
		return gdl_true;
	
	gdl_hashtable * bytype = gdl_hashtable_lookup (selected_feature, rec_type);
	
	if (!bytype)
		return gdl_false;
	if (gdl_hashtable_size (bytype)==0)
		return gdl_true;
	if (gdl_hashtable_lookup (bytype, ".")) // means that all ids are needed
		return gdl_true;
	if (gdl_hashtable_lookup (bytype, gdl_hashtable_lookup (rec->attributes, "ID")))
		return gdl_true;
	else if (rec->nparent)
	{
		size_t p;
		for(p = 0; p < rec->nparent; p++)
		{
			if (gdl_hashtable_lookup (bytype, gdl_hashtable_lookup (rec->parents[p]->attributes, "ID")))
				return gdl_true;
		}
	}
	return gdl_false;
}

static gdl_hashtable *
read_selected_features (const gdl_string * filename)
{
	FILE * stream;
	size_t i,j,n;
	gdl_string * key, * value, * line = 0;
	gdl_hashtable * selected, * bytype;
	
	stream = gdl_fileopen (filename, "r");
	
	selected = gdl_hashtable_alloc (gdl_hash_default, 0);
	
	while (gdl_getline (&line, &n, stream) != -1)
	{
		i = j = 0;
		key    = gdl_string_next_token (line, n, &i, &j);
		value  = gdl_string_next_token (line, n, &i, &j);
		bytype = gdl_hashtable_lookup (selected, key);
		if (!bytype)
		{
			bytype = gdl_hashtable_alloc (gdl_string_interface, 0);
			gdl_hashtable_add (selected, key, bytype, 0);
		}
		gdl_hashtable_add (bytype, value, value, 1);
		gdl_string_free (key);
		gdl_string_free (line);line=0;
	}
	
	gdl_fileclose (filename, stream);
	
	return selected;	
}
typedef struct
{
	size_t size;
	gdl_string ** seqid;
	long * start;
	long * end;
} _rm_region;

static _rm_region *
read_rm_region (const gdl_string * filename)
{
	FILE * stream;
	size_t i,j,n,b;
	gdl_string * tok, * line = 0;
	_rm_region * batch;
	
	batch = GDL_CALLOC (_rm_region, 1);
	
	stream = gdl_fileopen (filename, "r");
	while (gdl_getline (&line, &n, stream) != -1)
	{
		gdl_string_free (line);
		line=0;
		(batch->size)++;
	}
   gdl_fileclose (filename, stream);
   
   stream = gdl_fileopen (filename, "r");
	
	batch->seqid  = GDL_MALLOC (gdl_string *, batch->size);
	batch->start  = GDL_MALLOC (long, batch->size);
	batch->end    = GDL_MALLOC (long, batch->size);
	b=0;
   line=0;
   while (gdl_getline (&line, &n, stream) != -1)
	{ 
		i = j = 0;
		batch->seqid[b] = gdl_string_next_token (line, n, &i, &j);
		tok = gdl_string_next_token (line, n, &i, &j);
		batch->start[b]  = atol (tok);
		gdl_string_free (tok);
		tok = gdl_string_next_token (line, n, &i, &j);
		batch->end[b]    = atol (tok);
		gdl_string_free (tok);
		gdl_string_free (line);line=0;
		b++;
	}
	gdl_fileclose (filename, stream);
	
	return batch;
}

static void
_rm_region_perform (align_nucleo_alignment * nucleo_align, _rm_region * batch)
{
	size_t i,j;
	size_t nr = 0, * rix;
	
	for(i = 0; i < batch->size; i++)
	{
		if (!strcmp(batch->seqid[i],nucleo_align->ref_chrom->data->name))
			nr++;
	}
	if (!nr)
	{
		return;
	}	
	rix = GDL_MALLOC (size_t, nr);
	for(nr = i = 0; i < batch->size; i++)
	{
		if (!strcmp(batch->seqid[i],nucleo_align->ref_chrom->data->name))
		{
			rix[nr++]=i;
		}
	}
	for(j = 0; j < nr; j++)
	{
		long start = batch->start[rix[j]];
		long end   = batch->end[rix[j]];
		for(i = 0; i < nucleo_align->size; i++)
		{
			if (nucleo_align->ref[i] && (nucleo_align->ref[i]->ref_start_pos >= start && nucleo_align->ref[i]->ref_end_pos <= end))
			{
				nucleo_align->ref[i]->ignore = 'y';
				if (nucleo_align->ref[i]->align)
				{
					nucleo_align->ref[i]->align->ignore = 'y';
					if (nucleo_align->ref[i]->align->next_linker)
					{
						nucleo_align->ref[i]->align->next_linker->outlier = 1;	
					}
				}
				if (nucleo_align->ref[i]->next_linker)
				{
					nucleo_align->ref[i]->next_linker->outlier = 1;	
				}
			}
			if (nucleo_align->qry[i] && (nucleo_align->qry[i]->ref_start_pos >= start && nucleo_align->qry[i]->ref_end_pos <= end))
			{
				nucleo_align->qry[i]->ignore = 'y';
				if (nucleo_align->qry[i]->align)
				{
					nucleo_align->qry[i]->align->ignore = 'y';
					if (nucleo_align->qry[i]->align->next_linker)
					{
						nucleo_align->qry[i]->align->next_linker->outlier = 1;	
					}
				}
				if (nucleo_align->qry[i]->next_linker)
				{
					nucleo_align->qry[i]->next_linker->outlier = 1;	
				}
			}
		}
	}
	GDL_FREE (rix);
}

